import React, { Component, PropTypes } from 'react';
import { findDOMNode } from 'react-dom';
import { DragSource, DropTarget } from 'react-dnd';
import ItemTypes from './ItemTypes';

import TextInput from './TextInput';

const style = {
  fontSize: '25px',
  cursor: 'move',
};

const filterSource = {
  beginDrag(props) {
    return {
      id: props.id,
      index: props.index,
    };
  },
  endDrag(props, monitor) {
    $.ajax({
      url: '/admin/_move_feature_position',
      dataType: 'json',
      type: 'POST',
      data: { name: props.name, newPosition: props.position, link: props.link },
    });
  },
};

const filterTarget = {
  hover(props, monitor, component) {
    const dragIndex = monitor.getItem().index;
    const hoverIndex = props.index;

    // Don't replace items with themselves
    if (dragIndex === hoverIndex) {
      return;
    }

    // Determine rectangle on screen
    const hoverBoundingRect = findDOMNode(component).getBoundingClientRect();

    // Get vertical middle
    const hoverMiddleY = (hoverBoundingRect.bottom - hoverBoundingRect.top) / 2;

    // Determine mouse position
    const clientOffset = monitor.getClientOffset();

    // Get pixels to the top
    const hoverClientY = clientOffset.y - hoverBoundingRect.top;

    // Only perform the move when the mouse has crossed half of the items height
    // When dragging downwards, only move when the cursor is below 50%
    // When dragging upwards, only move when the cursor is above 50%

    // Dragging downwards
    if (dragIndex < hoverIndex && hoverClientY < hoverMiddleY) {
      return;
    }

    // Dragging upwards
    if (dragIndex > hoverIndex && hoverClientY > hoverMiddleY) {
      return;
    }

    // Time to actually perform the action
    props.moveFilter(dragIndex, hoverIndex);

    // Note: we're mutating the monitor item here!
    // Generally it's better to avoid mutations,
    // but it's good here for the sake of performance
    // to avoid expensive index searches.
    monitor.getItem().index = hoverIndex;
  },
};

@DropTarget(ItemTypes.FILTERROW, filterTarget, connect => ({
  connectDropTarget: connect.dropTarget(),
}))
@DragSource(ItemTypes.FILTERROW, filterSource, (connect, monitor) => ({
  connectDragSource: connect.dragSource(),
  isDragging: monitor.isDragging(),
}))
export default class FilterRow extends Component {
  static propTypes = {
    connectDragSource: PropTypes.func.isRequired,
    connectDropTarget: PropTypes.func.isRequired,
    index: PropTypes.number.isRequired,
    isDragging: PropTypes.bool.isRequired,
    id: PropTypes.any.isRequired,
    name: PropTypes.string.isRequired,
    moveFilter: PropTypes.func.isRequired,
    position: PropTypes.number.isRequired,
  };

  constructor(props) {
    super(props);
    this.state = {
      filter_name: '',
      filter_link: '',
    };
    this.onDelEvent = this.onDelEvent.bind(this);
    this.onSaveFilter = this.onSaveFilter.bind(this);
  }

  onDelEvent() {
    this.props.onDelEvent({ name: this.props.name, id: this.props.link });
  }

  onSaveFilter() {
    this.props.onSaveFilter({ name: this.iFilterName.value, id: this.iFilterId.value });
  }

  render() {
    const { name, isDragging, connectDragSource, connectDropTarget, link } = this.props;
    const opacity = isDragging ? 0 : 1;
    const isName = name !== '';

    return connectDragSource(connectDropTarget(
      <tr style={{ ...style, opacity }}>
        <td style={{ verticalAlign: 'middle' }} className="col-md-3 active">
          <form>
            <TextInput
              iref={(input) => { this.iFilterName = input; }}
              defaultValue={name || 'Input filter name'}
            />
          </form>
        </td>
        <td style={{ verticalAlign: 'middle' }} className="col-md-3 active">
          <form>
            <TextInput
              iref={(input) => { this.iFilterId = input; }}
              defaultValue={isName ? link : 'Input filter id'}
            />
          </form>
        </td>
        <td style={{ verticalAlign: 'middle' }} className="col-md-3 active">
          <a onClick={this.onSaveFilter} className="btn btn-sm btn-default">
            <i className="fa fa-floppy-o fa-2x" />
          </a>
        </td>
        <td style={{ verticalAlign: 'middle' }} className="col-md-3 active">
          <a onClick={this.onDelEvent} className="btn btn-sm btn-danger">
            <i className="fa fa-times fa-2x" />
          </a>
        </td>
      </tr>,
    ));
  }
}
