import _ from 'lodash';
import React from 'react';
import moment from 'moment';
import d3 from 'd3';
import { connect } from 'react-redux';
import { createSelector } from 'reselect';
import * as BS from 'react-bootstrap';

import { fetchAcceptanceSummary } from '../actions';
import { safeDiv } from '../../reports/base/utils';
import Widget from '../components/Widget';
import consts from '../../consts';
import NVD3Chart from '../../reports/base/NVD3Chart';
import ProgressBar from '../components/ProgressBar';


class AcceptanceSummary extends React.Component {

  static propTypes = {
    milestone: React.PropTypes.string.isRequired,
    isFetching: React.PropTypes.bool.isRequired,
    data: React.PropTypes.object.isRequired,
  };

  componentWillMount() {
    this.props.fetchAcceptanceSummary(this.props.milestone);
  }

  componentWillReceiveProps(nextProps) {
    if (nextProps.milestone !== this.props.milestone) {
      this.props.fetchAcceptanceSummary(nextProps.milestone);
    }
  }

  renderWidgetBody() {
    const data = this.props.data;

    const tempestCounts = data.tempest || {};
    const tempestTotal = _(tempestCounts).values().sum();

    const swarmCounts = data.swarm || {};
    const swarmTotal = _(swarmCounts).values().sum();

    const acceptanceCounts = data.acceptance || {};
    const acceptanceTotal = _(acceptanceCounts).values().sum();

    const totalCounts = data.total || {};
    const totalPassed = totalCounts.passed || 0;
    const totalTotal = _(totalCounts).values().sum();

    const totalChart = () => {
      if (!_.isEmpty(totalCounts)) {
        const totalPieData = consts.testrail.StatusOrder.map(status => ({
          label: status,
          value: totalCounts[status],
        }));
        return (
          <NVD3Chart
            type="pieChart"
            height={300}
            datum={totalPieData}
            configure={(chart) => {
              chart
              .margin({ top: 0, right: 0, bottom: 0, left: 0 })
              .x(d => d.label)
              .y(d => d.value)
              .color((d) => {
                const col = d3.scale.ordinal()
                .domain(consts.testrail.statusDomain)
                .range(consts.testrail.statusRange);
                return col(d.label);
              })
              .showLabels(false)
              .showLegend(false)
              .labelType('percent');
            }}
            configureD3={(selection) => {
              const pie = selection.selectAll('g.nv-slice:last-child');
              pie
              .insert('text', 'g')
              .text(`${safeDiv(totalPassed * 100, totalTotal).toFixed(1)}%`)
              .attr('class', 'middlePerc')
              .attr('text-anchor', 'middle')
              .attr('dy', '-.1em');
              pie
              .insert('text', 'g')
              .text(`${totalPassed} of ${totalTotal}`)
              .attr('class', 'middleText')
              .style('font-size', '16px')
              .attr('text-anchor', 'middle')
              .attr('dy', '1.11em');
              pie
              .insert('text', 'g')
              .text('cases passed')
              .attr('class', 'middleText')
              .style('font-size', '16px')
              .attr('text-anchor', 'middle')
              .attr('dy', '2.22em');
            }}
          />
        );
      }
    };

    return (
      <div className="container-fluid">
        <BS.Row>
          <BS.Col md={3} style={{ paddingLeft: 0, paddingRight: 0, marginTop: '0px' }}>
            <div style={{ position: 'absolute', fontSize: 26, fontWeight: 'bold', left: '50%', WebkitTransform: 'translateX(-50%)', transform: 'translateX(-50%)' }}>All suites</div>
            { totalChart() }
          </BS.Col>
          <BS.Col md={9} style={{ paddingLeft: 0 }}>
            <ProgressBar data={swarmCounts} max={swarmTotal} name="Swarm" />
            <ProgressBar data={tempestCounts} max={tempestTotal} name="Tempest" />
            <ProgressBar data={acceptanceCounts} max={acceptanceTotal} name="Acceptance" />
          </BS.Col>
        </BS.Row>
      </div>
    );
  }

  render() {
    let body = null;
    let title = 'Recent Acceptance Testing results';

    if (this.props.isFetching) {
      body = <h3>Fetching data ... </h3>;
    } else if (_.isEmpty(this.props.data)) {
      body = <h3>No acceptance runs yet</h3>;
    } else {
      body = this.renderWidgetBody();
      title += ` ${this.props.data.plan_name} (${moment(this.props.data.date).format('MMM D')})`;
    }

    return (
      <Widget
        title={title}
        expandUrl="report/acceptance-testing-progress"
        fill
      >
        {body}
      </Widget>
    );
  }
}

const lastRecordSelector = createSelector(
  state => state.data,
  (data) => {
    if (_.isEmpty(data)) return {};
    if (_(data.tempest).values().sum() +
      _(data.swarm).values().sum() +
      _(data.acceptance).values().sum() === 0) {
      return {};
    }
    return data;
  }
);

function mapStateToProps(state) {
  return {
    isFetching: state.dashboard.acceptanceSummary.isFetching,
    data: lastRecordSelector(state.dashboard.acceptanceSummary) || {},
  };
}

export default connect(mapStateToProps, {
  fetchAcceptanceSummary,
})(AcceptanceSummary);
