/**
 * Created by farhad on 07.06.16.
 */
import _ from 'lodash';
import React from 'react';
import d3 from 'd3';
import { connect } from 'react-redux';
import * as BS from 'react-bootstrap';
import moment from 'moment';

import { fetchBugsVerificationSummary } from '../actions';
import consts from '../../consts';
import Widget from '../components/Widget';
import NVD3Chart from '../../reports/base/NVD3Chart';


class BugsVerificationSummary extends React.Component {

  static propTypes = {
    milestone: React.PropTypes.string.isRequired,
    isFetching: React.PropTypes.bool.isRequired,
    data: React.PropTypes.array.isRequired,
  };

  componentWillMount() {
    this.props.fetchBugsVerificationSummary(this.props.milestone);
  }

  componentWillReceiveProps(nextProps) {
    if (nextProps.milestone !== this.props.milestone) {
      this.props.fetchBugsVerificationSummary(nextProps.milestone);
    }
  }

  renderCell(bugType, importance) {
    const data = _(this.props.data)
      .filter(bugType)
      .filter(x => !importance || x.importance === importance)
      .groupBy('status')
      .mapValues(items => _.sumBy(items, 'count'))
      .value();

    const cntFixReleased = data[consts.launchpad.STATUS_FIX_RELEASED] || 0;
    const cntFixCommitted = data[consts.launchpad.STATUS_FIX_COMMITTED] || 0;
    const verificationRate = cntFixReleased / (cntFixReleased + cntFixCommitted) || 0;
    return (
      <span
        title={`${cntFixReleased} FixReleased / ${cntFixCommitted} FixCommitted`}
      >
        {`${Math.round(verificationRate * 100)}%`}
      </span>
    );
  }

  renderWidgetBody() {
    const statuses = [consts.launchpad.IMPORTANCE_CRITICAL, consts.launchpad.IMPORTANCE_HIGH, null];

    const result = statuses.map(status =>
      _(this.props.data)
        .filter('is_product')
        .filter(x => !status || x.importance === status)
        .groupBy('status')
        .mapValues(items => _.sumBy(items, 'count'))
        .set('importance', status)
        .value()
    );
    const chartData = result.map((status) => {
      const cntFixReleased = status[consts.launchpad.STATUS_FIX_RELEASED] || 0;
      const cntFixCommitted = status[consts.launchpad.STATUS_FIX_COMMITTED] || 0;
      const verificationRate = cntFixReleased / (cntFixReleased + cntFixCommitted) || 0;
      const val = Math.round(verificationRate * 100);

      return [
        {
          label: _.lowerCase(status.importance) || 'all',
          value: val,
          fix_released: cntFixReleased,
          fix_committed: cntFixCommitted,
        },
        {
          label: _.lowerCase(status.importance) || 'all',
          value: 100 - val,
          fix_released: cntFixReleased,
          fix_committed: cntFixCommitted,
        },
      ];
    }
  );
    return (
      <div>
        <BS.Row>
          {chartData.map((data, index) =>
            <BS.Col key={index} sm={6} md={4}>
              <NVD3Chart
                type="pieChart"
                key={index}
                height={210}
                datum={data}
                configure={(chart) => {
                  chart
                  .margin({ top: 0, right: 0, bottom: 0, left: 0 })
                  .x(d => d.label)
                  .y(d => d.value)
                  .color((d, i) => {
                    const col = (num) => {
                      if (num <= 50) {
                        return d3.rgb(238, Math.round(221 * (num / 50)), 15);
                      }
                      if (num === 100) {
                        return d3.rgb(20, 137, 44);
                      }
                      return d3.rgb(
                      Math.round(255 * ((50 - (num % 50)) / 50)),
                      175,
                      80);
                    };
                    return i === 0 ? col(d.value) : '#254061';
                  })
                  .showLabels(false)
                  .showLegend(false)
                  .labelType('percent');
                  chart.tooltip.contentGenerator((sliceData) => {
                    const fixCommittedRow = (
                      '<tr>' +
                      '<td class="key">Fix Committed:</td>' +
                      `<td class="x-value">${sliceData.data.fix_committed}</td>` +
                      '</tr>'
                    );
                    const fixReleasedRow = (
                      '<tr>' +
                      '<td class="key">Fix Released:</td>' +
                      `<td class="x-value">${sliceData.data.fix_released}</td>` +
                      '</tr>'
                    );
                    return `<table><tbody>${fixCommittedRow}${fixReleasedRow}</tbody></table>`;
                  });
                }}
                configureD3={(selection) => {
                  const pie = selection.selectAll('g.nv-slice:last-child');
                  pie
                  .insert('text', 'g')
                  .text(`${_.get(data, '0.value')}%`)
                  .attr('class', 'middlePerc')
                  .attr('text-anchor', 'middle')
                  .attr('dy', '-.1em');
                  pie
                  .insert('text', 'g')
                  .text(_.get(data, '0.label'))
                  .attr('class', 'middleText')
                  .attr('text-anchor', 'middle')
                  .attr('dy', '.99em');
                }}
              />
            </BS.Col>
          )}
        </BS.Row>

        <BS.Table striped bordered condensed hover>
          <thead>
            <tr>
              <th>{}</th>
              <th>Product customer-affecting</th>
              <th>Release-related</th>
              <th>Internal QA</th>
            </tr>
          </thead>
          <tbody>
            <tr>
              <th>Crit</th>
              <td>{this.renderCell('is_product', consts.launchpad.IMPORTANCE_CRITICAL)}</td>
              <td>{this.renderCell('is_related', consts.launchpad.IMPORTANCE_CRITICAL)}</td>
              <td>{this.renderCell('is_internalqa', consts.launchpad.IMPORTANCE_CRITICAL)}</td>
            </tr>
            <tr>
              <th>High</th>
              <td>{this.renderCell('is_product', consts.launchpad.IMPORTANCE_HIGH)}</td>
              <td>{this.renderCell('is_related', consts.launchpad.IMPORTANCE_HIGH)}</td>
              <td>{this.renderCell('is_internalqa', consts.launchpad.IMPORTANCE_HIGH)}</td>
            </tr>
            <tr>
              <th>All</th>
              <td>{this.renderCell('is_product', null)}</td>
              <td>{this.renderCell('is_related', null)}</td>
              <td>{this.renderCell('is_internalqa', null)}</td>
            </tr>
          </tbody>
        </BS.Table>
      </div>
);
  }

  render() {
    return (
      <Widget
        title="Bugs verification progress"
        expandUrl="report/bugs-verification-progress"
        widgetTitle={'Bugs verification'}
        widgetLabel={` - product related bugs (${moment().format('MMMM D')})`}
        fill
      >
        {this.props.isFetching ? null : this.renderWidgetBody()}
      </Widget>
    );
  }
}

function mapStateToProps(state) {
  return {
    isFetching: state.dashboard.bugsVerificationSummary.isFetching,
    data: state.dashboard.bugsVerificationSummary.data || [],
  };
}

export default connect(mapStateToProps, {
  fetchBugsVerificationSummary,
})(BugsVerificationSummary);
