import React from 'react';
import _ from 'lodash';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import ReportBox from './base/ReportBox';
import TempestChart from './D3Charts/TempestChart';
import SimpleRangePicker from '../components/SimpleRangePicker';
import CheckBox from '../components/CheckBox';
import consts from '../consts';
import { safeDiv, isEqual } from './base/utils';
import getStatusColumns,
{ statsDefaultFieldMap } from './base/CommonColumns';

const GRID_DATA_URL = '/api/v1.0/testrail/acceptance_progress';

const statsFieldMap = {
  ...statsDefaultFieldMap,
};

function getMostFrequentPlanName(runs) {
  const [frequentPlanName] = _(runs)
    .map('data.plan_name')
    .countBy()
    .toPairs()
    .maxBy(item => item[1]);
  return frequentPlanName;
}

let acceptanceMilestones = null;

fetch('/api/v1.0/testrail/acceptance_runs', { credentials: 'same-origin' })
  .then(response => response.json())
  .then(data => (acceptanceMilestones = _(data).map('milestone_id').value()));

const getColumnDefs = () => ([
  {
    headerName: 'Date',
    field: 'date',
    rowGroupIndex: 0,
    cellRenderer: {
      renderer: 'group',
    },
    cellStyle: { 'font-weight': 'bold' },
    width: 150,
    pinned: true,
  },
  {
    headerName: 'Name',
    field: 'plan_name',
    width: 150,
  },
  {
    headerName: 'Test suite',
    field: 'config_name',
    width: 300,
  },
  {
    field: 'qa_team_id',
    hide: true,
  },
].concat(getStatusColumns(statsFieldMap)));

class AcceptanceTestingProgress extends ReportBox {
  static defaultProps = Object.assign({}, ReportBox.defaultProps, {
    title: 'Acceptance Testing Progress',
    columnDefs: getColumnDefs,
    csvExportFilename: 'acceptance_testing_progress.csv',
    gridDataUrl: GRID_DATA_URL,
    gridHeight: 400,
    chartHeight: 800,
    lastSyncInfoUrl: '/api/v1.0/testrail/sync/last',
    milestonesListUrl: '/api/v1.0/testrail/milestone_names',
    defaultSortModel: [{ colId: 'date', sort: 'desc' }],
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
    },
  });

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;
    this.state.teams = [];
    this.state.selectedTeam = [];
    this.state.datesRange = [
      moment().subtract(7, 'days').startOf('day'),
      moment().endOf('day'),
    ];

    this.state.minimise = false;

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
    this.handleTeamSelection = this.handleTeamSelection.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
    this.onToggleMinimiseChart = this.onToggleMinimiseChart.bind(this);
  }

  componentDidMount() {
    this.fetchInitialData();
    this.setConfiguration(this.initialConfiguration);
  }

  componentDidUpdate(prevProps, prevState) {
    if (!isEqual(this.state.selectedMilestones, prevState.selectedMilestones)) {
      this.fetchInitialData();
    }
    if (!isEqual(this.state.datesRange, prevState.datesRange)) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      selectedMilestones: [_(window.commonStore.testrail_milestones)
        .filter(o => _.includes(acceptanceMilestones, o.id))
        .map('id')
        .value()[0]
        .toString()],
    };
  }

  getConfiguration() {
    return {
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  onToggleMinimiseChart(e) {
    this.setState({ minimise: e.target.checked });
  }

  filterTeams() {
    const teamFilter = this.gridApi.getFilterApi('qa_team_id');
    teamFilter.setModel(this.state.selectedTeam);
    this.gridApi.onFilterChanged();
  }

  handleTeamSelection() {
    const selectedTeam = [];
    this.teamSelector.$multiselect
      .find('option:selected').each(function getVal() {
        selectedTeam.push($(this).val());
      });
    this.setState({ selectedTeam });
    this.filterTeams();
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each(
      (i, elem) => selectedMilestones.push($(elem).val()));
    this.setStateIf({ selectedMilestones });
  }

  updateGridData(data) {
    const preparedData = this.prepareGridData(data);
    const selection = _(preparedData).map('qa_team_id').uniq()
      .map(qaTeam => ({
        value: qaTeam,
        label: _(window.commonStore.testrail_qa_teams).find({ id: qaTeam }).name,
        selected: true,
      })
      )
      .value();
    this.setState({ teams: selection });
    this.setState({ selectedTeams: _(selection).map('value').value() });
    this.gridApi.setRowData(preparedData);
    this.setState({ rowData: preparedData });
  }

  prepareGridData(data) {
    _.each(data, (run) => {
      Object.assign(run, _.mapKeys(run.status_counts, (v, k) => `cnt_${k}`));
      run.cnt_total = _(run.status_counts).values().sum() || 0;
      run.date = moment(run.date).format('YYYY-MM-DD');
    });
    return data;
  }

  fetchInitialData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      milestone_id: this.state.selectedMilestones.join(' '),
    })
    .done((data) => {
      this.updateGridData(data);
    })
    .always(() => { this.gridApi.hideOverlay(); });
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      milestone_id: this.state.selectedMilestones.join(' '),
    })
      .done((date) => {
        this.updateGridData(date);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  onDatesRangeChange(newRange) {
    this.setStateIf({ datesRange: newRange });
  }

  getChartDatum() {
    const ret = [];
    if (!this.gridApi || _.isEmpty(this.state.rowData)) {
      return ret;
    }

    const model = this.gridApi.getModel();
    const groupColDef = _.get(this.columnApi.getAllDisplayedColumns(), '0.colDef');
    if (!groupColDef) {
      return ret;
    }

    const dates = {};
    model.forEachNodeAfterFilterAndSort((node) => {
      if (node.group && node.level === 0) {
        dates[node.key] = _.get(node, 'data');
        dates[node.key].plan_name = getMostFrequentPlanName(node.children);
      }
    });

    const statusResult = _.values(consts.testrail.Statuses);
    const statusColumns = _(this.columnApi.getAllColumns()).filter((obj) => {
      function isStatus(status) {
        if (obj.colDef.field === 'pct_blocked') { return false; }
        if (obj.colDef.field === 'cnt_total_failed') { return false; }
        return obj.colDef.field.indexOf(status) > -1;
      }
      return statusResult.some(isStatus);
    }).value();
    const succesRateCol = {
      key: '%Success Rate',
      type: 'line',
      yAxis: 2,
      values: _(dates).mapValues((v, k) => ({
        x: moment(k), y: safeDiv(v.cnt_passed, v.cnt_total),
      })).map().sortBy('x').value(),
      color: '#444141',
    };
    const result = _(statusColumns).map(obj => ({
      key: obj.colDef.headerName,
      type: 'bar',
      yAxis: 1,
      values: _(dates).mapValues((v, k) => ({
        x: moment(k),
        y: v[obj.colDef.field],
        plan_name: v.plan_name,
      })).map().sortBy('x').value(),
      color: obj.colDef.chart.color,
    })).value();
    return result.concat(succesRateCol);
  }

  renderChartCmp() {
    const datum = this.getChartDatum();
    return (
      <TempestChart
        height={this.props.chartHeight}
        minimise={this.state.minimise}
        datum={datum}
      />
    );
  }

  renderToolBarCmp() {
    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        onChange={this.onDatesRangeChange}
        startDate={this.state.datesRange[0]}
        endDate={this.state.datesRange[1]}
      />
    );
    const milestoneData = window.commonStore.testrail_milestones
    .filter(m => _.includes(acceptanceMilestones, m.id))
    .map(m => ({
      value: m.id,
      label: m.name,
      selected: _.includes(this.state.selectedMilestones, m.id.toString()),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        nonSelectedText="select milestone"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );
    const teamSelect = (
      <Multiselect
        key="team.select"
        ref={(c) => { this.teamSelector = c; }}
        multiple
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        includeSelectAllOption
        nonSelectedText="select team"
        allSelectedText="all selected"
        disableIfEmpty
        disabledText="No teams"
        onDropdownHide={this.handleTeamSelection}
        data={this.state.teams}
      />
    );
    const checkBox = (
      <CheckBox
        label=" Minimise chart"
        key="show.minim"
        checked={this.state.minimise}
        onChange={this.onToggleMinimiseChart}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );
    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect, teamSelect, checkBox] });
  }
}

export default AcceptanceTestingProgress;
