import React from 'react';
import _ from 'lodash';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import ReportBox from './base/ReportBox';
import AutomationChart from './D3Charts/AutomationChart';
import SimpleRangePicker from '../components/SimpleRangePicker';
import CheckBox from '../components/CheckBox';
import consts from '../consts';
import './base/multiChartExtended';
import { getComponentsSum, toFixedCelRenderer, safeDiv,
         getNodeValue, makeDatesSequence, isEqual } from './base/utils';


const FIELD_NA = 'n/a';
const FIELD_ADOPTION_IN_PROGRESS = 'in_progress';
const FIELD_ADOPTED = 'adopted';
const FIELD_NOT_ADOPTED = 'not_adopted';
const FIELD_PCT_AUTOMATED = 'pct_automated';

function renderSuiteLink(suiteID, text) {
  return `<a href="${consts.testrail.TESTRAIL_BASE_URL}/index.php?/suites/view/${suiteID}">${text}</a>`;
}

function getRanges() {
  return {
    'Last 7 Days': [
      moment().subtract(6, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 30 Days': [
      moment().subtract(29, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 3 Month': [
      moment().subtract(2, 'month').startOf('month'),
      moment(),
    ],
  };
}

const getColumnDefs = () => ([
  {
    headerName: 'Suites',
    field: '__all_teams__',
    rowGroupIndex: 0,
    valueGetter: () => 'Suites',
    hide: true,
  },
  {
    headerName: 'Suite name',
    rowGroupIndex: 1,
    field: 'suite_name',
    filter: 'set',
    cellRenderer(params) {
      const suiteId = getNodeValue(params, 'suite_id');
      if (params.value && suiteId) {
        return renderSuiteLink(suiteId, params.value);
      }
      return params.value;
    },
    width: 110,
  },
  {
    headerName: 'Milestone',
    field: 'milestone_name',
    filter: 'set',
    width: 110,
  },
  {
    headerName: 'QA Team',
    field: 'qa_team_name',
    cellRenderer: {
      renderer: 'group',
    },
    valueGetter(params) {
      return window.commonStore.getTestRailQATeamName(
          params.data.qa_team) || consts.testrail.DEFAULT_TEAM_NAME;
    },
    cellStyle: { 'font-weight': 'bold' },
    filter: 'set',
    width: 200,
    pinned: true,
  },

  {
    headerName: 'Not adopted',
    field: FIELD_NOT_ADOPTED,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    aggFunc: 'sum',
    chart: { color: '#E06666', type: 'bar', yAxis: 1 },
    width: 110,
  },
  {
    headerName: 'Adoption in progress',
    field: FIELD_ADOPTION_IN_PROGRESS,
    aggFunc: 'sum',
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#A6D49D', type: 'bar', yAxis: 1 },
    width: 180,
  },
  {
    headerName: 'Adopted',
    field: FIELD_ADOPTED,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    aggFunc: 'sum',
    chart: { color: '#69BA69', type: 'bar', yAxis: 1 },
    width: 110,
  },
  {
    headerName: 'n/a',
    width: 150,
    field: FIELD_NA,
    filter: 'number',
    aggFunc: 'sum',
    chart: { color: '#FFFF00', type: 'bar', yAxis: 1 },
  },
  {
    headerName: '% Adopted',
    field: FIELD_PCT_AUTOMATED,
    valueGetter(params) {
      return 100 * safeDiv(
          params.getValue(FIELD_ADOPTED),
          getComponentsSum([
            FIELD_NOT_ADOPTED,
            FIELD_ADOPTION_IN_PROGRESS,
            FIELD_ADOPTED, FIELD_NA,
          ], params),
        );
    },
    cellRenderer: toFixedCelRenderer,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
    width: 120,
  },
]);

export default class AdoptionProgress extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs: getColumnDefs,
    chartHeight: 600,
    title: 'TestRail: Adoption Progress',
    csvExportFilename: 'testrail_adoption_progress.csv',
    gridDataUrl: consts.testrail.ADOPTION_PROGRESS_RESOURCE,
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
      groupDefaultExpanded: 1,
    },
  };

  constructor(props) {
    super(props);
    this.state.datesRange = this.initialConfiguration.datesRange.map(d => moment(d));
    this.state.intervalType = this.initialConfiguration.intervalType;
    this.state.minimise = false;

    this.onIntervalTypeChange = this.onIntervalTypeChange.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
    this.onToggleMinimiseChart = this.onToggleMinimiseChart.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'intervalType'];
    if (!isEqual(_.pick(this.state, fetchParams), _.pick(prevState, fetchParams))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      datesRange: getRanges()['Last 7 Days'].map(d => d.toJSON()),
      intervalType: 'days',
    };
  }

  getConfiguration() {
    const conf = super.getConfiguration();
    return {
      ...conf,
      datesRange: this.state.datesRange.map(d => d.toJSON()),
      intervalType: this.state.intervalType,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf(_.pick(conf, ['intervalType']));
    if (conf.datesRange) {
      this.setStateIf({ datesRange: conf.datesRange.map(d => moment(d)) });
    }
  }

  getChartDates() {
    return makeDatesSequence(
      moment(this.state.datesRange[0]).endOf('day'),
      moment(this.state.datesRange[1]).endOf('day'),
      this.state.intervalType,
    );
  }

  prepareGridData(data) {
    const chartData = [];
    const gridData = [];

    const chartDates = _.fromPairs(this.getChartDates().map(v => [+v, true]));

    _.each(data, (obj) => {
      Object.assign(obj, obj.adoption_counts);
      obj.total = _.sum(_.values(obj.adoption_counts)) || 0;
      obj.date = new Date(obj.date);
      obj.milestone_name = obj.milestone_name || 'Unknown';

      if (+this.state.datesRange[1] === +obj.date) {
        gridData.push(obj);
      }

      if (chartDates[+obj.date]) {
        chartData.push(obj);
      }
    });
    this.setState({ chartData });
    return gridData;
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    const endOfToday = moment(this.state.datesRange[1]).endOf('day');
    const requestDates = [...this.getChartDates(), endOfToday];

    return $.get(this.props.gridDataUrl, {
      dates: requestDates.map(d => d.toISOString()).join(),
    }).done((data) => {
      this.updateGridData(data);
    })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  onDatesRangeChange(newRange) {
    this.setStateIf({ datesRange: newRange });
  }

  onIntervalTypeChange() {
    this.setStateIf({ intervalType: this.intervalTypeSelect.$multiselect.val() });
  }

  onToggleMinimiseChart(e) {
    this.setState({ minimise: e.target.checked });
  }

  getChartDatum() {
    if (!this.state.chartData) { return []; }

    const teamsFilter = this.gridApi.getFilterApi('qa_team_name');
    const milestonesFilter = this.gridApi.getFilterApi('milestone_name');
    const suiteName = this.gridApi.getFilterApi('suite_name');

    const teamsSelection = {};
    _.forEach(window.commonStore.testrail_qa_teams, (team) => {
      teamsSelection[team.id] = _.includes(teamsFilter.getModel(), team.name);
    });

    const grouped = _(this.state.chartData)
      .filter(obj => (
        (teamsFilter.isEverythingSelected() || teamsSelection[obj.qa_team]) &&
        (milestonesFilter.isEverythingSelected() ||
        _.includes(milestonesFilter.getModel(), obj.milestone_name)) &&
        (suiteName.isEverythingSelected() || _.includes(suiteName.getModel(), obj.suite_name))
      ))
      .groupBy('date')
      .values()
      .sortBy('date')
      .value();

    function getValues(field) {
      return _.map(grouped, rows => ({
        x: moment(rows[0].date),
        y: _.sumBy(rows, field),
      }));
    }

    return [
      {
        key: 'Manual',
        type: 'bar',
        yAxis: 1,
        values: getValues(FIELD_NOT_ADOPTED),
        color: '#E06666',
      },
      {
        key: 'Will not Automate',
        type: 'bar',
        yAxis: 1,
        values: getValues(FIELD_NA),
        color: '#FFFF00',
      },
      {
        key: 'Automation in Progress',
        type: 'bar',
        yAxis: 1,
        values: getValues(FIELD_ADOPTION_IN_PROGRESS),
        color: '#A6D49D',
      },
      {
        key: 'Automated',
        type: 'bar',
        yAxis: 1,
        values: getValues(FIELD_ADOPTED),
        color: '#69BA69',
      },
      {
        key: '% Adoption',
        color: '#5DA4B6',
        type: 'line',
        values: _.map(grouped, (rows) => {
          const date = rows[0].date;
          const cntM = _.sumBy(rows, FIELD_NOT_ADOPTED);
          const cntAiP = _.sumBy(rows, FIELD_ADOPTION_IN_PROGRESS);
          const cntAut = _.sumBy(rows, FIELD_ADOPTED);
          return {
            x: date,
            y: cntAut / (cntM + cntAiP + cntAut),
          };
        }),
        yAxis: 2,
      },
    ];
  }

  renderChartCmp() {
    const datum = this.getChartDatum();

    return (
      <AutomationChart
        adoption
        minimise={this.state.minimise}
        height={this.props.chartHeight}
        datum={datum}
      />
    );
  }

  renderToolBarCmp() {
    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        ranges={getRanges()}
        onChange={this.onDatesRangeChange}
        startDate={this.state.datesRange[0]}
        endDate={this.state.datesRange[1]}
      />
    );

    const intervalTypeChoices = ['days', 'weeks', 'months'].map(
      v => ({ value: v, selected: v === this.state.intervalType }),
    );

    const intervalTypeSelect = (
      <Multiselect
        key="intervalType.select"
        ref={(c) => { this.intervalTypeSelect = c; }}
        onChange={this.onIntervalTypeChange}
        data={intervalTypeChoices}
      />
    );
    const checkBox = (
      <CheckBox
        label=" Minimise chart"
        key="show.minim"
        checked={this.state.minimise}
        onChange={this.onToggleMinimiseChart}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    return super.renderToolBarCmp({ main: [datePicker, intervalTypeSelect, checkBox] });
  }
}
