/**
 * Created by farhad on 20.02.16.
 */
import React from 'react';
import _ from 'lodash';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import ReportBox from './base/ReportBox';
import AutomationChart from './D3Charts/AutomationChart';
import SimpleRangePicker from '../components/SimpleRangePicker';
import CheckBox from '../components/CheckBox';
import consts from '../consts';
import './base/multiChartExtended';
import { getComponentsSum, toFixedCelRenderer, safeDiv,
         getNodeValue, makeDatesSequence, isEqual } from './base/utils';
import AutomationProgressAutomatedFilter from './base/filters';

function renderSuiteLink(suiteID, text) {
  return `<a href="${consts.testrail.TESTRAIL_BASE_URL}/index.php?/suites/view/${suiteID}">${text}</a>`;
}

function getRanges() {
  return {
    'Last 7 Days': [
      moment().subtract(6, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 30 Days': [
      moment().subtract(29, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 3 Month': [
      moment().subtract(2, 'month').startOf('month'),
      moment(),
    ],
  };
}

const getColumnDefs = () => ([
  {
    headerName: 'All teams',
    field: '__all_teams__',
    rowGroupIndex: 0,
    valueGetter: () => 'All teams',
    hide: true,
  },
  {
    headerName: 'QA Team',
    field: 'qa_team_name',
    rowGroupIndex: 1,
    cellRenderer: {
      renderer: 'group',
    },
    valueGetter(params) {
      return window.commonStore.getTestRailQATeamName(
          params.data.qa_team) || consts.testrail.DEFAULT_TEAM_NAME;
    },
    cellStyle: { 'font-weight': 'bold' },
    width: 200,
    pinned: true,

  },
  {
    headerName: 'Milestone',
    field: 'milestone_name',
    filter: 'set',
    filterParams: {
      values: _.map(window.commonStore.testrail_milestones, 'name'),
      newRowsAction: 'keep',
      comparator: (a, b) => (a < b ? 1 : -1),
    },
    width: 110,
  },
  {
    headerName: 'Suite name',
    rowGroupIndex: 2,
    field: 'suite_name',
    filterParams: {
      newRowsAction: 'keep',
    },
    cellRenderer(params) {
      const suiteId = getNodeValue(params, 'suite_id');
      if (params.value && suiteId) {
        return renderSuiteLink(suiteId, params.value);
      }
      return params.value;
    },
    width: 110,
  },
  {
    headerName: 'Manual',
    field: consts.testrail.FIELD_MANUAL,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    aggFunc: 'sum',
    chart: { color: '#E06666', type: 'bar', yAxis: 1 },
    width: 110,
  },
  {
    headerName: 'Automation in Progress',
    field: consts.testrail.FIELD_AUTOMATION_IN_PROGRESS,
    aggFunc: 'sum',
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#A6D49D', type: 'bar', yAxis: 1 },
    width: 180,
  },
  {
    headerName: 'Automated',
    field: consts.testrail.FIELD_AUTOMATED,
    filter: AutomationProgressAutomatedFilter,
    aggFunc: 'sum',
    chart: { color: '#69BA69', type: 'bar', yAxis: 1 },
    width: 110,
  },
  {
    headerName: 'Will not Automate',
    width: 150,
    field: consts.testrail.FIELD_WILL_NOT_AUTOMATE,
    filter: 'number',
    aggFunc: 'sum',
    chart: { color: '#FFFF00', type: 'bar', yAxis: 1 },
  },
  {
    headerName: '% Automated',
    field: consts.testrail.FIELD_PCT_AUTOMATED,
    valueGetter(params) {
      return 100 * safeDiv(
          params.getValue(consts.testrail.FIELD_AUTOMATED),
          getComponentsSum([
            consts.testrail.FIELD_MANUAL,
            consts.testrail.FIELD_AUTOMATION_IN_PROGRESS,
            consts.testrail.FIELD_AUTOMATED,
          ], params),
        );
    },
    cellRenderer: toFixedCelRenderer,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
    width: 120,
  },
]);

export default class AutomationProgress extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs: getColumnDefs,
    chartHeight: 600,
    title: 'TestRail: Automation Progress',
    csvExportFilename: 'testrail_automation_progress.csv',
    gridDataUrl: consts.testrail.AUTOMATION_PROGRESS_RESOURCE,
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
      groupDefaultExpanded: 1,
    },
  };

  constructor(props) {
    super(props);
    this.state.isHideDuplicates = this.initialConfiguration.isHideDuplicates;
    this.state.datesRange = this.initialConfiguration.datesRange.map(d => moment(d));
    this.state.intervalType = this.initialConfiguration.intervalType;
    this.state.minimise = false;

    this.onIntervalTypeChange = this.onIntervalTypeChange.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
    this.onToggleHideDuplicates = this.onToggleHideDuplicates.bind(this);
    this.onToggleMinimiseChart = this.onToggleMinimiseChart.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'isHideDuplicates', 'intervalType'];
    if (!isEqual(_.pick(this.state, fetchParams), _.pick(prevState, fetchParams))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      isHideDuplicates: false,
      datesRange: getRanges()['Last 7 Days'].map(d => d.toJSON()),
      intervalType: 'days',
    };
  }

  getConfiguration() {
    const conf = super.getConfiguration();
    return {
      ...conf,
      isHideDuplicates: this.state.isHideDuplicates,
      datesRange: this.state.datesRange.map(d => d.toJSON()),
      intervalType: this.state.intervalType,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf(_.pick(conf, ['isHideDuplicates', 'intervalType']));
    if (conf.datesRange) {
      this.setStateIf({ datesRange: conf.datesRange.map(d => moment(d)) });
    }
  }

  getChartDates() {
    return makeDatesSequence(
      moment(this.state.datesRange[0]).endOf('day'),
      moment(this.state.datesRange[1]).endOf('day'),
      this.state.intervalType
    );
  }

  prepareGridData(data) {
    const chartData = [];
    const gridData = [];

    const chartDates = _.fromPairs(this.getChartDates().map(v => [+v, true]));

    _.each(data, (obj) => {
      Object.assign(obj, obj.type_counts);
      obj.total = _.sum(_.values(obj.type_counts)) || 0;
      obj.date = new Date(obj.date);

      if (+this.state.datesRange[1] === +obj.date) {
        gridData.push(obj);
      }

      if (chartDates[+obj.date]) {
        chartData.push(obj);
      }
    });
    this.setState({ chartData });
    return gridData;
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    const endOfToday = moment(this.state.datesRange[1]).endOf('day');
    const requestDates = [...this.getChartDates(), endOfToday];

    return $.get(this.props.gridDataUrl, {
      dates: requestDates.map(d => d.toISOString()).join(),
      distinct: this.state.isHideDuplicates,
    }).done((data) => {
      this.updateGridData(data);
    })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  onDatesRangeChange(newRange) {
    this.setStateIf({ datesRange: newRange });
  }

  onToggleHideDuplicates(e) {
    this.setStateIf({ isHideDuplicates: e.target.checked });
  }

  onIntervalTypeChange() {
    this.setStateIf({ intervalType: this.intervalTypeSelect.$multiselect.val() });
  }

  onToggleMinimiseChart(e) {
    this.setState({ minimise: e.target.checked });
  }

  getChartDatum() {
    if (!this.state.chartData) { return []; }

    const milestonesFilter = this.gridApi.getFilterApi('milestone_name');
    const suiteName = this.gridApi.getFilterApi('suite_name');

    const teamsSelection = {};
    window.commonStore.testrail_qa_teams.forEach((team) => {
      teamsSelection[team.id] = false;
    });

    const model = this.gridApi.getModel();
    model.forEachNodeAfterFilterAndSort((node) => {
      if (node.group && node.level === 1) {
        let teamQa = _.find(window.commonStore.testrail_qa_teams, team => team.name === node.key);
        if (!teamQa && node.key === 'Unknown') {
          teamQa = { id: 555, name: 'Unknown' };
        }
        teamsSelection[teamQa.id] = true;
      }
    });

    const grouped = _(this.state.chartData)
      .filter(obj => (
        teamsSelection[obj.qa_team || 555] &&
        (milestonesFilter.isEverythingSelected() ||
        _.includes(milestonesFilter.getModel(), obj.milestone_name)) &&
        (suiteName.isEverythingSelected() ||
        _.includes(suiteName.getModel(), obj.suite_name))
      ))
      .groupBy('date')
      .values()
      .sortBy('date')
      .value();

    function getValues(field) {
      return _.map(grouped, rows => ({
        x: moment(rows[0].date),
        y: _.sumBy(rows, field),
      }));
    }

    return [
      {
        key: 'Manual',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_MANUAL),
        color: '#E06666',
      },
      {
        key: 'Will not Automate',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_WILL_NOT_AUTOMATE),
        color: '#FFFF00',
      },
      {
        key: 'Automation in Progress',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_AUTOMATION_IN_PROGRESS),
        color: '#A6D49D',
      },
      {
        key: 'Automated',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_AUTOMATED),
        color: '#69BA69',
      },
      {
        key: '% Automated',
        color: '#5DA4B6',
        type: 'line',
        values: _.map(grouped, (rows) => {
          const date = rows[0].date;
          const cntM = _.sumBy(rows, consts.testrail.FIELD_MANUAL);
          const cntAiP = _.sumBy(rows, consts.testrail.FIELD_AUTOMATION_IN_PROGRESS);
          const cntAut = _.sumBy(rows, consts.testrail.FIELD_AUTOMATED);
          return {
            x: date,
            y: cntAut / (cntM + cntAiP + cntAut),
          };
        }),
        yAxis: 2,
      },
    ];
  }

  renderChartCmp() {
    const datum = this.getChartDatum();

    return (
      <AutomationChart
        minimise={this.state.minimise}
        height={this.props.chartHeight}
        datum={datum}
      />
    );
  }

  renderToolBarCmp() {
    const showDuplicates = (
      <CheckBox
        label=" Do not show duplicates"
        key="show.duplicates"
        checked={this.state.isHideDuplicates}
        onChange={this.onToggleHideDuplicates}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        ranges={getRanges()}
        onChange={this.onDatesRangeChange}
        startDate={this.state.datesRange[0]}
        endDate={this.state.datesRange[1]}
      />
    );

    const intervalTypeChoices = ['days', 'weeks', 'months'].map(
      v => ({ value: v, selected: v === this.state.intervalType })
    );

    const intervalTypeSelect = (
      <Multiselect
        key="intervalType.select"
        ref={(c) => { this.intervalTypeSelect = c; }}
        onChange={this.onIntervalTypeChange}
        data={intervalTypeChoices}
      />
    );
    const checkBox = (
      <CheckBox
        label=" Minimise chart"
        key="show.minim"
        checked={this.state.minimise}
        onChange={this.onToggleMinimiseChart}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    return super.renderToolBarCmp({
      main: [datePicker, intervalTypeSelect, showDuplicates, checkBox],
    });
  }
}
