import React from 'react';
import _ from 'lodash';
import * as BS from 'react-bootstrap';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import DatesRangeReportBox from './base/DatesRangeReportBox';
import VerificationProgressChart from './D3Charts/VerificationProgressChart';
import consts from '../consts';
import './base/multiChartExtended';
import { toFixedCelRenderer, safeDiv, isEqual, getDefaultDatesRanges, isCurrentWeek } from './base/utils';


const {
    FAKE_STATUS_DEV_TO_VERIFY,
    FAKE_STATUS_OPEN,
    STATUS_FIX_COMMITTED,
    STATUS_FIX_RELEASED,
    IMPORTANCE_CRITICAL,
    IMPORTANCE_HIGH,
    LP_URI,
} = consts.launchpad;

const importanceMap = {
  crit: IMPORTANCE_CRITICAL,
  high: IMPORTANCE_HIGH,
};

const statusColors = {
  FIX_RELEASED: '#69BA69',
  FIX_COMMITTED: '#E06666',
  OPEN: '#5DA4B6',
  DEV_TO_VERIFY: '#8062A0',
};

function getLink(prefix, value, status, milestones) {
  const [bugType, importance] = prefix.split('_');
  let url = 'https://lpreports.infra.mirantis.net/reports/custom_report?project=mos&project=fuel';
  if (milestones.length) {
    url += milestones.map(milestone => `&milestone=${milestone}`).join('');
  }
  if (typeof status !== 'string') {
    url += status.map(item => `&status=${item}`).join('');
  } else {
    url += `&status=${status}`;
  }
  url += `&importance=${importanceMap[importance]}`;
  _.mapValues(LP_URI[bugType], (tags, key) => {
    if (tags.length) {
      url += tags.map(item => `&${key}=${item}`).join('');
    }
  });
  if (prefix.indexOf('dev_to_verify') !== -1) {
    url += '&tags=dev-to-verify';
  }
  return `<a style="color: inherit;" target="_blank" href="${url}">${value}</a>`;
}

function getGroupCols(groupPrefix, milestones) {
  function getSubColumns(prefix) {
    const width = 100;
    return [
      {
        headerName: 'Fix Released',
        field: `${prefix}_fix_released`,
        filter: 'number',
        cellStyle: { color: statusColors.FIX_RELEASED },
        cellRenderer: (params) => {
          if (params.rowIndex === 0 && prefix.indexOf('all') === -1) {
            return getLink(prefix, params.value, 'Fix Released', milestones);
          }
          return params.value;
        },
        width,
      },
      {
        headerName: 'Fix Committed',
        field: `${prefix}_fix_committed`,
        filter: 'number',
        cellStyle: { color: statusColors.FIX_COMMITTED },
        cellRenderer: (params) => {
          if (params.rowIndex === 0 && prefix.indexOf('all') === -1) {
            return getLink(prefix, params.value, 'Fix Committed', milestones);
          }
          return params.value;
        },
        width,
      },
      {
        headerName: 'Open',
        field: `${prefix}_open`,
        filter: 'number',
        cellStyle: { color: statusColors.OPEN },
        cellRenderer: (params) => {
          if (params.rowIndex === 0 && prefix.indexOf('all') === -1) {
            return getLink(
              prefix,
              params.value,
              ['Confirmed', 'Triaged', 'Incomplete', 'In Progress'],
              milestones
            );
          }
          return params.value;
        },
        width,
      },
      {
        headerName: 'Dev-to-verify',
        field: `${prefix}_dev_to_verify`,
        filter: 'number',
        cellStyle: { color: statusColors.DEV_TO_VERIFY },
        cellRenderer: (params) => {
          if (params.rowIndex === 0 && prefix.indexOf('all') === -1) {
            return getLink(
              `${prefix}_dev_to_verify`,
              params.value,
              'Fix Committed',
              milestones
            );
          }
          return params.value;
        },
        width,
      },
      {
        headerName: 'Verification rate',
        field: `${prefix}_verification_rate_pct`,
        valueGetter(params) {
          return 100 * params.data[(`${prefix}_verification_rate`)];
        },
        cellRenderer: toFixedCelRenderer,
        filter: 'number',
        chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
        width: 60,
      },
    ];
  }

  return [
    {
      headerName: 'All',
      headerClass: 'header-center-aligned',
      children: getSubColumns(`${groupPrefix}_all`),
    },
    {
      headerName: 'Crit',
      headerClass: 'header-center-aligned',
      children: getSubColumns(`${groupPrefix}_crit`),
    },
    {
      headerName: 'High',
      headerClass: 'header-center-aligned',
      children: getSubColumns(`${groupPrefix}_high`),
    },
  ];
}

const columnDefs = [
  {
    headerName: 'Date',
    field: 'date',
    cellRenderer: (params) => {
      if (isCurrentWeek(params.data.date)) {
        return 'Current';
      }
      return params.data.date.format('YYYY-MM-DD');
    },
    valueGetter(p) {
      return p.data.date && p.data.date.format('YYYY-MM-DD');
    },
    filter: 'text',
    filterParams: {
      newRowsAction: 'keep',
    },
    suppressSizeToFit: true,
    cellStyle: { 'font-weight': 'bold' },
    width: 120,
    pinned: true,
  },
];

export default class extends DatesRangeReportBox {

  static defaultProps = {
    ...DatesRangeReportBox.defaultProps,
    getDefaultDatesRanges,
    title: 'Launchpad: Bugs Verification Progress',
    csvExportFilename: 'lp_bugs_verification_progress.csv',
    gridDataUrl: consts.launchpad.VERIFICATION_PROGRESS_RESOURCE,
    lastSyncInfoUrl: consts.launchpad.LAST_SYNC_INFO_URL,
    chartHeight: 400,
    defaultSortModel: [{ colId: 'date', sort: 'desc' }],
  };

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedMilestones'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  buildColumns() {
    const columns = [
      {
        headerName: 'Product customer-affecting',
        headerClass: 'header-center-aligned',
        children: getGroupCols('prod', this.state.selectedMilestones),
      },
      {
        headerName: 'Release-related',
        headerClass: 'header-center-aligned',
        children: getGroupCols('rel', this.state.selectedMilestones),
      },
      {
        headerName: 'Internal QA',
        headerClass: 'header-center-aligned',
        children: getGroupCols('intqa', this.state.selectedMilestones),
      },
    ];
    this.setState({
      columnDefs: [...columnDefs, ...columns],
    });
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedMilestones.push($(elem).val());
    });
    this.setStateIf({ selectedMilestones });
    this.buildColumns();
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: [window.commonStore.lp_milestone_in_focus.name],
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  prepareGridData(data) {
    function get_statuses_by(prefix, data) {
      const ret = {
        [`${prefix}_fix_released`]: data.filter({ status: STATUS_FIX_RELEASED }).sumBy('count'),
        [`${prefix}_fix_committed`]: data.filter({ status: STATUS_FIX_COMMITTED }).sumBy('count'),
        [`${prefix}_open`]: data.filter({ status: FAKE_STATUS_OPEN }).sumBy('count'),
        [`${prefix}_dev_to_verify`]: data.filter(
          { status: FAKE_STATUS_DEV_TO_VERIFY }
        ).sumBy('count'),
      };
      ret[`${prefix}_verification_rate`] = safeDiv(
        ret[`${prefix}_fix_released`],
        ret[`${prefix}_fix_released`] + ret[`${prefix}_fix_committed`]
      );
      return ret;
    }

    const preparedData = _(data).groupBy('date').map((values, date) => {
      const prod_bugs = _(values).filter({ is_product: true });
      const rel_bugs = _(values).filter({ is_related: true });
      const intqa_bugs = _(values).filter({ is_internalqa: true });

      const ret = Object.assign(
        {
          date: moment(date),
        },
        get_statuses_by('prod_all', prod_bugs),
        get_statuses_by('rel_all', rel_bugs),
        get_statuses_by('intqa_all', intqa_bugs),
        get_statuses_by('prod_crit', prod_bugs.filter({ importance: IMPORTANCE_CRITICAL })),
        get_statuses_by('rel_crit', rel_bugs.filter({ importance: IMPORTANCE_CRITICAL })),
        get_statuses_by('intqa_crit', intqa_bugs.filter({ importance: IMPORTANCE_CRITICAL })),
        get_statuses_by('prod_high', prod_bugs.filter({ importance: IMPORTANCE_HIGH })),
        get_statuses_by('rel_high', rel_bugs.filter({ importance: IMPORTANCE_HIGH })),
        get_statuses_by('intqa_high', intqa_bugs.filter({ importance: IMPORTANCE_HIGH }))
      );
      return ret;
    }).sortBy('date')
    .value();
    this.buildColumns();
    return preparedData;
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      milestones: this.state.selectedMilestones.join(' '),
    })
      .done((data) => {
        this.updateGridData(data);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  getChartDatum(prefix) {
    const rowData = this.state.rowData;
    if (!rowData) { return []; }

    function getValues(field) {
      return _.map(rowData, row => ({
        x: row.date,
        y: row[`${prefix}_${field}`] || 0,
      }));
    }

    return [
      {
        key: 'Fix Released',
        type: 'bar',
        yAxis: 1,
        values: getValues('fix_released'),
        color: statusColors.FIX_RELEASED,
      },
      {
        key: 'Fix Committed',
        type: 'bar',
        yAxis: 1,
        values: getValues('fix_committed'),
        color: statusColors.FIX_COMMITTED,
      },
      {
        key: 'Open',
        type: 'bar',
        yAxis: 1,
        values: getValues('open'),
        color: statusColors.OPEN,
      },
      {
        key: 'Dev-to-verify',
        type: 'bar',
        yAxis: 1,
        values: getValues('dev_to_verify'),
        color: statusColors.DEV_TO_VERIFY,
      },
      {
        key: '% Verification rate',
        color: '#444141',
        type: 'line',
        values: getValues('verification_rate'),
        yAxis: 2,
      },
    ];
  }

  renderOneChartCmp(prefix, title) {
    const datum = this.getChartDatum(prefix);

    return (
        <VerificationProgressChart
          title={title}
          height={this.props.chartHeight}
          datum={datum}
        />
    );
  }

  renderChartCmp() {
    return (
      <div>
        <BS.Row>
          <BS.Col md={4}>{this.renderOneChartCmp('prod_all', 'Product All')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('prod_crit', 'Product Crit')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('prod_high', 'Product High')}</BS.Col>
        </BS.Row>
        <BS.Row>
          <BS.Col md={4}>{this.renderOneChartCmp('rel_all', 'Release-related All')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('rel_crit', 'Release-related Crit')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('rel_high', 'Release-related High')}</BS.Col>
        </BS.Row>
        <BS.Row>
          <BS.Col md={4}>{this.renderOneChartCmp('intqa_all', 'Internal QA All')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('intqa_crit', 'Internal QA Crit')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('intqa_high', 'Internal QA High')}</BS.Col>
        </BS.Row>
      </div>
    );
  }

  renderToolBarCmp() {
    const datePicker = this.renderDatesRangePicker();

    const milestoneData = _.map(commonStore.launchpad_milestones, v => ({
      value: v, selected: _.includes(this.state.selectedMilestones, v),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        multiple
        buttonWidth="250px"
        maxHeight={300}
        includeSelectAllOption
        enableFiltering
        nonSelectedText="select milestone"
        allSelectedText="all selected"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );

    const resizeBnt = (
      <BS.Button
        key="resize.columns"
        bsStyle="link"
        bsSize="xsmall"
        onClick={() => {
          this.columnApi.autoSizeColumns(
            _.map(this.columnApi.getAllColumns().slice(1), 'colDef.field')
          );
        }}
      >
        Auto size columns
      </BS.Button>
    );

    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect, resizeBnt] });
  }

}
