import React from 'react';
import _ from 'lodash';
import * as BS from 'react-bootstrap';
import moment from 'moment';
import $ from 'jquery';
import Multiselect from 'react-bootstrap-multiselect';

import DatesRangeReportBox from './base/DatesRangeReportBox';
import ProgressChartHighCrit from './D3Charts/ProgressChartHighCrit';
import './base/multiChartExtended';
import { toFixedCelRenderer, safeDiv, isEqual, getDefaultDatesRanges, isCurrentWeek } from './base/utils';
import consts from '../consts';

const GRID_DATA_URL = '/api/v1.0/lp/verification_progress_teams';
const FAKE_STATUS_OPEN = 'Open';
const STATUS_FIX_COMMITTED = 'Fix Committed';
const STATUS_FIX_RELEASED = 'Fix Released';

const VERIFICATION_TYPE = {
  all: 'All bugs',
  prod: 'Product customer-affecting bugs',
  rel: 'Release-related issues',
  inter: 'Internal QA issues',
};

const statusColors = {
  FIX_RELEASED: '#69BA69',
  FIX_COMMITTED: '#E06666',
  OPEN: '#5DA4B6',
  DEV_TO_VERIFY: '#8062A0',
};


function getGroupTeamCols() {
  const width = 100;
  const getSubColumns = (teamPrefix, type) => {
    return [
      {
        headerName: 'Bugs to verify',
        field: `${teamPrefix}_fix_committed_${type}`,
        width,
      },
      {
        headerName: 'Bugs verified',
        field: `${teamPrefix}_fix_released_${type}`,
        width,
      },
      {
        headerName: 'Open',
        field: `${teamPrefix}_open_${type}`,
        width,
      },
      {
        headerName: 'Verification rate',
        field: `${teamPrefix}_verification_rate_${type}`,
        valueGetter: params => 100 * params.data[`${teamPrefix}_verification_rate_${type}`],
        cellRenderer: toFixedCelRenderer,
        filter: 'number',
        chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
        width: 60,
      },
    ];
  }
  const groupStatus = (teamPrefix) => {
    return [
      {
        headerName: 'All',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'all'),
      },
      {
        headerName: 'Crit',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'crit'),
      },
      {
        headerName: 'High',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'high'),
      },
    ];
  }

  return _.map(_.uniq(_.values(consts.launchpad.LAUNCHPAD_QATEAM_MAP)), v => ({
    headerName: v,
    children: groupStatus(_.snakeCase(v)),
    headerClass: 'header-center-aligned',
  }));
}

const getColumnDefs = () => [
  {
    headerName: 'Date',
    field: 'date',
    cellRenderer: (params) => {
      if (isCurrentWeek(params.data.date)) {
        return 'Current';
      }
      return params.data.date.format('YYYY-MM-DD');
    },
    valueGetter: params =>
      params.data.date && params.data.date.format('YYYY-MM-DD'),
    filter: 'text',
    filterParams: {
      newRowsAction: 'keep',
    },
    cellStyle: { 'font-weight': 'bold' },
    width: 100,
    pinned: true,
  },
  {
    headerName: 'Teams',
    children: getGroupTeamCols(),
    headerClass: 'header-center-aligned',
  },
];

export default class extends DatesRangeReportBox {

  static defaultProps = {
    ...DatesRangeReportBox.defaultProps,
    columnDefs: getColumnDefs,
    getDefaultDatesRanges,
    title: 'LaunchPad: Bugs Verification Progress - Teams',
    csvExportFilename: 'lp_bugs_verification_progress_teams.csv',
    gridDataUrl: GRID_DATA_URL,
    chartHeight: 400,
    lastSyncInfoUrl: '/api/v1.0/lp/sync/last',
    milestonesListUrl: '/api/v1.0/lp/milestones',
    defaultSortModel: [{ colId: 'date', sort: 'desc' }],
    agGridOptions: {
      ...DatesRangeReportBox.defaultProps.agGridOptions,
    },
  };

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;
    this.state.selectedVerificationType = this.initialConfiguration.selectedVerificationType;

    this.handleVerificationSelection = this.handleVerificationSelection.bind(this);
    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedMilestones', 'selectedVerificationType'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: [window.commonStore.lp_milestone_in_focus.name],
      selectedVerificationType: _.head(_.keys(VERIFICATION_TYPE)),
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
      selectedVerificationType: this.state.selectedVerificationType,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
    this.setStateIf({ selectedVerificationType: conf.selectedVerificationType });
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedMilestones.push($(elem).val());
    });
    this.setStateIf({ selectedMilestones });
  }

  handleVerificationSelection() {
    let selectedVerificationType = null;
    this.verificationSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedVerificationType = $(elem).val();
    });
    this.setStateIf({ selectedVerificationType });
  }

  getGridTitle() {
    return VERIFICATION_TYPE[this.state.selectedVerificationType];
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      verf_status: this.state.selectedVerificationType,
      milestones: this.state.selectedMilestones.join(' '),
    })
    .done(data => this.updateGridData(data))
    .always(() => this.gridApi.hideOverlay());
  }

  prepareGridData(data) {
    function getStats(teamPrefix, teamData) {
      const result = {};
      [STATUS_FIX_COMMITTED, STATUS_FIX_RELEASED, FAKE_STATUS_OPEN].forEach((status) => {
        const statusData = _(teamData).filter({ status });

        const allSum = _(statusData).sumBy('count');
        const hightSum = _(statusData).filter({ importance: 'High' }).sumBy('count');
        const criticalSum = _(statusData).filter({ importance: 'Critical' }).sumBy('count');

        result[`${teamPrefix}_${_.snakeCase(status)}_all`] = allSum;
        result[`${teamPrefix}_${_.snakeCase(status)}_crit`] = criticalSum;
        result[`${teamPrefix}_${_.snakeCase(status)}_high`] = hightSum;
      });
      result[`${teamPrefix}_verification_rate_all`] = safeDiv(
        result[`${teamPrefix}_fix_released_all`],
        result[`${teamPrefix}_fix_released_all`] + result[`${teamPrefix}_fix_committed_all`]
        );
      result[`${teamPrefix}_verification_rate_crit`] = safeDiv(
        result[`${teamPrefix}_fix_released_crit`],
        result[`${teamPrefix}_fix_released_crit`] + result[`${teamPrefix}_fix_committed_crit`]
        );
      result[`${teamPrefix}_verification_rate_high`] = safeDiv(
        result[`${teamPrefix}_fix_released_high`],
        result[`${teamPrefix}_fix_released_high`] + result[`${teamPrefix}_fix_committed_high`]
      );
      return result;
    }

    const prepareGridData = _(data).groupBy('date').map((values, date) => {
      const teamsData = _(values).groupBy('team_id').map((teamVal, id) => {
        const teamPrefix = _.snakeCase(consts.launchpad.LAUNCHPAD_QATEAM_MAP[id]);
        return getStats(teamPrefix, teamVal);
      }).value();
      const result = Object.assign(
        {
          date: moment(date),
        },
        ...teamsData,
      );
      return result;
    }).sortBy('date')
      .value();
    return prepareGridData;
  }

  getChartDatum(prefix) {
    const rowData = this.state.rowData;
    if (!rowData) { return []; }

    const getValues = (field, type) =>
      _.map(rowData, row => ({ x: row.date, y: row[`${prefix}_${field}_${type}`] || 0 }));
    const res = [];
    ['all', 'crit', 'high'].forEach((type) => {
      res.push(
        {
          key: 'Bugs verified',
          type: 'bar',
          yAxis: 1,
          groupName: type,
          values: getValues('fix_released', type),
          color: statusColors.FIX_RELEASED,
        },
        {
          key: 'Bugs to verify',
          type: 'bar',
          yAxis: 1,
          groupName: type,
          values: getValues('fix_committed', type),
          color: statusColors.FIX_COMMITTED,
        },
        {
          key: 'Open',
          type: 'bar',
          yAxis: 1,
          groupName: type,
          values: getValues('open', type),
          color: statusColors.OPEN,
        },
        {
          key: '% Verification rate',
          color: '#444141',
          type: 'line',
          groupName: type,
          values: getValues('verification_rate', type),
          yAxis: 2,
        },
      );
    });
    return res;
  }

  renderToolBarCmp() {
    const datePicker = this.renderDatesRangePicker();

    const milestoneData = _.map(window.commonStore.launchpad_milestones, v => ({
      value: v, selected: _.includes(this.state.selectedMilestones, v),
    }));

    const typeVerificationData = _.mapValues(VERIFICATION_TYPE, (v, k) => ({
      label: v, value: k, selected: _.includes(this.state.selectedVerificationType, k),
    }));


    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        multiple
        buttonWidth="250px"
        maxHeight={300}
        includeSelectAllOption
        enableFiltering
        nonSelectedText="select milestone"
        allSelectedText="all selected"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );
    const typeVerificationSelect = (
      <Multiselect
        key="verification.type.select"
        ref={(c) => { this.verificationSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        onDropdownHide={this.handleVerificationSelection}
        data={typeVerificationData}
      />
  );

    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect, typeVerificationSelect] });
  }

  renderOneChartCmp(prefix, title) {
    const datum = this.getChartDatum(prefix);

    return (
      <ProgressChartHighCrit
        title={title}
        height={this.props.chartHeight}
        datum={datum}
      />
    );
  }

  renderChartCmp() {
    return (
      <div>
        <BS.Row>
          <BS.Col md={6}>{this.renderOneChartCmp('fuel_qa', 'Fuel QA')}</BS.Col>
          <BS.Col md={6}>{this.renderOneChartCmp('mos_qa', 'MOS QA')}</BS.Col>
        </BS.Row>
        <BS.Row>
          <BS.Col md={6}>{this.renderOneChartCmp('pce_qa', 'PCE QA')}</BS.Col>
          <BS.Col md={6}>{this.renderOneChartCmp('telco_qa', 'Telco QA')}</BS.Col>
        </BS.Row>
        <BS.Row>
          <BS.Col md={6}>{this.renderOneChartCmp('performance_qa', 'Performance QA')}</BS.Col>
          <BS.Col md={6} />
        </BS.Row>
        <BS.Row>
          <BS.Col md={6}>{this.renderOneChartCmp('devs', 'Devs')}</BS.Col>
          <BS.Col md={6}>{this.renderOneChartCmp('infra', 'Infra')}</BS.Col>
        </BS.Row>
        <BS.Row>
          <BS.Col md={6}>{this.renderOneChartCmp('maintenance_qa', 'Maintenance QA')}</BS.Col>
          <BS.Col md={6} />
        </BS.Row>
      </div>
    );
  }
}
