import React from 'react';
import _ from 'lodash';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';
import d3 from 'd3';

import ReportBox from './base/ReportBox';
import BvtAndSwarmChart from './D3Charts/BvtAndSwarmChart';
import SimpleRangePicker from '../components/SimpleRangePicker';
import CheckBox from '../components/CheckBox';
import './base/multiChartExtended';
import { getComponentsSum, toFixedCelRenderer, safeDiv, isEqual } from './base/utils';

let RANGE = [];


const GRID_DATA_URL = '/api/v1.0/testrail/swarm_metrics';


export const columnDefs = [
  {
    headerName: 'Date',
    field: 'date',
    cellStyle: { 'font-weight': 'bold' },
    width: 100,
    pinned: true,
  },
  {
    headerName: 'Name',
    field: 'milestone_name',
    width: 150,
  },
  {
    headerName: 'Total',
    field: 'cnt_total',
    aggFunc: 'sum',
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    width: 80,
    chart: { nonStackable: true },
  },
  {
    headerName: 'Success Rate',
    field: 'success_rate',
    valueGetter(params) {
      const total = params.getValue('cnt_total');
      return 100 * safeDiv(params.getValue('cnt_passed'), total);
    },
    cellRenderer: toFixedCelRenderer,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    width: 80,
  },
  {
    headerName: 'Passed',
    field: 'cnt_passed',
    width: 100,
  },
  {
    headerName: 'Skipped',
    field: 'cnt_skipped',
    width: 100,
  },
  {
    headerName: 'Blocked',
    field: 'cnt_blocked',
    width: 100,
  },
  {
    headerName: 'In progress',
    field: 'cnt_in_progress',
    width: 100,
  },
  {
    headerName: 'Fixed',
    field: 'cnt_fixed',
    width: 100,
  },
  {
    headerName: 'Regression',
    field: 'cnt_regression',
    width: 100,
  },
  {
    headerName: 'Untested',
    field: 'cnt_untested',
    width: 100,
  },
  {
    headerName: 'Other',
    field: 'cnt_other',
    width: 100,
  },
  {
    headerName: 'Failed',
    children: [
      {
        headerName: 'InfraFailed',
        field: 'cnt_infra_failed',
        columnGroupShow: 'open',
        width: 100,
      },
      {
        headerName: 'TestFailed',
        field: 'cnt_test_failed',
        columnGroupShow: 'open',
        width: 100,
      },
      {
        headerName: 'ProdFailed',
        field: 'cnt_prod_failed',
        columnGroupShow: 'open',
        width: 100,
      },
      {
        headerName: 'Failed',
        field: 'cnt_failed',
        columnGroupShow: 'open',
        width: 100,
      },
      {
        headerName: 'TotalFailed',
        aggFunc: 'sum',
        filter: 'number',
        valueGetter: params => getComponentsSum([
          'cnt_infra_failed',
          'cnt_test_failed',
          'cnt_prod_failed',
          'cnt_failed',
        ], params),
        columnGroupShow: 'closed',
      },
    ],
  },
];


export default class BvtAndSwarmMetrics extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs,
    title: 'BVT and SWARM',
    csvExportFilename: 'testrail_bvt_smart_metrics.csv',
    gridDataUrl: GRID_DATA_URL,
    chartHeight: 800,
    lastSyncInfoUrl: '/api/v1.0/testrail/sync/last',
    milestonesListUrl: '/api/v1.0/testrail/milestone_names',
    defaultSortModel: [{ colId: 'date', sort: 'desc' }],
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
    },
  };

  constructor(props) {
    super(props);
    this.state.datesRange = [
      moment().subtract(7, 'days').startOf('day'),
      moment().endOf('day'),
    ];
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;
    this.state.minimise = false;

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
    this.onToggleMinimiseChart = this.onToggleMinimiseChart.bind(this);
  }

  componentDidMount() {
    this.fetchInitialGridData();
  }

  componentDidUpdate(prevProps, prevState) {
    if (!isEqual(this.state.selectedMilestones, prevState.selectedMilestones)) {
      this.fetchInitialGridData();
    }
    if (!isEqual(this.state.datesRange, prevState.datesRange)) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      selectedMilestones: [window.commonStore.testrail_milestones.filter(
        m => m.name === window.commonStore.tr_milestone_in_focus.name
      )[0].id],
    };
  }

  getConfiguration() {
    return {
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  onToggleMinimiseChart(e) {
    this.setState({ minimise: e.target.checked });
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each(
      (i, elem) => selectedMilestones.push($(elem).val()));
    this.setStateIf({ selectedMilestones });
  }

  prepareGridData(data) {
    super.prepareGridData(data);
    _.each(data, (run) => {
      Object.assign(run, _.mapKeys(run.status_counts, (v, k) => `cnt_${k}`));
      run.cnt_total = _(run.status_counts).values().sum() || 0;
      run.date = moment(run.date).format('YYYY-MM-DD');
    });
    return data;
  }

  fetchInitialGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get('/api/v1.0/testrail/swarm_recent_seven', {
      milestone_id: this.state.selectedMilestones.join(' '),
    })
      .done((data) => {
        this.updateGridData(data);
        this.setDateRange(data);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  setDateRange(data) {
    RANGE = d3.extent(data, d => d.date).map(d => moment(d));
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      milestone_id: this.state.selectedMilestones.join(' '),
    })
      .done((date) => {
        this.updateGridData(date);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  onDatesRangeChange(newRange) {
    this.setStateIf({ datesRange: newRange });
  }

  getChartDatum() {
    const rowData = this.state.rowData;
    if (_.isEmpty(rowData)) { return []; }

    const result = [];
    const model = this.gridApi.getModel();
    model.forEachNodeAfterFilterAndSort((node) => {
      result.push(node.data);
    });

    return result;
  }


  renderChartCmp() {
    const datum = this.getChartDatum();
    return (
      <div>
        <BvtAndSwarmChart
          height={this.props.chartHeight}
          minimise={this.state.minimise}
          datum={datum}
        />
      </div>
    );
  }

  renderToolBarCmp() {
    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        onChange={this.onDatesRangeChange}
        startDate={RANGE[0]}
        endDate={RANGE[1]}
      />
    );

    const milestoneData = window.commonStore.testrail_milestones.map(m => ({
      value: m.id,
      label: m.name,
      selected: _.includes(
        this.state.selectedMilestones,
        _.isString(this.state.selectedMilestones[0]) ? '' + m.id : m.id),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        nonSelectedText="select milestone"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );
    const checkBox = (
      <CheckBox
        label=" Minimise chart"
        key="show.minim"
        checked={this.state.minimise}
        onChange={this.onToggleMinimiseChart}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect, checkBox] });
  }
}
