import React from 'react';
import _ from 'lodash';
import d3 from 'd3';
import moment from 'moment';

const oneDayWidth = 127;
const tableRowHeight = 21;
const tableMainRowHeight = 23;

const fontSize = 20;

function getMinimiseChartWidth(countDays, maxWidth) {
  const result = oneDayWidth * countDays;
  return result >= maxWidth ? maxWidth : result + 500;
}

function getPercentFontSize(isMin) {
  if (isMin) {
    return `${fontSize - 2}px`;
  }
  return `${fontSize + 5}px`;
}

function getTableLegend(key, adoption) {
  const adoptionMap = {
    'Manual': 'Not Adopted',
    'Automation in Progress': 'Adoption in Progress',
    'Automated': 'Adopted',
    'Will not Automate': 'n/a',
  };
  if (adoption) {
    return adoptionMap[key];
  }
  return key;
}


export default class AutomationChart extends React.Component {

  static defaultProps = {
    margin: { top: 30, right: 50, bottom: 155, left: 255 },
  }

  static propTypes = {
    datum: React.PropTypes.arrayOf(React.PropTypes.object),
    height: React.PropTypes.number,
    minimise: React.PropTypes.bool,
    adoption: React.PropTypes.bool,
    integration: React.PropTypes.bool,
    margin: React.PropTypes.shape({
      top: React.PropTypes.number,
      right: React.PropTypes.number,
      left: React.PropTypes.number,
      bottom: React.PropTypes.number,
    }),
  };

  constructor(props) {
    super(props);
    this.renderChart = this.renderChart.bind(this);
    this.handleResize = this.handleResize.bind(this);
  }

  componentDidMount() {
    window.addEventListener('resize', this.handleResize);
  }

  shouldComponentUpdate(nextProps) {
    return this.props.datum !== nextProps.datum || this.props.minimise !== nextProps.minimise;
  }

  componentDidUpdate() {
    this.renderChart();
  }

  componentWillUnmount() {
    window.removeEventListener('resize', this.handleResize);
  }

  handleResize() {
    this.renderChart();
  }

  renderChart() {
    if (!this.props.datum.length) {
      const noData = d3.select(this.chart);
      return noData.append('text').text('No Data');
    }

    const myNode = this.chart;
    while (myNode.firstChild) {
      myNode.removeChild(myNode.firstChild);
    }

    const svg = d3.select(this.chart);
    const margin = this.props.margin;
    const maxWidth = d3.select(this.chartRoot).node().getBoundingClientRect().width;
    let width = null;
    if (!this.props.minimise) {
      width = maxWidth - margin.left - margin.right;
    } else {
      width = getMinimiseChartWidth(this.props.datum[0].values.length, maxWidth)
        - margin.left
        - margin.right;
    }
    const height = this.props.height - margin.top - margin.bottom;
    const g = svg.append('g')
      .attr('transform', `translate(${margin.left},${margin.top})`);

    const x = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const x2 = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const xAxis = d3.svg.axis()
      .scale(x)
      .orient('bottom')
      .tickFormat(d3.time.format('%d-%b'))
      .tickSize(1);

    const y = d3.scale.linear()
      .rangeRound([height, 0]);

    const y2 = d3.scale.linear()
      .rangeRound([height, 1]);

    const yAxis = d3.svg.axis()
    .scale(y)
    .orient('left')
    .tickSize(1);

    const yAxis2 = d3.svg.axis()
    .scale(y2)
    .orient('right')
    .tickFormat(d3.format('.1%'))
    .tickSize(1);

    const [lineDataSet, stackedData] = _.partition(this.props.datum, obj => obj.type === 'line');

    const layers = d3.layout.stack()(
      stackedData.map(layer =>
        layer.values.map(layerData => ({
          x: layerData.x.toDate(),
          y: layerData.y,
          name: _.snakeCase(layer.key),
          color: layer.color,
        })),
      ),
    );

    const lineData = lineDataSet[0].values.map(
      data => (
        {
          date: moment(data.x).toDate(),
          percentage: data.y || 0,
        }
      ),
    );

    if (!layers[0].length) { return null; }

    x.domain(layers[0].map(d => d.x));
    x2.domain(layers[0].map(d => d.x));
    y.domain([0, d3.max(layers[layers.length - 1], d => d.y0 + d.y)]).nice();
    y2.domain([0, 1]);

    const valueline = d3.svg.line()
        .interpolate('monotone')
        .x(d => (x2(d.date) + (x2.rangeBand() / 2)) - 20)
        .y(d => y2(d.percentage));

    g.selectAll('.layer')
        .data(layers)
        .enter()
        .append('g')
        .attr('class', 'layer')
        .selectAll('rect')
        .data(d => d)
        .enter()
        .append('rect')
        .style('fill', d => d.color)
        .attr('x', d => x(d.x))
        .attr('y', d => y(d.y + d.y0))
        .attr('height', d => y(d.y0) - y(d.y + d.y0))
        .attr('width', x.rangeBand() - 50);

    svg.append('g')
        .attr('class', 'axis axis--x')
        .attr('transform', `translate(${margin.left - 20},${height + margin.top})`)
        .style('font-size', `${fontSize}px`)
        .call(xAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr('transform', `translate(${margin.left - 20}, ${margin.top})`)
        .call(yAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr(
          'transform',
          `translate(${(x.rangeBand() * lineData.length) + margin.left}, ${margin.top})`,
        )
        .call(yAxis2);

    const lineGroup = g.selectAll('.line-group')
        .data([1])
        .enter().append('g')
        .attr('class', 'line-group');

    lineGroup.append('path')
        .attr('class', 'line')
        .style('fill', 'none')
        .style('stroke', 'black')
        .style('stroke-width', '2.5px')
        .attr('d', valueline(lineData));

    const point = lineGroup.append('g')
        .attr('class', 'line-point');

    point.selectAll('circle')
        .data(lineData)
        .enter().append('circle')
        .attr('cx', d => (x2(d.date) + (x2.rangeBand() / 2)) - 20)
        .attr('cy', d => y2(d.percentage))
        .attr('r', 5);

    const pointText = lineGroup.append('g')
        .attr('class', 'pointText');

    pointText.selectAll('text')
        .data(lineData)
        .enter().append('text')
        .attr('x', d => (x2(d.date) + (x2.rangeBand() / 2)) - 20)
        .attr('y', d => y2(d.percentage))
        .attr('dy', -10)
        .attr('text-anchor', 'middle')
        .style('font-size', getPercentFontSize(this.props.minimise))
        .style('font-weight', 'bold')
        .text(d => `${(d.percentage * 100).toFixed(1)}%`);

    if (this.props.integration) {
      const lineDataIntegr = lineDataSet[1].values.map(
        data => ({
          date: moment(data.x).toDate(),
          percentage: data.y || 0,
        }),
      );
      const lineGroupIntegr = g.selectAll('.line-group-integr')
          .data([1])
          .enter().append('g')
          .attr('class', 'line-group-integr');

      lineGroupIntegr.append('path')
          .attr('class', 'line')
          .style('fill', 'none')
          .style('stroke', '#4141f3')
          .style('stroke-width', '2.5px')
          .attr('d', valueline(lineDataIntegr));

      const pointInteg = lineGroupIntegr.append('g')
          .attr('class', 'line-group-integr');

      pointInteg.selectAll('circle')
          .data(lineDataIntegr)
          .enter().append('circle')
          .attr('cx', d => (x2(d.date) + (x2.rangeBand() / 2)) - 20)
          .attr('cy', d => y2(d.percentage))
          .attr('fill', '#4141f3')
          .attr('r', 5);

      const pointTextIntegr = lineGroupIntegr.append('g')
          .attr('class', 'pointText');

      pointTextIntegr.selectAll('text')
          .data(lineDataIntegr)
          .enter().append('text')
          .attr('x', d => (x2(d.date) + (x2.rangeBand() / 2)) - 20)
          .attr('y', d => y2(d.percentage))
          .attr('dy', -10)
          .attr('text-anchor', 'middle')
          .style('font-size', getPercentFontSize(this.props.minimise))
          .style('font-weight', 'bold')
          .text(d => `${(d.percentage * 100).toFixed(1)}%`);
    }

// Table
    const ticks = svg.selectAll('.axis--x .tick');

    ticks.append('rect')
        .attr('x', 0 - (x.rangeBand() / 2))
        .attr('y', 0)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    const rowCount = stackedData.length + lineDataSet.length + 1;

    const rowPositionY = _.range(
      tableMainRowHeight,
      (tableMainRowHeight - tableRowHeight) + (rowCount * tableRowHeight) + 1,
      tableRowHeight,
    );

    ticks.selectAll('rect')
        .data(rowPositionY, d => d)
        .enter()
        .append('rect')
        .attr('x', 0 - (x.rangeBand() / 2))
        .attr('y', d => d)
        .attr('height', tableRowHeight)
        .attr('class', (d, i) => i)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    let percentageLegend = '% Automated';
    let percentageIntegration = [];
    if (this.props.adoption) {
      percentageLegend = '% Adoption';
    }
    if (this.props.integration) {
      percentageIntegration = ['% CI-Integration'];
    }

    const legendLabel = [
      'All TCs',
      ...stackedData.map(layer => layer.key),
      percentageLegend,
      ...percentageIntegration,
    ];

    const legend = svg.append('g')
      .attr('class', 'legend')
      .attr('transform', `translate(0,${height + margin.top + tableMainRowHeight})`);

    const legendPositionY = _.range(
      0,
      0 + ((rowCount - 1) * tableRowHeight) + 1,
      tableRowHeight,
    );

    legend.selectAll('rect')
      .data(legendPositionY, d => d)
      .enter()
      .append('rect')
      .attr('x', 0)
      .attr('y', d => d)
      .attr('height', tableRowHeight)
      .attr('class', (d, i) => i)
      .attr('width', margin.left - 18)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    legendLabel.forEach((val, index) => {
      let values = [];
      let valuesObjIndex = null;
      valuesObjIndex = _.findIndex(stackedData, v => v.key.indexOf(val) !== -1);

      if (val !== 'All TCs' && val !== percentageLegend && val !== _.head(percentageIntegration)) {
        legend
          .append('text')
          .attr('x', 25)
          .attr('y', legendPositionY[index] + 17)
          .style('font-size', `${fontSize}px`)
          .text(getTableLegend(val, this.props.adoption));
      } else if (val === percentageLegend || val === _.head(percentageIntegration)) {
        valuesObjIndex = -2;
        legend
          .append('text')
          .attr('x', 25)
          .attr('y', legendPositionY[index] + 17)
          .style('font-size', `${fontSize}px`)
          .text(val);
      } else {
        legend
          .append('text')
          .attr('x', 25)
          .attr('y', legendPositionY[index] + 17)
          .style('font-size', `${fontSize}px`)
          .text(val);
      }


      if (valuesObjIndex > -1) {
        values = stackedData[valuesObjIndex].values;
        const color = stackedData[valuesObjIndex].color;

        legend
          .append('rect')
          .attr('x', 5)
          .attr('y', legendPositionY[index] + 5)
          .attr('width', 15)
          .attr('height', 10)
          .style('fill', color);

        values.forEach((value, valuePosition) => {
          d3.select(`.axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - (x.rangeBand() / 2))
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 17)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(value.y);
        });
      } else if (valuesObjIndex === -2) {
        const lineIndex = _.findIndex(lineDataSet, v => v.key.indexOf(val) !== -1);
        values = lineDataSet[lineIndex].values;
        values.forEach((value, valuePosition) => {
          d3.select(`.axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - (x.rangeBand() / 2))
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 17)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(`${((value.y || 0) * 100).toFixed(2)}%`);
        });
      } else {
        values = layers[0].map(d => d.x);
        values.forEach((value, valuePosition) => {
          d3.select(`.axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - (x.rangeBand() / 2))
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 17)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(
              layers[layers.length - 1][valuePosition].y
              + layers[layers.length - 1][valuePosition].y0,
              );
        });
      }
    });
    return true;
  }

  render() {
    if (!this.props.datum.length) {
      return <div />;
    }
    return (
      <div ref={(c) => { this.chartRoot = c; }} style={{ textAlign: 'center' }}>
        <svg
          ref={(c) => { this.chart = c; }}
          height={this.props.height}
          width={this.props.minimise ? getMinimiseChartWidth(_.get(this.props.datum, '0.values').length) : '100%'}
        />
      </div>
    );
  }
}
