import React from 'react';
import _ from 'lodash';
import d3 from 'd3';

import { safeDiv } from '../base/utils';
import consts from '../../consts';

const oneDayWidth = 127;
const tableRowHeight = 21;
const tableMainRowHeight = 60;

const donutWidth = 95;
const donutHeight = 95;
const radius = donutWidth / 2;
const donutInnerWidth = 20;

const fontSize = 20;

function getMinimiseChartWidth(countDays, maxWidth) {
  const result = oneDayWidth * countDays;
  return result >= maxWidth ? maxWidth : result;
}

function getPercentFontSize(isMin) {
  if (isMin) {
    return `${fontSize - 2}px`;
  }
  return `${fontSize + 5}px`;
}


export default class BvtAndSwarmChart extends React.Component {

  static propTypes = {
    datum: React.PropTypes.arrayOf(React.PropTypes.object),
    height: React.PropTypes.number,
    minimise: React.PropTypes.bool,
  };

  constructor(props) {
    super(props);
    this.renderChart = this.renderChart.bind(this);
    this.handleResize = this.handleResize.bind(this);
  }

  componentDidMount() {
    window.addEventListener('resize', this.handleResize);
  }

  shouldComponentUpdate(nextProps) {
    return this.props.datum !== nextProps.datum || this.props.minimise !== nextProps.minimise;
  }

  componentDidUpdate() {
    this.renderChart();
  }

  componentWillUnmount() {
    window.removeEventListener('resize', this.handleResize);
  }

  handleResize() {
    this.renderChart();
  }

  renderChart() {
    if (!this.props.datum.length) {
      const noData = d3.select(this.chart);
      return noData.append('text').text('No Data');
    }

    const myNode = this.chart;
    while (myNode.firstChild) {
      myNode.removeChild(myNode.firstChild);
    }

    const parseDate = d3.time.format('%Y-%m-%d').parse;

    const svg = d3.select(this.chart);
    const margin = { top: 150, right: 50, bottom: 334, left: 150 };
    const maxWidth = d3.select(this.chartRoot).node().getBoundingClientRect().width;
    let width = null;
    if (!this.props.minimise) {
      width = maxWidth - margin.left - margin.right;
    } else {
      width = getMinimiseChartWidth(this.props.datum.length, maxWidth) - margin.left - margin.right;
    }
    const height = this.props.height - margin.top - margin.bottom;
    const g = svg.append('g')
      .attr('transform', `translate(${margin.left},${margin.top})`);

    const x = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const x2 = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const xAxis = d3.svg.axis()
      .scale(x)
      .orient('bottom')
      .tickFormat(d3.time.format('%d-%b'))
      .tickSize(1);

    const y = d3.scale.linear()
      .rangeRound([height, 0]);

    const y2 = d3.scale.linear()
      .rangeRound([height, 1]);

    const yAxis = d3.svg.axis()
    .scale(y)
    .orient('left')
    .tickSize(1);

    const yAxis2 = d3.svg.axis()
    .scale(y2)
    .orient('right')
    .tickFormat(d3.format('.1%'))
    .tickSize(1);

    const z = d3.scale.ordinal()
      .domain(
        ['blocked', 'failed', 'fixed', 'in_progress',
        'infra_failed', 'other', 'passed', 'prod_failed',
        'regression', 'skipped', 'test_failed', 'untested']
      )
      .range(
        ['#8497B0', '#E6B9B8', '#548235', '#4472C4',
        '#FF3399', '#806000', '#92D050', '#FF0000',
        '#1F4E78', '#984807', '#FF7C80', '#595959']
      );


    const sortedData = _.sortBy(this.props.datum, 'date');

    const layers = d3.layout.stack()(consts.testrail.StatusOrder
      .map(status =>
        sortedData.map(d => ({
          x: parseDate(d.date),
          y: d.status_counts[status],
          name: status })
         )
       ));

    const lineData = sortedData.map(
      data => (
        {
          date: parseDate(data.date),
          percentage: safeDiv(data.status_counts.passed, _(data.status_counts).values().sum()),
        }
      )
    );

    const donutData = sortedData.map(
      data => (
        {
          date: parseDate(data.date),
          bvtStat: data.bvt_counts,
        }
      )
    );

    x.domain(layers[0].map(d => d.x));
    x2.domain(layers[0].map(d => d.x));
    y.domain([0, d3.max(layers[layers.length - 1], d => d.y0 + d.y)]).nice();
    y2.domain([0, 1]);

    const valueline = d3.svg.line()
        .interpolate('monotone')
        .x(d => (x2(d.date) + x2.rangeBand() / 2) - 20)
        .y(d => y2(d.percentage));

    g.selectAll('.layer')
        .data(layers)
        .enter().append('g')
        .attr('class', 'layer')
        .selectAll('rect')
        .data(d => d)
        .enter().append('rect')
        .style('fill', d => z(d.name))
        .attr('x', d => x(d.x))
        .attr('y', d => y(d.y + d.y0))
        .attr('height', d => y(d.y0) - y(d.y + d.y0))
        .attr('width', x.rangeBand() - 50);

    const names = _.map(sortedData, 'milestone_name');

    svg.append('g')
        .attr('class', 'axis axis--x')
        .attr('transform', `translate(${margin.left - 20},${height + margin.top})`)
        .style('font-size', `${fontSize}px`)
        .call(xAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr('transform', `translate(${margin.left - 20}, ${margin.top})`)
        .call(yAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr(
          'transform',
          `translate(${x.rangeBand() * lineData.length + margin.left}, ${margin.top})`
        )
        .call(yAxis2);

    const lineGroup = g.selectAll('.line-group')
        .data([1])
        .enter().append('g')
        .attr('class', 'line-group');

    lineGroup.append('path')
        .attr('class', 'line')
        .style('fill', 'none')
        .style('stroke', 'black')
        .style('stroke-width', '2.5px')
        .attr('d', valueline(lineData));

    const point = lineGroup.append('g')
        .attr('class', 'line-point');

    point.selectAll('circle')
        .data(lineData)
        .enter().append('circle')
        .attr('cx', d => (x2(d.date) + x2.rangeBand() / 2) - 20)
        .attr('cy', d => y2(d.percentage))
        .attr('r', 5);

    const pointText = lineGroup.append('g')
        .attr('class', 'valori');

    pointText.selectAll('text')
        .data(lineData)
        .enter().append('text')
        .attr('x', d => (x2(d.date) + x2.rangeBand() / 2) - 20)
        .attr('y', d => y2(d.percentage))
        .attr('dy', -10)
        .attr('text-anchor', 'middle')
        .style('font-size', getPercentFontSize(this.props.minimise))
        .style('font-weight', 'bold')
        .text(d => `${(d.percentage * 100).toFixed(1)}%`);

// Donuts
    const arc = d3.svg.arc()
      .outerRadius(radius)
      .innerRadius(radius - donutInnerWidth);

    const pie = d3.layout.pie()
      .sort(null)
      .value(d => d.arcVal);

    const donutColor = d3.scale.ordinal()
      .domain([2, 1, 0])
      .range(['grey', 'red', 'green']);

    donutData.forEach((val) => {
      const count = val.bvtStat.length;
      const weight = 100 / count;
      const dateArc = val.bvtStat.map(bvtValObj => {
        if (bvtValObj.untested > 0 && _(bvtValObj).omit('untested').values().sum() === 0) {
          return {
            arcVal: weight,
            colorVal: 2,
          };
        } else if (_(bvtValObj).omit('passed').values().sum() > 0) {
          return {
            arcVal: weight,
            colorVal: 1,
          };
        }
        return {
          arcVal: weight,
          colorVal: 0,
        };
      });

      g.selectAll('.donut-group')
        .call(renderDonuts, val, dateArc, arc);
    });

// Table
    const ticks = svg.selectAll('.axis--x .tick');

    const delimiter = (text) => {
      if (text.indexOf('snapshot') > -1) {
        return 'snapshot';
      }
      if (text.indexOf('mos iso') > -1) {
        return 'mos iso';
      }
      if (text.indexOf('iso') > -1) {
        return 'iso';
      }
      return ' ';
    };

    ticks.append('rect')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', 0)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');


    ticks.append('text')
        .attr('class', 'delimiter')
        .attr('x', -50)
        .attr('y', 3)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('dx', 50)
        .attr('dy', 32)
        .attr('text-anchor', 'middle')
        .style('font-size', `${fontSize}px`);

    ticks.data(names)
        .select('text.delimiter')
        .text(d => delimiter(d));

    ticks.append('text')
        .attr('class', 'run-name')
        .attr('x', -50)
        .attr('y', 5)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('dx', 50)
        .attr('dy', 50)
        .style('font-size', `${fontSize}px`)
        .attr('text-anchor', 'middle');

    ticks.data(names)
        .select('text.run-name')
        .text(d => d.split(delimiter(d))[1].trim());

    const rowCount = _.keys(sortedData[0].status_counts).length + ['total'].length;

    const rowPositionY = _.range(
      tableMainRowHeight,
      (tableMainRowHeight - tableRowHeight) + (rowCount * tableRowHeight) + 1,
      tableRowHeight
    );

    const statuses = _(sortedData)
        .map('status_counts')
        .values().value();

    const legendLabel = [
      'All TCs',
      'Passed',
      'Skipped',
      'Infra failed',
      'Test failed',
      'Prod failed',
      'Failed',
      'Untested',
      'Blocked',
      'In progress',
      'Other',
      'Fixed',
      'Regression',
    ];

    const statusKeys = [];

    legendLabel.forEach((label, index) => {
      const statusName = _.snakeCase(label);
      if (statusName in statuses[0]) {
        statusKeys.push({ value: statusName, index });
      }
    });

    ticks.selectAll('rect')
        .data(rowPositionY, d => d)
        .enter()
        .append('rect')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', d => d)
        .attr('height', tableRowHeight)
        .attr('class', (d, i) => i)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    const total = statuses.map(status => _.sum(_.values(status)));

    statusKeys.forEach((val, index) => {
      const statusVal = [];
      sortedData.forEach((s) => {
        statusVal.push(s.status_counts[val.value]);
      });

      statusVal.forEach((statusValue, statusPosition) => {
        d3.select(`.axis--x>g:nth-child(${statusPosition + 1})`).append('text')
          .attr('x', 0 - x.rangeBand() / 2)
          .attr('y', rowPositionY[index + 1])
          .attr('dx', x.rangeBand() / 2)
          .attr('dy', fontSize - 2)
          .attr('height', tableRowHeight)
          .attr('width', x.rangeBand())
          .attr('text-anchor', 'middle')
          .text(statusValue);
      });
    });

    [...Array(sortedData.length).keys()].map(num =>
      d3.select(`.axis--x>g:nth-child(${num + 1})`).append('text')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', tableMainRowHeight)
        .attr('dx', x.rangeBand() / 2)
        .attr('dy', fontSize - 2)
        .attr('height', tableRowHeight)
        .attr('width', x.rangeBand())
        .attr('text-anchor', 'middle')
        .text(total[num])
    );

    const legend = svg.append('g')
      .attr('class', 'legend')
      .attr('transform', `translate(0,${height + margin.top + tableMainRowHeight})`);

    const legendPositionY = _.range(
      0,
      0 + ((rowCount - 1) * tableRowHeight) + 1,
      tableRowHeight
    );

    legend.selectAll('rect')
      .data(legendPositionY, d => d)
      .enter()
      .append('rect')
      .attr('x', 0)
      .attr('y', d => d)
      .attr('height', tableRowHeight)
      .attr('class', (d, i) => i)
      .attr('width', margin.left - 18)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    legendLabel.forEach((label, index) => {
      const statusKey = _.snakeCase(label);
      if (_.findKey(statusKeys, { value: statusKey })) {
        d3.select('.legend')
          .append('rect')
          .attr('x', 5)
          .attr('y', legendPositionY[index] + 5)
          .attr('width', 15)
          .attr('height', 10)
          .style('fill', z(statusKey));
      }
    });

    legendLabel.forEach((label, index) => {
      d3.select('.legend')
        .append('text')
        .attr('x', 25)
        .attr('y', legendPositionY[index] + 17)
        .style('font-size', `${fontSize}px`)
        .text(label);
    });

    const surround = svg.append('g')
      .attr('class', 'around')
      .attr('transform', `translate(0,${margin.top - (margin.top - donutHeight) / 4})`);

    surround.append('line')
      .attr('x', 0)
      .attr('y', 0)
      .attr('x2', width + margin.left + margin.right)
      .attr('y2', 0)
      .style('stroke', 'black')
      .style('stroke-width', '0.5');

    surround.append('text')
      .attr('transform', `translate(${margin.left / 2},${-(margin.top / 2) + 30})rotate(-90)`)
      .style('font-weight', 'bold')
      .style('font-size', 'xx-large')
      .text('BVT');

    surround.append('text')
      .attr('transform', `translate(${margin.left / 2},${height / 2 + 100})rotate(-90)`)
      .style('font-weight', 'bold')
      .style('font-size', 'xx-large')
      .text('SWARM');

    function renderDonuts(selection, val, dateArc) {
      selection
            .data([val])
            .enter().append('g')
            .attr(
              'transform',
              d => `translate(${((x2(d.date) + x2.rangeBand() / 2) - 20)}, ${-margin.top / 2})`
            )
            .attr('class', '.donut-group')
            .selectAll('path')
            .data(pie(dateArc))
            .enter()
            .append('path')
            .attr('d', arc)
            .style('stroke', 'white')
            .style('stroke-width', 2)
            .attr('fill', d => donutColor(d.data.colorVal));
    }

    return true;
  }


  render() {
    return (
      <div ref={(c) => { this.chartRoot = c; }} style={{ textAlign: 'center' }}>
        <svg
          ref={(c) => { this.chart = c; }}
          height={this.props.height}
          width={this.props.minimise ? getMinimiseChartWidth(this.props.datum.length) : '100%'}
        />
      </div>
    );
  }
}
