import React from 'react';
import _ from 'lodash';
import d3 from 'd3';


const statusColors = {
  FIX_RELEASED: '#69BA69',
  FIX_COMMITTED: '#E06666',
  OPEN: '#5DA4B6',
  DEV_TO_VERIFY: '#8062A0',
};

export default class VerificationProgressChart extends React.Component {

  static propTypes = {
    datum: React.PropTypes.arrayOf(React.PropTypes.object),
    title: React.PropTypes.string,
  };

  constructor(props) {
    super(props);
    this.renderChart = this.renderChart.bind(this);
    this.handleResize = this.handleResize.bind(this);
  }

  componentDidMount() {
    window.addEventListener('resize', this.handleResize);
  }

  shouldComponentUpdate(nextProps) {
    return this.props.datum !== nextProps.datum;
  }

  componentDidUpdate() {
    this.renderChart();
  }

  componentWillUnmount() {
    window.removeEventListener('resize', this.handleResize);
  }

  handleResize() {
    this.renderChart();
  }

  renderChart() {
    if (!this.props.datum.length) {
      const noData = d3.select(this.chart);
      return noData.append('text').text('No Data');
    }

    const myNode = this.chart;
    while (myNode.firstChild) {
      myNode.removeChild(myNode.firstChild);
    }

    const margin = { top: 40, right: 55, bottom: 150, left: 150 };

    const svg = d3
      .select(this.chart)
      .attr('id', _.camelCase(this.props.title))
      .append('g')
      .attr('transform', `translate(${margin.left}, ${margin.top})`)
      ;

    const width = +d3.select(this.chart).style('width').replace('px', '') - margin.left - margin.right;
    const height = +d3.select(this.chart).attr('height') - margin.top - margin.bottom;

    const xDates = d3.scale.ordinal()
      .rangeRoundBands([0, width], 0.1);

    const xBars = d3.scale.ordinal();

    const xAxis = d3.svg.axis()
      .scale(xDates)
      .orient('bottom')
      .tickFormat(d3.time.format('%d-%b'))
      .tickSize(1);

    const yValues = d3.scale.linear()
      .rangeRound([height, 0]);

    const yPercents = d3.scale.linear()
      .rangeRound([height, 1]);

    const yAxis = d3.svg.axis()
      .scale(yValues)
      .orient('left')
      .tickSize(1);

    const yAxisPercent = d3.svg.axis()
      .scale(yPercents)
      .orient('right')
      .tickFormat(d3.format('.1%'))
      .tickSize(1);

    const barGroup = [{ groupName: 'all', color: 'green' }, { groupName: 'crit', color: 'red' }, { groupName: 'high', color: 'blue' }];
    const partitionData = _.partition(this.props.datum, obj => obj.type === 'line');

    const lineDataSet = partitionData[0];
    const stackedData = partitionData[1];

    if (!lineDataSet[0].values.length) { return null; }

    const dateDomain = _.map(lineDataSet[0].values, 'x');
    const dateMap = dateDomain.map(d => ({ date: d.toDate() }));

    xDates.domain(dateMap.map(o => o.date));
    xBars.domain(barGroup.map(o => o.groupName)).rangeRoundBands([0, xDates.rangeBand()], 0.4);
    yPercents.domain([0, 1]);

    const z = d3.scale.ordinal()
      .domain(
        ['Fix Released', 'Fix Committed',
        'Open', 'Dev-to-verify']
      )
      .range(
        [statusColors.FIX_RELEASED, statusColors.FIX_COMMITTED,
        statusColors.OPEN, statusColors.DEV_TO_VERIFY]
      );

    const stack = d3.layout.stack()
      .values(d => d.values);

    const rects = _.flatten(
      barGroup.map((layer) => {
        const layers = stack(_(stackedData).filter({ groupName: layer.groupName }).value());
        return _.flatten(
          layers.map(elem => elem.values.map(val => ({
            date: val.x.toDate(),
            key: elem.key,
            groupName: elem.groupName,
            y: val.y,
            y0: val.y0,
          })))
        );
      })
    );

    yValues.domain([0, d3.max(rects, d => d.y0 + d.y)]);

    barGroup.forEach((bar) => {
      const barData = _(lineDataSet).filter({ groupName: bar.groupName }).value();
      bar.lineGroup = barData[0].values.map(data => ({
        date: data.x.toDate(),
        percentage: data.y,
        bugType: bar.groupName,
        color: bar.color,
      }));
    });

    const flattenLineData = _(barGroup).map('lineGroup').flatten();

    dateMap.forEach((d) => {
      const data = _(rects).filter({ date: d.date });
      const lineData = flattenLineData.filter({ date: d.date });
      d.rects = data.value();
      d.all = [
        ...data.filter({ groupName: 'all' }).value(),
        ...lineData.filter({ bugType: 'all' }).value(),
        {}, {},
      ];
      d.crit = [
        ...data.filter({ groupName: 'crit' }).value(),
        {},
        ...lineData.filter({ bugType: 'crit' }).value(),
        {},
      ];
      d.high = [
        ...data.filter({ groupName: 'high' }).value(),
        {},
        {}, ...lineData.filter({ bugType: 'high' }).value(),
      ];
    });

    const valueline = d3.svg.line()
        .interpolate('monotone')
        .x(d => xDates(d.date) + xBars(d.bugType) + (xBars.rangeBand() / 2))
        .y(d => yPercents(d.percentage));


    svg.append('g')
        .attr('class', 'axis axis--x')
        .attr('transform', `translate(${0},${height})`)
        .call(xAxis);

    const dates = svg.selectAll('.date')
      .data(dateMap)
      .enter()
      .append('g')
      .attr('class', 'date')
      .attr('transform', d => `translate(${xDates(d.date)}, 0)`);


    dates
        .selectAll('rect')
        .data(d => d.rects)
        .enter()
        .append('rect')
        .attr('width', xBars.rangeBand())
        .attr('x', d => xBars(d.groupName))
        .attr('y', d => yValues(d.y + d.y0))
        .style('fill', d => z(d.key))
        .attr('height', d => yValues(d.y0) - yValues(d.y + d.y0));


    svg.append('g')
        .attr('class', 'axis axis--y')
        .call(yAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr(
          'transform',
          `translate(${width}, ${0})`
        )
        .call(yAxisPercent);


    const lineGroup = svg.selectAll('.line-group')
        .data(barGroup)
        .enter()
        .append('g')
        .attr('class', 'line-group');

    lineGroup
        .append('path')
        .attr('class', 'line')
        .style('fill', 'none')
        .style('stroke', d => d.color)
        .style('stroke-width', '2.5px')
        .attr('d', d => valueline(d.lineGroup));

    const linePoints = lineGroup
      .append('g')
      .attr('class', 'line-point');

    linePoints
      .selectAll('circle')
      .data(d => d.lineGroup)
      .enter()
      .append('circle')
      .attr('cx', d => xDates(d.date) + xBars(d.bugType) + (xBars.rangeBand() / 2))
      .attr('cy', d => yPercents(d.percentage))
      .attr('r', 5)
      .style('fill', d => d.color)
      .style('stroke', d => d.color);

    const pointText = lineGroup.append('g')
        .attr('class', 'point-text');

    pointText
      .selectAll('text')
      .data(d => d.lineGroup)
      .enter()
      .append('text')
      .attr('x', d => xDates(d.date) + xBars(d.bugType) + (xBars.rangeBand() / 2))
      .attr('y', d => yPercents(d.percentage))
      .attr('dy', -10)
      .attr('text-anchor', 'middle')
      .style('font-weight', 'bold')
      .style('fill', 'black')
      .text(d => `${(d.percentage * 100).toFixed(1)}%`);

    const tableRowHeight = 18;
    const tableMainRowHeight = 20;
    const tableInfoHead = 20;

    const tablePosition = xDates.range().map((startBand, index) => {
      const paddingHalf = (xDates.range()[1] - (xDates.range()[0] + xDates.rangeBand())) / 2;
      if (index === 0) {
        return {
          x: 0,
          width: startBand + xDates.rangeBand() + paddingHalf,
        };
      }
      if (index === xDates.range().length - 1) {
        return {
          x: startBand - paddingHalf,
          width: (width - startBand) + paddingHalf,
        };
      }
      return {
        x: startBand - paddingHalf,
        width: xDates.rangeBand() + (2 * paddingHalf),
      };
    });

    const tableHeadDate = svg.selectAll('.table-head')
      .data(tablePosition)
      .enter()
      .append('rect')
      .attr('class', 'table-head');

    tableHeadDate
      .attr('x', d => d.x)
      .attr('y', height)
      .attr('height', tableMainRowHeight)
      .attr('width', d => d.width)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    _(dateMap).sortBy('date').each((o, i) => { o.tableP = tablePosition[i]; });

    const tableHeadInfo = svg.selectAll('.table-info')
      .data(dateMap)
      .enter()
      .append('g')
      .attr('class', 'table-info');

    tableHeadInfo
      .append('rect')
      .attr('x', d => d.tableP.x)
      .attr('y', height + tableMainRowHeight)
      .attr('height', tableMainRowHeight)
      .attr('width', d => d.tableP.width / 3)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    tableHeadInfo
      .append('rect')
      .attr('x', d => d.tableP.x + (d.tableP.width / 3))
      .attr('y', height + tableMainRowHeight)
      .attr('height', tableMainRowHeight)
      .attr('width', d => d.tableP.width / 3)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    tableHeadInfo
      .append('rect')
      .attr('x', d => d.tableP.x + (2 * (d.tableP.width / 3)))
      .attr('y', height + tableMainRowHeight)
      .attr('height', tableMainRowHeight)
      .attr('width', d => d.tableP.width / 3)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    tableHeadInfo
      .append('text')
      .attr('x', d => d.tableP.x)
      .attr('y', height + tableMainRowHeight)
      .attr('text-anchor', 'middle')
      .attr('dx', d => (d.tableP.width / 3) / 2)
      .attr('dy', 15)
      .text('all');

    tableHeadInfo
      .append('text')
      .attr('x', d => d.tableP.x + (d.tableP.width / 3))
      .attr('y', height + tableMainRowHeight)
      .attr('text-anchor', 'middle')
      .attr('dx', d => (d.tableP.width / 3) / 2)
      .attr('dy', 15)
      .text('crit');

    tableHeadInfo
      .append('text')
      .attr('x', d => d.tableP.x + (2 * (d.tableP.width / 3)))
      .attr('y', height + tableMainRowHeight)
      .attr('text-anchor', 'middle')
      .attr('dx', d => (d.tableP.width / 3) / 2)
      .attr('dy', 15)
      .text('high');


    const rowCount = lineDataSet.length + _(stackedData).uniqBy('key').value().length;

    const rowPositionY = _.range(
      height + tableMainRowHeight + tableInfoHead,
      ((height + tableMainRowHeight + tableInfoHead) - tableRowHeight)
      + (rowCount * tableRowHeight) + 1,
      tableRowHeight
    );

    dateMap.forEach((date) => {
      date.all.forEach((o, index) => {
        o.xPosition = date.tableP.x;
        o.yPosition = rowPositionY[index];
        o.width = date.tableP.width / 3;
      });
      date.crit.forEach((o, index) => {
        o.xPosition = date.tableP.x + (date.tableP.width / 3);
        o.yPosition = rowPositionY[index];
        o.width = date.tableP.width / 3;
      });
      date.high.forEach((o, index) => {
        o.xPosition = date.tableP.x + (2 * (date.tableP.width / 3));
        o.yPosition = rowPositionY[index];
        o.width = date.tableP.width / 3;
      });
    });

    const tableDataAll = tableHeadInfo
      .append('g')
      .attr('class', 'table-data-all');

    tableDataAll
      .selectAll('rect')
      .data(d => d.all)
      .enter()
      .append('rect')
      .attr('class', 'table-data-all')
      .attr('x', d => d.xPosition)
      .attr('y', d => d.yPosition)
      .attr('height', 18)
      .attr('width', d => d.width)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    tableDataAll
      .selectAll('text')
      .data(d => d.all)
      .enter()
      .append('text')
      .attr('class', 'table-data-all')
      .attr('x', d => d.xPosition)
      .attr('y', d => d.yPosition)
      .attr('height', 18)
      .attr('width', d => d.width)
      .attr('text-anchor', 'middle')
      .attr('dx', d => d.width / 2)
      .attr('dy', 15)
      .text((d) => {
        if (d.y || d.y === 0) {
          return d.y;
        }
        if (d.percentage || d.percentage === 0) {
          return `${(d.percentage * 100).toFixed(1)}%`;
        }
        return '-';
      });

    const tableDataCrit = tableHeadInfo
      .append('g')
      .attr('class', 'table-data-crit');

    tableDataCrit
      .selectAll('rect')
      .data(d => d.crit)
      .enter()
      .append('rect')
      .attr('class', 'table-data-crit')
      .attr('x', d => d.xPosition)
      .attr('y', d => d.yPosition)
      .attr('height', 18)
      .attr('width', d => d.width)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    tableDataCrit
      .selectAll('text')
      .data(d => d.crit)
      .enter()
      .append('text')
      .attr('class', 'table-data-crit')
      .attr('x', d => d.xPosition)
      .attr('y', d => d.yPosition)
      .attr('height', 18)
      .attr('width', d => d.width)
      .attr('text-anchor', 'middle')
      .attr('dx', d => d.width / 2)
      .attr('dy', 15)
      .text((d) => {
        if (d.y || d.y === 0) {
          return d.y;
        }
        if (d.percentage || d.percentage === 0) {
          return `${(d.percentage * 100).toFixed(1)}%`;
        }
        return '-';
      });

    const tableDataHigh = tableHeadInfo
      .append('g')
      .attr('class', 'table-data-high');

    tableDataHigh
      .selectAll('rect')
      .data(d => d.high)
      .enter()
      .append('rect')
      .attr('class', 'table-data-high')
      .attr('x', d => d.xPosition)
      .attr('y', d => d.yPosition)
      .attr('height', 18)
      .attr('width', d => d.width)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    tableDataHigh
      .selectAll('text')
      .data(d => d.high)
      .enter()
      .append('text')
      .attr('class', 'table-data-high')
      .attr('x', d => d.xPosition)
      .attr('y', d => d.yPosition)
      .attr('height', 18)
      .attr('width', d => d.width)
      .attr('text-anchor', 'middle')
      .attr('dx', d => d.width / 2)
      .attr('dy', 15)
      .text((d) => {
        if (d.y || d.y === 0) {
          return d.y;
        }
        if (d.percentage || d.percentage === 0) {
          return `${(d.percentage * 100).toFixed(1)}%`;
        }
        return '-';
      });

    const legendLabel = [
      ...dateMap[0].all.filter(o => o.groupName).map(layer => ({ label: layer.key })),
      ...barGroup.map(layer => ({
        label: `Verif. rate ${layer.groupName}`,
        color: layer.color,
      })),
    ];

    const legend = svg.selectAll('g.legend')
      .data(legendLabel)
      .enter()
      .append('g')
      .attr('class', 'legend');

    legend
      .append('rect')
      .attr('x', `-${margin.left - 1}`)
      .attr('y', (d, i) => rowPositionY[i])
      .attr('height', 18)
      .attr('width', margin.left - 1)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    legend
      .append('text')
      .attr('x', `-${margin.left - (margin.left * 0.22)}`)
      .attr('y', (d, i) => rowPositionY[i])
      .attr('dy', 15)
      .text(d => d.label);


    legend
      .filter((d, i) => i < 3)
      .append('rect')
      .attr('x', `-${margin.left - 10}`)
      .attr('y', (d, i) => rowPositionY[i] + 4)
      .attr('width', 15)
      .attr('height', 10)
      .style('fill', d => z(d.label));

    legend
      .filter((d, i) => i > 2)
      .append('line')
      .attr('x1', `-${margin.left - 5}`)
      .attr('y2', (d, i) => rowPositionY[3 + i] + 9)
      .attr('x2', `-${margin.left - 29}`)
      .attr('y1', (d, i) => rowPositionY[3 + i] + 9)
      .attr('stroke-width', '2.5')
      .attr('stroke', d => d.color);

    legend
      .filter((d, i) => i > 2)
      .append('circle')
      .attr('cx', `-${margin.left - 17}`)
      .attr('cy', (d, i) => rowPositionY[3 + i] + 9)
      .attr('r', 5)
      .style('fill', d => d.color);


    svg.append('text')
      .attr('x', `-${margin.left}`)
      .attr('y', 0)
      .style('text-decoration', 'underline')
      .text(this.props.title);

    return true;
  }

  render() {
    return (
      <div>
        <svg style={{ width: '100%' }} ref={(c) => { this.chart = c; }} {..._.pick(this.props, ['width', 'height'])} />
      </div>
    );
  }
  }
