import React from 'react';
import _ from 'lodash';
import d3 from 'd3';
import moment from 'moment';

const oneDayWidth = 120;
const legendWidth = 132;
const tableRowHeight = 21;
const tableMainRowHeight = 60;

const fontSize = 20;

function getMinimiseChartWidth(countDays, maxWidth) {
  if (countDays === 1) {
    return oneDayWidth + legendWidth + 50;
  }
  const result = countDays > 1 ? (oneDayWidth * countDays) + legendWidth : countDays + legendWidth;
  return result >= maxWidth ? maxWidth : result;
}

function getPercentFontSize(isMin) {
  if (isMin) {
    return `${fontSize - 2}px`;
  }
  return `${fontSize + 5}px`;
}


export default class TempestChart extends React.Component {

  static propTypes = {
    datum: React.PropTypes.arrayOf(React.PropTypes.object),
    height: React.PropTypes.number,
    minimise: React.PropTypes.bool,
  };

  constructor(props) {
    super(props);
    this.renderChart = this.renderChart.bind(this);
    this.handleResize = this.handleResize.bind(this);
  }

  componentDidMount() {
    window.addEventListener('resize', this.handleResize);
  }

  shouldComponentUpdate(nextProps) {
    return this.props.datum !== nextProps.datum || this.props.minimise !== nextProps.minimise;
  }

  componentDidUpdate() {
    this.renderChart();
  }

  componentWillUnmount() {
    window.removeEventListener('resize', this.handleResize);
  }

  handleResize() {
    this.renderChart();
  }

  renderChart() {
    if (!this.props.datum.length) {
      const noData = d3.select(this.chart);
      return noData.append('text').text('No Data');
    }

    const myNode = this.chart;
    while (myNode.firstChild) {
      myNode.removeChild(myNode.firstChild);
    }

    const svg = d3.select(this.chart);
    const margin = { top: 30, right: 50, bottom: 334, left: legendWidth + 18 };
    const maxWidth = d3.select(this.chartRoot).node().getBoundingClientRect().width;
    let width = null;
    if (!this.props.minimise) {
      width = maxWidth - margin.left - margin.right;
    } else {
      width = getMinimiseChartWidth(this.props.datum[0].values.length, maxWidth) - margin.left - margin.right;
    }
    const height = this.props.height - margin.top - margin.bottom;
    const g = svg.append('g')
      .attr('transform', `translate(${margin.left},${margin.top})`);

    const x = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const x2 = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const xAxis = d3.svg.axis()
      .scale(x)
      .orient('bottom')
      .tickFormat(d3.time.format('%d-%b'))
      .tickSize(1);

    const y = d3.scale.linear()
      .rangeRound([height, 0]);

    const y2 = d3.scale.linear()
      .rangeRound([height, 1]);

    const yAxis = d3.svg.axis()
    .scale(y)
    .orient('left')
    .tickSize(1);

    const yAxis2 = d3.svg.axis()
    .scale(y2)
    .orient('right')
    .tickFormat(d3.format('.1%'))
    .tickSize(1);

    const z = d3.scale.ordinal()
      .domain(
      ['blocked', 'failed', 'fixed', 'in_progress',
        'infra_failed', 'other', 'passed', 'prod_failed',
        'regression', 'skipped', 'test_failed', 'untested']
      )
      .range(
      ['#8497B0', '#E6B9B8', '#548235', '#4472C4',
        '#FF3399', '#806000', '#92D050', '#FF0000',
        '#1F4E78', '#984807', '#FF7C80', '#595959']
      );


    const partitionData = _.partition(this.props.datum, obj => obj.type === 'line');

    const lineDataSet = partitionData[0];
    const stackedData = partitionData[1];

    const layerOrder = [
      'other', 'regression', 'prod_failed', 'failed',
      'test_failed', 'infra_failed', 'in_progress', 'untested',
      'fixed', 'blocked', 'skipped', 'passed',
    ];

    const layers = d3.layout.stack()(
      stackedData.map(layer =>
        layer.values.map(layerData => ({
          x: layerData.x.toDate(),
          y: layerData.y,
          name: _.snakeCase(layer.key),
        }))
      )
    );

    const lineData = lineDataSet[0].values.map(
      data => (
        {
          date: data.x.toDate(),
          percentage: data.y,
        }
      )
    );

    if (!layers[0].length) { return null; }

    x.domain(layers[0].map(d => d.x));
    x2.domain(layers[0].map(d => d.x));
    y.domain([0, d3.max(layers[layers.length - 1], d => d.y0 + d.y)]).nice();
    y2.domain([0, 1]);

    const valueline = d3.svg.line()
        .interpolate('monotone')
        .x(d => (x2(d.date) + x2.rangeBand() / 2) - 20)
        .y(d => y2(d.percentage));

    g.selectAll('.layer')
        .data(layers)
        .enter().append('g')
        .attr('class', 'layer')
        .selectAll('rect')
        .data(d => d)
        .enter()
        .append('rect')
        .style('fill', d => z(d.name))
        .attr('x', d => x(d.x))
        .attr('y', d => y(d.y + d.y0))
        .attr('height', d => y(d.y0) - y(d.y + d.y0))
        .attr('width', x.rangeBand() - 50);

    const names = _.map(stackedData[0].values, 'plan_name');

    svg.append('g')
        .attr('class', 'axis axis--x')
        .attr('transform', `translate(${margin.left - 20},${height + margin.top})`)
        .style('font-size', `${fontSize}px`)
        .call(xAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr('transform', `translate(${margin.left - 20}, ${margin.top})`)
        .call(yAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr(
          'transform',
          `translate(${x.rangeBand() * lineData.length + margin.left}, ${margin.top})`
        )
        .call(yAxis2);

    const lineGroup = g.selectAll('.line-group')
        .data([1])
        .enter().append('g')
        .attr('class', 'line-group');

    lineGroup.append('path')
        .attr('class', 'line')
        .style('fill', 'none')
        .style('stroke', 'black')
        .style('stroke-width', '2.5px')
        .attr('d', valueline(lineData));

    const point = lineGroup.append('g')
        .attr('class', 'line-point');

    point.selectAll('circle')
        .data(lineData)
        .enter().append('circle')
        .attr('cx', d => (x2(d.date) + x2.rangeBand() / 2) - 20)
        .attr('cy', d => y2(d.percentage))
        .attr('r', 5);

    const pointText = lineGroup.append('g')
        .attr('class', 'pointText');

    pointText.selectAll('text')
        .data(lineData)
        .enter().append('text')
        .attr('x', d => (x2(d.date) + x2.rangeBand() / 2) - 20)
        .attr('y', d => y2(d.percentage))
        .attr('dy', -10)
        .attr('text-anchor', 'middle')
        .style('font-size', getPercentFontSize(this.props.minimise))
        .style('font-weight', 'bold')
        .text(d => `${(d.percentage * 100).toFixed(1)}%`);

// Table
    const ticks = svg.selectAll('.axis--x .tick');

    ticks.append('rect')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', 0)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    const rowCount = stackedData.length + lineDataSet.length;

    const delimiter = (text) => {
      if (text.indexOf('snapshot') > -1) {
        return 'snapshot';
      }
      if (text.indexOf('mos iso') > -1) {
        return 'mos iso';
      }
      if (text.indexOf('iso') > -1) {
        return 'iso';
      }
      if (text.indexOf('-') > -1) {
        return '-';
      }
      return false;
    };

    ticks.append('text')
        .attr('class', 'delimiter')
        .attr('x', -50)
        .attr('y', 3)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('dx', 50)
        .attr('dy', 32)
        .attr('text-anchor', 'middle')
        .style('font-size', `${fontSize}px`);

    ticks.data(names)
      .select('text.delimiter')
      .text((d) => {
        const delim = delimiter(d);
        if (delim === false) {
          return '';
        }
        return delim;
      });

    ticks.append('text')
        .attr('class', 'run-name')
        .attr('x', -50)
        .attr('y', 5)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('dx', 50)
        .attr('dy', 50)
        .style('font-size', `${fontSize}px`)
        .attr('text-anchor', 'middle');

    ticks.data(names)
        .select('text.run-name')
        .text((d) => {
          if (moment(d).isValid()) {
            return d;
          }
          const delim = delimiter(d);
          if (delim === false) {
            if (d.length >= 25) {
              return d.slice(0, 24);
            }
            return d;
          }
          return d.split(delim)[1].trim();
        });


    const rowPositionY = _.range(
      tableMainRowHeight,
      (tableMainRowHeight - tableRowHeight) + (rowCount * tableRowHeight) + 1,
      tableRowHeight
    );

    ticks.selectAll('rect')
        .data(rowPositionY, d => d)
        .enter()
        .append('rect')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', d => d)
        .attr('height', tableRowHeight)
        .attr('class', (d, i) => i)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    const legendLabel = [
      'All TCs',
      ...stackedData.map(layer => layer.key),
    ];

    const legend = svg.append('g')
      .attr('class', 'legend')
      .attr('transform', `translate(0,${height + margin.top + tableMainRowHeight})`);

    const legendPositionY = _.range(
      0,
      0 + ((rowCount - 1) * tableRowHeight) + 1,
      tableRowHeight
    );

    legend.selectAll('rect')
      .data(legendPositionY, d => d)
      .enter()
      .append('rect')
      .attr('x', 0)
      .attr('y', d => d)
      .attr('height', tableRowHeight)
      .attr('class', (d, i) => i)
      .attr('width', legendWidth)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    legendLabel.forEach((val, index) => {
      let values = [];
      let valuesObjIndex = null;
      valuesObjIndex = _.findIndex(stackedData, v => v.key.indexOf(val) !== -1);

      if (val !== 'All TCs') {
        legend
          .append('text')
          .attr('x', 25)
          .attr('y', legendPositionY[index] + 17)
          .style('font-size', `${fontSize}px`)
          .text(_.startCase(val));
      } else {
        legend
          .append('text')
          .attr('x', 25)
          .attr('y', legendPositionY[index] + 17)
          .style('font-size', `${fontSize}px`)
          .text(val);
      }


      if (valuesObjIndex > -1) {
        values = stackedData[valuesObjIndex].values;

        legend
          .append('rect')
          .attr('x', 5)
          .attr('y', legendPositionY[index] + 5)
          .attr('width', 15)
          .attr('height', 10)
          .style('fill', z(_.snakeCase(val)));

        values.forEach((value, valuePosition) => {
          d3.select(`.axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - x.rangeBand() / 2)
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 17)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(value.y);
        });
      } else {
        values = layers[0].map(d => d.x);
        values.forEach((value, valuePosition) => {
          d3.select(`.axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - x.rangeBand() / 2)
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 17)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(layers[layers.length - 1][valuePosition].y + layers[layers.length - 1][valuePosition].y0);
        });
      }
    });
    return true;
  }

  render() {
    return (
      <div ref={(c) => { this.chartRoot = c; }} style={{ textAlign: 'center' }}>
        <svg
          ref={(c) => { this.chart = c; }}
          height={this.props.height}
          width={this.props.minimise ? getMinimiseChartWidth(this.props.datum[0].values.length) : '100%'}
        />
      </div>
    );
  }
}
