import React from 'react';
import _ from 'lodash';
import d3 from 'd3';


const statusColors = {
  FIX_RELEASED: '#69BA69',
  FIX_COMMITTED: '#E06666',
  OPEN: '#5DA4B6',
  DEV_TO_VERIFY: '#8062A0',
};

const LAUNCHPAD_DOMAIN = [
  'Fix Released',
  'Fix Committed',
  'Open',
  'Dev-to-verify',
];

const JIRA_DOMAIN = [
  'Done',
  'Fixed',
  'Open',
];

const LAUNCHPAD_RANGE = [
  statusColors.FIX_RELEASED,
  statusColors.FIX_COMMITTED,
  statusColors.OPEN,
  statusColors.DEV_TO_VERIFY,
];

const JIRA_RANGE = [
  statusColors.FIX_RELEASED,
  statusColors.FIX_COMMITTED,
  statusColors.OPEN,
];

export default class VerificationProgressChart extends React.Component {

  static propTypes = {
    datum: React.PropTypes.arrayOf(React.PropTypes.object),
    title: React.PropTypes.string,
  };

  constructor(props) {
    super(props);
    this.renderChart = this.renderChart.bind(this);
    this.handleResize = this.handleResize.bind(this);
  }

  componentDidMount() {
    window.addEventListener('resize', this.handleResize);
  }

  shouldComponentUpdate(nextProps) {
    return this.props.datum !== nextProps.datum;
  }

  componentDidUpdate() {
    this.renderChart();
  }

  componentWillUnmount() {
    window.removeEventListener('resize', this.handleResize);
  }

  handleResize() {
    this.renderChart();
  }

  renderChart() {
    if (!this.props.datum.length) {
      const noData = d3.select(this.chart);
      return noData.append('text').text('No Data');
    }

    const myNode = this.chart;
    while (myNode.firstChild) {
      myNode.removeChild(myNode.firstChild);
    }

    const svg = d3.select(this.chart).attr('id', _.camelCase(this.props.title));
    const margin = { top: 40, right: 40, bottom: 120, left: 150 };
    const width = +svg.style('width').replace('px', '') - margin.left - margin.right;
    const height = +svg.attr('height') - margin.top - margin.bottom;
    const g = svg.append('g')
      .attr('transform', `translate(${margin.left},${margin.top})`);

    const x = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const x2 = d3.scale.ordinal()
      .rangeRoundBands([0, width]);

    const xAxis = d3.svg.axis()
      .scale(x)
      .orient('bottom')
      .tickFormat(d3.time.format('%d-%b'))
      .tickSize(1);

    const y = d3.scale.linear()
      .rangeRound([height, 0]);

    const y2 = d3.scale.linear()
      .rangeRound([height, 1]);

    const yAxis = d3.svg.axis()
      .scale(y)
      .orient('left')
      .tickSize(1);

    const yAxis2 = d3.svg.axis()
      .scale(y2)
      .orient('right')
      .tickFormat(d3.format('.1%'))
      .tickSize(1);

    const partitionData = _.partition(this.props.datum, obj => obj.type === 'line');

    const lineDataSet = partitionData[0];
    const stackedData = partitionData[1];

    const z = d3.scale.ordinal()
      .domain(
        this.props.jira ? JIRA_DOMAIN : LAUNCHPAD_DOMAIN
      )
      .range(
        this.props.jira ? JIRA_RANGE : LAUNCHPAD_RANGE
      );

    const layers = d3.layout.stack()(
      stackedData.map(layer =>
        layer.values.map(layerData => ({
          x: layerData.x.toDate(),
          y: layerData.y,
          name: layer.key,
        }))
      )
    );

    const lineData = lineDataSet[0].values.map(
      data => (
        {
          date: data.x.toDate(),
          percentage: data.y,
        }
      )
    );

    if (!layers[0].length) { return null; }

    x.domain(layers[0].map(d => d.x));
    x2.domain(layers[0].map(d => d.x));
    y.domain([0, d3.max(layers[layers.length - 1], d => d.y0 + d.y)]).nice();
    y2.domain([0, 1]);

    const valueline = d3.svg.line()
        .interpolate('monotone')
        .x(d => (x2(d.date) + x2.rangeBand() / 2) - 25)
        .y(d => y2(d.percentage));


    g.selectAll('.layer')
        .data(layers)
        .enter().append('g')
        .attr('class', 'layer')
        .selectAll('rect')
        .data(d => d)
        .enter().append('rect')
        .style('fill', d => z(d.name))
        .attr('x', d => x(d.x))
        .attr('y', d => y(d.y + d.y0))
        .attr('height', d => y(d.y0) - y(d.y + d.y0))
        .attr('width', x.rangeBand() - 50);

    svg.append('g')
        .attr('class', 'axis axis--x')
        .attr('transform', `translate(${margin.left - 20},${height + margin.top})`)
        .call(xAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr('transform', `translate(${margin.left - 20}, ${margin.top})`)
        .call(yAxis);

    svg.append('g')
        .attr('class', 'axis axis--y')
        .attr(
          'transform',
          `translate(${x.rangeBand() * stackedData[0].values.length + (margin.left - margin.right / 2)}, ${margin.top})`
        )
        .call(yAxis2);

    const lineGroup = g.selectAll('.line-group')
        .data([1])
        .enter().append('g')
        .attr('class', 'line-group');

    lineGroup.append('path')
        .attr('class', 'line')
        .style('fill', 'none')
        .style('stroke', 'black')
        .style('stroke-width', '2.5px')
        .attr('d', valueline(lineData));

    const point = lineGroup.append('g')
        .attr('class', 'line-point');

    point.selectAll('circle')
        .data(lineData)
        .enter().append('circle')
        .attr('cx', d => (x2(d.date) + x2.rangeBand() / 2) - 25)
        .attr('cy', d => y2(d.percentage))
        .attr('r', 5);

    const pointText = lineGroup.append('g')
        .attr('class', 'valori');

    pointText.selectAll('text')
        .data(lineData)
        .enter().append('text')
        .attr('x', d => (x2(d.date) + x2.rangeBand() / 2) - 25)
        .attr('y', d => y2(d.percentage))
        .attr('dy', -10)
        .attr('text-anchor', 'middle')
        .style('font-weight', 'bold')
        .text(d => `${(d.percentage * 100).toFixed(1)}%`);

    const tableRowHeight = 18;
    const tableMainRowHeight = 20;
    const ticks = svg.selectAll('.axis--x .tick');

    ticks.append('rect')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', 0)
        .attr('height', tableMainRowHeight)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    const rowCount = stackedData.length + lineDataSet.length;

    const rowPositionY = _.range(
      tableMainRowHeight,
      (tableMainRowHeight - tableRowHeight) + (rowCount * tableRowHeight) + 1,
      tableRowHeight
    );

    ticks.selectAll('rect')
        .data(rowPositionY, d => d)
        .enter()
        .append('rect')
        .attr('x', 0 - x.rangeBand() / 2)
        .attr('y', d => d)
        .attr('height', 18)
        .attr('class', (d, i) => i)
        .attr('width', x.rangeBand())
        .attr('stroke-width', '0.5')
        .style('stroke', 'black')
        .style('fill', 'none');

    const legendLabel = [
      ..._.reverse(stackedData.map(layer => layer.key)),
      ...lineDataSet.map(layer => layer.key),
    ];

    const legend = svg.append('g')
      .attr('class', 'legend')
      .attr('transform', `translate(0,${height + margin.top + tableMainRowHeight})`);

    const legendPositionY = _.range(
      0,
      0 + ((rowCount - 1) * tableRowHeight) + 1,
      tableRowHeight
    );

    legend.selectAll('rect')
      .data(legendPositionY, d => d)
      .enter()
      .append('rect')
      .attr('x', 0)
      .attr('y', d => d)
      .attr('height', tableRowHeight)
      .attr('class', (d, i) => i)
      .attr('width', margin.left - 18)
      .attr('stroke-width', '0.5')
      .style('stroke', 'black')
      .style('fill', 'none');

    legendLabel.forEach((val, index) => {
      let values = [];
      let valuesObjIndex = null;
      valuesObjIndex = _.findIndex(stackedData, v => v.key.indexOf(val) !== -1);

      if (val.indexOf('% Verification rate') > -1) {
        d3.select(`#${_.camelCase(this.props.title)}>.legend`)
          .append('text')
          .attr('x', 10)
          .attr('y', legendPositionY[index] + 15)
          .text(val);
      } else {
        d3.select(`#${_.camelCase(this.props.title)}>.legend`)
          .append('text')
          .attr('x', 25)
          .attr('y', legendPositionY[index] + 15)
          .text(val);
      }

      if (valuesObjIndex === -1) {
        valuesObjIndex = _.findIndex(lineDataSet, v => v.key.indexOf(val) !== -1);
        values = lineDataSet[valuesObjIndex].values;
        values.forEach((value, valuePosition) => {
          d3.select(`#${_.camelCase(this.props.title)} .axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - x.rangeBand() / 2)
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 15)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(`${(value.y * 100).toFixed(1)}%`);
        });
      } else {
        d3.select(`#${_.camelCase(this.props.title)}>.legend`)
          .append('rect')
          .attr('x', 5)
          .attr('y', legendPositionY[index] + 5)
          .attr('width', 15)
          .attr('height', 10)
          .style('fill', z(val));

        values = stackedData[valuesObjIndex].values;
        values.forEach((value, valuePosition) => {
          d3.select(`#${_.camelCase(this.props.title)} .axis--x>g:nth-child(${valuePosition + 1})`)
            .append('text')
            .attr('x', 0 - x.rangeBand() / 2)
            .attr('y', rowPositionY[index])
            .attr('dx', x.rangeBand() / 2)
            .attr('dy', 15)
            .attr('height', tableRowHeight)
            .attr('width', x.rangeBand())
            .attr('text-anchor', 'middle')
            .text(value.y);
        });
      }
    });

    svg.append('text')
      .attr('x', 50)
      .attr('y', 10)
      .style('text-decoration', 'underline')
      .text(this.props.title);

    return true;
  }

  render() {
    return (
      <div>
        <svg style={{ width: '100%' }} ref={(c) => { this.chart = c; }} {..._.pick(this.props, ['width', 'height'])} />
      </div>
    );
  }
  }
