import React from 'react';
import _ from 'lodash';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import ReportBox from './base/ReportBox';
import SimpleRangePicker from '../components/SimpleRangePicker';
import { jiraBugLink, isEqual } from './base/utils';

const getColumnDefs = () => ([
  {
    headerName: 'QA Team',
    field: 'qa_team',
    width: 150,
    pinned: true,
  },
  {
    headerName: 'Delivery team',
    field: 'delivery_team',
    width: 150,
  },
  {
    headerName: 'Components',
    field: 'bug_component',
    valueGetter: (params) => {
      if (params.data.bug_component !== null) {
        return params.data.bug_component
          .map(component => component)
          .join(', ');
      }
      return params.value;
    },
    width: 120,
  },
  {
    headerName: 'Type',
    valueGetter: (params) => {
      if (params.data.bug_types) {
        const val = _(params.data.bug_types).pickBy(item => item === true).keys().value();
        return val.map(type => type).join(', ');
      }
      return params.value;
    },
    width: 100,
  },
  {
    headerName: 'Assignee',
    field: 'bug_assignee',
    width: 120,
  },
  {
    headerName: 'ID',
    field: 'bug_id',
    width: 100,
  },
  {
    headerName: 'Issue description',
    field: 'bug_title',
    cellRenderer: jiraBugLink,
    width: 600,
  },
  {
    headerName: 'Status',
    field: 'bug_status',
    width: 100,
  },
  {
    headerName: 'Priority',
    field: 'bug_priority',
    width: 100,
    cellClassRules: {
      'rag-red-outer': params => (params.value === 'Blocker' && params.data.bug_status !== 'Done'),
      'bug-critical': params => (params.value === 'Critical' && params.data.bug_status !== 'Done'),
    },
    cellRenderer: params =>
            `<span class="rag-element">${params.value}</span>`,
  },
  {
    headerName: 'Index',
    field: 'bug_types.counter',
    width: 100,
  },
  {
    headerName: 'Bug_submitted',
    field: 'bug_created',
    width: 100,
  },
  {
    headerName: 'Bug_escalated',
    field: 'bug_escalation',
    width: 100,
  },
]);

export default class FailureDriversJira extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs: getColumnDefs,
    gridHeight: 600,
    title: 'Jira: Failure Drivers',
    csvExportFilename: 'failure_drivers_jira.csv',
    gridDataUrl: '/api/v1.0/jira/jira_metrics',
    lastSyncInfoUrl: '/api/v1.0/jira/sync/last',
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
    },
  };

  constructor(props) {
    super(props);
    this.state.datesRange = this.initialConfiguration.datesRange.map(d => moment(d));
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedMilestones'];
    if (!isEqual(_.pick(this.state, fetchParams), _.pick(prevState, fetchParams))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      datesRange: [
        moment().subtract(30, 'days').startOf('day'),
        moment().endOf('day'),
      ].map(d => d.toJSON()),
      selectedMilestones: ['58'],
    };
  }

  getConfiguration() {
    const conf = super.getConfiguration();
    return {
      ...conf,
      datesRange: this.state.datesRange.map(d => d.toJSON()),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf(_.pick(conf, ['selectedMilestones']));
    if (conf.datesRange) {
      this.setStateIf({ datesRange: conf.datesRange.map(d => moment(d)) });
    }
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each(
      (i, elem) => selectedMilestones.push($(elem).val()));
    this.setState({ selectedMilestones });
  }

  onDatesRangeChange(newRange) {
    this.setState({ datesRange: newRange });
  }

  prepareGridData(data) {
    _.each(data, (bug) => {
      bug.bug_created = moment(bug.bug_created).format('YYYY-MM-DD');
      bug.bug_escalation = moment(bug.bug_escalation).format('YYYY-MM-DD');
    });
    return data;
  }


  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      milestone_ids: this.state.selectedMilestones.join(','),
    }).done((data) => {
      this.updateGridData(data);
    })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  renderChartCmp() {
    // not needed implement
  }

  renderToolBarCmp() {
    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        onChange={this.onDatesRangeChange}
        startDate={this.state.datesRange[0]}
        endDate={this.state.datesRange[1]}
      />
    );

    const milestoneData = commonStore.testrail_milestones
    .filter(m => _.includes(m.name.toLowerCase(), 'mk')
      || _.includes(m.name.toLowerCase(), 'mcp')
      || _.includes(m.name.toLowerCase(), 'stacklight'))
    .map(m => ({
      value: m.id,
      label: m.name,
      selected: _.includes(
        this.state.selectedMilestones,
        typeof m.id === 'string' ? m.id : '' + m.id),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        includeSelectAllOption
        allSelectedText="all selected"
        nonSelectedText="select milestone"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );
    return super.renderToolBarCmp({ main: [milestonesSelect, datePicker] });
  }
}
