/**
 * Created by farhad on 19.04.16.
 */

import _ from 'lodash';
import React from 'react';
import Multiselect from 'react-bootstrap-multiselect';
import { reactCellRendererFactory } from 'ag-grid-react';

import ReportBox from './base/ReportBox';
import './base/multiChartExtended';
import getStatusColumns,
  { statsDefaultFieldMap, makeStatusFieldKey } from './base/CommonColumns';
import { isEqual, safeDiv, toFixedCelRenderer } from './base/utils';
import consts from '../consts';
import CheckBox from '../components/CheckBox';
import PopoverCell from './base/PopoverCell';


const GRID_DATA_URL = '/api/v1.0/feature_testing_progress';
const JIRA_BASE_URL = 'https://mirantis.jira.com';

function _iconify(val) {
  val = val.replace('(/)', `<img src="${JIRA_BASE_URL}/images/icons/emoticons/check.png" />`);
  val = val.replace('(!)', `<img src="${JIRA_BASE_URL}/images/icons/emoticons/warning.png" />`);
  val = val.replace('(x)', `<img src="${JIRA_BASE_URL}/images/icons/emoticons/error.png" />`);
  return val;
}

function _renderIssueLink(issueId) {
  return `<a target="_blank" href="${JIRA_BASE_URL}/browse/${issueId}">${issueId}</a>`;
}

function getRowStyle(params) {
  if (params.data.issuetype === 'Epic') {
    return { 'background-color': 'rgba(3,169,244,0.58)' };
  }
  return null;
}

function timeToNumber(time) {
  if (time === undefined || time === null) {
    return null;
  }
  if (time.indexOf('day') > -1) {
    return time.split(' ')[0] * 86400;
  }
  if (time.indexOf('hour') > -1) {
    return time.split(' ')[0] * 3600;
  }
  if (time.indexOf('minute') > -1) {
    return time.split(' ')[0] * 60;
  }
  return 0;
}

function timeComparator(time1, time2) {
  const time1Number = timeToNumber(time1);
  const time2Number = timeToNumber(time2);
  if (time1Number === null && time2Number === null) {
    return 0;
  }
  if (time1Number === null) {
    return -1;
  }
  if (time2Number === null) {
    return 1;
  }
  return time1Number - time2Number;
}

const statsFieldMap = {
  ...statsDefaultFieldMap,
  TOTAL: 'status_total',
};

const getColumnDefs = () => ([
  {
    headerName: 'Epic',
    field: 'type',
    cellStyle: { 'font-weight': 'bold' },
    cellRenderer(params) {
      if (params.value && params.data.issuetype === 'Epic') {
        return _renderIssueLink(params.value);
      } else if (params.data.visiable_epic_link) {
        return _renderIssueLink(params.data.epic_link);
      }
      return '';
    },
    filterParams: {
      newRowsAction: 'keep',
    },
    width: 200,
    pinned: true,
  },
  {
    headerName: 'Issuetype',
    field: 'issuetype',
    width: 90,
  },
  {
    headerName: 'ID',
    field: 'key',
    cellRenderer(params) {
      if (params.value) {
        return _renderIssueLink(params.value);
      }
      return params.value;
    },
    filter: 'text',
    filterParams: { newRowsAction: 'keep' },
    width: 100,
  },
  {
    headerName: 'Summary',
    field: 'summary',
    filter: 'text',
    filterParams: { newRowsAction: 'keep' },
    width: 150,
  },
  {
    headerName: 'Status',
    field: 'status',
    filterParams: { newRowsAction: 'keep' },
    width: 90,
  },
  {
    headerName: 'Feature Health Update',
    field: 'health_time',
    width: 80,
    comparator: timeComparator,
  },
  {
    headerName: 'Feature health',
    cellRenderer(params) {
      if (params.value) {
        return _iconify(params.value);
      }
      return params.value;
    },
    field: 'feature_health',
    filter: 'text',
    filterParams: { newRowsAction: 'keep' },
    width: 150,
  },
  {
    headerName: 'QA status',
    cellRenderer(params) {
      if (params.value) {
        return _iconify(params.value);
      }
      return params.value;
    },
    field: 'status_qa',
    filter: 'text',
    filterParams: { newRowsAction: 'keep' },
    width: 150,
  },
  {
    headerName: 'Delivery team',
    field: 'delivery_team',
    filterParams: { newRowsAction: 'keep' },
    width: 150,
  },
  {
    headerName: 'QA Team',
    valueGetter: (params) => {
      if (params.data.qa_team_name) {
        const teamName = _.findKey(
          consts.launchpad.TeamsURIParams,
          obj => _.values(obj).find(el => _.includes(params.data.qa_team_name, el))
        );
        if (teamName) {
          return teamName;
        }
        return params.value;
      }
    },
    filterParams: { newRowsAction: 'keep' },
    width: 150,
  },
  {
    headerName: 'QA engineers',
    valueGetter: (params) => {
      if (params.data.qa_engineers.length) {
        return params.data.qa_engineers
          .map(engineer => engineer.display_name)
          .join(', ');
      }
      return params.value;
    },
    width: 150,
  },
  {
    headerName: 'Assignee',
    field: 'assignee.display_name',
    filterParams: { newRowsAction: 'keep' },
    width: 140,
  },
  {
    headerName: 'Bugs',
    headerClass: 'header-center-aligned',
    children: [
      ...consts.launchpad.BUG_STATUSES.map(status => ({
        headerName: status,
        columnGroupShow: 'open',
        valueGetter: params => params.data[_.snakeCase(status)].count,
        cellRenderer: reactCellRendererFactory(PopoverCell),
        width: 80,
      })),
      {
        headerName: 'Total bugs',
        aggFunc: 'sum',
        filter: 'number',
        valueGetter: (params) => {
          return params.data.totalBugs;
        },
        columnGroupShow: 'closed',
      },
      {
        headerName: 'Fix Rate',
        field: 'fix_rate',
        width: 100,
        cellRenderer: toFixedCelRenderer,
        valueGetter: params => (
          100 * safeDiv(params.data.fix_released.count || 0, params.data.totalBugs)
        ),
      },
      {
        headerName: 'Verification Rate',
        field: 'verification_rate',
        width: 100,
        cellRenderer: toFixedCelRenderer,
        valueGetter: (params) => {
          const sum = (params.data.fix_committed.count || 0) + (params.data.fix_released.count || 0);
          return 100 * safeDiv(params.data.fix_released.count || 0, sum);
        },
      },
    ],
  },
  {
    headerName: 'Test results',
    headerClass: 'header-center-aligned',
    children: getStatusColumns(statsFieldMap, true),
  },

]);

function getCountsByCases(cases) {
  const counts = {};
  _.each(cases, (case_obj) => {
    if (case_obj.last_result_status) {
      const field = makeStatusFieldKey(case_obj.last_result_status);
      counts[field] || (counts[field] = 0);
      counts[field] += 1;
    }
  });
  return counts;
}

function getIssueType(issue) {
  const issueType = {};
  if (issue.epic_link) {
    issueType.type = issue.epic_link;
  } else {
    issueType.type = issue.key;
  }
  return issueType;
}

export default class FeatureTestingProgress extends ReportBox {

  static defaultProps = Object.assign({}, ReportBox.defaultProps, {
    title: 'Feature Testing Progress',
    csvExportFilename: 'feature_testing_progress.csv',
    gridDataUrl: GRID_DATA_URL,
    gridHeight: 600,
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
      getRowStyle,
    },
  })

  constructor(props) {
    super(props);
    this.state.productVersion = this.initialConfiguration.productVersion;
    this.state.suppress = false;
    this.title = 'Statistics';

    this.handleProductSelection = this.handleProductSelection.bind(this);
    this.suppressBugsColumns = this.suppressBugsColumns.bind(this);
  }

  componentDidMount() {
    this.loadColumns();
    this.fetchGridData();
    this.setConfiguration(this.initialConfiguration);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['productVersion'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  loadColumns() {
    this.setState({ columnDefs: getColumnDefs() });
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      productVersion: _(window.commonStore.jira_filter_ids)
        .filter({ position: 0 })
        .map('name')
        .value(),
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      productVersion: this.state.productVersion,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ productVersion: conf.productVersion });
  }

  suppressBugsColumns(e) {
    this.setState({ suppress: e.target.checked });
    if (e.target.checked) {
      const columnDefs = _.filter(getColumnDefs(), d => d.headerName !== 'Bugs');
      this.setState({ columnDefs });
    } else {
      this.loadColumns();
    }
  }

  prepareGridData(data) {
    const prepared_issues = [];

    const { issues, cases, bugs } = data;
    const issuesByKey = _(issues).groupBy('key').mapValues('0').value();

    _.each(issues, issue => (issue.direct_cases = []));

    // join cases on issues by refs.
    _.each(cases, (case_obj) => {
      _.each(case_obj.refs, (ref) => {
        const issue = issuesByKey[ref];
        if (issue) {
          issue.direct_cases.push(case_obj);
        }
      });
    });

    const statusesPlaceholder = {}

    consts.launchpad.BUG_STATUSES.forEach((status) => {
      statusesPlaceholder[_.snakeCase(status)] = 0;
    });

    _.each(issues, (issue) => {
      issue.unique_cases = {};
      _.each(issue.direct_cases, c => (issue.unique_cases[c.id] = c));

      Object.assign(issue, statusesPlaceholder);

      issue.cnt_created_cases = _.size(issue.unique_cases);

      const bugsStatuses = {};

      issue.totalBugsId = [];
      issue.totalBugs = 0;

      if (bugs[issue.key]) {
        const uniqStatuses = _.uniq(bugs[issue.key].map(el => el.status));
        uniqStatuses.forEach((status) => {
          const statusSnake = _.snakeCase(status);
          const bugCount = _.filter(bugs[issue.key], el => (el.status === status));
          const bugIds = bugCount.map(el => el.bug_id);
          issue.totalBugsId = [...issue.totalBugsId, ...bugIds];
          bugsStatuses[statusSnake] = {
            count: bugCount.length,
            bugIds,
          };
          issue.totalBugs += bugCount.length;
        });
      }

      issue.covered = issue.cnt_created_cases ? 1 : 0;
      const counts = getCountsByCases(issue.direct_cases);
      const issueType = getIssueType(issue);
      Object.assign(issue, counts, issueType, bugsStatuses);
      issue.status_total = _.sum(_.values(counts));
      issue.issues_count = 1;

      prepared_issues.push(issue);
    });

    const epics = _(prepared_issues)
      .filter({ issuetype: 'Epic' })
      .sortBy(['priority_id', 'status_id']).value();

    if (!epics.length) {
      _.each(prepared_issues, (issue) => {
        if (issue.epic_link) {
          issue.visiable_epic_link = true;
        }
      });
      return _(prepared_issues).sortBy(['priority_id', 'status_id']).value();
    }

    let result = [];
    _.each(epics, (issue) => {
      if (issue.issues_in_epic === []) {
        result.push(epics);
      } else {
        const subIssue = _.filter(
          prepared_issues, o => issue.issues_in_epic.indexOf(o.key) > -1
        );
        result = [...result, issue, ...subIssue];
      }
    });

    return result;
  }

  handleProductSelection() {
    const selected = [];
    this.prodSelector.$multiselect.find(':selected').each(
      (i, elem) => selected.push($(elem).val()));
    this.setState({ productVersion: selected });
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      prod_version: this.state.productVersion.join(' '),
    })
      .done((date) => {
        this.updateGridData(date);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  renderChartCmp() {
    if (!this.gridApi || _.isEmpty(this.state.rowData)) {
      return;
    }

    let total = 0;
    let passed = 0;
    let failed = 0;
    let blocked = 0;
    let untested = 0;

    const model = this.gridApi.getModel();
    model.forEachNodeAfterFilterAndSort((node) => {
      if (node.data.issuetype === 'Epic') {
        total += node.data.status_total;
        passed += node.data.cnt_passed || 0;
        failed += _.sum([
          node.data.cnt_failed,
          node.data.cnt_prod_failed,
          node.data.cnt_test_failed,
          node.data.cnt_infra_failed,
        ]) || 0;
        blocked += node.data.cnt_blocked || 0;
        untested += node.data.cnt_untested || 0;
      }
    });

    return (
      <div>
        <h3>Epic's statistic of test results</h3>
        <p />
        <strong>Total: {total}</strong>
        <p />
        <strong>Passed: {passed}</strong>
        <p />
        <strong>Failed: {failed}</strong>
        <p />
        <strong>Blocked: {blocked}</strong>
        <p />
        <strong>Untested: {untested}</strong>
        <p />
        <strong>% Blocked: {(100 * safeDiv(blocked, total)).toFixed(2)}%</strong>
        <p />
        <strong>% Complete: {(100 * safeDiv((passed + failed + blocked), total)).toFixed(2)}%</strong>
        <p />
        <strong>% Success Rate: {(100 * safeDiv(passed, (passed + failed + blocked))).toFixed(2)}%</strong>
      </div>
    );
  }

  renderToolBarCmp() {
    const prodactVersion = _(window.commonStore.jira_filter_ids)
      .sortBy('position')
      .map(v => ({
        value: v.name, selected: _.includes(this.state.productVersion, v.name),
      })).value();

    const productSelect = (
      <Multiselect
        key="prod.select"
        ref={(c) => { this.prodSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        nonSelectedText="select version"
        onDropdownHide={this.handleProductSelection}
        data={prodactVersion}
      />
    );
    const checkBox = (
      <CheckBox
        label=" Suppress Bugs"
        key="suppress"
        checked={this.state.suppress}
        onChange={this.suppressBugsColumns}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );
    return super.renderToolBarCmp({ main: [productSelect, checkBox] });
  }

}
