import React from 'react';
import _ from 'lodash';
import * as BS from 'react-bootstrap';
import moment from 'moment';
import $ from 'jquery';
import d3 from 'd3';
import Multiselect from 'react-bootstrap-multiselect';

import DatesRangeReportBox from './base/DatesRangeReportBox';
import NVD3Chart from './base/NVD3Chart';
import './base/multiChartExtended';
import { isEqual } from './base/utils';

const GRID_DATA_URL = '/api/v1.0/jira/bug_trends';

const IMPORTANCE_CODES = ['Blocker', 'Critical', 'Major', 'Nice to have'];

function getColumnKey(iterationIndex, weekIndex, importance) {
  return `cnt_${iterationIndex}_${weekIndex}_${importance}`;
}


function getDefaultDatesRanges() {
  return {
    'Last 3 week': [
      moment().subtract(7 * 3 - 1, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 30 Days': [
      moment().subtract(29, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 3 Month': [
      moment().subtract(2, 'month').startOf('month'),
      moment().endOf('day'),
    ],
    'Last 6 Month': [
      moment().subtract(5, 'month').startOf('month'),
      moment().endOf('day'),
    ],
  };
}


function getBaseColumns() {
  return [
    {
      headerName: 'All teams',
      field: '__allteams__',
      rowGroupIndex: 0,
      valueGetter(params) { return 'All teams'; },
      hide: true,
    },
    {
      headerName: 'Bug type',
      field: 'bug_type',
      valueGetter(params) {
        if (params.data.bug_type) {
          return {
            'team_data': 'Delivery Teams',
          }[params.data.bug_type];
        }
      },
      rowGroupIndex: 1,
      hide: true,
    },
    {
      headerName: 'Team',
      field: 'qa_team_name',
      cellRenderer: {
        renderer: 'group',
      },
      cellStyle: { 'font-weight': 'bold' },
      filter: 'set',
      filterParams: {
        newRowsAction: 'keep',
        suppressRemoveEntries: true,
      },
      width: 200,
      pinned: true,
    },

  ];
}


export default class extends DatesRangeReportBox {

  static defaultProps = {
    ...DatesRangeReportBox.defaultProps,
    getDefaultDatesRanges,
    title: 'Jira: Bug Trends',
    csvExportFilename: 'jira_bug_trends.csv',
    gridDataUrl: GRID_DATA_URL,
    chartHeight: 500,
    lastSyncInfoUrl: '/api/v1.0/lp/sync/last',
    agGridOptions: {
      ...DatesRangeReportBox.defaultProps.agGridOptions,
      groupDefaultExpanded: 2,
    },
  };

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedMilestones'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  buildColumnDefs(weeks) {
    const countColumns = weeks.map((item) => {
      const mdate = moment(item.week.week_end_date);
      return {
        headerName: `Iter.${item.iteration.index} week ${item.week.week_index} (${mdate.utc().format('MMM DD')})`,
        headerClass: 'header-center-aligned',
          // extra info used in chart
        ext: {
          week: item.week,
          iteration: item.iteration,
        },
        children: IMPORTANCE_CODES.map(importance => ({
          headerName: importance,
          field: getColumnKey(item.iteration.index, item.week.week_index, importance),
          aggFunc: 'sum',
          filter: 'number',
          ext: {
            iteration: item.iteration,
          },
          filterParams: {
            newRowsAction: 'keep',
          },
          width: 85,
        })),
      };
    });
    this.setState({
      'columnDefs': [
        ...getBaseColumns(),
        ...countColumns,
      ],
    });
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each(
      (i, elem) => selectedMilestones.push($(elem).val()));
    this.setStateIf({ selectedMilestones });
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: ['MK22'],
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  prepareGridData(data) {
    const teamsData = {};
    const weeks = [];

    data.forEach((itRecord) => {
      itRecord.data.forEach((weekRecord) => {
        weeks.push({
          iteration: itRecord.iteration,
          week: weekRecord.week,
        });
        const bugTypesData = { team_data: weekRecord.delivery_team_data };

        _.each(bugTypesData, (weekData, bugType) => {
          _.each(weekData, (teamCounts, teamName) => {
            const teamKey = `${bugType}_${teamName}`;
            if (teamsData[teamKey] === undefined) {
              teamsData[teamKey] = { qa_team_name: teamName, bug_type: bugType };
            }
            _.each(teamCounts, (cnt, importance) => {
              const colKey = getColumnKey(itRecord.iteration.index, weekRecord.week.week_index, importance);
              teamsData[teamKey][colKey] = cnt;
            });
          });
        });
      });
    });

    this.buildColumnDefs(weeks);
    return _.values(teamsData);
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(
      this.props.gridDataUrl, {
        date_from: moment(this.state.datesRange[0]).utc().format('YYYY-MM-DD'),
        date_to: moment(this.state.datesRange[1]).utc().format('YYYY-MM-DD'),
        milestone: this.state.selectedMilestones.join(' '),
      })
      .done((data) => {
        this.updateGridData(data);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  getChartDatum(bugType) {
    if (!this.gridApi || _.isEmpty(this.state.rowData)) { return []; }

    const data = [];
    this.gridApi.getModel().forEachNodeAfterFilterAndSort((node) => {
      if (!node.group && node.data.bug_type == bugType) {
        data.push(node.data);
      }
    });
    const teamsMap = _(data).groupBy('qa_team_name').mapValues('0').value();

    const getSeries = (importance) => {
      const series = [];
      _(this.state.columnDefs)
        .filter(c => _.get(c, 'ext.week'))
        .each((colDef) => {
          _.each(teamsMap, (obj, teamName) => {
            const key = getColumnKey(colDef.ext.iteration.index, colDef.ext.week.week_index, importance);
            series.push({
              x: `${teamName} | Iter ${colDef.ext.iteration.index}. week ${colDef.ext.week.week_index}`,
              y: (teamsMap[teamName] || {})[key] || 0,
            });
          });
        });
      return series;
    };

    return IMPORTANCE_CODES.map(importance => ({
      values: getSeries(importance),
      key: importance,
      color: {
        'Blocker': 'red',
        'Critical': '#E28F00',
        'Major': '#00BAE0',
        'Nice to have': 'grey',
      }[importance],
    }));
  }

  renderOneChartCmp(bugType, title) {
    const datum = this.getChartDatum(bugType);
    return (
      <NVD3Chart
        type="multiBarChart"
        height={this.props.chartHeight}
        datum={datum}
        configure={(chart) => {
          chart.margin({ top: 150, right: 10, bottom: 180, left: 40 });
          chart.showXAxis(false);
          chart.showControls(false);
          chart.multibar.stacked(true);
          chart.yAxis.tickFormat(d3.format(',d'));

          function renderAxis() {
            if (datum.length === 0) {
              return;
            }
            const xdata = _.map(datum[0].values, 'x');
            const scale = chart.multibar.xScale();

            const parent = d3.select(chart.container).select('.nvd3.nv-multibar');
            const customAxisG = parent.selectAll('g.custom').data([0]);
            customAxisG.enter().append('g').attr('class', 'custom');
            customAxisG.attr('transform', `translate(0, ${chart.multibar.height()})`);

            const text = customAxisG.selectAll('text.team-label').data(xdata);
            text.enter().append('text').attr('class', 'team-label');
            text.exit().remove();
            text
              .attr('transform', d =>
                `translate(${scale(d) + scale.rangeBand() / 2 + 5}, 5) rotate(-90)`)
              .attr('text-anchor', 'end')
              .text(d => d.split('|')[0]);

            const groups = _(xdata).groupBy(d => d.split('|')[1]).map(v => v[0]).value();

            const weekG = customAxisG.selectAll('g.week').data(groups);
            const weekGEnter = weekG.enter().append('g').attr('class', 'week');
            weekGEnter.append('line');
            weekGEnter.append('text').attr('class', 'week-label');
            weekG.exit().remove();

            weekG.attr('transform', d => `translate(${scale(d)}, 0)`);

            weekG.select('line')
              .filter((d, i) => i != 0)
              .attr('x1', 0)
              .attr('x2', 0)
              .attr('y1', -chart.multibar.height()).attr('y2', 150)
              .style('stroke', 'black')
              .style('stroke-width', 1)
              .style('opacity', 0.6);

            const pad = chart.multibar.width() / groups.length / 2;
            weekG.select('text.week-label')
              .attr('transform', d => `translate(${pad}, 150)`)
              .style('font-size', '1.3em')
              .attr('text-anchor', 'middle')
              .text(d => d.split('|')[1]);
          }

          chart.dispatch.on('renderEnd', renderAxis);
        }}
        configureD3={(selection) => {
          selection
            .selectAll('text.chart-title')
            .data([0])
            .enter()
            .append('text')
            .attr('x', 50)
            .attr('y', 20)
            .attr('class', 'chart-title')
            .style('font-size', '16px')
            .style('text-decoration', 'underline')
            .text(title);
        }}
      />
    );
  }

  renderChartCmp() {
    return (
      <div>
        <BS.Row>
          <BS.Col md={12}>{this.renderOneChartCmp('team_data', 'Delivery team defects')}</BS.Col>
        </BS.Row>
      </div>
    );
  }

  renderToolBarCmp() {
    const datePicker = this.renderDatesRangePicker();

    const milestoneData = _.map(commonStore.jira_fix_versions, v => ({
      value: v, selected: _.includes(this.state.selectedMilestones, v),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        nonSelectedText="select milestone"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );

    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect] });
  }

}
