import React, { Component } from 'react';
import update from 'react/lib/update';
import { Panel, Row, Col, Button, Form, ButtonGroup } from 'react-bootstrap';
import Multiselect from 'react-bootstrap-multiselect';
import _ from 'lodash';
import moment from 'moment';

import CheckBox from '../components/CheckBox';
import AgGridExtended from '../components/AgGridExtended';
import { jiraBugLink } from './base/utils';

const GRID_DATA_URL = '/api/v1.0/testrail/jira_failure_metrics';
const PLAN_LIST_URL = '/api/v1.0/testrail/plans';
const RUN_LIST_URL = '/api/v1.0/testrail/runs';

const getColumnDefs = () => ([
  {
    headerName: 'Team',
    field: 'team',
    width: 150,
    pinned: true,
  },
  {
    headerName: 'Assignee',
    field: 'bug_assignee',
    width: 120,
  },
  {
    headerName: 'ID',
    field: 'bug_id',
    width: 100,
  },
  {
    headerName: 'ISO',
    field: 'iso_name',
    width: 90,
    hide: true,
  },
  {
    headerName: 'Milestone_id',
    field: 'milestone_id',
    width: 80,
    hide: true,
  },
  {
    headerName: 'Issue description',
    field: 'summary',
    cellRenderer: jiraBugLink,
    width: 600,
  },
  {
    headerName: 'Status',
    field: 'bug_status',
    width: 100,
  },
  {
    headerName: 'Priority',
    field: 'bug_priority',
    width: 100,
    cellClassRules: {
      'rag-red-outer': params => params.value === 'Blocker',
      'bug-critical': params => params.value === 'Critical',
    },
    cellRenderer: params =>
            `<span class="rag-element">${params.value}</span>`,
  },
  {
    headerName: 'Bug_submitted',
    field: 'bug_created',
    width: 100,
  },
  {
    headerName: 'Bug_escalated',
    field: 'bug_escalation',
    width: 100,
  },
  {
    headerName: 'Total',
    field: 'total',
    width: 80,
    sort: 'desc',
  },
]);

const defaultGridOption = columnDefs => ({
  columnDefs,
  rowSelection: 'multiple',
  enableSorting: true,
  enableFilter: true,
  enableColResize: true,
  rowHeight: 28,
  groupSuppressAutoColumn: true,
  groupSelectsChildren: true,
  groupUseEntireRow: false,
  agGridOptions: {
    icons: {
      menu: '<i class="fa fa-bars"></i>',
      filter: '<i class="fa fa-filter"></i>',
      sortAscending: '<i class="fa fa-sort-amount-asc"></i>',
      sortDescending: '<i class="fa fa-sort-amount-desc"></i>',
      sortUnSort: '<i class="fa fa-sort"></i>',
      groupExpanded: '<i class="fa fa-minus-square-o"></i>',
      groupContracted: '<i class="fa fa-plus-square-o"></i>',
      columnGroupOpened: '<i class="fa fa-minus-square"/>',
      columnGroupClosed: '<i class="fa fa-plus-square"/>',
      columnVisible: '<i class="fa fa-check-square"></i>',
      columnHidden: '<i class="fa fa-square"></i>',
      columnRemoveFromGroup: '<i class="fa fa-times"></i>',
    },
  },
});

export default class extends Component {

  constructor(props) {
    super(props);

    this.state = {
      selectedMilestone: ['Mk22_with_mitaka'],
      selectedPlan: '',
      milestonePlans: [],
      columnDefs: getColumnDefs(),
    };

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
    this.handlePlanSelection = this.handlePlanSelection.bind(this);
    this.onGridReady = this.onGridReady.bind(this);
    this.onRefreshGrid = this.onRefreshGrid.bind(this);
    this.onExportCSV = this.onExportCSV.bind(this);
  }

  componentDidMount() {
    this.setlastSyncDate();
    this.getPlans();
  }


  onGridReady({ api, columnApi }) {
    this.gridApi = api;
    this.columnApi = columnApi;
  }

  handleMilestoneSelection() {
    const selectedMilestone = [];
    this.milestonesSelector.$multiselect.find(':selected').each(
      (i, elem) => selectedMilestone.push($(elem).text()));
    this.setState({ selectedMilestone });
    this.getPlans();
  }

  handlePlanSelection() {
    const plan_id = this.planSelector.$multiselect.find(':selected').val();
    this.setState({ selectedPlan: parseInt(plan_id, 10) });
    this.getColumns(plan_id);
    this.fetchGridData({ plan_id });
  }

  getColumns(plan_id) {
    $.get(RUN_LIST_URL, { plan_id })
      .done((runs) => {
        const addColumnDefs = runs.map(col => ({
          headerName: col,
          field: _.snakeCase(col),
          width: 100 + col.length,
        }));
        const columnDefs = update(getColumnDefs(), { $push: addColumnDefs });
        this.setState({ columnDefs });
      });
  }

  onRefreshGrid() {
    const plan_id = this.planSelector.$multiselect.find(':selected').val();
    this.fetchGridData({ plan_id });
  }

  onExportCSV() {
    this.gridApi.exportDataAsCsv({
      fileName: 'jira_failure_drivers.csv',
    });
  }

  getPlans() {
    const milestone_id = window.commonStore.testrail_milestones
      .filter(m => m.name === this.state.selectedMilestone[0])[0].id;
    $.get(PLAN_LIST_URL, { milestone_id })
      .done((milestonePlans) => {
        this.setState({ milestonePlans });
        this.getColumns(_.last(_.head(milestonePlans)));
        this.fetchGridData({ plan_id: _.last(_.head(milestonePlans)) });
      });
  }

  setlastSyncDate() {
    this.setState({ lastSyncDate: 'fetching... ' });

    return $.get('/api/v1.0/testrail/sync/last')
      .done((info) => {
        if (info) {
          const date = new Date(info.started);
          this.setState({ lastSyncDate: moment(date).fromNow() });
        } else {
          this.setState({ lastSyncDate: '[unknown]' });
        }
      })
      .fail(() => { this.setState({ lastSyncDate: 'cannot fetch data' }); });
  }

  fetchGridData(query) {
    this.gridApi.showLoadingOverlay();

    return $.get(GRID_DATA_URL, query)
      .done(this.updateGridData.bind(this))
      .always(() => { this.gridApi.hideOverlay(); });
  }

  updateGridData(data) {
    const preparedData = this.prepareGridData(data);
    this.gridApi.setRowData(preparedData);
  }

  prepareGridData(data) {
    _.each(data, (bug) => {
      Object.assign(bug, _.mapKeys(bug.rate, (v, k) => _.snakeCase(k)));
      bug.bug_escalation = moment(bug.bug_escalation).format('YYYY-MM-DD');
      bug.bug_created = moment(bug.bug_created).format('YYYY-MM-DD');
      bug.total = _(bug.rate).values().sum();
    });
    return data;
  }

  render() {
    const milestoneData = window.commonStore.testrail_milestones
    .filter(m => _.includes(m.name.toLowerCase(), 'mk')
      || _.includes(m.name.toLowerCase(), 'mcp')
      || _.includes(m.name.toLowerCase(), 'stacklight'))
    .map(m => ({
      value: m.id,
      label: m.name,
      selected: _.includes(
        this.state.selectedMilestone,
        m.name),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        nonSelectedText="select milestone"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );

    const planData = _(this.state.milestonePlans)
      .map((plan, index) => ({
        value: _.last(plan),
        label: _.head(plan),
        selected: this.state.selectedPlan === '' ? index === 0 : this.state.selectedPlan === _.last(plan),
      })).value();

    const planSelect = (
      <Multiselect
        ref={(c) => { this.planSelector = c; }}
        buttonWidth="250px"
        maxHeight={300}
        enableFiltering
        nonSelectedText="select plan"
        onDropdownHide={this.handlePlanSelection}
        data={planData}
      />
    )

    return (
      <Panel header={<h3>{'Jira Failure Drivers'}</h3>}>
        <Panel footer={`Updated ${this.state.lastSyncDate}`}>
          <Row>
            <Col md={10}>
              <Form inline role="form">
                <ButtonGroup bsSize="sm" bsStyle="raised">
                  <Button bsStyle="info" bsSize="sm" onClick={this.onRefreshGrid}>
                    <i className="fa fa-refresh" /> Refresh
                  </Button>
                  <Button onClick={this.onExportCSV}>
                    <i className="fa fa-download" /> Download as CSV
                  </Button>
                </ButtonGroup>
                {milestonesSelect}
                {planSelect}
              </Form>
            </Col>
            <Col md={2}>
              <CheckBox
                label=" Tool Panel"
                groupClassName="pull-right"
                onChange={() => {}}
              />
            </Col>
          </Row>
          <AgGridExtended
            style={{ height: 600 }}
            gridOptions={{
              ...defaultGridOption(this.state.columnDefs),
              onGridReady: this.onGridReady,
            }}
          />
        </Panel>
      </Panel>
    );
  }
}
