import _ from 'lodash';
import React from 'react';
import Multiselect from 'react-bootstrap-multiselect';
import d3 from 'd3';
import moment from 'moment';

import { isEqual, jiraBugLink } from './base/utils';
import ReportBox from './base/ReportBox';
import NVD3Chart from './base/NVD3Chart';
import consts from '../consts';

const getColumnDefs = () => ([
  {
    headerName: 'ID',
    field: 'bug_id',
    width: 130,
    cellRenderer: jiraBugLink,
  },
  {
    headerName: 'Subject',
    field: 'bug_subject',
    width: 500,
  },
  {
    headerName: 'Status',
    field: 'bug_status',
    width: 100,
  },
  {
    headerName: 'Priority',
    field: 'bug_priority',
    width: 100,
  },
  {
    headerName: 'Assignee',
    field: 'bug_assignee',
    width: 100,
  },
  {
    headerName: 'Reopened date',
    field: 'bug_date',
    width: 150,
  },
]);

export default class JiraReopenedBugs extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs: getColumnDefs,
    title: 'Jira: Reopened Bugs',
    gridHeight: 400,
    csvExportFilename: 'jira_reopened_bugs.csv',
    gridDataUrl: consts.jira.REOPENED_BUGS_URL,
    lastSyncInfoUrl: '/api/v1.0/lp/sync/last',
    defaultSortModel: [{ colId: 'bug_date', sort: 'asc' }],
    refresh: React.PropTypes.func,
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
    },
  };

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;
    this.state.chartDatum = [];

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['selectedMilestones'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: ['MK22'],
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }


  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedMilestones.push($(elem).val());
    });
    this.setStateIf({ selectedMilestones });
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      milestone: this.state.selectedMilestones.join(' '),
    })
    .done(data => this.updateGridData(data))
    .always(() => this.gridApi.hideOverlay());
  }

  prepareGridData(data) {
    const rowChartDate = _.omit(data, ['current_bugs']);
    const chartDatum = [
      {
        color: 'red',
        key: 'Count',
        values: _(rowChartDate).map((val, key) => ({
          key: 'Count',
          x: moment(key),
          y: val.count,
          bug_ids: val.bug_ids,
        })).sortBy('x').value(),
      },
    ];
    this.setState({ chartDatum });
    return data.current_bugs;
  }

  renderChartCmp() {
    return (
      <NVD3Chart
        type="discreteBarChart"
        height={this.props.chartHeight}
        datum={this.state.chartDatum}
        configure={(chart) => {
          chart
            .margin({ top: 80, right: 20, bottom: 20, left: 20 });
          chart.color(['#1f77b4']);
          chart.tooltip.contentGenerator((week) => {
            if (week.data === null) return;
            const bottomRange = moment(week.data.x).subtract(7, 'days');
            const rows = week.data.bug_ids.map(id => (
              '<tr>' +
              '<td class="key">ID:</td>' +
              `<td class="x-value">${id}</td>` +
              '</tr>'
            ));
            const header =
              '<thead>' +
              '<tr>' +
              '<td class="key">' +
              '<strong>Range: </strong></td>' +
              `<td class="key"><strong>${moment(bottomRange).format('MMM D')}` +
              ` - ${moment(week.data.x).format('MMM D')}</strong></td>` +
              '</tr>' +
              '</thead>';
            const footer =
              '<tr>' +
              '<td class="key">Amount:' +
              `<td class="x-value">${week.data.bug_ids.length}</td>` +
              '</tr>';
            const result = `<table>${header}<tbody>${rows.join('')}${footer}</tbody></table>`;
            return result;
          });
          chart.yAxis
             .tickFormat(d3.format(',d'));
          chart.xAxis
               .tickFormat((d) => (d3.time.format('%d-%b')(new Date(d))))
               .showMaxMin(false)
               .tickPadding(10);
        }}
      />
    );
  }

  renderToolBarCmp() {
    const milestoneData = _.map(commonStore.jira_fix_versions, v => ({
      value: v, selected: _.includes(this.state.selectedMilestones, v),
    }));

    const milestonesSelect = (
        <Multiselect
          key="milestone.select"
          ref={(c) => { this.milestonesSelector = c; }}
          buttonWidth="250px"
          maxHeight={300}
          enableFiltering
          nonSelectedText="select milestone"
          allSelectedText="all selected"
          onDropdownHide={this.handleMilestoneSelection}
          data={milestoneData}
        />
      );
    return super.renderToolBarCmp({ main: [milestonesSelect] });
  }
}
