import _ from 'lodash';
import React from 'react';
import Multiselect from 'react-bootstrap-multiselect';

import ReportBox from './base/ReportBox';
import { isEqual } from './base/utils';
import consts from '../consts';

const getColumnDefs = () => ([
  {
    headerName: 'ID',
    field: 'bug_id',
    width: 130,
    cellRenderer: (params) => {
      if (params.data === null) {
        return params.value;
      }
      return `<a target="_blank" href="https://launchpad.net/bugs/${params.value}">${params.value}</a>`;
    },
  }, {
    headerName: 'Subject',
    field: 'bug_subject',
    width: 500,
  }, {
    headerName: 'Status',
    field: 'bug_status',
    width: 100,
  }, {
    headerName: 'Importance',
    field: 'bug_importance',
    width: 100,
  }, {
    headerName: 'Assignee',
    field: 'bug_assignee',
    width: 100,
  }, {
    headerName: 'Reporter of the bug',
    field: 'bug_reporter',
    width: 150,
  }, {
    headerName: 'One of the cases',
    field: 'lase_case_title',
    cellRenderer: (params) => {
      if (params.data === null) {
        return params.value;
      }
      return `<a target="_blank" href="https://mirantis.testrail.com/index.php?/cases/view/${params.data.last_case_id}">${params.value}</a>`;
    },
    width: 550,
  }, {
    headerName: 'Regression/Floating',
    field: 'regression',
    cellRenderer: (params) => {
      if (params.value) {
        return 'Regression';
      }
      return 'Floating';
    },
  },
]);

export default class RegressionBugsReport extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs: getColumnDefs,
    title: 'Regression Bugs',
    gridHeight: 600,
    csvExportFilename: 'regression_bugs.csv',
    gridDataUrl: consts.launchpad.REGRESSION_BUGS_URL,
    lastSyncInfoUrl: '/api/v1.0/lp/sync/last',
    defaultSortModel: [
      {
        colId: 'regression',
        sort: 'desc',
      },
    ],
    refresh: React.PropTypes.func,
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
    },
  };

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;
    this.state.chartDatum = [];

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['selectedMilestones'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: [window.commonStore.lp_milestone_in_focus.name],
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedMilestones.push($(elem).val());
    });
    this.setStateIf({ selectedMilestones });
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, { milestone: this.state.selectedMilestones.join(' ') }).done(data => this.updateGridData(data)).always(() => this.gridApi.hideOverlay());
  }

  renderChartCmp() {}

  renderToolBarCmp() {
    const milestoneData = _.map(window.commonStore.launchpad_milestones, v => ({
      value: v,
      selected: _.includes(this.state.selectedMilestones, v),
    }));
    const milestonesSelect = (<Multiselect key="milestone.select" ref={(c) => {
      this.milestonesSelector = c;
    }} buttonWidth="250px" maxHeight={300} enableFiltering nonSelectedText="select milestone" allSelectedText="all selected" onDropdownHide={this.handleMilestoneSelection} data={milestoneData}
    />);
    return super.renderToolBarCmp({ main: [milestonesSelect] });
  }
}
