/**
 * Created by farhad on 20.02.16.
 */
import React from 'react';
import _ from 'lodash';
import { reactFilterFactory } from 'ag-grid-react';
import FixedValuesFilter from '../components/FixedValuesFilter.jsx';
import ReportBox from './base/ReportBox';
import getStatusColumns, { statsDefaultFieldMap } from './base/CommonColumns.jsx';
import CheckBox from '../components/CheckBox';
import consts from '../consts';
import { getNodeValue } from './base/utils';

const DELETE_CHAR_CODE = 127;

const statsFieldMap = {
  ...statsDefaultFieldMap,
};

const primaryStatusesFieldMap = _.pick(statsFieldMap, _.keys(consts.testrail.Statuses));


function _renderRunLink(runID, text) {
  return `<a href="${consts.testrail.TESTRAIL_BASE_URL}/index.php?/runs/view/${runID}">${text}</a>`;
}

const getColumnDefs = () => ([
  {
    headerName: 'Teams and Runs',
    children: [
      {
        headerName: 'QA Team',
        field: 'qa_team_name',
        rowGroupIndex: 0,
        cellRenderer: {
          renderer: 'group',
        },
        valueGetter(params) {
          return commonStore.getTestRailQATeamName(
              params.data.qa_team) || consts.testrail.DEFAULT_TEAM_NAME;
        },
        filter: 'set',
        filterParams: {
          values: _.map(commonStore.testrail_qa_teams, 'name'),
          newRowsAction: 'keep',
        },
        cellStyle: { 'font-weight': 'bold' },
        pinned: true,
      },
      {
        headerName: 'ID',
        field: 'id',
        width: 70,
        filter: reactFilterFactory(FixedValuesFilter),
        filterParams: { mode: 'exclude' },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Plan Name',
        field: 'plan_name',
        width: 100,
        filter: 'text',
        filterParams: { newRowsAction: 'keep' },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Created On',
        field: 'plan_created_on',
        width: 160,
        filter: 'text',
        filterParams: { newRowsAction: 'keep' },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Run Name',
        field: 'name',
        filter: 'text',
        filterParams: { newRowsAction: 'keep' },
        rowGroupIndex: 1,
        width: 150,
        columnGroupShow: 'open',
        cellRenderer(params) {
          const run_id = getNodeValue(params, 'id');
          if (params.value && run_id) {
            return _renderRunLink(run_id, params.value);
          }
          return params.value;
        },
        pinned: true,
      },
      {
        headerName: 'Type',
        field: 'type_name',
        filter: 'set',
        width: 80,
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Is last',
        hide: true,
        field: 'is_last',
        filter: 'set',
        filterParams: {
          values: [true, false],
          newRowsAction: 'keep',
        },
        width: 80,
        columnGroupShow: 'open',
        pinned: true,
      },
    ],
  },
  {
    headerName: 'Milestone',
    field: 'milestone_name',
    filter: 'set',
    filterParams: {
      values: _.map(commonStore.testrail_milestones, 'name'),
      newRowsAction: 'keep',
      comparator: (a, b) => (a < b ? 1 : -1),
    },
    width: 110,
    columnGroupShow: 'open',
  },
  ...getStatusColumns(statsFieldMap),
  {
    headerName: 'Case Types',
    headerClass: 'header-center-aligned',
    children: [
      {
        headerName: 'Manual',
        field: `casetype_${consts.testrail.CaseTypes.MANUAL_ID}`,
        filter: 'number',
        filterParams: { newRowsAction: 'keep' },
        aggFunc: 'sum',
        width: 110,
      },
      {
        headerName: 'Automation in Progress',
        field: `casetype_${consts.testrail.CaseTypes.AUTOMATION_IN_PROGRESS_ID}`,
        aggFunc: 'sum',
        filter: 'number',
        filterParams: { newRowsAction: 'keep' },
        width: 180,
      },
      {
        headerName: 'Automated',
        field: `casetype_${consts.testrail.CaseTypes.AUTOMATED_ID}`,
        filter: 'number',
        filterParams: { newRowsAction: 'keep' },
        aggFunc: 'sum',
        width: 110,
      },
    ],
  },

]);


export default class extends ReportBox {

  static defaultProps = Object.assign({}, ReportBox.defaultProps, {
    columnDefs: getColumnDefs,
    title: 'Custom Test Runs Report',
    csvExportFilename: 'testrail_run_statuses.csv',
    gridDataUrl: consts.testrail.RUNS_STATUSES_RESOURCE,
    defaultSortModel: [
      { colId: 'qa_team_name', sort: 'asc' },
      { colId: 'id', sort: 'desc' },
    ],
  });

  constructor(props) {
    super(props);

    this.onToggleLatestRuns = this.onToggleLatestRuns.bind(this);
    this.onGridKeyPress = this.onGridKeyPress.bind(this);
  }

  getColumnsForChart() {
    const fieldNames = [statsFieldMap.TOTAL, ..._.values(primaryStatusesFieldMap)];
    return _(this.columnApi.getAllColumns())
      .filter((c) => (c.visible && _.includes(fieldNames, c.colDef.field)))
      .sortBy('left')
      .value();
  }

  onToggleLatestRuns(e) {
    const filterApi = this.gridApi.getFilterApi('is_last');
    if (e.target.checked) {
      filterApi.selectNothing();
      filterApi.selectValue(true);
    } else {
      filterApi.selectEverything();
    }
    this.gridApi.onFilterChanged();
    this.forceUpdate();
  }

  onGridKeyPress(e) {
    if (e.charCode === DELETE_CHAR_CODE) {
      const runIds = _.map(this.gridApi.getSelectedNodes(), 'data.id');
      if (!_.isEmpty(runIds)) {
        const filterApi = this.gridApi.getFilterApi('id');
        const model = filterApi.getModel() || { content: '', mode: 'exclude' };
        model.content = _.filter([model.content, ...runIds]).join();
        filterApi.setModel(model);
        this.gridApi.onFilterChanged();
      }
    }
  }

  renderGridCmp() {
    return React.cloneElement(super.renderGridCmp(), {
      onKeyPress: this.onGridKeyPress,
    });
  }

  prepareGridData(data) {
    super.prepareGridData(data);

    _.each(data, (run) => {
      run[statsFieldMap.TOTAL] = 0;

      _.each(consts.testrail.Statuses, (fieldName, status) => {
        const gridColumnFieldName = statsFieldMap[status];
        const cnt = run.status_counts[fieldName] || 0;
        run[gridColumnFieldName] = cnt;
        run[statsFieldMap.TOTAL] += cnt;
      });

      _.each(consts.testrail.CaseTypes, (caseTypeId) => {
        run[`casetype_${caseTypeId}`] = (run.casetype_counts || {})[caseTypeId] || 0;
      });

      run.type_name = run.run_type_name || 'Unknown';
      run.milestone_name = run.milestone.name;
      run.plan_name = run.plan.name;
      run.plan_created_on = moment(run.plan.created_on).format(consts.MOMENT_DATE_TIME_FORMAT);
    });

    _(data).groupBy('name').each((runsGroup) => {
      const lastRunId = _.maxBy(runsGroup, 'id').id;
      _.each(runsGroup, (run) => { run.is_last = run.id === lastRunId; });
    });

    return data;
  }

  renderToolBarCmp() {
    let latestRunsOnly = false;
    if (this.gridApi) {
      latestRunsOnly = _.isEqual(this.gridApi.getFilterApi('is_last').getModel(), ['true']);
    }

    const latestRunsOnlyTogglebutton = (
      <CheckBox label=" Latest Runs Only"
        key="latestRunsOnly"
        checked={latestRunsOnly}
        onChange={this.onToggleLatestRuns}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    return super.renderToolBarCmp({ main: latestRunsOnlyTogglebutton });
  }
}
