import React from 'react';
import _ from 'lodash';
import { reactFilterFactory } from 'ag-grid-react';
import moment from 'moment';

import FixedValuesFilter from '../components/FixedValuesFilter';
import ReportBox from './base/ReportBox';
import getStatusColumns, { statsDefaultFieldMap } from './base/CommonColumns';
import CheckBox from '../components/CheckBox';
import consts from '../consts';
import { getNodeValue, isEqual, safeDiv } from './base/utils';
import SimpleRangePicker from '../components/SimpleRangePicker';
import TempestChart from './D3Charts/TempestChart';

const DELETE_CHAR_CODE = 127;

const statsFieldMap = {
  ...statsDefaultFieldMap,
};

const primaryStatusesFieldMap = _.pick(statsFieldMap, _.keys(consts.testrail.Statuses));


function _renderRunLink(runID, text) {
  return `<a href="${consts.testrail.TESTRAIL_BASE_URL}/index.php?/runs/view/${runID}">${text}</a>`;
}

const getColumnDefs = () => ([
  {
    headerName: 'Teams and Runs',
    children: [
      {
        headerName: 'Date',
        field: 'date',
        rowGroupIndex: 0,
        cellRenderer: {
          renderer: 'group',
        },
        width: 160,
        filterParams: { newRowsAction: 'keep' },
        cellStyle: { 'font-weight': 'bold' },
        pinned: true,
      },
      {
        headerName: 'QA Team',
        field: 'qa_team_name',
        valueGetter(params) {
          return commonStore.getTestRailQATeamName(
              params.data.qa_team) || consts.testrail.DEFAULT_TEAM_NAME;
        },
        filter: 'set',
        filterParams: {
          values: _.map(commonStore.testrail_qa_teams, 'name'),
          newRowsAction: 'keep',
        },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'ID',
        field: 'id',
        width: 70,
        filter: reactFilterFactory(FixedValuesFilter),
        filterParams: { mode: 'exclude' },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Plan Name',
        field: 'plan_name',
        width: 100,
        filter: 'text',
        filterParams: { newRowsAction: 'keep' },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Created On',
        field: 'plan_created_on',
        width: 160,
        filter: 'text',
        filterParams: { newRowsAction: 'keep' },
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Run Name',
        field: 'name',
        filter: 'text',
        filterParams: { newRowsAction: 'keep' },
        width: 150,
        columnGroupShow: 'open',
        cellRenderer(params) {
          const run_id = getNodeValue(params, 'id');
          if (params.value && run_id) {
            return _renderRunLink(run_id, params.value);
          }
          return params.value;
        },
        pinned: true,
      },
      {
        headerName: 'Type',
        field: 'type_name',
        filter: 'set',
        width: 80,
        columnGroupShow: 'open',
        pinned: true,
      },
      {
        headerName: 'Is last',
        hide: true,
        field: 'is_last',
        filter: 'set',
        filterParams: {
          values: [true, false],
          newRowsAction: 'keep',
        },
        width: 80,
        columnGroupShow: 'open',
        pinned: true,
      },
    ],
  },
  {
    headerName: 'Milestone',
    field: 'milestone_name',
    filter: 'set',
    filterParams: {
      values: _.map(commonStore.testrail_milestones, 'name'),
      newRowsAction: 'keep',
      comparator: (a, b) => (a < b ? 1 : -1),
    },
    width: 110,
    columnGroupShow: 'open',
  },
  ...getStatusColumns(statsFieldMap),
]);


export default class extends ReportBox {

  static defaultProps = Object.assign({}, ReportBox.defaultProps, {
    columnDefs: getColumnDefs,
    title: 'Runs Report by Date',
    chartHeight: 800,
    csvExportFilename: 'testrail_run_statuses.csv',
    gridDataUrl: consts.testrail.RUNS_STATUSES_RESOURCE,
    defaultSortModel: [
      { colId: 'date', sort: 'desc' },
    ],
  });

  constructor(props) {
    super(props);

    this.state.datesRange = [
      moment().subtract(6, 'days').startOf('day'),
      moment().endOf('day'),
    ];
    this.state.minimise = false;

    this.onGridKeyPress = this.onGridKeyPress.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
    this.onToggleMinimiseChart = this.onToggleMinimiseChart.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    if (!isEqual(this.state.datesRange, prevState.datesRange)) {
      this.fetchGridData();
    }
  }

  getColumnsForChart() {
    const fieldNames = [statsFieldMap.TOTAL, ..._.values(primaryStatusesFieldMap)];
    return _(this.columnApi.getAllColumns())
      .filter((c) => (c.visible && _.includes(fieldNames, c.colDef.field)))
      .sortBy('left')
      .value();
  }

  onGridKeyPress(e) {
    if (e.charCode === DELETE_CHAR_CODE) {
      const runIds = _.map(this.gridApi.getSelectedNodes(), 'data.id');
      if (!_.isEmpty(runIds)) {
        const filterApi = this.gridApi.getFilterApi('id');
        const model = filterApi.getModel() || { content: '', mode: 'exclude' };
        model.content = _.filter([model.content, ...runIds]).join();
        filterApi.setModel(model);
        this.gridApi.onFilterChanged();
      }
    }
  }

  renderGridCmp() {
    return React.cloneElement(super.renderGridCmp(), {
      onKeyPress: this.onGridKeyPress,
    });
  }

  onDatesRangeChange(newRange) {
    this.setStateIf({ datesRange: newRange });
  }

  onToggleMinimiseChart(e) {
    this.setState({ minimise: e.target.checked });
  }


  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
    })
      .done((date) => {
        this.updateGridData(date);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }


  prepareGridData(data) {
    super.prepareGridData(data);

    _.each(data, (run) => {
      run[statsFieldMap.TOTAL] = 0;

      _.each(consts.testrail.Statuses, (fieldName, status) => {
        const gridColumnFieldName = statsFieldMap[status];
        const cnt = run.status_counts[fieldName] || 0;
        run[gridColumnFieldName] = cnt;
        run[statsFieldMap.TOTAL] += cnt;
      });

      _.each(consts.testrail.CaseTypes, (caseTypeId) => {
        run[`casetype_${caseTypeId}`] = (run.casetype_counts || {})[caseTypeId] || 0;
      });

      run.type_name = run.run_type_name || 'Unknown';
      run.milestone_name = run.milestone.name;
      run.plan_name = run.plan.name;
      run.plan_created_on = moment(run.plan.created_on).format(consts.MOMENT_DATE_TIME_FORMAT);
      run.date = moment(run.created_on).format('YYYY-MM-DD');
    });

    _(data).groupBy('name').each((runsGroup) => {
      const lastRunId = _.maxBy(runsGroup, 'id').id;
      _.each(runsGroup, (run) => { run.is_last = run.id === lastRunId; });
    });

    return data;
  }

  getChartDatum() {
    const rowData = this.state.rowData;
    if (_.isEmpty(rowData)) { return []; }

    const dates = {};
    const model = this.gridApi.getModel();
    model.forEachNodeAfterFilterAndSort((node) => {
      if (node.group && node.level === 0) {
        dates[node.key] = _.get(node, 'data');
        dates[node.key].plan_name = node.key;
      }
    });
    const statusResult = _.values(consts.testrail.Statuses);
    const statusColumns = _(this.columnApi.getAllColumns()).filter((obj) => {
      function isStatus(status) {
        if (obj.colDef.field === 'pct_blocked') { return false; }
        if (obj.colDef.field === 'cnt_total_failed') { return false; }
        return obj.colDef.field.indexOf(status) > -1;
      }
      return statusResult.some(isStatus);
    }).value();
    const succesRateCol = {
      key: '%Success Rate',
      type: 'line',
      yAxis: 2,
      values: _(dates).mapValues((v, k) => ({
        x: moment(k), y: safeDiv(v.cnt_passed, v.cnt_total),
      })).map().sortBy('x')
      .value(),
      color: '#444141',
    };
    const result = _(statusColumns).map(obj => ({
      key: obj.colDef.headerName,
      type: 'bar',
      yAxis: 1,
      values: _(dates).mapValues((v, k) => ({
        x: moment(k),
        y: v[obj.colDef.field],
        plan_name: v.plan_name,
      })).map().sortBy('x')
      .value(),
      color: obj.colDef.chart.color,
    })).value();
    return result.concat(succesRateCol);
  }

  renderChartCmp() {
    const datum = this.getChartDatum();
    return (
      <TempestChart
        height={this.props.chartHeight}
        minimise={this.state.minimise}
        datum={datum}
      />
    );
  }

  renderToolBarCmp() {
    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        onChange={this.onDatesRangeChange}
        startDate={this.state.datesRange[0]}
        endDate={this.state.datesRange[1]}
      />
    );

    const checkBox = (
      <CheckBox
        label=" Minimise chart"
        key="show.minim"
        checked={this.state.minimise}
        onChange={this.onToggleMinimiseChart}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    return super.renderToolBarCmp({ main: [datePicker, checkBox] });
  }
}
