import React from 'react';
import _ from 'lodash';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import ReportBox from './base/ReportBox';
import AutomationChart from './D3Charts/AutomationChart';
import SimpleRangePicker from '../components/SimpleRangePicker';
import CheckBox from '../components/CheckBox';
import consts from '../consts';
import './base/multiChartExtended';
import { getComponentsSum, toFixedCelRenderer, safeDiv,
         makeDatesSequence, getNodeValue, isEqual } from './base/utils';


function renderSuiteLink(suiteID, text) {
  return `<a href="${consts.testrail.TESTRAIL_BASE_URL}/index.php?/suites/view/${suiteID}">${text}</a>`;
}

function getRanges() {
  return {
    'Last 7 Days': [
      moment().subtract(6, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 30 Days': [
      moment().subtract(29, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 3 Month': [
      moment().subtract(2, 'month').startOf('month'),
      moment(),
    ],
  };
}

const getColumnDefs = () => ([
  {
    headerName: 'Suites',
    field: '__all_teams__',
    rowGroupIndex: 0,
    valueGetter: () => 'Suites',
    hide: true,
  },
  {
    headerName: 'Suite name',
    rowGroupIndex: 1,
    field: 'suite_name',
    filter: 'set',
    cellRenderer(params) {
      const suiteId = getNodeValue(params, 'suite_id');
      if (params.value && suiteId) {
        return renderSuiteLink(suiteId, params.value);
      }
      return params.value;
    },
    width: 110,
  },
  {
    headerName: 'Milestone',
    field: 'milestone_name',
    filter: 'set',
    width: 110,
  },
  {
    headerName: 'Sections',
    field: 'section_name',
    cellRenderer: {
      renderer: 'group',
    },
    cellStyle: { 'font-weight': 'bold' },
    filter: 'set',
    width: 200,
    pinned: true,
  },

  {
    headerName: 'Manual',
    field: consts.testrail.FIELD_MANUAL,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    aggFunc: 'sum',
    chart: { color: '#E06666', type: 'bar', yAxis: 1 },
    width: 110,
  },
  {
    headerName: 'Automation in progress',
    field: consts.testrail.FIELD_AUTOMATION_IN_PROGRESS,
    aggFunc: 'sum',
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#A6D49D', type: 'bar', yAxis: 1 },
    width: 180,
  },
  {
    headerName: 'Automated',
    field: consts.testrail.FIELD_AUTOMATED,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    aggFunc: 'sum',
    chart: { color: '#69BA69', type: 'bar', yAxis: 1 },
    width: 110,
  },
  {
    headerName: 'Will not Automate',
    width: 150,
    field: consts.testrail.FIELD_WILL_NOT_AUTOMATE,
    filter: 'number',
    aggFunc: 'sum',
    chart: { color: '#FFFF00', type: 'bar', yAxis: 1 },
  },
  {
    headerName: 'CI-integration',
    width: 150,
    field: consts.testrail.FIELD_CI_INTEGRATION,
    filter: 'number',
    aggFunc: 'sum',
    chart: { color: '#FFFF00', type: 'bar', yAxis: 1 },
  },
  {
    headerName: '% Automated',
    field: consts.testrail.FIELD_PCT_AUTOMATED,
    valueGetter(params) {
      return 100 * safeDiv(
          params.getValue(consts.testrail.FIELD_AUTOMATED),
          getComponentsSum([
            consts.testrail.FIELD_MANUAL,
            consts.testrail.FIELD_AUTOMATION_IN_PROGRESS,
            consts.testrail.FIELD_AUTOMATED,
            consts.testrail.FIELD_CI_INTEGRATION,
          ], params),
        );
    },
    cellRenderer: toFixedCelRenderer,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
    width: 120,
  },
  {
    headerName: '% CI-Integration',
    field: consts.testrail.FIELD_PCT_AUTOMATED,
    valueGetter(params) {
      return 100 * safeDiv(
          params.getValue(consts.testrail.FIELD_CI_INTEGRATION),
          getComponentsSum([
            consts.testrail.FIELD_MANUAL,
            consts.testrail.FIELD_AUTOMATION_IN_PROGRESS,
            consts.testrail.FIELD_AUTOMATED,
            consts.testrail.FIELD_CI_INTEGRATION,
          ], params),
        );
    },
    cellRenderer: toFixedCelRenderer,
    filter: 'number',
    filterParams: { newRowsAction: 'keep' },
    chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
    width: 120,
  },
]);

export default class SuiteProgress extends ReportBox {

  static defaultProps = {
    ...ReportBox.defaultProps,
    columnDefs: getColumnDefs,
    chartHeight: 600,
    title: 'TestRail: Suite Statistic',
    csvExportFilename: 'testrail_adoption_progress.csv',
    gridDataUrl: consts.testrail.SUITE_PROGRESS_RESOURCE,
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
      groupDefaultExpanded: 1,
    },
  };

  constructor(props) {
    super(props);
    this.state.selectedProject = this.initialConfiguration.selectedProject;
    this.state.datesRange = this.initialConfiguration.datesRange.map(d => moment(d));
    this.state.intervalType = this.initialConfiguration.intervalType;
    this.state.minimise = false;
    this.state.chartData = [];

    this.onProjectChange = this.onProjectChange.bind(this);
    this.onDatesRangeChange = this.onDatesRangeChange.bind(this);
    this.onToggleMinimiseChart = this.onToggleMinimiseChart.bind(this);
    this.onIntervalTypeChange = this.onIntervalTypeChange.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedProject', 'intervalType'];
    if (!isEqual(_.pick(this.state, fetchParams), _.pick(prevState, fetchParams))) {
      this.fetchGridData();
    }
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      datesRange: getRanges()['Last 7 Days'].map(d => d.toJSON()),
      selectedProject: commonStore.testrail_projects[7].id,
      intervalType: 'days',
    };
  }

  getConfiguration() {
    const conf = super.getConfiguration();
    return {
      ...conf,
      datesRange: this.state.datesRange.map(d => d.toJSON()),
      selectedProject: this.state.selectedProject,
      intervalType: this.state.intervalType,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf(_.pick(conf, ['selectedProject', 'intervalType']));
    if (conf.datesRange) {
      this.setStateIf({ datesRange: conf.datesRange.map(d => moment(d)) });
    }
  }

  onProjectChange() {
    const selectedProject = parseInt(this.trProject.$multiselect.find(':selected').val(), 10);
    this.setState({ selectedProject });
  }

  getChartDates() {
    return makeDatesSequence(
      moment(this.state.datesRange[0]).endOf('day'),
      moment(this.state.datesRange[1]).endOf('day'),
      this.state.intervalType,
    );
  }

  prepareGridData(data) {
    const chartData = [];
    const gridData = [];

    const chartDates = _.fromPairs(this.getChartDates().map(v => [+v, true]));

    _.each(data, (obj) => {
      obj.date = new Date(obj.date);
      Object.assign(obj, obj.type_counts);

      if (+this.state.datesRange[1] === +obj.date) {
        obj.total = _.sum(_.values(obj.type_counts)) || 0;
        gridData.push(obj);
      }
      if (chartDates[+obj.date]) {
        chartData.push(obj);
      }
    });

    this.setState({ chartData });
    return gridData;
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    const endOfToday = moment(this.state.datesRange[1]).endOf('day');
    const requestDates = [...this.getChartDates(), endOfToday];


    return $.get(this.props.gridDataUrl, {
      dates: requestDates.map(d => d.toISOString()).join(),
      project_id: this.state.selectedProject,
    }).done((data) => {
      this.updateGridData(data);
    })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  onDatesRangeChange(newRange) {
    this.setStateIf({ datesRange: newRange });
  }

  onToggleMinimiseChart(e) {
    this.setState({ minimise: e.target.checked });
  }

  onIntervalTypeChange() {
    this.setStateIf({ intervalType: this.intervalTypeSelect.$multiselect.val() });
  }

  getChartDatum() {
    if (!this.gridApi || _.isEmpty(this.state.rowData)) {
      return [];
    }

    const milestonesFilter = this.gridApi.getFilterApi('milestone_name');
    const selectionName = this.gridApi.getFilterApi('section_name');
    const suiteName = this.gridApi.getFilterApi('suite_name');

    const teamsSelection = {};
    window.commonStore.testrail_qa_teams.forEach((team) => {
      teamsSelection[team.id] = false;
    });

    const grouped = _(this.state.chartData)
      .filter(obj => (
        (suiteName.isEverythingSelected() ||
        _.includes(suiteName.getModel(), obj.suite_name)) &&
        (milestonesFilter.isEverythingSelected() ||
        _.includes(milestonesFilter.getModel(), obj.milestone_name)) &&
        (selectionName.isEverythingSelected() ||
        _.includes(selectionName.getModel(), obj.section_name))
      ))
      .groupBy('date')
      .values()
      .sortBy('date')
      .value();

    function getValues(field) {
      return _.map(grouped, rows => ({
        x: moment(rows[0].date),
        y: _.sumBy(rows, field),
      }));
    }

    return [
      {
        key: 'Manual',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_MANUAL),
        color: '#E06666',
      },
      {
        key: 'Will not Automate',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_WILL_NOT_AUTOMATE),
        color: '#FFFF00',
      },
      {
        key: 'Automation in Progress',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_AUTOMATION_IN_PROGRESS),
        color: '#A6D49D',
      },
      {
        key: 'Automated',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_AUTOMATED),
        color: '#69BA69',
      },
      {
        key: 'CI-Integration',
        type: 'bar',
        yAxis: 1,
        values: getValues(consts.testrail.FIELD_CI_INTEGRATION),
        color: '#1BCABB',
      },
      {
        key: '% CI-Integration',
        type: 'line',
        yAxis: 2,
        values: _.map(grouped, (rows) => {
          const date = rows[0].date;
          const cntM = _.sumBy(rows, consts.testrail.FIELD_MANUAL);
          const cntAiP = _.sumBy(rows, consts.testrail.FIELD_AUTOMATION_IN_PROGRESS);
          const cntAut = _.sumBy(rows, consts.testrail.FIELD_AUTOMATED);
          const cntCi = _.sumBy(rows, consts.testrail.FIELD_CI_INTEGRATION);
          return {
            x: date,
            y: cntCi / (cntM + cntAiP + cntAut + cntCi),
          };
        }),
      },
      {
        key: '% Automated',
        type: 'line',
        yAxis: 2,
        values: _.map(grouped, (rows) => {
          const date = rows[0].date;
          const cntM = _.sumBy(rows, consts.testrail.FIELD_MANUAL);
          const cntAiP = _.sumBy(rows, consts.testrail.FIELD_AUTOMATION_IN_PROGRESS);
          const cntAut = _.sumBy(rows, consts.testrail.FIELD_AUTOMATED);
          const cntCi = _.sumBy(rows, consts.testrail.FIELD_CI_INTEGRATION);
          return {
            x: date,
            y: cntAut / (cntM + cntAiP + cntAut + cntCi),
          };
        }),
      },
    ];
  }

  renderChartCmp() {
    const datum = this.getChartDatum();

    return (
      <AutomationChart
        minimise={this.state.minimise}
        height={this.props.chartHeight}
        width={400}
        datum={datum}
        integration
        margin={{ top: 30, right: 50, bottom: 195, left: 255 }}
      />
    );
  }

  renderToolBarCmp() {
    const datePicker = (
      <SimpleRangePicker
        key="dates.pick"
        ranges={getRanges()}
        onChange={this.onDatesRangeChange}
        startDate={this.state.datesRange[0]}
        endDate={this.state.datesRange[1]}
      />
    );

    const intervalTypeChoices = ['days', 'weeks', 'months'].map(
      v => ({ value: v, selected: v === this.state.intervalType }),
    );

    const intervalTypeSelect = (
      <Multiselect
        key="intervalType.select"
        ref={(c) => { this.intervalTypeSelect = c; }}
        onChange={this.onIntervalTypeChange}
        data={intervalTypeChoices}
      />
    );

    const projectChoices = commonStore.testrail_projects.map(
      v => ({ value: v.id, label: v.name, selected: v.id === this.state.selectedProject }),
    );

    const trProject = (
      <Multiselect
        key="trProject.select"
        ref={(c) => { this.trProject = c; }}
        onChange={this.onProjectChange}
        data={projectChoices}
      />
    );
    const checkBox = (
      <CheckBox
        label=" Minimise chart"
        key="show.minim"
        checked={this.state.minimise}
        onChange={this.onToggleMinimiseChart}
        groupClassName="checkbox-inline"
        togglebutton
      />
    );

    return super.renderToolBarCmp({ main: [datePicker, intervalTypeSelect, trProject, checkBox] });
  }
}
