import React from 'react';
import _ from 'lodash';
import * as BS from 'react-bootstrap';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import DatesRangeReportBox from './base/DatesRangeReportBox';
import VerificationProgressChart from './D3Charts/VerificationProgressChart';
import consts from '../consts';
import './base/multiChartExtended';
import { toFixedCelRenderer, safeDiv, isEqual, getDefaultDatesRanges, isCurrentWeek } from './base/utils';

const statusColors = {
  FIX_RELEASED: '#69BA69',
  FIX_COMMITTED: '#E06666',
  OPEN: '#5DA4B6',
  DEV_TO_VERIFY: '#8062A0',
};

function getSubColumns(prefix) {
  const width = 100;
  return [
    {
      headerName: 'Done',
      field: `${prefix}_done`,
      filter: 'number',
      cellStyle: { color: statusColors.FIX_RELEASED },
      width,
    },
    {
      headerName: 'Fixed',
      field: `${prefix}_fixed`,
      filter: 'number',
      cellStyle: { color: statusColors.FIX_COMMITTED },
      width,
    },
    {
      headerName: 'Open',
      field: `${prefix}_open`,
      filter: 'number',
      cellStyle: { color: statusColors.OPEN },
      width,
    },
    {
      headerName: 'Total',
      valueGetter: params => params.data[`${prefix}_done`] + params.data[`${prefix}_fixed`] + params.data[`${prefix}_open`],
      filter: 'number',
      cellStyle: { color: statusColors.OPEN },
      width: 70,
    },
    {
      headerName: 'Verification rate',
      field: `${prefix}_verification_rate_pct`,
      valueGetter(params) {
        return 100 * params.data[(`${prefix}_verification_rate`)];
      },
      cellRenderer: toFixedCelRenderer,
      filter: 'number',
      chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
      width: 60,
    },
  ];
}

const columnDefs = [
  {
    headerName: 'Date',
    field: 'date',
    cellRenderer: (params) => {
      if (isCurrentWeek(params.data.date)) {
        return 'Current';
      }
      return params.data.date.format('YYYY-MM-DD');
    },
    valueGetter(p) {
      return p.data.date && p.data.date.format('YYYY-MM-DD');
    },
    filter: 'text',
    filterParams: {
      newRowsAction: 'keep',
    },
    suppressSizeToFit: true,
    cellStyle: { 'font-weight': 'bold' },
    width: 120,
    pinned: true,
  },
  {
    headerName: 'All',
    headerClass: 'header-center-aligned',
    children: getSubColumns('all'),
  },
  {
    headerName: 'Blocker',
    headerClass: 'header-center-aligned',
    children: getSubColumns('blocker'),
  },
  {
    headerName: 'Critical',
    headerClass: 'header-center-aligned',
    children: getSubColumns('crit'),
  },
  {
    headerName: 'Major',
    headerClass: 'header-center-aligned',
    children: getSubColumns('major'),
  },
];

export default class extends DatesRangeReportBox {

  static defaultProps = {
    ...DatesRangeReportBox.defaultProps,
    columnDefs,
    getDefaultDatesRanges,
    title: 'Jira: Bugs Verification Progress by Fix Version',
    csvExportFilename: 'jira_bugs_verification_progress.csv',
    gridDataUrl: consts.jira.VERIFICATION_TCP_MCP_BUGS,
    lastSyncInfoUrl: consts.launchpad.LAST_SYNC_INFO_URL,
    chartHeight: 400,
    defaultSortModel: [{ colId: 'date', sort: 'desc' }],
  };

  constructor(props) {
    super(props);
    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedMilestones'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedMilestones.push($(elem).val());
    });
    this.setStateIf({ selectedMilestones });
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: window.commonStore.jira_fix_versions,
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  prepareGridData(data) {
    function get_statuses_by(prefix, data) {
      const ret = {
        [`${prefix}_done`]: data.filter({ status: 'Done' }).sumBy('count'),
        [`${prefix}_fixed`]: data.filter({ status: 'Fixed' }).sumBy('count'),
        [`${prefix}_open`]: data.filter({ status: 'Open' }).sumBy('count'),
      };
      ret[`${prefix}_verification_rate`] = safeDiv(
        ret[`${prefix}_done`],
        ret[`${prefix}_done`] + ret[`${prefix}_fixed`]
      );
      return ret;
    }

    const preparedData = _(data).groupBy('date').map((values, date) => {
      const all = _(values);
      const blocker = _(values).filter({ priority: 'Blocker' });
      const crit = _(values).filter({ priority: 'Critical' });
      const major = _(values).filter({ priority: 'Major' });

      const ret = Object.assign(
        {
          date: moment(date),
        },
        get_statuses_by('all', all),
        get_statuses_by('blocker', blocker),
        get_statuses_by('crit', crit),
        get_statuses_by('major', major)
      );
      return ret;
    })
    .sortBy('date')
    .value();

    return preparedData;
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      milestones: this.state.selectedMilestones.map(d => d).join(),
    })
      .done((data) => {
        this.updateGridData(data);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  getChartDatum(prefix) {
    const rowData = this.state.rowData;
    if (!rowData) { return []; }

    function getValues(field) {
      return _.map(rowData, row => ({
        x: row.date,
        y: row[`${prefix}_${field}`] || 0,
      }));
    }

    return [
      {
        key: 'Done',
        type: 'bar',
        yAxis: 1,
        values: getValues('done'),
        color: statusColors.FIX_RELEASED,
      },
      {
        key: 'Fixed',
        type: 'bar',
        yAxis: 1,
        values: getValues('fixed'),
        color: statusColors.FIX_COMMITTED,
      },
      {
        key: 'Open',
        type: 'bar',
        yAxis: 1,
        values: getValues('open'),
        color: statusColors.OPEN,
      },
      {
        key: '% Verification rate',
        color: '#444141',
        type: 'line',
        values: getValues('verification_rate'),
        yAxis: 2,
      },
    ];
  }

  renderOneChartCmp(prefix, title) {
    const datum = this.getChartDatum(prefix);

    return (
     <VerificationProgressChart
       title={title}
       height={this.props.chartHeight}
       datum={datum}
       jira
     />
    );
  }

  renderChartCmp() {
    return (
      <div>
        <BS.Row>
          <BS.Col md={4}>{this.renderOneChartCmp('all', 'All')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('blocker', 'Blocker')}</BS.Col>
          <BS.Col md={4}>{this.renderOneChartCmp('crit', 'Critical')}</BS.Col>
        </BS.Row>
        <BS.Row>
          <BS.Col md={4}>{this.renderOneChartCmp('major', 'Major')}</BS.Col>
        </BS.Row>
      </div>
    );
  }

  renderToolBarCmp() {
    const datePicker = this.renderDatesRangePicker();

    const milestoneData = _.map(commonStore.jira_fix_versions, v => ({
      value: v, selected: _.includes(this.state.selectedMilestones, v),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        multiple
        buttonWidth="250px"
        maxHeight={300}
        includeSelectAllOption
        enableFiltering
        nonSelectedText="select milestone"
        allSelectedText="all selected"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );
    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect] });
  }

}
