import React from 'react';
import _ from 'lodash';
import * as BS from 'react-bootstrap';
import moment from 'moment';
import Multiselect from 'react-bootstrap-multiselect';

import DatesRangeReportBox from './base/DatesRangeReportBox';
import VerificationProgressChart from './D3Charts/VerificationProgressChart';
import consts from '../consts';
import './base/multiChartExtended';
import { toFixedCelRenderer, safeDiv, isEqual, getDefaultDatesRanges, isCurrentWeek } from './base/utils';

const statusColors = {
  FIX_RELEASED: '#69BA69',
  FIX_COMMITTED: '#E06666',
  OPEN: '#5DA4B6',
  DEV_TO_VERIFY: '#8062A0',
};

const teamSet = new Set();

function getGroupTeamCols(teamList) {
  const width = 100;
  const getSubColumns = (teamPrefix, type) =>
    [
      {
        headerName: 'Done',
        field: `${teamPrefix}_done_${type}`,
        filter: 'number',
        cellStyle: { color: statusColors.FIX_RELEASED },
        width,
      },
      {
        headerName: 'Fixed',
        field: `${teamPrefix}_fixed_${type}`,
        filter: 'number',
        cellStyle: { color: statusColors.FIX_COMMITTED },
        width,
      },
      {
        headerName: 'Open',
        field: `${teamPrefix}_open_${type}`,
        filter: 'number',
        cellStyle: { color: statusColors.OPEN },
        width,
      },
      {
        headerName: 'Verification rate',
        field: `${teamPrefix}_verification_rate_${type}`,
        valueGetter(params) {
          return 100 * params.data[`${teamPrefix}_verification_rate_${type}`];
        },
        cellRenderer: toFixedCelRenderer,
        filter: 'number',
        chart: { color: '#5DA4B6', type: 'line', yAxis: 2 },
        width: 60,
      },
    ]
  ;

  const groupStatus = teamPrefix =>
    [
      {
        headerName: 'All',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'all'),
      },
      {
        headerName: 'Blocker',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'blocker'),
      },
      {
        headerName: 'Critical',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'critical'),
      },
      {
        headerName: 'Major',
        headerClass: 'header-center-aligned',
        children: getSubColumns(teamPrefix, 'major'),
      },
    ]
  ;

  return _.map(teamList, v => ({
    headerName: v,
    children: groupStatus(_.snakeCase(v)),
    headerClass: 'header-center-aligned',
  }));
}

export default class extends DatesRangeReportBox {

  static defaultProps = {
    ...DatesRangeReportBox.defaultProps,
    getDefaultDatesRanges,
    title: 'Jira: Bugs Verification Progress by Teams',
    csvExportFilename: 'jira_bugs_verification_progress_teams.csv',
    gridDataUrl: consts.jira.VERIFICATION_TCP_MCP_BUGS_TEAMS,
    chartHeight: 500,
    lastSyncInfoUrl: '/api/v1.0/lp/sync/last',
    defaultSortModel: [{ colId: 'date', sort: 'desc' }],
  };

  constructor(props) {
    super(props);

    this.state.selectedMilestones = this.initialConfiguration.selectedMilestones;
    this.state.teams = [];
    this.state.allTeams = [];
    this.state.visibleTeams = [];

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
    this.handleTeamsSelection = this.handleTeamsSelection.bind(this);
  }

  componentDidUpdate(prevProps, prevState) {
    const fetchParams = ['datesRange', 'selectedMilestones'];
    if (!fetchParams.every(x => isEqual(this.state[x], prevState[x]))) {
      this.fetchGridData();
    }
    if (!isEqual(this.state.visibleTeams, prevState.visibleTeams)) {
      this.buildColumnDefs();
    }
  }

  handleMilestoneSelection() {
    const selectedMilestones = [];
    this.milestonesSelector.$multiselect.find(':selected').each((i, elem) => {
      selectedMilestones.push($(elem).val());
    });
    this.setStateIf({ selectedMilestones });
  }

  handleTeamsSelection() {
    const visibleTeams = [];
    this.teamsSelector.$multiselect.find(':selected').each((i, elem) => {
      visibleTeams.push($(elem).val());
    });
    this.setState({ visibleTeams });
  }

  getDefaultConfiguration() {
    return {
      ...super.getDefaultConfiguration(),
      selectedMilestones: window.commonStore.jira_fix_versions,
    };
  }

  getConfiguration() {
    return {
      ...super.getConfiguration(),
      selectedMilestones: this.state.selectedMilestones,
    };
  }

  setConfiguration(conf) {
    super.setConfiguration(conf);
    this.setStateIf({ selectedMilestones: conf.selectedMilestones });
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl, {
      date_from: this.state.datesRange[0].toISOString(),
      date_to: this.state.datesRange[1].toISOString(),
      milestones: this.state.selectedMilestones.map(d => d).join(),
    })
      .done((data) => {
        this.updateGridData(data);
      })
      .always(() => { this.gridApi.hideOverlay(); });
  }

  buildColumnDefs() {
    const teams = this.state.visibleTeams;
    if (!teams.length) {
      return;
    }
    const columnDefs = getGroupTeamCols(teams);
    const dateColumn = {
      headerName: 'Date',
      field: 'date',
      cellRenderer: (params) => {
        if (isCurrentWeek(params.data.date)) {
          return 'Current';
        }
        return params.data.date.format('YYYY-MM-DD');
      },
      valueGetter: params =>
        params.data.date && params.data.date.format('YYYY-MM-DD'),
      filter: 'text',
      filterParams: {
        newRowsAction: 'keep',
      },
      cellStyle: { 'font-weight': 'bold' },
      width: 100,
      pinned: true,
      sort: 'desc',
    };
    const teamColums = {
      headerName: 'Teams',
      children: columnDefs,
      headerClass: 'header-center-aligned',
    };

    this.setState({
      columnDefs: [
        dateColumn,
        teamColums,
      ],
    });
  }

  prepareGridData(data) {
    function getStats(teamPrefix, teamData) {
      const result = {};
      ['Fixed', 'Done', 'Open'].forEach((status) => {
        const statusData = _(teamData).filter({ status });

        const allSum = _(statusData).sumBy('count');
        const blockerSum = _(statusData).filter({ priority: 'Blocker' }).sumBy('count');
        const majorSum = _(statusData).filter({ priority: 'Major' }).sumBy('count');
        const criticalSum = _(statusData).filter({ priority: 'Critical' }).sumBy('count');

        result[`${teamPrefix}_${_.lowerCase(status)}_all`] = allSum;
        result[`${teamPrefix}_${_.lowerCase(status)}_critical`] = criticalSum;
        result[`${teamPrefix}_${_.lowerCase(status)}_blocker`] = blockerSum;
        result[`${teamPrefix}_${_.lowerCase(status)}_major`] = majorSum;
      });
      ['all', 'critical', 'blocker', 'major'].forEach((priority) => {
        result[`${teamPrefix}_verification_rate_${priority}`] = safeDiv(
          result[`${teamPrefix}_done_${priority}`],
          result[`${teamPrefix}_done_${priority}`] + result[`${teamPrefix}_fixed_${priority}`]
          );
      });
      return result;
    }


    const prepareGridData = _(data).groupBy('date').map((values, date) => {
      const teamsData = _(values).groupBy('team').map((teamVal, id) => {
        teamSet.add(id);
        const teamPrefix = _.snakeCase(id);
        return getStats(teamPrefix, teamVal);
      }).value();
      const result = Object.assign(
        {
          date: moment(date),
        },
        ...teamsData,
      );
      return result;
    }).sortBy('date')
      .value();
    this.setState({ allTeams: [...teamSet] });
    this.setState({ visibleTeams: [this.state.allTeams[0]] });
    this.buildColumnDefs();
    return prepareGridData;
  }

  renderToolBarCmp() {
    const datePicker = this.renderDatesRangePicker();

    const milestoneData = _.map(commonStore.jira_fix_versions, v => ({
      value: v, selected: _.includes(this.state.selectedMilestones, v),
    }));

    const teamData = _.map(this.state.allTeams, team => ({
      value: team, selected: _.includes(this.state.visibleTeams, team),
    }));

    const milestonesSelect = (
      <Multiselect
        key="milestone.select"
        ref={(c) => { this.milestonesSelector = c; }}
        multiple
        buttonWidth="250px"
        maxHeight={300}
        includeSelectAllOption
        enableFiltering
        nonSelectedText="select milestone"
        allSelectedText="all selected"
        onDropdownHide={this.handleMilestoneSelection}
        data={milestoneData}
      />
    );
    const teamSelect = (
      <Multiselect
        key="team.select"
        ref={(c) => { this.teamsSelector = c; }}
        multiple
        buttonWidth="250px"
        maxHeight={300}
        includeSelectAllOption
        enableFiltering
        nonSelectedText="select milestone"
        allSelectedText="all selected"
        onDropdownHide={this.handleTeamsSelection}
        data={teamData}
      />
    );
    return super.renderToolBarCmp({ main: [datePicker, milestonesSelect, teamSelect] });
  }

  getChartDatum(prefix, type) {
    const rowData = this.state.rowData;
    if (!rowData) { return []; }

    const getValues = field =>
      _.map(rowData, row => ({ x: row.date, y: row[`${_.snakeCase(prefix)}_${field}_${_.lowerCase(type)}`] || 0 }));

    return [
      {
        key: 'Done',
        type: 'bar',
        yAxis: 1,
        values: getValues('done'),
        color: statusColors.FIX_RELEASED,
      },
      {
        key: 'Fixed',
        type: 'bar',
        yAxis: 1,
        values: getValues('fixed'),
        color: statusColors.FIX_COMMITTED,
      },
      {
        key: 'Open',
        type: 'bar',
        yAxis: 1,
        values: getValues('open'),
        color: statusColors.OPEN,
      },
      {
        key: '% Verification rate',
        color: '#444141',
        type: 'line',
        values: getValues('verification_rate'),
        yAxis: 2,
      },
    ];
  }

  renderOneChartCmp(prefix, title) {
    const datum = this.getChartDatum(prefix, title);

    return (
      <VerificationProgressChart
        title={`${title} ${prefix}`}
        height={this.props.chartHeight}
        datum={datum}
        jira
      />
    );
  }

  renderChartCmp() {
    return (
      <div>
        {_.map(this.state.visibleTeams, team => (
          <div key={team}>
            <BS.Row>
              <BS.Col md={6}>{this.renderOneChartCmp(team, 'All')}</BS.Col>
              <BS.Col md={6}>{this.renderOneChartCmp(team, 'Blocker')}</BS.Col>
            </BS.Row>
            <BS.Row>
              <BS.Col md={6}>{this.renderOneChartCmp(team, 'Critical')}</BS.Col>
              <BS.Col md={6}>{this.renderOneChartCmp(team, 'Major')}</BS.Col>
            </BS.Row>
          </div>
        ))}
      </div>
    );
  }
}
