import React, { PropTypes } from 'react';
import _ from 'lodash';
import Multiselect from 'react-bootstrap-multiselect';
import * as BS from 'react-bootstrap';

import ReportBox, {
  getConfigurationFromUrl,
  setConfigurationToUrl,
 } from './ReportBox';
import CheckBox from '../../components/CheckBox';
import AgGridExtended from '../../components/AgGridExtended';
import LastSyncInfo from './LastSyncInfo';

const propTypes = {
  title: PropTypes.string.isRequired,
  rowData: PropTypes.shape({
    data: PropTypes.array,
    isFetching: PropTypes.bool,
  }).isRequired,
  columnDefs: PropTypes.array.isRequired,
  gridHeight: PropTypes.number.isRequired,
  refresh: PropTypes.func.isRequired,
  agGridOptions: PropTypes.any,

};


class FailureDriversTable extends React.Component {

  static defaultProps = {
    agGridOptions: {
      ...ReportBox.defaultProps.agGridOptions,
    },
  };

  constructor(props) {
    super(props);

    const conf = getConfigurationFromUrl();
    if (conf === null) {
      this.flag = true;
    } else {
      this.initialConfiguration = conf;
      this.flag = false;
    }

    this.state = {
      selectedMilestone: null,
      selectedIsoNames: null,
    };

    this.milestoneSelectorData = [];
    this.isoSelectorData = [];

    this.onGridReady = this.onGridReady.bind(this);
    this.onRefreshGrid = this.onRefreshGrid.bind(this);

    this.handleMilestoneSelection = this.handleMilestoneSelection.bind(this);
    this.handleIsoSelection = this.handleIsoSelection.bind(this);

    this.milestoneFilter = null;
    this.isoNamesFilter = null;

    this.isExternalFilterPresent = this.isExternalFilterPresent.bind(this);
    this.doesExternalFilterPass = this.doesExternalFilterPass.bind(this);
  }

  componentWillReceiveProps(nextProps) {
    if (nextProps.rowData.isFetching) {
      this.gridApi.showLoadingOverlay();
    } else {
      this.gridApi.hideOverlay();
    }
    if (!nextProps.rowData.isFetching && this.props.rowData.isFetching) {
      this.setDefaultMilestone(nextProps.rowData.data);
    }
  }

  componentWillUpdate(nextProps, nextState) {
    if (nextState.selectedMilestone !== this.state.selectedMilestone) {
      this.setIsoSelectorData(nextState.selectedMilestone, nextProps.rowData.data);
    }
  }

  componentDidUpdate(prevProps, prevState) {
    if (this.props.rowData.data.length) {
      if (this.props.defaultSortModel) {
        this.gridApi.setSortModel(this.props.defaultSortModel);
      }
      const conf = this.getConfiguration();
      if (this.state.selectedMilestone !== prevState.selectedMilestone
        || !_.isEqual(this.state.selectedIsoNames, prevState.selectedIsoNames)) {
        const conf = this.getConfiguration();
        setConfigurationToUrl(conf);
      }
    }
  }

  onGridReady({ api, columnApi }) {
    this.gridApi = api;
    this.columnApi = columnApi;
    this.gridApi.showLoadingOverlay();
  }

  onRefreshGrid() {
    if (typeof this.props.refresh === 'function') {
      this.props.refresh();
    }
  }

  setDefaultMilestone(data) {
    const milestones = _(data)
      .map('milestone_id')
      .uniq().value();

    this.milestoneSelectorData = _(window.commonStore.testrail_milestones)
    .filter(obj => milestones.indexOf(obj.id) !== -1)
    .map(obj => ({
      value: obj.id,
      label: obj.name,
      selected: false,
    }))
    .value();

    if (!this.flag) {
      this.setState({ selectedMilestone: this.initialConfiguration.selectedMilestone });
      return;
    }

    let defaultTrMilestoneId = _.head(
      window.commonStore.testrail_milestones.filter(
        m => m.name === window.commonStore.tr_milestone_in_focus.name,
      ));

    if (defaultTrMilestoneId) {
      defaultTrMilestoneId = defaultTrMilestoneId.id;
      if (milestones.indexOf(defaultTrMilestoneId) > -1) {
        this.setState({ selectedMilestone: defaultTrMilestoneId });
        return;
      }
    }

    this.setState({ selectedMilestone: _.head(milestones) });
  }

  setIsoSelectorData(milestoneId, data) {
    let selectedIsoNames;
    let initialMilestoneId;

    if (!this.flag) {
      selectedIsoNames = this.initialConfiguration.selectedIsoNames;
      initialMilestoneId = this.initialConfiguration.selectedMilestone;
    }

    this.isoSelectorData = _(data)
      .filter(item => item.milestone_id === (initialMilestoneId || milestoneId))
      .map('iso_name')
      .uniq()
      .sortBy(item => _.parseInt(_.last(_.split(item, '#'))))
      .reverse()
      .map(item => ({ value: item, selected: true }))
      .value();

    if (initialMilestoneId) {
      this.setState({ selectedIsoNames });
      this.flag = true;
      return;
    }

    this.setState({ selectedIsoNames: _(this.isoSelectorData).map('value').value() });
  }

  getDefaultConfiguration() {
    return {
      grid_filter: {},
      grid_sort: this.props.defaultSortModel || [],
      selectedMilestone: this.state.selectedMilestone,
      selectedIsoNames: this.state.selectedIsoNames,
    };
  }

  getConfiguration() {
    const filterModel = this.gridApi.getFilterModel();
    const sortModel = this.gridApi.getSortModel();
    return {
      grid_filter: filterModel,
      grid_sort: sortModel,
      selectedMilestone: this.state.selectedMilestone,
      selectedIsoNames: this.state.selectedIsoNames,
    };
  }

  getGridOptions() {
    return {
      onGridReady: this.onGridReady,
      columnDefs: this.props.columnDefs,
      rowData: this.props.rowData.data,
      rowSelection: 'multiple',
      enableSorting: true,
      enableFilter: true,
      enableColResize: true,
      rowHeight: 28,
      groupSuppressAutoColumn: true,
      groupSelectsChildren: true,
      groupUseEntireRow: false,
      isExternalFilterPresent: this.isExternalFilterPresent,
      doesExternalFilterPass: this.doesExternalFilterPass,
      ...this.props.agGridOptions,
    };
  }

  isExternalFilterPresent() {
    return this.milestoneFilter !== null || this.isoNamesFilter !== null;
  }

  doesExternalFilterPass(node) {
    if (this.milestoneFilter && !this.isoNamesFilter) {
      return node.data.milestone_id === this.milestoneFilter;
    }
    if (this.milestoneFilter && this.isoNamesFilter) {
      return this.isoNamesFilter.indexOf(node.data.iso_name) > -1
        && node.data.milestone_id === this.milestoneFilter;
    }
    return true;
  }

  handleMilestoneSelection() {
    const selectedMilestone = parseInt(
      this.milestonesSelector.$multiselect.find(':selected').val(),
      10,
    );
    this.setState({ selectedMilestone });
  }

  handleIsoSelection() {
    const selected = [];
    this.isoSelector.$multiselect
      .find('option:selected').each(function getVal() {
        selected.push($(this).val());
      });
    this.setState({ selectedIsoNames: selected });
  }

  render() {
    this.milestoneFilter = this.state.selectedMilestone;
    this.isoNamesFilter = this.state.selectedIsoNames;
    if (this.gridApi) {
      this.gridApi.onFilterChanged();
    }

    const getMilestones = data => data.map((item) => {
      if (item.value === this.state.selectedMilestone) {
        return {
          ...item,
          selected: true,
        };
      }
      return item;
    });

    const getIsos = data => data.map((item) => {
      if (!this.state.selectedIsoNames) {
        return {};
      }
      if (this.state.selectedIsoNames.indexOf(item.value) !== -1) {
        return {
          ...item,
          selected: true,
        };
      }
      return {
        ...item,
        selected: false,
      };
    });

    const milestones = getMilestones(this.milestoneSelectorData);
    const isos = getIsos(this.isoSelectorData);

    return (
      <BS.Panel header={<h3>{this.props.title}</h3>}>
        <BS.Panel footer={<LastSyncInfo module="lp" />}>
          <BS.Row>
            <BS.Col md={10}>
              <form className="form-inline" role="form">
                <BS.ButtonGroup bsSize="sm" bsStyle="raised">

                  <BS.Button bsStyle="info" bsSize="sm" onClick={this.onRefreshGrid}>
                    <i className="fa fa-refresh" /> Refresh
                  </BS.Button>

                  <BS.Button onClick={this.onExportCSV}>
                    <i className="fa fa-download" /> Download as CSV
                  </BS.Button>

                </BS.ButtonGroup>

                <Multiselect
                  key="milestone.select"
                  ref={(c) => { this.milestonesSelector = c; }}
                  buttonWidth="250px"
                  maxHeight={300}
                  enableFiltering
                  nonSelectedText="select milestone"
                  onDropdownHide={this.handleMilestoneSelection}
                  data={milestones}
                />

                <Multiselect
                  key="iso.select"
                  ref={(c) => { this.isoSelector = c; }}
                  multiple
                  buttonWidth="250px"
                  maxHeight={300}
                  enableFiltering
                  includeSelectAllOption
                  nonSelectedText="select iso"
                  allSelectedText="all selected"
                  disableIfEmpty
                  onDropdownHide={this.handleIsoSelection}
                  data={isos}
                />
              </form>

            </BS.Col>

            <BS.Col md={2}>
              <CheckBox
                label=" Tool Panel"
                groupClassName="pull-right"
                onChange={this.onToggleToolPanel}
              />
            </BS.Col>
          </BS.Row>

          <AgGridExtended
            style={{ height: this.props.gridHeight }}
            gridOptions={this.getGridOptions()}
          />

        </BS.Panel>
      </BS.Panel>
    );
  }
}


FailureDriversTable.propTypes = propTypes;

export default FailureDriversTable;
