/**
 * Created by farhad on 05.02.16.
 */
import React from 'react';
import _ from 'lodash';
import d3 from 'd3';
import moment from 'moment';
import rison from 'rison';
import { browserHistory } from 'react-router';
import * as BS from 'react-bootstrap';

import CheckBox from '../../components/CheckBox';
import AgGridExtended from '../../components/AgGridExtended';
import NVD3Chart from './NVD3Chart';
import { isEqual } from './utils';


const LAST_SYNC_INFO_URL = '/api/v1.0/testrail/sync/last';


export function getConfigurationFromUrl() {
  let conf = null;
  const hash = location.hash;

  if (hash && hash.startsWith('#conf=')) {
    try {
      conf = rison.decode_object(decodeURI(hash.slice(6)));
    } catch (Error) {
      console.warn('invalid config in hash string!');
    }
  }
  return conf;
}


export function setConfigurationToUrl(conf, delay = 0) {
  const encodedConfig = encodeURI(rison.encode_object(conf));
  const newHash = `#conf=${encodedConfig}`;
  if (location.hash != newHash) {
    browserHistory.push({
      pathname: location.pathname,
      search: location.search,
      hash: newHash,
    });
  }
}


export default class ReportBox extends React.Component {

  static propTypes = {
    title: React.PropTypes.node.isRequired,
    csvExportFilename: React.PropTypes.string,
    agGridOptions: React.PropTypes.object,
    defaultSortModel: React.PropTypes.arrayOf(React.PropTypes.object),
    columnDefs: React.PropTypes.any,
    gridDataUrl: React.PropTypes.string,
    gridHeight: React.PropTypes.number,
    chartHeight: React.PropTypes.number,
    data: React.PropTypes.array,
  };

  static defaultProps = {
    title: 'Report',
    csvExportFilename: 'report.csv',
    defaultSortModel: null,
    gridHeight: 410,
    chartHeight: 500,
    agGridOptions: {
      icons: {
        menu: '<i class="fa fa-bars"></i>',
        filter: '<i class="fa fa-filter"></i>',
        sortAscending: '<i class="fa fa-sort-amount-asc"></i>',
        sortDescending: '<i class="fa fa-sort-amount-desc"></i>',
        sortUnSort: '<i class="fa fa-sort"></i>',
        groupExpanded: '<i class="fa fa-minus-square-o"></i>',
        groupContracted: '<i class="fa fa-plus-square-o"></i>',
        columnGroupOpened: '<i class="fa fa-minus-square"/>',
        columnGroupClosed: '<i class="fa fa-plus-square"/>',
        columnVisible: '<i class="fa fa-check-square"></i>',
        columnHidden: '<i class="fa fa-square"></i>',
        columnRemoveFromGroup: '<i class="fa fa-times"></i>',
      },
    },
  };

  constructor(props) {
    super(props);

    let columnDefs = this.props.columnDefs;
    if (_.isFunction(columnDefs)) {
      columnDefs = columnDefs();
    }
    this.title = 'Grid Visualization';

    this.initialConfiguration = getConfigurationFromUrl() || this.getDefaultConfiguration();

    this.initialRowData = this.props.data || [];

    this.state = {
      rowData: this.initialRowData,
      columnDefs,
      showToolPanel: false,
      isChartPanelOpen: true,
    };

    this.onToggleChart = this.onToggleChart.bind(this);
    this.onToggleToolPanel = this.onToggleToolPanel.bind(this);
    this.onGridReady = this.onGridReady.bind(this);
    this.onRefreshGrid = this.onRefreshGrid.bind(this);
    this.onExportCSV = this.onExportCSV.bind(this);
    this.onGridConfChanged = this.onGridConfChanged.bind(this);

    this.debouncedUpdateComponent = _.debounce(this.debouncedUpdateComponent.bind(this), 250);
    this.debouncedSaveConfigurationToUrl =
      _.debounce(this.debouncedSaveConfigurationToUrl.bind(this), 800);
  }

  /**
   * This method is same as standard setState() but applies only changed
   * properties. It performs deep objects comparison so avoid use it for
   * modifying state of big nested objects or arrays.
   */
  setStateIf(state, ...args) {
    const updState = _.omitBy(state, (val, key) => isEqual(this.state[key], val));
    if (!_.isEmpty(updState)) {
      this.setState(updState, ...args);
    }
  }

  componentDidMount() {
    this.fetchGridData();
    this.setConfiguration(this.initialConfiguration);
  }

  componentWillReceiveProps(nextProps) {
    if (process.env.FRONTEND_DEV) {
      console.log('Will receive props!');
    }

    if (nextProps.data && nextProps.data !== this.state.rowData) {
      this.updateGridData(nextProps.data);
    }

    let conf = getConfigurationFromUrl();
    if (conf === null) {
      // if there is no any config then use default configuration
      conf = this.getDefaultConfiguration();
    }
    this.setConfiguration(conf);
  }

  /**
   * Returns default (JSON serializable) report configuration
   */
  getDefaultConfiguration() {
    return {
      grid_filter: {},
      grid_sort: this.props.defaultSortModel || [],
    };
  }

  /**
   * Returns current (JSON serializable) report configuration
   */
  getConfiguration() {
    const filterModel = this.gridApi.getFilterModel();
    const sortModel = this.gridApi.getSortModel();
    return {
      grid_filter: filterModel,
      grid_sort: sortModel,
    };
  }

  /**
   * This method should apply conf object to the report
   */
  setConfiguration(conf) {
    if (this.gridApi) {
      if (conf.grid_filter) {
        this._suppressGridEvents = true;

        // Workaround for set filter. Without it after grid reload filter
        // will be reset even if filterParam.newRowsAction set to 'keep'.
        _.each(conf.grid_filter, (obj, colId) => {
          const colFilterApi = this.gridApi.getFilterApi(colId);
          if (colFilterApi && _.isFunction(colFilterApi.selectNothing)) {
            colFilterApi.selectNothing();
          }
        });

        this.gridApi.setFilterModel(conf.grid_filter);
        this._suppressGridEvents = false;
      }

      if (conf.grid_sort) {
        this._suppressGridEvents = true;
        this.gridApi.setSortModel(conf.grid_sort);
        this._suppressGridEvents = false;
      }
    }
  }

  saveConfigurationToUrl() {
    setConfigurationToUrl(this.getConfiguration());
  }

  debouncedSaveConfigurationToUrl() {
    this.saveConfigurationToUrl();
  }

  onToggleToolPanel(e) {
    this.setState({ showToolPanel: e.target.checked });
  }

  onToggleChart() {
    this.setState({ isChartPanelOpen: !this.state.isChartPanelOpen });
  }

  onGridReady({ api, columnApi }) {
    this.gridApi = api;
    this.columnApi = columnApi;

    if (this.props.defaultSortModel) {
      this.gridApi.setSortModel(this.props.defaultSortModel);
    }

    this.gridApi.addEventListener('modelUpdated', this.onGridConfChanged);
    this.gridApi.addEventListener('columnVisible', this.onGridConfChanged);
    this.gridApi.addEventListener('columnRowGroupChanged', this.onGridConfChanged);
    this.gridApi.addEventListener('afterFilterChanged', this.onGridConfChanged);
    this.gridApi.addEventListener('cellValueChanged', this.onGridConfChanged);
    this.gridApi.addEventListener('columnMoved', this.onGridConfChanged);
    this.gridApi.addEventListener('columnValueChanged', this.onGridConfChanged);

    if (process.env.FRONTEND_DEV) {
      window.ga = this.gridApi;
      window.ca = this.columnApi;
    }
  }

  debouncedUpdateComponent() {
    this.forceUpdate();
  }

  onGridConfChanged() {
    if (this._suppressGridEvents) {
      return;
    }

    this.debouncedUpdateComponent();
    this.debouncedSaveConfigurationToUrl();
  }

  onRefreshGrid() {
    this.fetchGridData();
  }

  onExportCSV() {
    this.gridApi.exportDataAsCsv({
      fileName: this.props.csvExportFilename,
    });
  }

  setlastSyncDate() {
    this.setState({ lastSyncDate: 'fetching... ' });

    return $.get(this.props.lastSyncInfoUrl || LAST_SYNC_INFO_URL)
      .done((info) => {
        if (info) {
          const date = new Date(info.started);
          this.setState({ lastSyncDate: moment(date).fromNow() });
        } else {
          this.setState({ lastSyncDate: '[unknown]' });
        }
      })
      .fail(() => { this.setState({ lastSyncDate: 'cannot fetch data' }); });
  }

  getChartDatum() {
    const ret = [];

    if (!this.gridApi || _.isEmpty(this.state.rowData)) {
      return ret;
    }

    const model = this.gridApi.getModel();
    const groupColDef = _.get(this.columnApi.getRowGroupColumns(), '0.colDef');

    if (!groupColDef) {
      return ret;
    }

    const topNodes = [];
    model.forEachNodeAfterFilterAndSort((node) => {
      if (node.group && node.level === 0) {
        topNodes.push(node);
      }
    });

    _(this.getColumnsForChart())
      .each((valueCol) => {
        const values = _.map(topNodes, topNode => ({
          x: topNode.key,
          y: this.gridApi.getValue(valueCol.colDef, topNode.data),
        }));
        ret.push({
          values,
          key: valueCol.colDef.headerName,
          ..._.get(valueCol.colDef, 'chart', {}),
        });
      });
    return ret;
  }

  getColumnsForChart() {
    return _(this.columnApi.getValueColumns())
      .filter(c => (c.visible && c.left !== undefined))
      .sortBy('left')
      .value();
  }

  fetchGridData() {
    this.gridApi.showLoadingOverlay();
    this.setlastSyncDate();

    return $.get(this.props.gridDataUrl)
      .done(this.updateGridData.bind(this))
      .always(() => { this.gridApi.hideOverlay(); });
  }

  updateGridData(data) {
    const prevRowData = this.state.rowData;

    const preparedData = this.prepareGridData(data);
    this.gridApi.setRowData(preparedData);
    this.setState({ rowData: preparedData });

    if (prevRowData === this.initialRowData) {
      this.setConfiguration(this.initialConfiguration);
    }
  }

  prepareGridData(data) {
    return data;
  }

  getGridTitle(title) {
    return title || this.title;
  }

  getGridOptions() {
    return {
      onGridReady: this.onGridReady,
      showToolPanel: this.state.showToolPanel,
      columnDefs: this.state.columnDefs,
      rowData: this.state.rowData,
      rowSelection: 'multiple',
      enableSorting: true,
      enableFilter: true,
      enableColResize: true,
      rowHeight: 28,
      groupSuppressAutoColumn: true,
      groupSelectsChildren: true,
      groupUseEntireRow: false,
      ...this.props.agGridOptions,
    };
  }

  renderToolBarCmp(extraControls = {}) {
    return (
      <div>
        {extraControls.top}

        <BS.Row>
          <BS.Col md={10}>

            <form className="form-inline" role="form">

              <BS.ButtonGroup bsSize="sm" bsStyle="raised">
                <BS.Button bsStyle="info" bsSize="sm" onClick={this.onRefreshGrid}>
                  <i className="fa fa-refresh" /> Refresh
                </BS.Button>

                <BS.Button onClick={this.onExportCSV}>
                  <i className="fa fa-download" /> Download as CSV
                </BS.Button>
              </BS.ButtonGroup>

              {extraControls.main}

            </form>

          </BS.Col>

          <BS.Col md={2}>
            <CheckBox label=" Tool Panel"
              groupClassName="pull-right"
              onChange={this.onToggleToolPanel}
            />
          </BS.Col>
        </BS.Row>

      </div>

    );
  }

  renderGridCmp() {
    return (
      <AgGridExtended
        style={{ height: this.props.gridHeight }}
        gridOptions={this.getGridOptions()}
      />
    );
  }

  renderChartCmp() {
    return (
      <NVD3Chart
        type="multiBarChart"
        height={this.props.chartHeight}
        datum={this.getChartDatum()}
        configure={(chart) => {
          chart
            .margin({ top: 80, right: 20, bottom: 50, left: 80 });
          chart.yAxis
             .tickFormat(d3.format(',d'));
        }}
      />
    );
  }

  renderExtraBlocks() {
  }

  render() {
    const gridPanel = (
      <BS.Panel footer={`Updated ${this.state.lastSyncDate}`}>
        {this.renderToolBarCmp()}
        {this.renderGridCmp()}
      </BS.Panel>
    );

    const chart = this.renderChartCmp();

    const chartPanel = chart ? (
      <BS.Panel
        header={<span onClick={this.onToggleChart}>{this.getGridTitle()}</span>}
        expanded={this.state.isChartPanelOpen}
      >
        {chart}
      </BS.Panel>
    ) : null;
    return (
      <BS.Panel header={<h3>{this.props.title}</h3>}>
        {gridPanel}
        {chartPanel}
        {this.renderExtraBlocks()}
      </BS.Panel>
    );
  }
}
