/**
 * Created by farhad on 04.03.16.
 */
import _ from 'lodash';
import moment from 'moment';

import consts from '../../consts';

export function isCurrentWeek(m) {
  return moment(m).isSame(new Date(), 'day');
}

export function getComponentsSum(fields, params) {
  let ret = 0;
  fields.map(i => (ret += params.getValue(i) || 0));
  return ret;
}

export function toFixedCelRenderer(params) {
  let val = 0;
  if (isFinite(params.value)) {
    val = params.value;
  }
  return `${val.toFixed(2)}%`;
}

export function safeDiv(num1, num2) {
  const a = num1 || 0;
  const b = num2 || 0;
  if (a === 0 || b === 0) {
    return 0;
  }
  return a / b;
}

/**
 * Find node data value for field `field`.
 * If node is grouped then function will recursively visit
 * its child nodes and look for field value in them.
 * If child nodes have different data values for specified key
 * then function returns undefined.
 */
export function getNodeValue(nodeOrParams, field, childrenLimit = 10, expectedValue = undefined) {
  const node = nodeOrParams.node || nodeOrParams;

  if (node.data) {
    if (expectedValue === undefined) {
      expectedValue = node.data[field];
    } else {
      if (node.data[field] !== expectedValue) {
        return undefined;
      }
    }
  }

  if (node.group &&
     node.children.length > 0 &&
     node.children.length <= childrenLimit
  ) {
    _.each(node.children, (childNode) => {
      let value;
      if (childNode.group) {
        value = getNodeValue(childNode, field, childrenLimit, expectedValue);
      } else {
        value = childNode.data[field];
      }

      if (expectedValue === undefined) {
        expectedValue = value;
      }

      if (value !== undefined && value !== expectedValue) {
        return undefined;
      }
    });
  }

  return expectedValue;
}

export function makeDatesSequence(dateFrom, dateTo, step) {
  const dates = [];
  for (let m = moment(dateFrom); m.isSameOrBefore(dateTo); m.add(1, step)) {
    dates.push(moment(m));
  }
  return dates;
}

function compareCustomizer(objValue, othValue) {
  if (moment.isMoment(objValue)) {
    return objValue.isSame(othValue);
  }
}

export function isEqual(value, other) {
  return _.isEqualWith(value, other, compareCustomizer);
}

export function getDefaultDatesRanges() {
  return {
    'Last 30 Days': [
      moment().subtract(29, 'days').startOf('day'),
      moment().endOf('day'),
    ],
    'Last 3 Month': [
      moment().subtract(2, 'month').startOf('month'),
      moment().endOf('day'),
    ],
    'Last 6 Month': [
      moment().subtract(5, 'month').startOf('month'),
      moment().endOf('day'),
    ],
  };
}

export function jiraBugLink(params) {
  if (params.data === null) {
    return params.value;
  }
  return `<a target="_blank" href="https://mirantis.jira.com/browse/${params.data.bug_id}">${params.value}</a>`;
}

export function lpBugLink(params) {
  if (params.data === null) {
    return params.value;
  }
  return `<a target="_blank" href="https://launchpad.net/bugs/${params.data.bug_id}">${params.value}</a>`;
}

const isEditedValue = ({ oldValue, newValue }) => {
  if (((oldValue === null || oldValue === undefined) && newValue === '')
  || oldValue === newValue) {
    return false;
  }
  return true;
};

export function failureDriversHooverOther(params) {
  if (params.data === null) {
    return params.value;
  }
  if (params.data.runs === undefined) {
    return 0;
  }
  if (params.data.runs.other !== [null]) {
    const html = `<a href="#" html=true data-toggle="tooltip"
      title="${params.data.runs.other
                .map(id => `${id}`)
                .filter(val => val !== 'null')
                .join('\n')}">${params.value}</a>`;
    return html;
  }
  return params.value;
}

export function failureDriversHooverSwarmBvt(params) {
  if (params.data === null) {
    return params.value;
  }
  if (params.data.runs === undefined) {
    return 0;
  }
  if (!_.isNull(params.data.runs.swarm_bvt)) {
    const html = `<a href="#" html=true data-toggle="tooltip"
      title="${params.data.runs.swarm_bvt}">${params.value}</a>`;
    return html;
  }
  return params.value;
}

export function failureDriversHooverTempest(params) {
  if (params.data === null) {
    return params.value;
  }
  if (params.data.runs === undefined) {
    return 0;
  }
  if (!_.isNull(params.data.runs.tempest)) {
    const html = `<a href="#" html=true data-toggle="tooltip"
      title="${params.data.runs.tempest}">${params.value}</a>`;
    return html;
  }
  return params.value;
}

export function failureDriversAddComment(params) {
  if (!isEditedValue(params)) {
    return;
  }
  $.ajax({
    url: '/add_comment_to_bug',
    dataType: 'json',
    type: 'POST',
    data: { comment: params.newValue, bug_id: params.data.bug_id },
    error(xhr, status, err) {
      console.error(status, err.toString());
    },
  });
  params.data[params.colDef.field] = params.newValue;
}

export function failureDriversAddETA(params) {
  if (!isEditedValue(params)) {
    return;
  }
  $.ajax({
    url: '/add_eta_to_bug',
    dataType: 'json',
    type: 'POST',
    data: { eta: params.newValue, bug_id: params.data.bug_id },
    error(xhr, status, err) {
      console.error(status, err.toString());
    },
  });
  params.data[params.colDef.field] = params.newValue;
}

export function failureDriversPrepareData(data) {
  const isNotFixedBug = bug =>
    consts.launchpad.FIXED_BUG_STATUSES.indexOf(bug.bug_status) === -1;
  const isSwarmBvt = bug => bug.bug_swarm_bvt >= 3;
  const isTempest = bug => bug.bug_tempest >= 5;
  const isBlockerTag = bug => _.some(bug.tags, tag =>
    consts.launchpad.BLOCKER_BUG_TAGS.indexOf(tag) > -1);
  const isBlockerBug = (bug) => {
    if (isNotFixedBug(bug)) {
      if (isSwarmBvt(bug) || isTempest(bug) || isBlockerTag(bug)) {
        return true;
      }
    }
    return false;
  };
  _.each(data, (bug) => {
    Object.assign(bug, _.mapKeys(bug.rate, (v, k) => `bug_${k}`));
    bug.bug_type = isBlockerBug(bug) ? 'BLOCKER' : '';
    bug.bug_reported = moment(bug.bug_reported).format('YYYY-MM-DD');
    bug.bug_escalation = moment(bug.bug_escalation).format('YYYY-MM-DD');
    bug.bug_fix_committed = moment(bug.bug_fix_committed).format('YYYY-MM-DD');
    if (bug.bug_ETA !== 'n/a') {
      bug.bug_ETA = moment(bug.bug_ETA, 'YYYY-MM-DD').format('YYYY-MM-DD');
    }
    bug.bug_assignee = bug.bug_assignee || 'Unassigned';
  });

  return data;
}
