import datetime

from flask import (
    Blueprint, render_template, request, redirect, url_for, jsonify
)
from flask import current_app as app
from flask_login import logout_user

from qareporting.admin_panel.models import User
from qareporting.auth.decorators import admin_role_required
from qareporting.auth.models import AuthUsers
from qareporting.integration.jira.models import (
    JiraFilter, JiraBugsFilter)
from qareporting.integration.launchpad.models import LpMilestoneInFocus
from qareporting.integration.models import DashboardMilestone
from qareporting.integration.testrail.models import (
    TrMilestoneInFocus, Run, Plan, TrAcceptanceDate, Milestone)
from qareporting.resources.acceptance_testing_progress import \
    get_acceptance_runs
from qareporting.resources.launchpad import LpMilestonesList
from qareporting.resources.testrail import MilestoneList
from qareporting.utils import dt_from_iso
from .forms import LoginForm

admin = Blueprint('admin', __name__, template_folder='templates')


def is_number(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def save_filters(filters, _id, name, mongo_object):
    if any(int(_id) == _filter['filter_id'] for _filter in filters):
        mongo_object.objects(
            filter_id=int(_id)
        ).modify(name=name)
    elif any(name == _filter['name'] for _filter in filters):
        mongo_object.objects(
            name=name
        ).modify(filter_id=int(_id))
    else:
        position = max(x['position'] for x in filters) + 1
        mongo_object(
            name=name,
            filter_id=int(_id),
            position=position or 0
        ).save()


@admin.route('/', methods=['GET', 'POST'])
@admin_role_required
def admin_show():
    print(request)
    lp_milestones = LpMilestonesList().get()
    tr_milestones = MilestoneList().get()
    return render_template(
        'admin/index.html',
        lp_milestones=lp_milestones,
        tr_milestones=tr_milestones,
    )


@admin.route('/login', methods=['GET', 'POST'])
def login():
    form = LoginForm(request.form)
    if form.validate_on_submit():
        user = User()
        user.id = form.email.data
        if form.password.data == app.config['ADMIN_PASSWORD']:
            AuthUsers.objects(email=form.email.data).modify(
                passed=True
            )
            return redirect(url_for('admin.admin_show'))
    return render_template('admin/login.html', form=form)


@admin.route('/logout', methods=['GET', 'POST'])
def logout():
    logout_user()
    return redirect(url_for('pages.index'))


@admin.route('/_set_lp_milestone')
def set_lp_milestone():
    milestone = request.args.get('milestone', '', type=str)
    cur = LpMilestoneInFocus.objects.all().first().pk
    LpMilestoneInFocus.objects(id=cur).modify(
        name=milestone
    )
    return "Done"


@admin.route('/_set_tr_milestone')
def set_tr_milestone():
    milestone = request.args.get('milestone', '', type=str)
    cur = TrMilestoneInFocus.objects.all().first().pk
    TrMilestoneInFocus.objects(id=cur).modify(
        name=milestone
    )
    return 'Done'


@admin.route('/_set_dashboard_milestone')
def set_dashboard_milestone():
    milestone = request.args.get('milestone', '', type=str)
    cur = DashboardMilestone.objects.all().first()
    if cur:
        DashboardMilestone.objects(id=cur.pk).modify(
            name=milestone
        )
    else:
        dashboard = DashboardMilestone()
        dashboard.name = milestone
        dashboard.save()

    return jsonify('Done!')


@admin.route('/_save_filter', methods=['POST'])
def save_filter():
    name = request.form['name']
    _id = request.form['id']

    if is_number(_id):
        filters = JiraFilter.objects.all().as_pymongo()

        if filters:
            save_filters(filters, _id, name, JiraFilter)

    data = [{'name': _filter['name'],
             'id': _filter['filter_id'],
             'position': _filter['position']}
            for _filter in JiraFilter.objects.all().as_pymongo()]
    return jsonify(data)


@admin.route('/_remove_filter', methods=['POST'])
def remove_filter():
    name = request.form['name']
    _id = request.form['id']
    if is_number(_id):
        position = JiraFilter.objects(
            name=name, filter_id=int(_id)
        ).first().position
        filters = JiraFilter.objects(position__gt=position)
        if filters.count() > 0:
            filters.update(dec__position=1)
        JiraFilter.objects(name=name, filter_id=int(_id)).delete()
    data = [{'name': _filter['name'],
             'id': _filter['filter_id'],
             'position': _filter['position']}
            for _filter in JiraFilter.objects.all().as_pymongo()]
    return jsonify(data)


@admin.route('/_add_acceptance_run', methods=['POST'])
def add_acceptance_run():
    milestone_id = request.form['milestone_id']
    run_names = request.form.getlist('run_names[]')
    if is_number(milestone_id):
        Run.objects(
            milestone_id=int(milestone_id),
            name__in=run_names,
        ).update(
            set__run_type=Run.RUN_TYPE_ACCEPTANCE_ID
        )
    elif '_' in milestone_id:
        plan_ids = Plan.objects(fake_milestone=milestone_id).distinct('id')
        Run.objects(
            milestone_id=int(milestone_id.split('_')[0]),
            plan_id__in=plan_ids,
            name__in=run_names,
        ).update(
            set__run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
            set__fake_milestone=milestone_id,
        )
    return jsonify(get_acceptance_runs())


@admin.route('/_remove_acceptance_run', methods=['POST'])
def remove_acceptance_run():
    milestone_id = request.form['milestone_id']
    run_names = request.form.getlist('run_names[]')
    qa_teams = Run.objects(
        name__in=run_names
    ).only('qa_team').distinct('qa_team')
    if is_number(milestone_id):
        Run.objects(
            milestone_id=int(milestone_id),
            name__in=run_names,
        ).update(
            set__run_type=None
        )
        TrAcceptanceDate.objects(
            milestone_id=milestone_id,
            qa_team_id__in=qa_teams,
        ).delete()
    elif '_' in milestone_id:
        plan_ids = Plan.objects(fake_milestone=milestone_id).distinct('id')
        Run.objects(
            milestone_id=int(milestone_id.split('_')[0]),
            plan_id__in=plan_ids,
            name__in=run_names,
        ).update(
            unset__run_type=1,
            unset__fake_milestone=1,
        )
        TrAcceptanceDate.objects(
            milestone_id=int(milestone_id.split('_')[0]),
            qa_team_id__in=qa_teams,
        ).delete()

    return jsonify(get_acceptance_runs())


@admin.route('/_set_acceptance_date', methods=['POST'])
def set_acceptance_date():
    milestone_id = request.form['milestone_id']
    team_id = request.form['team_id']
    start_date = request.form['start_date']

    is_fake = not is_number(milestone_id)
    date = datetime.datetime.combine(
        dt_from_iso(start_date), datetime.time.min)

    if is_fake:
        acceptance_date = TrAcceptanceDate.objects(
            milestone_id=milestone_id.split('_')[0],
            fake_postfix=milestone_id.split('_')[-1],
            is_fake_milestone=is_fake,
            qa_team_id=team_id,
        ).first()
    else:
        acceptance_date = TrAcceptanceDate.objects(
            milestone_id=milestone_id,
            qa_team_id=team_id,
        ).first()
    if acceptance_date:
        acceptance_date.date = date

    else:
        if is_fake:
            acceptance_date = TrAcceptanceDate(
                milestone_id=milestone_id.split('_')[0],
                fake_postfix=milestone_id.split('_')[-1],
                is_fake_milestone=is_fake,
                qa_team_id=team_id,
                date=date,
            )
        else:
            acceptance_date = TrAcceptanceDate(
                milestone_id=milestone_id,
                qa_team_id=team_id,
                date=date,
            )
    acceptance_date.save()

    return jsonify('Done')


@admin.route('/_move_feature_position', methods=['POST'])
def move_position():
    filter_name = request.form['name']
    filter_link = request.form['link']
    hover_index = request.form['newPosition']

    drag_index = JiraFilter.objects(
        name=filter_name,
        filter_id=filter_link,
    ).first().position

    if drag_index < int(hover_index):
        JiraFilter.objects(
            position__gt=drag_index,
            position__lte=hover_index
        ).update(dec__position=1)
        JiraFilter.objects(
            name=filter_name,
            filter_id=filter_link,
        ).modify(position=hover_index)
    else:
        JiraFilter.objects(
            position__lt=drag_index,
            position__gte=hover_index
        ).update(inc__position=1)
        JiraFilter.objects(
            name=filter_name,
            filter_id=filter_link,
        ).modify(position=hover_index)

    return jsonify({'response': 'ok'})


@admin.route('/_save_bug_filter', methods=['POST'])
def save_bug_filter():

    name = request.form['name']
    _id = request.form['id']

    if is_number(_id):
        filters = JiraBugsFilter.objects.all().as_pymongo()
        if filters:
            save_filters(filters, _id, name, JiraBugsFilter)

    data = [{'name': _filter['name'],
             'id': _filter['filter_id'],
             'position': _filter['position']}
            for _filter in JiraBugsFilter.objects.all().as_pymongo()]

    return jsonify(data)


@admin.route('/_remove_bug_filter', methods=['POST'])
def remove_bug_filter():
    name = request.form['name']
    _id = request.form['id']
    if is_number(_id):
        position = JiraBugsFilter.objects(
            name=name, filter_id=int(_id)
        ).first().position
        filters = JiraBugsFilter.objects(position__gt=position)
        if filters.count() > 0:
            filters.update(dec__position=1)
        JiraBugsFilter.objects(name=name, filter_id=int(_id)).delete()
    data = [{'name': _filter['name'],
             'id': _filter['filter_id'],
             'position': _filter['position']}
            for _filter in JiraBugsFilter.objects.all().as_pymongo()]
    return jsonify(data)


@admin.route('/_set_acceptance_any', methods=['POST'])
def set_acceptance_any():
    milestone_id = request.form['milestone_id']
    value = request.form['value']

    if is_number(milestone_id):
        if value == 'true':
            value = True
        else:
            value = False
        Milestone.objects(id=int(milestone_id)).modify(acceptance_any=value)

    return jsonify('Done!')
