import logging
from logging import handlers as log_handlers
import os
import sys

import mongoengine
from flask_security import MongoEngineUserDatastore
from flask import Flask
from flask import render_template

from qareporting.auth import models as auth_models
from qareporting.auth.views import auth
from qareporting.config import config
from qareporting import ext
from qareporting.pages.views import pages
from qareporting.resources import api_bp
from qareporting.admin_panel.views import admin


def configure_app(app, config_name):
    app.config.from_object(config[config_name])
    config_filename = 'qareporting.conf'
    app.config.from_pyfile('/etc/qareporting/' + config_filename, silent=True)
    app.config.from_pyfile(os.path.join(os.path.expanduser('~'),
                                        config_filename), silent=True)

    # This is needed to pass tz_aware param to pymongo connection.
    app.config['MONGODB_SETTINGS'] = {
        'db': app.config['MONGODB_DB'],
        'host': app.config['MONGODB_HOST'],
        'port': app.config['MONGODB_PORT'],
        'tz_aware': True,
    }


def configure_logging(app):
    if not app.debug and not app.testing:
        handler = log_handlers.RotatingFileHandler(
            app.config['LOGGING_LOCATION'],
            maxBytes=1024 * 1024 * 100,
            backupCount=20)

        logging.getLogger('aiohttp').setLevel(logging.ERROR)
        logging.getLogger('requests').setLevel(logging.ERROR)
    else:
        handler = logging.StreamHandler(sys.stdout)

    formatter = logging.Formatter(app.config['LOGGING_FORMAT'])
    handler.setLevel(app.config['LOGGING_LEVEL'])
    handler.setFormatter(formatter)

    # beause app.logger doesn't work when command (except Server) is running
    logging.basicConfig(stream=sys.stdout,
                        level=app.config['LOGGING_LEVEL'],
                        format=app.config['LOGGING_FORMAT'])

    app.logger.addHandler(handler)

    logging.getLogger('iso8601').setLevel(logging.ERROR)
    logging.getLogger('passlib').setLevel(logging.ERROR)


def configure_hook(app):

    @app.before_request
    def before_request():
        pass


def configure_extensions(app):
    ext.db.init_app(app)
    assert mongoengine.signals.signals_available, 'Signals must be available!'

    ext.cache.init_app(app, config=app.config['CACHING'])
    ext.oauth.init_app(app)

    # Setup Flask-Security
    user_datastore = MongoEngineUserDatastore(ext.db,
                                              auth_models.AuthUsers,
                                              auth_models.AuthRoles)
    ext.security.init_app(app, user_datastore)

    if app.debug:
        ext.toolbar.init_app(app)


def configure_blueprints(app):
    app.register_blueprint(auth)
    app.register_blueprint(api_bp)
    app.register_blueprint(pages)
    app.register_blueprint(admin, url_prefix='/admin')


def configure_templates(app):

    @app.template_filter()
    def format_date(value, format='%Y-%m-%d'):
        return value.strftime(format)

    @app.context_processor
    def is_dev():
        return dict(is_dev=app.config['DEBUG'])


def configure_error_handlers(app):
    def register_error_handler(code, page):
        app.register_error_handler(code, lambda error: (render_template(page),
                                                        code))

    register_error_handler(401, 'errors/401.html')
    register_error_handler(500, 'errors/500.html')


def configure_mail(app):
    app.config.update(
        MAIL_SERVER='smtp.gmail.com',
        MAIL_PORT=465,
        MAIL_USE_SSL=True
    )


def configure_webpack(app):
    ext.webpack.init_app(app)


def create_app(config_name='development'):
    app = Flask(__name__, static_url_path='/assets')

    configure_app(app, config_name)
    configure_logging(app)
    configure_hook(app)
    configure_extensions(app)
    configure_blueprints(app)
    configure_templates(app)
    configure_error_handlers(app)
    configure_mail(app)
    configure_webpack(app)
    return app
