import logging

from flask import current_app as app
from flask_mail import Message
from flask_script import Command
from flask_script import Option
from mongoengine import connect

from qareporting.utils import dt_now
from qareporting.integration.launchpad.utils import LaunchpadUsersFinder
from qareporting.integration.models import EventLog
from qareporting.ext import cache
from qareporting.ext import mail
from qareporting.integration.jira.models import JiraFilter, JiraBugsFilter
from qareporting.integration.launchpad.models import LpMilestoneInFocus
from qareporting.integration.testrail.models import TrMilestoneInFocus
from qareporting.integration.models import DashboardMilestone
from .gen_message import recent_swarm_stat
from .feature_health import get_recipients, send_notification
from .incomplete_bugs import get_incomplete_bugs

log = logging.getLogger(__name__)


class SendIncompleteBugNotification(Command):
    option_list = (
        Option('-m', '--milestone_name',
               dest='milestone_name',
               type=str,
               required=True,
               help='Launchpad milestone name'),
    )

    def run(self, milestone_name):
        incomplete_bugs = get_incomplete_bugs(milestone_name)

        lp_user_finder = LaunchpadUsersFinder()

        for bug_id, username in incomplete_bugs.items():
            email = None
            user = None

            if username:
                user = lp_user_finder.find_user(username=username)

            if user:
                email = user.email

            if email:
                if 'mirantis.com' in email:
                    mail.init_app(app)
                    msg = Message(
                        sender='qa-reporting@mirantis.com',
                        recipients=[email]
                    )

                    msg.html = (
                        '<!DOCTYPE html><head><style type="text/css">'
                        '.im {{color: #ff00 !important;}}</style></head>'
                        '<html><body><p>Hello, </p>'
                        '<p>Please provide more information about bug:</p>'
                        '<p><a href="https://launchpad.net/bugs/{0}">'
                        'https://launchpad.net/bugs/{0}</a></p>'
                        '<br>'
                        '<br>'
                        '<p>Thanks,</p>'
                        '<p>QA Reporting team</p>'.format(bug_id)
                    )
                    msg.subject = 'Incomplete notification'

                    mail.send(msg)


class SendNotificationCommand(Command):
    def run(self):
        send_box = get_recipients()
        for recipient in send_box:
            log.info('Send notification to {}'.format(recipient['email']))
            send_notification(recipient)
        log.info('The notification has been sent')


class SendMailCommand(Command):
    option_list = (
        Option('-m', '--milestone_id',
               dest='milestone_id',
               type=int,
               required=True,
               help='Testrail milestone id'),
    )

    def run(self, milestone_id):
        mail.init_app(app)
        msg = Message(
            sender='qa-reporting@mirantis.com',
            recipients=[
                'ibrikin@mirantis.com',
                'tborukhovich@mirantis.com',
                'musichenko@mirantis.com',
            ]
        )

        msg.html, msg.subject = recent_swarm_stat(milestone_id)
        mail.send(msg)
        return 'Sent'


class ClearCacheCommand(Command):
    def run(self):
        with app.app_context():
            cache.clear()
        log.info('Cache has been cleared!')


class DropDB(Command):
    def run(self):
        db_name = app.config['MONGODB_DB']
        client = connect(db_name)
        client.drop_database(db_name)
        log.info('DB {0} has been dropped!'.format(db_name))


class DropCollection(Command):
    option_list = (
        Option('-c', '--collection',
               dest='collection',
               required=True,
               help='Collection name'),
    )

    def run(self, collection):
        db_name = app.config['MONGODB_DB']
        client = connect(db_name)
        db = client[db_name]
        db.drop_collection(collection)
        log.info('Collection {0} has been dropped!'.format(collection))


class PopulateDB(Command):

    def run(self):
        default_lp_milestone = '9.1'
        default_tr_milestone = '9.1'
        default_jira_filter_id = 33870
        default_jira_filter_name = '9.1'
        default_dashboard_milestone = '9.1'
        default_jira_bug_filter_name = 'MK22'
        default_jira_bug_filter_id = 39000

        if LpMilestoneInFocus.objects().all().count() == 0:
            lp_milestone = LpMilestoneInFocus()
            lp_milestone.name = default_lp_milestone
            lp_milestone.save()
            log.info('Add Default Launchpad milestone')

        if TrMilestoneInFocus.objects().all().count() == 0:
            tr_milestone = TrMilestoneInFocus()
            tr_milestone.name = default_tr_milestone
            tr_milestone.save()
            log.info('Add Default Testrail milestone')

        if JiraFilter.objects().all().count() == 0:
            jira_filter = JiraFilter()
            jira_filter.name = default_jira_filter_name
            jira_filter.filter_id = default_jira_filter_id
            jira_filter.position = 0
            jira_filter.save()
            log.info('Add Default Jira Filter')

        if DashboardMilestone.objects().all().count() == 0:
            dashboard = DashboardMilestone()
            dashboard.name = default_dashboard_milestone
            dashboard.save()
            log.info('Add Default Dashboard Milestone')

        if JiraBugsFilter.objects().all().count() == 0:
            jira_bug_filters = JiraBugsFilter()
            jira_bug_filters.name = default_jira_bug_filter_name
            jira_bug_filters.filter_id = default_jira_bug_filter_id
            jira_bug_filters.position = 0
            jira_bug_filters.save()
            log.info('Add Default Jira Bugs Filter')

        log.info('DB data has been loaded!')


class CheckUnfinished(Command):
    def run(self):
        unfinished_events = EventLog.objects.filter(
            finished=None,
        )
        for event in unfinished_events:
            delta = dt_now() - event.started
            if delta.seconds > 14400:
                log.info('{} event was unfinished'.format(event.event))
                log.info('Started in {}'.format(
                    event.started.strftime('%Y-%m-%d-%H-%M-%S')))
                event.delete()

        log.info('Events have been checked!')
