import logging
import asyncio
import aiohttp
import urllib.parse

from flask import current_app


log = logging.getLogger(__name__)


class CollectionIterator:

    """Iteration of paginated jira api results."""

    MAX_CHUNK_SIZE = 100

    def __init__(self, url, client, *, entries_field):
        self._client = client
        self._cur_chunk = None
        self._curr_entries = []
        self._curr_entry_index = 0
        self._next_chunk_url = url
        self._entries_field = entries_field

    def _get_next_chunk_url(self, url):
        components = urllib.parse.urlparse(url)
        qs = urllib.parse.parse_qs(components.query)
        qs.setdefault('startAt', [0])
        qs.setdefault('maxResults', [self.MAX_CHUNK_SIZE])

        assert len(qs['startAt']) == 1
        start_at = int(qs['startAt'][0])

        start_at += len(self._curr_entries)
        qs['startAt'] = [str(start_at)]

        if self._cur_chunk is not None:
            if start_at >= self._cur_chunk.get('total'):
                return None

        qs_string = urllib.parse.urlencode(qs, doseq=True)
        new_components = list(components)
        new_components[4] = qs_string
        return urllib.parse.urlunparse(new_components)

    async def __aiter__(self):
        return self

    async def __anext__(self):
        if self._curr_entry_index >= len(self._curr_entries):

            self._next_chunk_url = self._get_next_chunk_url(
                self._next_chunk_url
            )

            if self._next_chunk_url:

                self._cur_chunk = await self._client.send_request(
                    self._next_chunk_url)

                self._curr_entries = self._cur_chunk[self._entries_field]

                if not self._curr_entries:
                    raise StopAsyncIteration

                self._curr_entry_index = 0
            else:
                raise StopAsyncIteration

        result = self._curr_entries[self._curr_entry_index]
        self._curr_entry_index += 1
        return result


class JiraClientError(Exception):

    """Error raised by JiraClient ."""


class JiraClient:

    API_URL = 'https://mirantis.jira.com/rest/api/2'
    SEND_REQUEST_ATTEMPTS = 4

    def __init__(self, session):

        self.http_session = session

        self.auth = aiohttp.helpers.BasicAuth(
            current_app.config['JIRA_USER'],
            current_app.config['JIRA_PASSWORD'],
        )

    def get_filter_url(self, filter_id):
        return '{}/filter/{}'.format(self.API_URL, filter_id)

    def get_issues_search_url(self, jql, fields=None, changelog=None):
        query_params = [('jql', jql)]
        if fields:
            query_params.append(('fields', ','.join(fields)))

        if changelog:
            query_params.append(changelog)

        return '{}/search?{}'.format(
            self.API_URL, urllib.parse.urlencode(query_params))

    def get_issues_remotelink_url(self, issue):
        return '{}/issue/{}/remotelink'.format(self.API_URL, issue)

    def get_issue_changelog_url(self, issue):
        query_params = [('expand', 'changelog'), ('fields', 'null')]
        return '{}/issue/{}?{}'.format(
            self.API_URL, issue, urllib.parse.urlencode(query_params)
        )

    async def _send_request(self, url, method='GET'):

        request = self.http_session.request(
            method,
            url,
            auth=self.auth,
            allow_redirects=False,
        )

        async with request as response:

            if response.status == 200:
                result = await response.json()
                return result
            else:
                log.debug('Response status is not equal to 200! '
                          'url: %s. response: %s', url, response)
                raise JiraClient('Request Failed')

    async def send_request(self, url, method='GET'):

        for attempt_index in range(1, self.SEND_REQUEST_ATTEMPTS + 1):
            is_last_attempt = (attempt_index == self.SEND_REQUEST_ATTEMPTS)
            if attempt_index > 1:
                log.info('Send request (attempt %s):  %s', attempt_index, url)

            try:
                return await self._send_request(url, method)
            except Exception:
                if not is_last_attempt:
                    asyncio.sleep(1)
                    continue
                log.exception('Error during sending request %s', url)
                raise
