import logging

import mongoengine

from qareporting.utils import prefixed_collection_name

log = logging.getLogger(__name__)


class JiraUser(mongoengine.EmbeddedDocument):

    username = mongoengine.StringField(required=True)
    email = mongoengine.StringField()
    display_name = mongoengine.StringField()

    # custom field for mapping to launchpad user
    launchpad_username = mongoengine.StringField(null=True)

    def __str__(self):
        return self.username


class JiraFilter(mongoengine.Document):

    name = mongoengine.StringField(required=True)
    filter_id = mongoengine.IntField(required=True)
    position = mongoengine.IntField(required=True)

    meta = {
        'collection': 'jira_filter_ids',
    }


class JiraBugsFilter(mongoengine.Document):

    name = mongoengine.StringField(required=True)
    filter_id = mongoengine.IntField(required=True)
    position = mongoengine.IntField(required=True)

    meta = {
        'collection': 'jira_bug_filters',
    }


class JiraDocument(mongoengine.DynamicDocument):

    id = mongoengine.IntField(primary_key=True)
    key = mongoengine.StringField(required=True)
    status = mongoengine.StringField(required=True)
    summary = mongoengine.StringField(required=True)
    delivery_team = mongoengine.StringField(null=True)

    meta = {
        'abstract': True,
        'collection': prefixed_collection_name('jira_')
    }


class Issue(JiraDocument):

    issuetype = mongoengine.StringField(required=True)
    assignee = mongoengine.EmbeddedDocumentField(JiraUser)
    epic_link = mongoengine.StringField(null=True)
    qa_engineers = mongoengine.EmbeddedDocumentListField(
        JiraUser,
        null=True
    )
    feature_health = mongoengine.StringField(null=True)
    status_qa = mongoengine.StringField(null=True)
    launchpad_bug_ids = mongoengine.ListField(
        field=mongoengine.IntField(),
        null=True
    )
    qa_team_name = mongoengine.StringField(null=True)
    issues_in_epic = mongoengine.ListField(
        field=mongoengine.StringField(),
        null=True
    )
    priority_id = mongoengine.IntField(null=True)
    status_id = mongoengine.IntField(null=True)
    health_time = mongoengine.StringField(null=True)

    filter_ids = mongoengine.ListField(
        field=mongoengine.IntField(required=True),
        required=True
    )

    def __str__(self):
        return 'Issue-{} filters-{}'.format(self.key, str(self.filter_ids))


class BugJiraDocument(JiraDocument):
    created_on = mongoengine.DateTimeField(required=True)
    reporter = mongoengine.StringField(required=True)
    program = mongoengine.StringField(null=True)
    compatible_versions = mongoengine.ListField(
        field=mongoengine.StringField(),
        null=True,
    )
    components = mongoengine.ListField(
        field=mongoengine.StringField(),
        null=True
    )
    assignee = mongoengine.StringField(null=True)
    priority = mongoengine.StringField(null=True)
    fix_versions = mongoengine.ListField(
        field=mongoengine.StringField(),
        null=True)
    major_team_ids = mongoengine.ListField(
        field=mongoengine.IntField(),
        null=True
    )
    qa_teams = mongoengine.StringField(null=True)

    meta = {
        'abstract': True,
    }

    def __str__(self):
        return '{} '.format(self.key)


class Bug(BugJiraDocument):
    pass


class BugHist(BugJiraDocument):

    STATUS = 'status'
    PRIORITY = 'priority'
    ASSIGNEE = 'assignee'
    FIX_VERSION = 'Fix Version'

    HISTORY_FIELDS = {
        STATUS: 'status',
        PRIORITY: 'priority',
        ASSIGNEE: 'assignee',
        FIX_VERSION: 'fix_versions',
    }

    id = mongoengine.ObjectIdField(primary_key=True)

    bug_id = mongoengine.IntField(required=True)
    filter_id = mongoengine.IntField(required=True)

    version_start = mongoengine.DateTimeField(required=True)
    version_end = mongoengine.DateTimeField(required=True)
