import asyncio
import logging

import aiohttp

from qareporting.integration.jira.jira_client import (
    JiraClient, CollectionIterator)
from qareporting.integration.jira.models import (
    JiraBugsFilter, Issue, JiraFilter, Bug, BugHist)
from qareporting.integration.jira.utils import (
    bug_from_raw, issue_from_raw, bugs_hist_from_raw, get_time_gap)
from qareporting.integration.launchpad.models import Bug as LPBug
from qareporting.integration.launchpad.utils import LaunchpadUsersFinder
from qareporting.integration.models import EventLog
from qareporting.integration.sync_base import SyncBase
from qareporting.integration.utils import save_if_modified
from qareporting.utils import dt_from_iso, dt_now

log = logging.getLogger(__name__)


ALL_BUGS_FILTER = 41906


def save_issue(new_obj, issue_filters):
    assert new_obj.id is not None, 'object must have id value'
    old_obj = new_obj._qs.with_id(new_obj.id)

    if not old_obj:
        new_obj.save()
        return new_obj, True

    old_filters = set(old_obj.filter_ids)
    new_filters = set(new_obj.filter_ids)

    filters = set(issue_filters)
    modified = None
    old_contain_new = new_filters.issubset(old_filters)
    filters_added = old_filters.issubset(filters)

    if not old_contain_new:
        new_obj.filter_ids = list(old_filters | new_filters)
        modified = True

    if old_contain_new and not filters_added:
        modified = True

    if modified:
        logging.debug(
            'Object %s is modified. saving....', repr(new_obj)
        )
        new_obj.save()
        return new_obj, True

    if filters_added:
        new_obj.filter_ids = list(old_filters)
        new_obj.save()
        return new_obj, True

    return new_obj, False


def save_bug_history(entries, bug_hist_len):
    assert entries, 'entries should have items'

    bug = entries[0]

    old_bug_hist = bug._qs.filter(bug_id=bug.bug_id)

    diff = abs(old_bug_hist.count() - bug_hist_len)

    old_bug_hist.delete()

    if diff > 0:
        logging.debug(
            '%s is modified. Add %d entries', repr(bug), diff
        )

    for entry in entries:
        entry.save()

    return True


class JiraSync(SyncBase):
    event_code = EventLog.EVENT_JIRA_SYNC

    PROJECT = 'PROD'
    ISSUE_TYPE_BUG = 'Bug'
    ISSUE_TYPE_SUBISSUE = 'SubIssue'

    ISSUES_IN_EPIC = (
        'User Story',
        'QA Task',
        ISSUE_TYPE_BUG,
        'Initiative',
        'Epic'
    )

    client = None

    FIELD_EPIC_LINK = 'customfield_11700'
    FIELD_QA_ENGINEERS = 'customfield_17101'
    FIELD_FEATURE_HEALTH = 'customfield_16909'
    FIELD_DELIVERY_TEAM = 'customfield_19000'
    FIELD_COMPATIBLE_VERSION = 'customfield_22302'
    FIELD_PROGRAM = 'customfield_23000'
    FIELD_QA_STATUS = 'customfield_20900'

    FH = 'Feature Health'

    _lp_users_finder = None

    SECONDARY_TEAMS = (
        'mos-glance',
        'mos-ironic',
        'mos-horizon',
        'mos-heat',
    )

    def __init__(self):
        self.__issue_filters = tuple(JiraFilter.objects.scalar('filter_id'))

    def _get_fields(self, issuetype):

        base_fields = (
            'status',
            'summary',
            'assignee',
            'issuetype',
            'priority',
            self.FIELD_DELIVERY_TEAM,
        )

        if issuetype == self.ISSUE_TYPE_BUG:
            return base_fields + (
                'created',
                'creator',
                'reporter',
                'project',
                'updated',
                'fixVersions',
                'components',
                self.FIELD_COMPATIBLE_VERSION,
                self.FIELD_PROGRAM,
            )
        elif issuetype == self.ISSUE_TYPE_SUBISSUE:
            return base_fields + (
                self.FIELD_QA_ENGINEERS,
                self.FIELD_FEATURE_HEALTH,
                self.FIELD_EPIC_LINK,
            )

        return base_fields + (
            self.FIELD_QA_ENGINEERS,
            self.FIELD_FEATURE_HEALTH,
            self.FIELD_EPIC_LINK,
            self.FIELD_QA_STATUS,
        )

    def _get_search_url(self, jql, issuetype):

        fields = self._get_fields(issuetype)
        changelog = None

        if issuetype == self.ISSUE_TYPE_BUG:
            changelog = ('expand', 'changelog')

        return self.client.get_issues_search_url(
            jql=jql,
            fields=fields,
            changelog=changelog
        )

    def _get_subissue_jql(self, issue_key):

        return 'project = {} '.format(self.PROJECT) \
              + 'AND (issuetype = "{}" ' \
                'OR issuetype = "{}" ' \
                'OR issuetype = "{}" ' \
                'OR issuetype = "{}" ' \
                'OR issuetype = "{}") '.format(*self.ISSUES_IN_EPIC) \
              + 'AND "Epic Link"={}'.format(issue_key)

    async def sync_issues_by_filter(self, jira_filter_id, issuetype=None):

        filter_data = await self.client.send_request(
            self.client.get_filter_url(jira_filter_id))

        search_url = self._get_search_url(filter_data['jql'], issuetype)

        return CollectionIterator(
            search_url,
            self.client,
            entries_field='issues'
        )

    async def sync_subissue_by_key(self, issue_key):

        jql = self._get_subissue_jql(issue_key)

        search_url = self._get_search_url(
            jql, issuetype=self.ISSUE_TYPE_SUBISSUE
        )

        return CollectionIterator(
            search_url,
            self.client,
            entries_field='issues'
        )

    async def _get_time_since_health_update(self, issue_key):

        issue_history_data = await self.client.send_request(
            self.client.get_issue_changelog_url(issue_key)
        )
        last_note = None

        if issue_history_data:
            history = issue_history_data['changelog']['histories']
            history.sort(key=lambda x: x['created'])
            history.reverse()

            for item in history:
                if any(field['field'] == self.FH for field in item['items']):
                    last_note = dt_from_iso(item['created'])
                    break

            if not last_note:
                return get_time_gap(
                    dt_now(), dt_from_iso(history[0]['created'])
                )
            else:
                return get_time_gap(dt_now(), last_note)

        return None

    async def _get_lp_bugs_by_key(self, issue_key):

        bug_ids = set()
        issue_links_data = await self.client.send_request(
            self.client.get_issues_remotelink_url(issue_key)
        )

        if issue_links_data:
            links = [link['object']['url'] for link in issue_links_data
                     if 'bugs.launchpad.net/' in link['object']['url']]

            for link in links:
                if 'tag' in link:
                    tag = link.rpartition('=')[-1]
                    bug_ids.update(list(LPBug.objects(tags=tag).scalar('id')))
                else:
                    bug_ids.add(int(link.rpartition('/')[-1]))

        return list(bug_ids)

    async def get_bugs(self):

        bugs = await self.sync_issues_by_filter(
            ALL_BUGS_FILTER,
            issuetype=self.ISSUE_TYPE_BUG
        )

        bug_ids = set()

        async for raw_bug in bugs:
            bug_ids.add(raw_bug['id'])
            bug = bug_from_raw(raw_bug)
            save_if_modified(bug)

        Bug.objects.filter(id__nin=tuple(bug_ids)).delete()

    async def _get_subissues(self, issue_key, filter_id):

        subissues = await self.sync_subissue_by_key(issue_key)

        sub_issue_ids = set()
        sub_issue_keys = set()

        async for raw_issue in subissues:
            sub_issue_ids.add(raw_issue['id'])
            sub_issue_keys.add(raw_issue['key'])

            health_time = None
            lp_bug_ids = await self._get_lp_bugs_by_key(raw_issue['key'])

            if raw_issue['fields'][self.FIELD_FEATURE_HEALTH] is not None:
                health_time = await self._get_time_since_health_update(
                    raw_issue['key']
                )

            subissue = issue_from_raw(
                raw_issue,
                filter_id,
                lp_bug_ids,
                health_time=health_time
            )

            save_issue(subissue, self.__issue_filters)

        Issue.objects.filter(
            epic_link=issue_key,
            id__nin=tuple(sub_issue_ids),
            filter_ids=filter_id
        ).delete()

        return sub_issue_ids, sub_issue_keys

    async def get_issues(self, issues, filter_id):

        issues_ids = set()
        total_sub_issue_ids = set()

        async for raw_issue in issues:
            issues_ids.add(raw_issue['id'])

            lp_bug_ids = await self._get_lp_bugs_by_key(raw_issue['key'])

            sub_issue_ids, sub_issue_keys = await self._get_subissues(
                raw_issue['key'], filter_id
            )

            total_sub_issue_ids.update(sub_issue_ids)

            health_time = None

            if raw_issue['fields'][self.FIELD_FEATURE_HEALTH] is not None:
                health_time = await self._get_time_since_health_update(
                    raw_issue['key']
                )

            issue = issue_from_raw(
                raw_issue,
                filter_id,
                list(lp_bug_ids),
                list(sub_issue_keys),
                health_time
            )
            save_issue(issue, self.__issue_filters)

        Issue.objects.filter(
            id__nin=tuple(issues_ids) + tuple(total_sub_issue_ids),
            filter_ids=filter_id
        ).delete()

    async def get_bug_history(self, issues, filter_id):

        issues_ids = set()

        async for bug in issues:
            issues_ids.add(bug['id'])
            entries = bugs_hist_from_raw(bug, filter_id)
            save_bug_history(entries, len(entries))

        BugHist.objects.filter(
            bug_id__nin=tuple(issues_ids),
            filter_id=filter_id
        ).delete()

    async def get_by_filter_id(self, filter_id, function, issuetype=None):

        issues = await self.sync_issues_by_filter(filter_id, issuetype)

        await function(issues, filter_id)

    async def marshal_syncs(self):

        bug_filters = tuple(JiraBugsFilter.objects.scalar('filter_id'))

        issue_coros = [self.get_by_filter_id(filter_set, self.get_issues)
                       for filter_set in self.__issue_filters]

        bugs_coros = [self.get_by_filter_id(
            filter_set, self.get_bug_history, issuetype=self.ISSUE_TYPE_BUG)
                      for filter_set in bug_filters]

        await self.get_bugs()

        if issue_coros:
            await asyncio.wait(issue_coros)

        if bugs_coros:
            await asyncio.wait(bugs_coros)

        Issue.objects.filter(
            filter_ids__nin=self.__issue_filters
        ).delete()

        BugHist.objects.filter(
            filter_id__nin=bug_filters
        ).delete()

    def _sync(self):

        loop = asyncio.get_event_loop()
        self._lp_users_finder = LaunchpadUsersFinder()

        with aiohttp.ClientSession(loop=loop) as session:
            self.client = JiraClient(session=session)

            loop.run_until_complete(self.marshal_syncs())
