import itertools

from bson import ObjectId

from qareporting.integration.jira import models
from qareporting.integration.launchpad.utils import LaunchpadUsersFinder
from .models import JiraUser, BugHist
from qareporting.utils import utc_max_date, dt_from_iso


def get_bug_defaults(raw_bug):

    from .sync import JiraSync
    fields = raw_bug['fields']

    result = {
        'key': raw_bug['key'],
        'status': fields['status']['name'],
        'priority': fields['priority']['name'],
        'summary': fields['summary'],
        'created_on': fields['created'],
        'assignee': None,
        'reporter': None,
        'delivery_team': None,
        'compatible_versions': None,
        'fix_versions': None,
        'program': None,
        'components': None,
    }

    if fields.get('assignee'):
        result['assignee'] = fields['assignee']['key']

    if fields.get('reporter'):
        result['reporter'] = fields['reporter']['key']

    if fields.get(JiraSync.FIELD_DELIVERY_TEAM):
        result['delivery_team'] = fields[JiraSync.FIELD_DELIVERY_TEAM].get(
            'value')

    if fields.get(JiraSync.FIELD_COMPATIBLE_VERSION):
        result['compatible_versions'] = [x['name']
                                         for
                                         x
                                         in
                                         fields[
                                             JiraSync.FIELD_COMPATIBLE_VERSION
                                         ]]

    if fields.get('fixVersions'):
        result['fix_versions'] = [x['name'] for x in fields['fixVersions']]

    if fields.get(JiraSync.FIELD_PROGRAM):
        result['program'] = fields[JiraSync.FIELD_PROGRAM]['value']

    if fields.get('components'):
        result['components'] = [x['name'] for x in fields['components']]

    return result


def bug_from_raw(raw_bug):

    defaults = get_bug_defaults(raw_bug)

    bug = models.Bug(
        **defaults,
        id=raw_bug['id'],
    )
    return bug


def get_qa_engineers(data):

    from .sync import JiraSync
    if not data.get(JiraSync.FIELD_QA_ENGINEERS):
        return []
    return [
        JiraUser(
            username=user['key'],
            email=user['emailAddress'],
            display_name=user['displayName'],
        ) for user in data.get(JiraSync.FIELD_QA_ENGINEERS, [])
        ]


def set_qa_team_name(issue):

    if not issue.qa_engineers:
        return

    from .sync import JiraSync
    teams = set()
    lp_user_finder = LaunchpadUsersFinder()

    for engineer in issue.qa_engineers:

        lp_user = lp_user_finder.find_user(
            email=engineer.email,
            username=engineer.username
        )

        if lp_user:
            engineer.launchpad_username = lp_user.username
            teams.update(lp_user.teams)

    if teams:
        issue.qa_team_name = ', '.join(sorted(
            teams, key=lambda s: s in JiraSync.SECONDARY_TEAMS))


def issue_from_raw(raw_issue, filter_id,
                   lp_bugs, sub_issue_keys=None,
                   health_time=None):

    from .sync import JiraSync
    fields = raw_issue['fields']
    delivery_team = None
    assignee = None
    if fields.get(JiraSync.FIELD_DELIVERY_TEAM):
        delivery_team = fields[JiraSync.FIELD_DELIVERY_TEAM].get('value')
    if fields.get('assignee'):
        assignee = JiraUser(
            username=fields['assignee']['key'],
            email=fields['assignee']['emailAddress'],
            display_name=fields['assignee']['displayName'],
        )
    issue = models.Issue(
        id=raw_issue['id'],
        key=raw_issue['key'],
        status=fields['status']['name'],
        issuetype=fields['issuetype']['name'],
        assignee=assignee,
        summary=fields['summary'],
        epic_link=fields.get(JiraSync.FIELD_EPIC_LINK),
        feature_health=fields.get(JiraSync.FIELD_FEATURE_HEALTH),
        status_qa=fields.get(JiraSync.FIELD_QA_STATUS),
        qa_engineers=get_qa_engineers(fields),
        launchpad_bug_ids=lp_bugs,
        issues_in_epic=sub_issue_keys,
        filter_ids=[filter_id],
        status_id=fields['status']['id'],
        priority_id=fields['priority']['id'],
        delivery_team=delivery_team,
        health_time=health_time,
    )

    set_qa_team_name(issue)

    return issue


def get_recent_changes(data):

    result = {}
    version_end = None

    for date in sorted(data, key=lambda x: x[0], reverse=True):
        _, records = date
        count = 0
        for record in sorted(records, key=lambda x: x['created'],
                             reverse=True):
            count += 1
            created = record['created']
            for item in record['items']:

                is_in_first = count <= 1 or item['field'] not in result
                is_hist_field = item['field'] in BugHist.HISTORY_FIELDS.keys()

                if is_hist_field and is_in_first:

                    if isinstance(item['toString'], str):
                        value = item['toString']

                        if item['field'] == BugHist.FIX_VERSION:
                            value = [value]

                        if version_end is None:
                            version_end = utc_max_date

                        result.update({
                            item['field']: {
                                'value': value,
                                'version_start': dt_from_iso(created),
                                'version_end': version_end
                            }
                        })
                        version_end = dt_from_iso(created)

                if not is_in_first:
                    result[item['field']].update({
                                    'version_start': dt_from_iso(created)
                                })

        if result.keys():
            yield result
            result = {}


def get_history(data, defaults):

    result = []
    args = {}

    count = 0

    for change in get_recent_changes(data):
        count += 1
        new_args = defaults.copy()
        for field in change.keys():
            if args.keys():

                field_start = change[field]['version_start']
                field_end = change[field]['version_end']

                start = (field_start
                         if field_start < args['version_start']
                         else args['version_start'])
                end = (field_end
                       if field_end > args['version_end']
                       else args['version_end'])

                args.update({
                    BugHist.HISTORY_FIELDS[field]: change[field]['value'],
                    'version_start': start,
                    'version_end': end,
                })
            else:
                args.update({
                    BugHist.HISTORY_FIELDS[field]: change[field]['value'],
                    'version_start': change[field]['version_start'],
                    'version_end': change[field]['version_end'],
                })

            if count == 1 and change[field]['version_end'] != utc_max_date:
                args.update({
                    'version_end': utc_max_date,
                })

            if field != 'status' and count != 1:
                args.update({'status': 'Open'})

        args.update({'id': ObjectId()})
        new_args.update(args)

        result.append(models.BugHist(
            **new_args
        ))

        args = {}

    return result


def bugs_hist_from_raw(raw_bug, filter_id):

    bug_history = raw_bug.get('changelog', {}).get('histories', [])

    defaults = get_bug_defaults(raw_bug)
    defaults.update({
        'filter_id': filter_id,
        'bug_id': raw_bug['id'],
        'version_start': raw_bug['fields']['created'],
        'version_end': utc_max_date,
    })

    data = []

    if bug_history:
        bugs_history_by_date = itertools.groupby(
            bug_history,
            key=lambda a: dt_from_iso(a['created']).date()
        )

        for key, group in bugs_history_by_date:
            data.append((key, tuple(group)))

        if data != 1:
            hist = get_history(data, defaults)
            if hist:
                return hist

    return [models.BugHist(
        **defaults,
        id=ObjectId()
    )]


def get_time_gap(date_from, date_to):

    gap = date_from - date_to

    if gap.days > 0:
        return '{} day(s)'.format(gap.days)
    elif gap.seconds < 3600:
        return '{} minutes'.format(round(gap.seconds / 60))
    else:
        return '{} hours'.format(round(gap.seconds / 3600))
