import copy
import itertools
import logging
from collections import OrderedDict

from . import consts
from .utils import get_focus_of_dev, split_target

log = logging.getLogger(__name__)


class HistoryBuilder:

    BUG_TASK_FIELDS = ('status', 'importance', 'assignee', 'milestone')
    BUG_TASK_DEFAULTS = {
        'status': consts.STATUS_NEW,
    }

    INTERVAL_TYPE_DAY = 'day'
    VALUABLE_TARGETS = ['mos', 'fuel']

    activity = None
    grouped_activity = None
    history = None

    def __init__(self, bug):
        self.bug = bug
        self._setup()

    def get_history(self):
        if self.history is None:
            self.history = self._build_bug_history()
        return self.history

    def _setup(self):
        self.activity = self._prepare_activity(self.bug.activity)

        self.grouped_activity = list(self._group_activity_by_interval(
            interval_type=self.INTERVAL_TYPE_DAY
        ).items())

    def _get_last_bug_state(self):

        state = {
            'interval': self.grouped_activity[-1],
            'bug': {
                'summary': self.bug.title,
                'description': self.bug.description,
                'tags': ' '.join(self.bug.tags),
                'duplicate_of': self.bug.duplicate_of,
                'date_last_updated': self.bug.date_last_updated,
            },
            'tasks': {
                bug_task.target: {
                    'status': bug_task.status,
                    'importance': bug_task.importance,
                    'assignee': bug_task.assignee,
                    'milestone': bug_task.milestone,
                }
                for bug_task in self.bug.tasks
            },
        }

        return state

    def _build_bug_history(self):
        last_state = self._get_last_bug_state()
        history = [copy.deepcopy(last_state)]

        for interval, interval_activity in reversed(self.grouped_activity):
            state_copy = copy.deepcopy(history[-1])
            state_copy['interval'] = interval

            for change in reversed(interval_activity):
                self._rewind_to_prev(state_copy, change)

            history.append(state_copy)

        history.reverse()
        assert len(history) == len(self.grouped_activity) + 1

        for i, (interval, interval_activity) in enumerate(self.grouped_activity):
            state = history[i]
            for change in interval_activity:
                self._rewind_to_next(state, change)

            next_state = history[i + 1]
            for target in state['tasks']:
                if target in next_state['tasks']:
                    next_state['tasks'][target].update(
                        state['tasks'][target])

        history.pop()
        self._clean_history(history, last_state)
        return history

    def _clean_history(self, history, actual_last_state):
        last_state = history[-1]
        last_state['bug'].update(actual_last_state['bug'])
        last_targets = set(last_state['tasks'].keys())
        actual_last_targets = set(actual_last_state['tasks'].keys())

        excessive_targets = last_targets - actual_last_targets
        missing_targets = actual_last_targets - last_targets

        if excessive_targets:
            log.debug('%s found excessive_targets', self.bug)
            for t in excessive_targets:
                del last_state['tasks'][t]

        if missing_targets:
            log.debug('%s found missing_targets', self.bug)
            for t in missing_targets:
                last_state['tasks'][t] = {}

        for target in actual_last_targets:
            bug_task = actual_last_state['tasks'][target]
            last_state['tasks'][target].update(bug_task)

        for state in history:
            for target in state['tasks']:
                state['tasks'][target].setdefault('status_tracked_in', None)
                state['tasks'][target].setdefault('dev_focus', None)

                proj_name, series = split_target(target)
                if proj_name == target:
                    date = state['interval'][1]
                    dev_focus = get_focus_of_dev(proj_name, date)
                    if dev_focus in state['tasks']:
                        state['tasks'][proj_name]['status_tracked_in'] = dev_focus
                        state['tasks'][dev_focus]['dev_focus'] = True

    def _is_valuable_target(self, target):
        for valuable_target in self.VALUABLE_TARGETS:
            if target.startswith(valuable_target):
                return True
        return False

    def _prepare_activity(self, activity):

        assert activity
        # activity items must be sorted by date_changed field
        _activity_dates = [a.date_changed for a in activity]
        assert _activity_dates == sorted(_activity_dates)

        bug_create_change = activity[0]
        assert (
            bug_create_change.what_changed == 'bug' and
            bug_create_change.message == 'added bug'
        )

        def fix_assignee(change):

            def fix(assignee):
                return assignee and assignee[:-1].rsplit('(')[-1].strip()

            if change.what_changed.endswith(': assignee'):
                change.old_value = fix(change.old_value)
                change.new_value = fix(change.new_value)

            return change

        prepared_activity = []
        for idx, change in enumerate(activity):
            assert self._is_known_change(change), change
            change_copy = copy.copy(change)
            change_copy.index = idx
            fix_assignee(change_copy)
            prepared_activity.append(change_copy)

        return prepared_activity

    def _group_activity_by_interval(self, interval_type='day'):

        result = OrderedDict()

        if interval_type == self.INTERVAL_TYPE_DAY:
            grouped = itertools.groupby(
                self.activity, key=lambda a: a.date_changed.date())

            for date, day_activity in grouped:
                day_activity = tuple(day_activity)
                interval = (
                    day_activity[0].date_changed,
                    day_activity[-1].date_changed,
                )
                result[interval] = day_activity

        else:
            raise AssertionError(
                'Unsupported interval type {}'.format(interval_type)
            )
        return result

    def _rewind_to_next(self, state, change):

        def set_extra(bug_task, field):
            bug_task['{}_change_date'.format(field)] = change.date_changed
            bug_task['{}_change_person'.format(field)] = change.person

        def setup_initial_task(bug_task):
            bug_task['task_created'] = change.date_changed
            for field in self.BUG_TASK_FIELDS:
                bug_task.setdefault(field, self.BUG_TASK_DEFAULTS.get(field))
                set_extra(bug_task, field)

        if change.what_changed == 'bug' and change.message == 'added bug':
            for target in state['tasks']:
                setup_initial_task(state['tasks'][target])
            state['bug']['date_last_updated'] = change.date_changed

        elif change.what_changed in ('summary', 'description', 'tags'):
            state['bug'][change.what_changed] = change.new_value
            state['bug']['date_last_updated'] = change.date_changed

        elif change.what_changed in (
            'marked as duplicate', 'changed duplicate marker'
        ):
            state['bug']['duplicate_of'] = int(change.new_value)
            state['bug']['date_last_updated'] = change.date_changed

        elif change.what_changed == 'removed duplicate marker':
            state['bug']['duplicate_of'] = None
            state['bug']['date_last_updated'] = change.date_changed

        elif change.what_changed == 'bug task added':
            target = change.new_value
            state['tasks'][target] = {}
            setup_initial_task(state['tasks'][target])

            proj_name, series = split_target(target)
            if (
                series is not None and proj_name in state['tasks'] and
                get_focus_of_dev(proj_name, change.date_changed) == target
            ):
                state['tasks'][target].update(state['tasks'][proj_name])
                state['tasks'][target]['task_created'] = change.date_changed

            state['bug']['date_last_updated'] = change.date_changed

        elif change.what_changed == 'bug task deleted':
            target = change.old_value
            state['tasks'].pop(target, None)
            state['bug']['date_last_updated'] = change.date_changed

        elif change.what_changed == 'affects':
            old_target, new_target = change.old_value, change.new_value
            state['tasks'][new_target] = state['tasks'].pop(old_target)
            state['bug']['date_last_updated'] = change.date_changed

        elif ': ' in change.what_changed:
            target, field = change.what_changed.split(': ')

            if target not in state['tasks']:
                state['tasks'][target] = {}
                setup_initial_task(state['tasks'][target])

            state['tasks'][target][field] = change.new_value
            set_extra(state['tasks'][target], field)

            proj_name, series = split_target(target)
            fod = get_focus_of_dev(proj_name, change.date_changed)
            if series:
                if fod == target and proj_name in state['tasks']:
                    state['tasks'][proj_name][field] = change.new_value
                    set_extra(state['tasks'][proj_name], field)
            else:
                assert proj_name == target
                if fod in state['tasks']:
                    state['tasks'][fod][field] = change.new_value
                    set_extra(state['tasks'][fod], field)

            state['bug']['date_last_updated'] = change.date_changed

        # Hook for invalid bug activity in launchpad.
        # There are bug_tasks, with broken activity log.
        # For example, target "fuel/future" in bug 1528535, doesn't have
        # the creation record (record having what_changed='bug task added')
        # in it's activity log list.
        for task in state['tasks'].values():
            if 'task_created' not in task:
                setup_initial_task(task)

    def _rewind_to_prev(self, state, change):

        if change.what_changed in ('summary', 'description', 'tags'):
            state['bug'][change.what_changed] = change.old_value

        elif change.what_changed == 'marked as duplicate':
            state['bug']['duplicate_of'] = None

        elif change.what_changed in (
            'changed duplicate marker', 'removed duplicate marker'
        ):
            state['bug']['duplicate_of'] = int(change.old_value)

        elif change.what_changed == 'bug task added':
            target = change.new_value
            state['tasks'].pop(target, None)

        elif change.what_changed == 'bug task deleted':
            target = change.old_value
            state['tasks'][target] = {}

        elif change.what_changed == 'affects':
            old_target, new_target = change.old_value, change.new_value
            state['tasks'][old_target] = state['tasks'].pop(new_target, {})

    def _is_known_change(self, change):

        if change.what_changed == 'bug':
            return (
                change.message == 'added bug' or
                change.message.startswith('added subscriber')
            )

        if change.what_changed in (
            'summary',
            'description',
            'tags',
            'bug task added',
            'bug task deleted',
            'nominated for series',
            'affects',
            'branch linked',
            'branch unlinked',
            'attachment added',
            'attachment removed',
            'marked as duplicate',
            'removed duplicate marker',
            'changed duplicate marker',
            'information type',
            'bug watch added',
            'cve linked',
            'cve unlinked',
            'converted to question',
        ):
            return True

        if change.what_changed.startswith('removed subscriber'):
            return True

        if ':' in change.what_changed:
            field = change.what_changed.split(': ')[-1]
            return field in self.BUG_TASK_FIELDS

        return False
