import datetime
import itertools
import logging

import pytz

from . import utils
from .history_builder import HistoryBuilder
from .models import BugTaskHistory

log = logging.getLogger(__name__)


utc_max_date = pytz.utc.localize(datetime.datetime.max)


class BugTaskHistoryUpdater:

    def __init__(self, bug, *, projects):
        self.bug = bug
        self.projects = tuple(projects)
        assert self.bug.date_last_updated

    def perform_update(self):

        builder = HistoryBuilder(self.bug)
        full_history = builder.get_history()
        filtered_history = self._filter_history(full_history)

        history_by_targets = self._get_history_by_targets(filtered_history)
        saved_tasks_by_target = self._get_saved_bug_tasks_by_targets()

        for target, saved_tasks in saved_tasks_by_target.items():
            if target not in history_by_targets:
                log.info('Deleting all task for target %s', target)
                # Remove tasks for missing targets
                for task in saved_tasks:
                    task.delete()

        for target, target_hist in history_by_targets.items():
            saved_tasks = saved_tasks_by_target.get(target, [])
            self._process_target_history(target, saved_tasks, target_hist)

    def _is_valuable_target(self, target):
        for project in self.projects:
            if target == project.name or target.startswith(project.name + '/'):
                return True
        return False

    def _filter_history(self, full_history):
        """
        Just uses algorithm described here:
        https://mirantis.jira.com/wiki/display/QA/LPR+Milestones+Filtering+Algorithm

        :return: The same history items but with some bug tasks excluded.
        """
        filtered_history = []
        milestones_to_series_map = {}

        for project in self.projects:
            milestones_to_series_map[project.name] = {
                m.name: m.series for m in project.milestones
            }

        for state in full_history:
            state = state.copy()
            state['tasks'] = {
                target: task for target, task in state['tasks'].items()
                if self._is_valuable_target(target)
            }

            tasks_map = {}
            for target, bug_task in sorted(state['tasks'].items()):
                proj_name, series = utils.split_target(target)
                mi_name = bug_task.get('milestone')
                mi_ser = milestones_to_series_map[proj_name].get(mi_name)
                if target == proj_name:
                    status_tracked_target = '{}/{}'.format(
                        proj_name, mi_ser) if mi_ser else None
                    if status_tracked_target in state['tasks']:
                        continue
                tasks_map[(proj_name, mi_name)] = (target, bug_task)

            if tasks_map:
                state['tasks'] = dict(tasks_map.values())
                filtered_history.append(state)

        return filtered_history

    def _get_saved_bug_tasks_by_targets(self):
        result = {}

        query = BugTaskHistory.objects.filter(bug_id=self.bug.id)

        for bug_task in query:
            result.setdefault(bug_task.target, []).append(bug_task)

        for target in result:
            result[target].sort(key=lambda x: x.version_begin)

        return result

    def _get_history_by_targets(self, history):
        result = {}
        hist_by_targets = self._split_by_targets(history)
        for target, target_history in hist_by_targets.items():
            assert self._is_valuable_target(target)
            result[target] = self._prune_one_target_history(target_history)
        return result

    @staticmethod
    def _prune_one_target_history(target_history):

        def key_fn(hist):
            if hist['task'] is None:
                return None

            # Ignore duplicates
            if hist['bug']['duplicate_of'] is not None:
                return None

            return {
                'task': hist['task'],
                'bug': dict(hist['bug'], date_last_updated=None)
            }

        # convert iterator to plain list
        grouped_history = [(k, tuple(v)) for k, v in
                           itertools.groupby(target_history, key=key_fn)]
        result = []
        for item, next_item in itertools.zip_longest(
            grouped_history, grouped_history[1:]
        ):
            cur_key, cur_grouped = item
            if cur_key is None:
                continue

            period_begin = cur_grouped[0]['interval'][1]

            if next_item is not None:
                next_key, next_grouped = next_item
                period = (period_begin, next_grouped[0]['interval'][1])
            else:
                period = (period_begin, utc_max_date)

            result.append({
                'period': period,
                'task': cur_key['task'],
                'bug': cur_key['bug']
            })

        return result

    @staticmethod
    def _split_by_targets(history):
        result = {}
        all_targets = set(t for h in history for t in h['tasks'])
        for hist in history:
            for target in all_targets:
                target_hist = result.setdefault(target, [])
                task = hist['tasks'].get(target)
                item = {
                    'interval':  hist['interval'],
                    'task': task,  # task may be None
                    'bug': hist['bug'],
                }
                target_hist.append(item)
        return result

    def _process_target_history(self, target, saved_tasks, history):
        assert sorted(history, key=lambda x: x['period']) == history
        assert sorted(saved_tasks, key=lambda x: x.version_begin) == saved_tasks

        for hist, saved_task in itertools.zip_longest(history, saved_tasks):

            if hist is None:
                log.info('Deleting saved_task %s', repr(saved_task))
                saved_task.delete()
                continue

            if saved_task is not None:
                assert saved_task.target == target
                # TODO: If task fetched form DB is already equal to built `hist`
                # entry we can skip saving(updating) task object
                task = saved_task
            else:
                task = BugTaskHistory(target=target)

            self._fill_task(task, hist)
            task.save()

    def _fill_task(self, task, hist):
        task.bug_id = self.bug.to_dbref()
        task.task_created = hist['task']['task_created']
        task.version_begin = hist['period'][0]
        task.version_end = hist['period'][1]
        task.bug_created = self.bug.date_created
        task.title = hist['bug']['summary']
        task.tags = hist['bug']['tags'].split()

        for field in HistoryBuilder.BUG_TASK_FIELDS:
            for suffix in ('', '_change_date', '_change_person'):
                full_field = field + suffix
                assert hasattr(task, full_field)
                setattr(task, full_field, hist['task'].get(full_field))

    @classmethod
    def update_focus_of_development_history(cls):
        from .models import Bug
        from collections import defaultdict
        query = Bug.objects.order_by('date_last_updated')
        data = defaultdict(set)

        search_fields = [
            'date_assigned',
            'date_confirmed',
            'date_triaged',
            'date_in_progress',
            'date_fix_committed',
            'date_fix_released',
            'date_incomplete',
            'date_closed',
            'date_left_new',
            'date_left_closed',
        ]

        for i, bug in enumerate(query):
            project_tasks = [t for t in bug.tasks if '/' not in t.target]
            for p_task in project_tasks:
                series_tasks = [t for t in bug.tasks
                                if t.target.startswith(p_task.target + '/')]
                for s_task in series_tasks:
                    for field in search_fields:
                        dt1 = getattr(p_task, field)
                        dt2 = getattr(s_task, field)
                        if (
                            dt1 and dt2 and dt1 == dt2 and
                            dt1 >= s_task.date_created
                        ):
                            data[p_task.target].add((dt1, s_task.target))

            if i % 100 == 0:
                log.info('processed %s', i)

        ret = defaultdict(list)
        for proj in data:
            sorted_dates = sorted(data[proj])
            for target, items in itertools.groupby(sorted_dates, lambda x: x[1]):
                items = tuple(items)
                first, last = items[0], items[-1]
                if (last[0] - first[0]).total_seconds() > 60:
                    ret[proj].extend([first, last])

        return ret
