import logging
import asyncio
import urllib.parse

from oauthlib import oauth1
from flask import current_app
from .consts import ALL_STATUSES

log = logging.getLogger(__name__)


class CollectionIterator:

    def __init__(self, url, client):
        self.client = client

        self._next_chunk_url = url
        self._curr_entries = []
        self._curr_entry_index = 0

    async def __aiter__(self):
        return self

    async def __anext__(self):
        if self._curr_entry_index >= len(self._curr_entries):
            if self._next_chunk_url:
                data = await self.client.send_request(self._next_chunk_url)
                self._curr_entries = data['entries']

                if not self._curr_entries:
                    raise StopAsyncIteration

                self._next_chunk_url = data.get('next_collection_link')
                self._curr_entry_index = 0
            else:
                raise StopAsyncIteration

        result = self._curr_entries[self._curr_entry_index]
        self._curr_entry_index += 1
        return result


class LaunchpadClientError(Exception):

    """Error raised by LaunchpadClient ."""


class LaunchpadClient:

    API_URL = 'https://api.launchpad.net/devel'
    SEND_REQUEST_ATTEMPTS = 4

    def __init__(self, session):

        self.http_session = session

        self.oauth_client = oauth1.Client(
            client_key=current_app.config['LAUNCHPAD_CONSUMER_KEY'],
            resource_owner_key=(
                current_app.config['LAUNCHPAD_OAUTH_ACCESS_TOKEN']),
            resource_owner_secret=(
                current_app.config['LAUNCHPAD_OAUTH_ACCESS_TOKEN_SECRET']),
            signature_method=oauth1.SIGNATURE_PLAINTEXT,
            signature_type=oauth1.SIGNATURE_TYPE_AUTH_HEADER,
        )

    def get_tasks_search_url(self, target, modified_since=None,
                             omit_duplicates=False):
        query_params = [
            ('ws.op', 'searchTasks'),
            ('omit_duplicates', 'true' if omit_duplicates else 'false'),
        ]

        if modified_since and modified_since.year > 1:
            query_params += [('modified_since', modified_since.isoformat())]

        query_params += [('status', status) for status in ALL_STATUSES]

        url = "{0}/{1}/?{2}".format(
            self.API_URL,
            target,
            urllib.parse.urlencode(query_params)
        )

        return url

    def get_bug_url(self, bug_id):
        return '{0}/bugs/{1}'.format(self.API_URL, bug_id)

    def get_bug_activity_url(self, bug_id):
        return '{0}/activity'.format(self.get_bug_url(bug_id))

    def get_bug_messages_url(self, bug_id):
        return '{0}/messages'.format(self.get_bug_url(bug_id))

    def get_bug_tasks_url(self, bug_id):
        return '{0}/bug_tasks'.format(self.get_bug_url(bug_id))

    def get_project_url(self, project):
        return '{0}/{1}'.format(self.API_URL, project)

    def get_person_url(self, username):
        return '{0}/~{1}'.format(self.API_URL, username)

    def collection(self, url):
        return CollectionIterator(url, client=self)

    async def _send_request(self, url, method='GET'):

        headers = self.oauth_client.sign(url)[1]

        request = self.http_session.request(
            method,
            url,
            allow_redirects=False,
            headers=headers,
        )

        async with request as response:

            if response.status == 200:
                result = await response.json()
                return result
            else:
                log.debug('Response status is not equal to 200! '
                          'url: %s. response: %s', url, response)
                raise LaunchpadClientError('Request Failed')

    async def send_request(self, url, method='GET'):

        for attempt_index in range(1, self.SEND_REQUEST_ATTEMPTS + 1):
            is_last_attempt = (attempt_index == self.SEND_REQUEST_ATTEMPTS)
            if attempt_index > 1:
                log.info('Send request (attempt %s):  %s', attempt_index, url)

            try:
                return await self._send_request(url, method)
            except Exception:
                if not is_last_attempt:
                    asyncio.sleep(1)
                    continue
                log.exception('Error during sending request %s', url)
                raise
