import mongoengine

from . import consts


class BugMessages(mongoengine.EmbeddedDocument):

    content = mongoengine.StringField(null=True)
    date_created = mongoengine.DateTimeField(null=True)
    subject = mongoengine.StringField(null=True)


class BugTask(mongoengine.EmbeddedDocument):

    target = mongoengine.StringField(required=True)

    status = mongoengine.StringField(null=True)
    importance = mongoengine.StringField(null=True)
    assignee = mongoengine.StringField(null=True)
    milestone = mongoengine.StringField(null=True)

    date_created = mongoengine.DateTimeField(required=True)
    date_confirmed = mongoengine.DateTimeField(null=True)
    date_triaged = mongoengine.DateTimeField(null=True)
    date_assigned = mongoengine.DateTimeField(null=True)
    date_in_progress = mongoengine.DateTimeField(null=True)
    date_fix_committed = mongoengine.DateTimeField(null=True)
    date_fix_released = mongoengine.DateTimeField(null=True)
    date_incomplete = mongoengine.DateTimeField(null=True)
    date_closed = mongoengine.DateTimeField(null=True)
    date_left_new = mongoengine.DateTimeField(null=True)
    date_left_closed = mongoengine.DateTimeField(null=True)


class BugActivity(mongoengine.EmbeddedDocument):

    what_changed = mongoengine.StringField(required=True)
    date_changed = mongoengine.DateTimeField(required=True)
    person = mongoengine.StringField(required=True)
    old_value = mongoengine.StringField(null=True, default=None)
    new_value = mongoengine.StringField(null=True, default=None)
    message = mongoengine.StringField(null=True)

    def __str__(self):
        return (
            '{a.date_changed:%Y-%m-%d %H:%M:%S} '
            '~{a.person} [{a.what_changed}] '
            '"{a.old_value}" => "{a.new_value}" ({a.message})'.format(a=self)
        )


class Bug(mongoengine.DynamicDocument):

    id = mongoengine.IntField(primary_key=True)
    title = mongoengine.StringField(null=True)
    description = mongoengine.StringField(null=True)

    date_created = mongoengine.DateTimeField(required=True)
    date_last_updated = mongoengine.DateTimeField(required=True)
    date_last_synced = mongoengine.DateTimeField(required=True)

    duplicate_of = mongoengine.IntField(null=True)
    verifier = mongoengine.StringField(null=True)
    tags = mongoengine.ListField(field=mongoengine.StringField())

    tasks = mongoengine.EmbeddedDocumentListField(BugTask)
    activity = mongoengine.EmbeddedDocumentListField(BugActivity)
    messages = mongoengine.EmbeddedDocumentListField(BugMessages)

    user_comment = mongoengine.StringField(null=True)
    eta = mongoengine.StringField(null=True)

    meta = {
        'collection': 'launchpad_bug',
        'indexes': [
            'date_created',
            'date_last_updated',
            'date_last_synced',
        ]
    }

    def __str__(self):
        return '({}: {})'.format(
            self.id, ' '.join(t.target for t in self.tasks)
        )


class BugTaskHistory(mongoengine.DynamicDocument):

    """
    History of bug tasks.

    NOTE: This collection does not include tasks of duplicated bugs.
    """

    bug_id = mongoengine.ReferenceField(Bug, required=True)

    target = mongoengine.StringField(required=True)

    task_created = mongoengine.DateTimeField(required=True)

    version_begin = mongoengine.DateTimeField(
        required=True,
        verbose_name='Version relevance begin date'
    )

    version_end = mongoengine.DateTimeField(
        required=True,
        verbose_name='Version relevance end date'
    )

    bug_created = mongoengine.DateTimeField(required=True)
    title = mongoengine.StringField(required=True)
    tags = mongoengine.ListField(field=mongoengine.StringField())

    status = mongoengine.StringField(required=True)
    status_change_date = mongoengine.DateTimeField(null=True)
    status_change_person = mongoengine.StringField(null=True)

    importance = mongoengine.StringField(null=True)
    importance_change_date = mongoengine.DateTimeField(null=True)
    importance_change_person = mongoengine.StringField(null=True)

    assignee = mongoengine.StringField(null=True)
    assignee_change_date = mongoengine.DateTimeField(null=True)
    assignee_change_person = mongoengine.StringField(null=True)

    milestone = mongoengine.StringField(null=True)
    milestone_change_date = mongoengine.DateTimeField(null=True)
    milestone_change_person = mongoengine.StringField(null=True)

    project = mongoengine.StringField(required=True)

    # This field is computed using assignee's major_team_ids.
    major_team_ids = mongoengine.SortedListField(
        mongoengine.IntField(),
    )
    dev_to_qa_team_id = mongoengine.IntField(null=True)

    meta = {
        'collection': 'launchpad_bug_task_history',
        'indexes': [
            'bug_id',
            'target',
            'status',
            'version_begin',
            'version_end',
        ]
    }

    def clean(self):
        if not self.version_begin < self.version_end:
            raise mongoengine.ValidationError(
                'version_end should be greater than version_begin')

    def get_major_team_ids(self):

        def for_user(username):
            if username:
                lp_user = LaunchpadUser.objects.with_id(username)
                if lp_user:
                    return lp_user.major_team_ids

        return for_user(self.assignee) or []

    def get_dev_to_qa_team_id(self):
        def for_user(username):
            if username:
                lp_user = LaunchpadUser.objects.with_id(username)
                if lp_user:
                    teams = lp_user.teams
                    if 'mos-qa' in teams:
                        return consts.MajorQATeams.TEAM_MOS_ID
                    elif any(team in teams for team in consts.FUEL_TEAMS):
                        return consts.MajorQATeams.TEAM_FUEL_ID
                    elif any(team in teams for team in consts.PCE_TEAMS):
                        return consts.MajorQATeams.TEAM_PCE_ID
        return for_user(self.assignee) or None

    def save(self, *args, **kwargs):
        self.project = self.target.split('/', 1)[0]
        self.major_team_ids = self.get_major_team_ids()
        self.dev_to_qa_team_id = self.get_dev_to_qa_team_id()
        super().save(*args, **kwargs)

    def __str__(self):
        return 'bug-{}:{}'.format(self.bug_id, self.target)


class LaunchpadProject(mongoengine.DynamicDocument):

    class Milestone(mongoengine.DynamicEmbeddedDocument):

        name = mongoengine.StringField(required=True, unique=True)
        title = mongoengine.StringField(required=True)
        series = mongoengine.StringField(required=True)
        is_active = mongoengine.BooleanField(required=True)
        date_targeted = mongoengine.DateTimeField(null=True)

    class Series(mongoengine.DynamicEmbeddedDocument):

        name = mongoengine.StringField(required=True, unique=True)
        title = mongoengine.StringField(required=True)
        is_active = mongoengine.BooleanField(required=True)
        status = mongoengine.StringField(required=True)

    name = mongoengine.StringField(primary_key=True)
    display_name = mongoengine.StringField(required=True)
    milestones = mongoengine.EmbeddedDocumentListField(Milestone)
    series = mongoengine.EmbeddedDocumentListField(Series)

    meta = {
        'collection': 'launchpad_project',
    }

    def __str__(self):
        return self.name


class LaunchpadUser(mongoengine.DynamicDocument):

    username = mongoengine.StringField(primary_key=True)
    email = mongoengine.StringField(
        null=True, verbose_name='Primary email address')
    display_name = mongoengine.StringField()

    # Launchpad users and teams actually is the same.
    is_team = mongoengine.BooleanField(required=True)

    # List of username strings.
    teams = mongoengine.ListField(
        mongoengine.StringField(),
        verbose_name='Teams where user is member'
    )

    major_team_ids = mongoengine.SortedListField(
        mongoengine.IntField(),
    )

    meta = {
        'collection': 'launchpad_user',
        'indexes': ['email', 'username'],
    }

    def __str__(self):
        return self.username


class LpMilestoneInFocus(mongoengine.Document):
    name = mongoengine.StringField(required=True)

    meta = {
        'collection': 'launchpad_milestone_infocus',
    }
