import asyncio
import datetime
import logging
import aiohttp

from flask import current_app

import qareporting.integration.utils
from qareporting.integration.utils import Pool
from qareporting.integration.models import EventLog
from qareporting.integration.sync_base import SyncBase
from qareporting.utils import dt_now

from . import consts
from . import models
from . import utils
from .history_updater import BugTaskHistoryUpdater
from .launchpad_client import LaunchpadClient


log = logging.getLogger(__name__)

LAST_SYNC = 'last_sync'


class LaunchpadBugsSync(SyncBase):

    event_code = EventLog.EVENT_BUGS_SYNC

    client = None
    targets = consts.PROJECTS
    BUG_PROCESSING_CONCURRENCY = 10
    BUG_PROCESSING_TIMEOUT = 20

    def __init__(self, *, modified_since=LAST_SYNC, update_history=True):

        self.bugs_to_sync = set()
        self.sync_failed_bugs = []
        self.loop = asyncio.get_event_loop()

        assert (modified_since == LAST_SYNC or
                isinstance(modified_since, datetime.datetime))

        self.modified_since = modified_since
        self.update_history = update_history
        self.projects = None

    async def _collect_bug_ids_for_target(self, target):
        url = self.client.get_tasks_search_url(target, self.modified_since)
        cnt = 0
        async for bug_task in self.client.collection(url):
            bug_id = int(bug_task['bug_link'].rpartition('/')[2])
            self.bugs_to_sync.add(bug_id)
            cnt += 1
            if cnt % 100 == 0:
                log.info('collected bugs for "%s":  %s', target, cnt)

    def _collect_bugs_to_sync(self):
        log.info('Begin collecting bug ids for projects %s', self.projects)

        tasks = [self._collect_bug_ids_for_target(proj.name)
                 for proj in self.projects]

        self.loop.run_until_complete(asyncio.wait(tasks))

        log.info('total collected bugs count %s', len(self.bugs_to_sync))

    async def _fetch_and_save_bug(self, bug_id):

        now = dt_now()

        raw_bug = await self.client.send_request(
            self.client.get_bug_url(bug_id))

        raw_bug_activity = await qareporting.integration.utils.to_list(
            self.client.collection(self.client.get_bug_activity_url(bug_id)))

        raw_bug_tasks = await qareporting.integration.utils.to_list(
            self.client.collection(self.client.get_bug_tasks_url(bug_id)))

        raw_bug_messages = await qareporting.integration.utils.to_list(
            self.client.collection(self.client.get_bug_messages_url(bug_id)))

        bug = utils.bug_from_raw(
            raw_bug,
            raw_bug_tasks,
            raw_bug_activity,
            raw_bug_messages
        )

        bug.date_last_synced = now

        saved_bug = models.Bug.objects(
            id=bug.id,
        ).modify(
            upsert=True, new=True,
            set__title=bug.title,
            set__description=bug.description,
            set__tags=bug.tags,
            set__date_created=bug.date_created,
            set__date_last_updated=bug.date_last_updated,
            set__activity=bug.activity,
            set__tasks=bug.tasks,
            set__duplicate_of=bug.duplicate_of,
            set__date_last_synced=now,
            set__messages=bug.messages,
        )
        return saved_bug

    async def _sync_bug(self, bug_id):

        try:
            with aiohttp.Timeout(self.BUG_PROCESSING_TIMEOUT):
                bug = await self._fetch_and_save_bug(bug_id)
        except Exception:
            log.exception('Bug %s sync error:', bug_id)
            self.sync_failed_bugs.append(bug_id)
            raise

        if self.update_history:
            try:
                updater = BugTaskHistoryUpdater(bug, projects=self.projects)
                updater.perform_update()
            except Exception:
                log.exception('Bug %s history updating error:', bug_id)
                raise

        return bug

    def _sync_collected_bugs(self):
        total = len(self.bugs_to_sync)

        def tasks_gen():
            processed_cnt = 0
            for bug_id in self.bugs_to_sync:
                yield self._sync_bug(bug_id)
                processed_cnt += 1
                if processed_cnt % 10 == 0:
                    log.info('processed bugs: %s/%s',  processed_cnt, total)

        pool = Pool(
            self.BUG_PROCESSING_CONCURRENCY, tasks_gen(), loop=self.loop)
        self.loop.run_until_complete(pool.run())

    def _setup(self):
        super()._setup()
        self.projects = tuple(models.LaunchpadProject.objects.all())
        if self.modified_since == LAST_SYNC:
            last_successful_sync = self._get_last_successful_sync()

            if last_successful_sync:
                assert last_successful_sync.finished <= self.sync_start_time
                self.modified_since = last_successful_sync.started
            else:
                self.modified_since = None

    def _sync(self):
        log.info(
             'Sync params modified_since=%s. update_history=%s.',
             self.modified_since,
             self.update_history,
         )

        with aiohttp.ClientSession(loop=self.loop) as session:
            self.client = LaunchpadClient(session=session)

            self._collect_bugs_to_sync()

            self.sync_log.result.update({
                'collected_bugs_count': len(self.bugs_to_sync),
            })

            self._sync_collected_bugs()

    def _sync_finally_action(self):
        if self.sync_failed_bugs:
            self.sync_log.result['sync_failed_bugs'] = self.sync_failed_bugs
            log.warn('There were %s bugs failed syncing!',
                     len(self.sync_failed_bugs))


class LaunchpadProjectsSync(SyncBase):

    event_code = EventLog.EVENT_PROJECTS_SYNC
    client = None
    projects = consts.PROJECTS

    async def _sync_project(self, project_name):

        log.info('Syncing project %s', project_name)

        raw_project = await self.client.send_request(
            self.client.get_project_url(project_name))

        raw_milestones = await qareporting.integration.utils.to_list(self.client.collection(
            raw_project['all_milestones_collection_link']))

        raw_series = await qareporting.integration.utils.to_list(self.client.collection(
            raw_project['series_collection_link']))

        milestones = [
            models.LaunchpadProject.Milestone(
                name=d['name'],
                title=d['title'],
                series=utils.parse_link(d['series_target_link']),
                is_active=d['is_active'],
                date_targeted=d.get('date_targeted'),
            ) for d in raw_milestones
            ]

        series = [
            models.LaunchpadProject.Series(
                name=d['name'],
                title=d['title'],
                is_active=d['active'],
                status=d['status'],
            ) for d in raw_series
            ]

        project = models.LaunchpadProject(
            name=raw_project['name'],
            display_name=raw_project['display_name'],
            milestones=milestones,
            series=series,
        )
        project.save()

    def _sync(self):
        loop = asyncio.get_event_loop()

        with aiohttp.ClientSession(loop=loop) as session:
            self.client = LaunchpadClient(session=session)
            coros = [self._sync_project(proj) for proj in self.projects]
            loop.run_until_complete(asyncio.wait(coros))


class TeamsMappingSync(SyncBase):

    event_code = EventLog.EVENT_TEAMS_MAPPING_SYNC

    def __init__(self):
        self.users = {}

    async def _get_teams_mapping(self):
        token = current_app.config['LPREPORTS_AUTH_TOKEN']
        headers = {'Authorization': 'Token {}'.format(token)}
        url = '{}/full_teams'.format(current_app.config['LPREPORTS_API_URL'])
        async with self.session.get(url, headers=headers) as resp:
            return await resp.json()

    async def _sync_teams(self):
        full_teams = await self._get_teams_mapping()

        code_to_id_map = {
            team_code: team_id for team_id, team_code in
            consts.MajorQATeams.CODES_IN_LP_REPORTS.items()
        }

        for major_team_name, subteam in full_teams.items():
            team_id = code_to_id_map.get(major_team_name)
            if team_id is None:
                log.warn(
                    'Unknown major_team "%s"!.'
                    'Fix MajorQATeams.CODES_IN_LP_REPORTS',
                    major_team_name,
                )
                continue

            for subteam_name, subteam_members in subteam.items():
                for username in subteam_members:
                    user_info = self.users.setdefault(username, {})
                    user_info['username'] = username
                    user_info.setdefault('teams', set()).add(subteam_name)
                    user_info.setdefault(
                        'major_team_ids', set()
                    ).add(team_id)

        return full_teams

    async def _sync_user(self, username):
        user_info = self.users[username]

        person_url = self.client.get_person_url(username)
        data = await self.client.send_request(person_url)

        email = None
        email_link = data['preferred_email_address_link']
        if email_link and '@' in email_link:
            email = email_link.rsplit('/', 1)[-1]

        user = models.LaunchpadUser(
            username=username,
            is_team=data['is_team'],
            display_name=data['display_name'],
            email=email,
            teams=tuple(user_info['teams']),
            major_team_ids=tuple(user_info['major_team_ids']),
        )
        user.save()

        log.info('Synced %s (email=%s)', username, email)

    async def _sync_users(self):
        tasks = (self._sync_user(username) for username in self.users)
        pool = Pool(15, tasks, loop=self.loop)
        await pool.run()

        models.LaunchpadUser.objects.filter(
            username__nin=self.users.keys()
        ).delete()

    def _sync(self):
        self.loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=self.loop) as session:
            self.session = session
            self.client = LaunchpadClient(session=session)
            self.loop.run_until_complete(self._sync_teams())
            self.loop.run_until_complete(self._sync_users())
