import bisect

from qareporting.integration.launchpad.models import LaunchpadUser
from qareporting.utils import dt_from_iso
from . import consts
from . import models


def parse_link(link):
    if link:
        ret = link.rpartition('/')[2]
        return ret.strip().strip('~')


def parse_date(value):
    if value:
        return dt_from_iso(value)


def bug_from_raw(raw_bug, raw_bug_tasks, raw_bug_activity, raw_bug_messages):

    bug_activity = [
        models.BugActivity(
            what_changed=d['whatchanged'],
            date_changed=parse_date(d['datechanged']),
            person=parse_link(d['person_link']),
            old_value=d['oldvalue'],
            new_value=d['newvalue'],
            message=d['message'],
        ) for d in raw_bug_activity
    ]

    bug_tasks = [
        models.BugTask(
            target=d['bug_target_name'],
            status=d['status'],
            importance=d['importance'],
            assignee=parse_link(d['assignee_link']),
            milestone=parse_link(d['milestone_link']),
            date_created=parse_date(d['date_created']),
            date_confirmed=parse_date(d['date_confirmed']),
            date_triaged=parse_date(d['date_triaged']),
            date_assigned=parse_date(d['date_assigned']),
            date_in_progress=parse_date(d['date_in_progress']),
            date_fix_committed=parse_date(d['date_fix_committed']),
            date_fix_released=parse_date(d['date_fix_released']),
            date_incomplete=parse_date(d['date_incomplete']),
            date_closed=parse_date(d['date_closed']),
            date_left_new=parse_date(d['date_left_new']),
            date_left_closed=parse_date(d['date_left_closed']),
        ) for d in raw_bug_tasks
    ]

    bug_messages = [
        models.BugMessages(
            content=d['content'],
            date_created=parse_date(d['date_created']),
            subject=d['subject']
        ) for d in raw_bug_messages
    ]

    duplicate_of = parse_link(raw_bug['duplicate_of_link'])
    if duplicate_of:
        duplicate_of = int(duplicate_of)

    bug = models.Bug(
        id=int(raw_bug['id']),
        title=raw_bug['title'],
        description=raw_bug['description'],
        tags=raw_bug['tags'],
        date_created=parse_date(raw_bug['date_created']),
        date_last_updated=parse_date(raw_bug['date_last_updated']),
        duplicate_of=duplicate_of,
        activity=bug_activity,
        tasks=bug_tasks,
        messages=bug_messages,
    )

    return bug


_focus_of_dev = None


def get_focus_of_dev(project, date):
    global _focus_of_dev
    if _focus_of_dev is None:
        _focus_of_dev = {
            proj: [(dt_from_iso(date), target) for date, target in hist]
            for proj, hist in consts.FOCUS_OF_DEVELOPMENT_HISTORY.items()
            }

    hist = _focus_of_dev.get(project)
    if hist is not None:
        i = bisect.bisect_left(hist, (date, '')) - 1
        i = max(0, i)
        i = min(len(hist) - 1, i)
        return hist[i][1]


def get_projects():
    query = models.LaunchpadProject.objects.all()
    return {proj.name: proj for proj in query}


def split_target(target):
    if '/' in target:
        proj, rest = target.split('/', 1)
    else:
        proj, rest = target, None
    return proj, rest


class LaunchpadUsersFinder:

    def __init__(self):
        self._mapping = self._get_mapping()

    @staticmethod
    def _get_mapping():
        mapping = {}
        for user in LaunchpadUser.objects.all():
            mapping[user.username] = user
            if user.email:
                mapping[user.email] = user
        return mapping

    def find_user(self, email=None, username=None):
        assert email or username

        if email and email in self._mapping:
            return self._mapping[email]

        if username and username in self._mapping:
            return self._mapping[username]
