import bisect
import datetime

from bson import DBRef
import mongoengine


class DenyMicrosValidator:

    def __call__(self, value):
        if (isinstance(value, datetime.datetime) and
                value.microsecond % 1000 != 0):
            return False
        return True


class ExtSetattrMixin:

    def _set_ref_field(self, field, value):
        ref = None
        if value is not None:
            coll_name = field.document_type_obj._meta['collection']
            ref = DBRef(coll_name, value)

        if field.name in self._data:
            obj = self._data[field.name]
            if obj is None or obj.id != value:
                setattr(self, field.name, ref)
        else:
            setattr(self, field.name, ref)

    def __setattr__(self, name, value):
        if name and name.endswith('_id'):
            field = self._fields.get(name[:-3])
            if isinstance(field, mongoengine.ReferenceField):
                self._set_ref_field(field, value)
                return

        super().__setattr__(name, value)

    def mark_as_changed(self, fields=None):
        if fields is None:
            fields = tuple(self)
        for field_name in fields:
            self._mark_as_changed(field_name)


class IntervalDocument(mongoengine.DynamicDocument):

    class OverlapError(mongoengine.ValidationError):

        def __init__(self, *args, **kwargs):
            self.overlapped = kwargs.pop('overlapped')
            super().__init__(*args, **kwargs)

    MINIMAL_INTERVAL = datetime.timedelta(seconds=1)
    INTERVAL_FIELDS = None
    DIMENSION_FIELDS = None
    OVERLAP_MESSAGE = 'Intervals overlap!'

    meta = {
        'abstract': True,
    }

    @property
    def interval_begin(self):
        return getattr(self, self.INTERVAL_FIELDS[0])

    @property
    def interval_end(self):
        return getattr(self, self.INTERVAL_FIELDS[1])

    def _ensure_no_overlaps(self, versions, exc_msg=None):
        exc_msg = exc_msg or self.OVERLAP_MESSAGE
        prev_obj = None
        for obj in versions:
            assert obj.interval_begin < obj.interval_end
            if prev_obj and not prev_obj.interval_end <= obj.interval_begin:
                raise self.OverlapError(exc_msg, overlapped=(prev_obj, obj))
            prev_obj = obj

    def _clean_overlaps(self):
        other_objects_in_dimension = tuple(
            self.get_dimension_query(include_self=False))

        self._ensure_no_overlaps(other_objects_in_dimension)

        if not other_objects_in_dimension:
            dim_objects = [self]
        else:
            pairs = [(x.interval_begin, x) for x in other_objects_in_dimension]
            bisect.insort_right(pairs, (self.interval_begin, self))
            dim_objects = [obj for interval_begin, obj in pairs]

        self._ensure_no_overlaps(dim_objects)

    def clean(self):
        super().clean()
        if not self.interval_begin < self.interval_end:
            raise mongoengine.ValidationError(
                '{} must be less than {}'.format(self.INTERVAL_FIELDS),
                field_name=self.INTERVAL_FIELDS[0]
            )
        delta = self.interval_end - self.interval_begin
        if not delta >= self.MINIMAL_INTERVAL:
            raise mongoengine.ValidationError(
                'Interval must be longer than {} seconds'.format(
                    self.MINIMAL_INTERVAL.total_seconds())
            )

        self._clean_overlaps()

    def get_dimension_query(self, include_self=True):
        filter_dict = {
            dim_field: getattr(self, dim_field)
            for dim_field in self.DIMENSION_FIELDS
        }

        if not all(filter_dict.values()):
            raise mongoengine.ValidationError(
                'Ensure that all dimensions fields {} are set.'.format(
                    self.DIMENSION_FIELDS)
            )

        if self.pk and not include_self:
            filter_dict['pk__ne'] = self.pk

        return self._qs.filter(**filter_dict).order_by(
            self.INTERVAL_FIELDS[0])

    def save(self, **kwargs):
        assert self.interval_begin < self.interval_end
        assert self.interval_end
        super().save(**kwargs)


class VersionedDocument(IntervalDocument):

    VERSIONED_FIELDS = None

    INTERVAL_FIELDS = ('version_begin', 'version_end')
    OVERLAP_MESSAGE = 'Versions are overlapped!'

    version_begin = mongoengine.DateTimeField(
        required=True,
        validation=DenyMicrosValidator(),
    )

    version_end = mongoengine.DateTimeField(
        required=True,
        validation=DenyMicrosValidator(),
    )

    meta = {
        'abstract': True,
    }

    @classmethod
    def has_difference(cls, old_instance, new_instance):
        for field_name in cls.VERSIONED_FIELDS:
            old_val = getattr(old_instance, field_name)
            new_val = getattr(new_instance, field_name)
            if old_val != new_val:
                return True
        return False

    @classmethod
    def push_new_version(cls, new_obj, *,
                         allow_merge=True, allow_replace=True):
        """
        Push `new_obj` at the end of versions history or updates previous.
        """

        if new_obj.pk is not None:
            raise mongoengine.ValidationError('new_obj must be without pk!')

        # Get latest already saved version of this object
        prev_obj = new_obj.get_dimension_query().order_by(
            '-version_begin').first()

        if prev_obj:
            if not prev_obj.version_begin <= new_obj.version_begin:
                raise mongoengine.ValidationError(
                    'new version must go after last version')
            if prev_obj.version_begin == new_obj.version_begin:
                if allow_replace:
                    new_obj.pk = prev_obj.pk
                    new_obj.save()
                else:
                    raise mongoengine.ValidationError('Cant replace')
            elif prev_obj.version_end > new_obj.version_begin:

                if allow_merge and not cls.has_difference(prev_obj, new_obj):
                    new_obj.pk = prev_obj.pk
                    new_obj.version_begin = prev_obj.version_begin
                    new_obj.save()
                else:
                    prev_obj.version_end = new_obj.version_begin
                    prev_obj.save()
                    new_obj.save()

            elif prev_obj.version_end <= new_obj.version_begin:
                new_obj.save()
            else:
                raise AssertionError('Unexpected case')
        else:
            new_obj.save()

        return new_obj


class EventLog(mongoengine.DynamicDocument):

    EVENT_BUGS_SYNC = 'LP_BUGS_SYNC'
    EVENT_PROJECTS_SYNC = 'LP_PROJECTS_SYNC'
    EVENT_JIRA_SYNC = 'JIRA_SYNC'
    EVENT_TEAMS_MAPPING_SYNC = 'LPREPORTS_TEAMS_MAPPING_SYNC'
    EVENT_TESTRAIL_SYNC = 'TESTRAIL_SYNC'
    EVENT_TESTRAIL_CASES_SYNC = 'TESTRAIL_CASES_SYNC'

    EVENTS = (
        EVENT_BUGS_SYNC,
        EVENT_PROJECTS_SYNC,
        EVENT_JIRA_SYNC,
        EVENT_TEAMS_MAPPING_SYNC,
        EVENT_TESTRAIL_SYNC,
        EVENT_TESTRAIL_CASES_SYNC,
    )

    event = mongoengine.StringField(
        required=True,
        choices=EVENTS,
    )

    started = mongoengine.DateTimeField(required=True)
    finished = mongoengine.DateTimeField(null=True)
    success = mongoengine.BooleanField(required=True, default=False)
    error = mongoengine.StringField(null=True)
    result = mongoengine.DictField(null=True)

    meta = {
        'collection': 'event_log',
        'indexes': [
            'started',
            'finished',
        ],
        'ordering': ['-id'],
    }

    def __str__(self):
        return '{0} - {1}'.format(
            self.started,
            self.finished or '...',
        )

    @classmethod
    def get_last(cls, event, *, finished=True, succeeded=True):
        query = cls.objects.filter(event=event)
        if finished:
            query = query.filter(finished__ne=None)
        if succeeded:
            query = query.filter(success=True)
        return query.order_by('-started').first()


class Iteration(IntervalDocument):

    INTERVAL_FIELDS = ['date_begin', 'date_end']
    DIMENSION_FIELDS = ['milestone']

    milestone = mongoengine.StringField(required=True)

    date_begin = mongoengine.DateTimeField(
        required=True, validation=DenyMicrosValidator()
    )
    date_end = mongoengine.DateTimeField(
        required=True, validation=DenyMicrosValidator()
    )

    index = mongoengine.IntField(required=True, unique_with='milestone')


class DashboardMilestone(mongoengine.Document):
    name = mongoengine.StringField(required=True)

    meta = {
        'collection': 'dashboard_milestone',
    }
