import logging
import traceback

from qareporting.integration.models import EventLog
from qareporting.utils import dt_now


log = logging.getLogger(__name__)


class SyncBase:

    event_code = None
    sync_start_time = None
    sync_log = None

    def _get_last_successful_sync(self):
        ret = EventLog.objects.filter(
            event=self.event_code,
            success=True,
        ).order_by('-started').first()
        if ret:
            assert ret.finished
        return ret

    def _setup(self):
        assert self.event_code is not None

        self.sync_start_time = dt_now()

        last_syncs_query = EventLog.objects.filter(
            event=self.event_code
        ).order_by('-started')

        # Prohibit concurrent syncing
        assert last_syncs_query.filter(finished=None).count() == 0

        self.sync_log = EventLog(
            event=self.event_code,
            success=False,
            started=self.sync_start_time,
            result={},
        )

        self.sync_log.save()

    def _sync_finally_action(self):
        """
        Action to perform after sync finished (regardless successful or not).
        """

    def _sync(self):
        raise NotImplementedError

    def run(self):
        self._setup()

        log.info('Start %s', self.event_code)

        try:
            self._sync()
        except Exception:
            self.sync_log.success = False
            self.sync_log.error = ''.join(traceback.format_exc())
            log.exception('Sync failed')
            raise
        else:
            self.sync_log.success = True
        finally:
            self.sync_log.finished = dt_now()
            self.sync_log.save()
            self._sync_finally_action()

            log.info(
                'Finish %s.%s',
                self.event_code,
                (' (with error)' if not self.sync_log.success else '')
            )
