import logging
from collections import OrderedDict
import re

import mongoengine

from qareporting.integration.models import DenyMicrosValidator
from qareporting.integration.models import ExtSetattrMixin
from qareporting.integration.models import VersionedDocument
from qareporting.utils import dt_now
from qareporting.utils import prefixed_collection_name
from qareporting.utils import utc_max_date

from .consts import QATeams


log = logging.getLogger(__name__)
mu_pattern = re.compile(r'MU-?(?P<postfix>\d)')


class TestRailDocument(ExtSetattrMixin, mongoengine.DynamicDocument):

    id = mongoengine.IntField(primary_key=True)

    meta = {
        'abstract': True,
        'collection': prefixed_collection_name('testrail_'),
    }

    def __str__(self):
        if hasattr(self, 'name'):
            return self.name
        return str(self.id)


class User(TestRailDocument):

    name = mongoengine.StringField(required=True)
    email = mongoengine.StringField(required=True)
    is_active = mongoengine.BooleanField(required=True)


class CaseType(TestRailDocument):

    AUTOMATED_ID = 1
    MANUAL_ID = 7
    AUTOMATION_IN_PROGRESS_ID = 8
    NOT_AUTOMATE_ID = 10
    CI_INTEGRATED = 11

    NAMES = OrderedDict([
        (MANUAL_ID, 'Manual'),
        (AUTOMATION_IN_PROGRESS_ID, 'Automation in Progress'),
        (AUTOMATED_ID, 'Automated'),
        (NOT_AUTOMATE_ID, 'Will Not Automate'),
        (CI_INTEGRATED, 'CI-Integrated'),
    ])

    name = mongoengine.StringField(required=True)


class Priority(TestRailDocument):

    name = mongoengine.StringField(required=True)
    priority = mongoengine.IntField(required=True)
    short_name = mongoengine.StringField(required=True)

    @classmethod
    def get_names(cls):
        return dict(cls.objects.scalar('id', 'name'))


class Status(TestRailDocument):

    STATUS_PASSED = 1
    STATUS_BLOCKED = 2
    STATUS_UNTESTED = 3
    STATUS_OTHER = 4   # In testrail actually named as "retest"
    STATUS_FAILED = 5
    # Statuses bellow are custom:
    STATUS_SKIPPED = 6
    STATUS_IN_PROGRESS = 7
    STATUS_PROD_FAILED = 8
    STATUS_TEST_FAILED = 9
    STATUS_INFRA_FAILED = 10
    STATUS_FIXED = 11
    STATUS_REGRESSION = 12

    INTERNAL_CODES = {
        STATUS_PASSED: 'passed',
        STATUS_BLOCKED: 'blocked',
        STATUS_UNTESTED: 'untested',
        STATUS_OTHER: 'other',
        STATUS_FAILED: 'failed',
        STATUS_SKIPPED: 'skipped',
        STATUS_IN_PROGRESS: 'in_progress',
        STATUS_PROD_FAILED: 'prod_failed',
        STATUS_TEST_FAILED: 'test_failed',
        STATUS_INFRA_FAILED: 'infra_failed',
        STATUS_FIXED: 'fixed',
        STATUS_REGRESSION: 'regression',
    }

    id = mongoengine.IntField(primary_key=True)
    name = mongoengine.StringField()

    internal_code = mongoengine.StringField(
        required=True, unique=True,
        verbose_name='Status identity code'
    )

    def save(self, **kwargs):
        self.internal_code = self.INTERNAL_CODES[self.id]
        super().save(**kwargs)


class Project(TestRailDocument):

    PROJECT_MOS = 3
    PROJECT_VMWARE_DVS_PLUGIN = 4
    PROJECT_VMWARE_NSX_PLUGIN = 5
    PROJECT_CONTRAIL_PLUGIN = 6

    PCE_TEAM_PROJECTS = (
        PROJECT_VMWARE_DVS_PLUGIN,
        PROJECT_VMWARE_NSX_PLUGIN,
        PROJECT_CONTRAIL_PLUGIN,
    )

    name = mongoengine.StringField(required=True)
    completed_on = mongoengine.DateTimeField(null=True)


class Milestone(TestRailDocument):

    project = mongoengine.ReferenceField(
        Project, db_field='project_id',
        reverse_delete_rule=mongoengine.CASCADE,
        verbose_name='The ID of the project the milestone belongs to',
    )

    completed_on = mongoengine.DateTimeField(
        null=True,
        verbose_name='The date/time when the milestone was marked as completed'
    )

    due_on = mongoengine.DateTimeField(
        null=True,
        verbose_name='The due date of the milestone'
    )

    acceptance_any = mongoengine.BooleanField(null=True)


class Config(TestRailDocument):

    project = mongoengine.ReferenceField(
        Project, db_field='project_id', required=True
    )


class Suite(TestRailDocument):

    project = mongoengine.ReferenceField(
        Project, db_field='project_id', required=True
    )

    completed_on = mongoengine.DateTimeField(null=True)


class Section(TestRailDocument):

    suite = mongoengine.ReferenceField(
        Suite,  db_field='suite_id', required=True
    )


class Case(TestRailDocument):

    COMPLEXITY_SMOKE = 1
    COMPLEXITY_CORE = 2
    COMPLEXITY_ADVANCED = 3

    CASE_COMPLEXITIES = (
        (COMPLEXITY_SMOKE, 'Smoke'),
        (COMPLEXITY_CORE, 'Core'),
        (COMPLEXITY_ADVANCED, 'Advanced'),
    )

    ADOPTION_NOT_ADOPTED = 1
    ADOPTION_IN_PROGRESS = 2
    ADOPTION_ADOPTED = 3

    CASE_ADOPTION = (
        (ADOPTION_NOT_ADOPTED, 'Not Adopted'),
        (ADOPTION_IN_PROGRESS, 'Adoption in Progress'),
        (ADOPTION_ADOPTED, 'Adopted'),
    )

    suite = mongoengine.ReferenceField(
        Suite,  db_field='suite_id', required=True
    )
    section = mongoengine.ReferenceField(
        Section,  db_field='section_id', required=True
    )
    casetype = mongoengine.ReferenceField(
        CaseType, db_field='casetype_id', required=True
    )
    createdby = mongoengine.ReferenceField(
        User, db_field='createdby_id', null=True
    )
    milestone = mongoengine.ReferenceField(
        Milestone, db_field='milestone_id', null=True
    )
    priority = mongoengine.ReferenceField(
        Priority, db_field='priority_id', null=True
    )
    updatedby = mongoengine.ReferenceField(
        User, db_field='updatedby_id', null=True
    )

    title = mongoengine.StringField(required=True)
    created_on = mongoengine.DateTimeField(
        required=True, validation=DenyMicrosValidator()
    )
    updated_on = mongoengine.DateTimeField(
        required=True, validation=DenyMicrosValidator()
    )

    refs = mongoengine.ListField(
        mongoengine.StringField(), null=True
    )

    custom_case_complexity = mongoengine.IntField(
        null=True, choices=CASE_COMPLEXITIES
    )
    custom_qa_team = mongoengine.IntField(
        null=True,  choices=tuple(QATeams.NAMES.items())
    )
    custom_test_adoption = mongoengine.IntField(
        null=True, choices=CASE_ADOPTION
    )
    custom_report_label = mongoengine.StringField(null=True)
    custom_test_case_description = mongoengine.StringField(null=True)
    custom_test_case_steps = mongoengine.ListField(null=True)
    custom_test_group = mongoengine.StringField(null=True)
    project_id = mongoengine.IntField(null=True)

    meta = {
        'indexes': [
            'suite',
            'casetype',
            'milestone',
            'refs',
        ]
    }

    def _update_case_hist_version(self):
        from .utils import case_history_from_case
        case_hist = case_history_from_case(self)
        case_hist.version_begin = self.updated_on
        case_hist.version_end = utc_max_date
        CaseHistory.push_new_version(case_hist)

    def _close_case_hist_version(self):
        from .utils import case_history_from_case
        case_hist = case_history_from_case(self)
        case_hist.version_begin = self.updated_on
        # set right interval to deletion date
        case_hist.version_end = dt_now().replace(microsecond=0)

        log.debug('Closing case hist version for case %s', repr(self))
        CaseHistory.push_new_version(case_hist)

    def save(self, **kwargs):
        super().save(**kwargs)
        self._update_case_hist_version()

    @classmethod
    def post_delete(cls, sender, *, document, **kwargs):
        """
        Update CaseHistory by closing last active version.

        We use `post_delete` signal instead of overriding `delete` method
        explicitly because of strange `QuerySet.delete` method behavior -
        it does not call `delete` method for each deleting object in queryset
        until there is at least one post/pre delete listener.
        """
        assert sender is Case and isinstance(document, Case)
        case = document
        case._close_case_hist_version()


# Connect Case.post_delete
mongoengine.signals.post_delete.connect(Case.post_delete, sender=Case)


class CaseHistory(VersionedDocument):

    DIMENSION_FIELDS = ('case_id',)

    VERSIONED_FIELDS = (
        'casetype_id',
        'milestone_id',
        'priority_id',
        'refs',
        'custom_case_complexity',
        'custom_qa_team',
        'custom_test_adoption',
        'section_id'
    )

    # dimension field
    case_id = mongoengine.IntField(required=True)

    # versioned fields:
    casetype_id = mongoengine.IntField(required=True)
    milestone_id = mongoengine.IntField(null=True)
    priority_id = mongoengine.IntField(null=True)
    refs = mongoengine.ListField(
        mongoengine.StringField(), null=True
    )
    custom_case_complexity = mongoengine.IntField(null=True)
    custom_qa_team = mongoengine.IntField(null=True)
    custom_test_adoption = mongoengine.IntField(null=True)
    section_id = mongoengine.IntField(null=True)

    # not versioned fields:
    suite_id = mongoengine.IntField(required=True)
    created_on = mongoengine.DateTimeField(
        required=True, validation=DenyMicrosValidator())
    updated_on = mongoengine.DateTimeField(
        required=True, validation=DenyMicrosValidator())
    title = mongoengine.StringField(required=True)

    meta = {
        'collection': 'testrail_case_history',
        'indexes': [
            'case_id',
            'version_begin',
            'version_end',
            'casetype_id',
        ],
    }

    def clean(self):
        super().clean()
        if not self.version_begin <= self.updated_on < self.version_end:
            msg = 'version interval must overlap updated_on date'
            raise mongoengine.ValidationError(msg, field_name='updated_on')

    def __str__(self):
        return 'hist-{} for case-{}'.format(self.pk, self.case_id)


class StatusCounts(mongoengine.EmbeddedDocument):

    passed = mongoengine.IntField(default=0)
    prod_failed = mongoengine.IntField(default=0)
    test_failed = mongoengine.IntField(default=0)
    infra_failed = mongoengine.IntField(default=0)
    skipped = mongoengine.IntField(default=0)
    other = mongoengine.IntField(default=0)
    failed = mongoengine.IntField(default=0)
    blocked = mongoengine.IntField(default=0)
    in_progress = mongoengine.IntField(default=0)
    fixed = mongoengine.IntField(default=0)
    regression = mongoengine.IntField(default=0)
    untested = mongoengine.IntField(default=0)


class PlanAndRunBase(TestRailDocument):

    project = mongoengine.ReferenceField(
        Project, db_field='project_id', required=True
    )

    assignedto = mongoengine.ReferenceField(
        User, db_field='assignedto_id', null=True
    )

    createdby = mongoengine.ReferenceField(
        User, db_field='createdby_id', required=True
    )

    milestone = mongoengine.ReferenceField(
        Milestone, db_field='milestone_id', null=True
    )

    created_on = mongoengine.DateTimeField(required=True)

    completed_on = mongoengine.DateTimeField(null=True)

    # NOTE: This field is not part of TestRail API. We track it manually.
    modified_on = mongoengine.DateTimeField(
        required=True,
        verbose_name='Moment of last modification'
    )

    status_counts = mongoengine.EmbeddedDocumentField(
        StatusCounts, default=StatusCounts)

    meta = {
        'abstract': True,
    }

    def save(self, **kwargs):
        self.modified_on = dt_now()
        super().save(**kwargs)


class Plan(PlanAndRunBase):

    NON_API_FIELDS = ('modified_on', 'fake_milestone')

    meta = {
        'indexes': [
            'project',
            'milestone',
            'created_on',
        ]
    }

    fake_milestone = mongoengine.StringField(null=True)

    @classmethod
    def _get_fake_milestone(cls, plan_name, milestone_id):

        if milestone_id is None:
            return None
        mu_match = mu_pattern.search(plan_name)
        if mu_match:
            return str(milestone_id) + '_' + mu_match.group('postfix')
        return None

    def clean(self):
        self.fake_milestone = self._get_fake_milestone(
            self.name,
            self.milestone.id if self.milestone else None,
        )
        super().clean()


class Run(PlanAndRunBase):

    NON_API_FIELDS = ('modified_on', 'qa_team', 'run_type', 'fake_milestone')

    RUN_TYPE_MANUAL_ID = 1
    RUN_TYPE_SWARM_ID = 2
    RUN_TYPE_TEMPEST_ID = 3
    RUN_TYPE_ACCEPTANCE_ID = 4
    RUN_TYPE_UNKNOWN_ID = 9

    RUN_TYPES = (
        (RUN_TYPE_MANUAL_ID, 'Manual'),
        (RUN_TYPE_SWARM_ID, 'SWARM'),
        (RUN_TYPE_TEMPEST_ID, 'Tempest'),
        (RUN_TYPE_ACCEPTANCE_ID, 'Acceptance'),
        (RUN_TYPE_UNKNOWN_ID, 'Unknown'),
    )

    RUN_NAME_SMOKE_BVT = 'Smoke/BVT'

    name = mongoengine.StringField(required=True)

    plan = mongoengine.ReferenceField(
        Plan, db_field='plan_id', null=True,
        reverse_delete_rule=mongoengine.CASCADE,
    )
    suite = mongoengine.ReferenceField(
        Suite, db_field='suite_id', null=True
    )
    config = mongoengine.StringField(null=True)

    # NOTE: Fields bellow are computed on save
    qa_team = mongoengine.IntField(
        null=True, choices=tuple(QATeams.NAMES.items())
    )
    run_type = mongoengine.IntField(null=True)

    fake_milestone = mongoengine.StringField(null=True)

    meta = {
        'indexes': [
            'project',
            'plan',
            'suite',
            'milestone',
        ]
    }

    @classmethod
    def _get_run_type(cls, obj):
        run_name = obj.name.lower()
        if obj.run_type:
            return obj.run_type
        if 'manual' in run_name:
            return cls.RUN_TYPE_MANUAL_ID
        if 'swarm' in run_name or 'bvt' in run_name:
            return cls.RUN_TYPE_SWARM_ID
        if 'tempest' in run_name:
            return cls.RUN_TYPE_TEMPEST_ID
        return None

    @classmethod
    def _get_qa_team_id(cls, run_name, run_type=None):
        run_name = run_name.lower()

        for team_id, conventional_name in QATeams.CONVENTIONAL_NAMES.items():
            if '[{}]'.format(conventional_name) in run_name:
                return team_id

        if run_type == cls.RUN_TYPE_SWARM_ID:
            return QATeams.TEAM_FUEL_ID

        if run_type == cls.RUN_TYPE_TEMPEST_ID:
            return QATeams.TEAM_MOS_ID

        return None

    def clean(self):
        self.run_type = self._get_run_type(self)
        self.qa_team = self._get_qa_team_id(self.name, self.run_type)
        super().clean()


class Test(TestRailDocument):

    class Result(mongoengine.EmbeddedDocument):

        id = mongoengine.IntField(required=True)
        assignedto_id = mongoengine.IntField(null=True)
        createdby_id = mongoengine.IntField(null=True)
        status_id = mongoengine.IntField(null=True)
        created_on = mongoengine.DateTimeField(required=True)
        custom_launchpad_bug = mongoengine.StringField(null=True)
        defects = mongoengine.StringField(null=True)

    case = mongoengine.ReferenceField(
        Case, db_field='case_id', required=True,
        reverse_delete_rule=mongoengine.CASCADE,
    )
    run = mongoengine.ReferenceField(
        Run, db_field='run_id', required=True,
        reverse_delete_rule=mongoengine.CASCADE,
    )
    assignedto = mongoengine.ReferenceField(
        User, db_field='assignedto_id', null=True
    )
    results = mongoengine.EmbeddedDocumentListField(
        Result, null=True
    )
    status_id = mongoengine.IntField(
        null=True,
        verbose_name='The ID of the current status of the test',
    )
    custom_test_group = mongoengine.StringField(null=True)
    custom_qa_team = mongoengine.IntField(null=True)

    # We denormalize casetype_id for performance reason.
    # Data in Case model may change over time and in order to get
    # actual (on date when last result was added to test) casetype_id
    # we should use CaseHistory model that makes computations even more
    # complicated. So we just set it directly here during syncing tests.
    casetype = mongoengine.ReferenceField(
        CaseType, db_field='casetype_id', required=True
    )

    meta = {
        'indexes': [
            'case',
            'run',
        ]
    }


class TrMilestoneInFocus(mongoengine.Document):
    name = mongoengine.StringField(required=True)

    meta = {
        'collection': 'testrail_milestone_infocus',
    }


class TrAcceptanceDate(mongoengine.Document):
    milestone_id = mongoengine.IntField(required=True)
    qa_team_id = mongoengine.IntField(required=True)
    date = mongoengine.DateTimeField(required=True)
    is_fake_milestone = mongoengine.BooleanField(null=True)
    fake_postfix = mongoengine.IntField(null=True)

    meta = {
        'collection': 'testrail_acceptance_date',
    }
