import asyncio
import logging
from itertools import chain

import aiohttp
from flask import current_app as app

from qareporting.ext import cache
from qareporting.integration.models import EventLog
from qareporting.integration.sync_base import SyncBase
from qareporting.integration.testrail import models
from qareporting.integration.utils import Pool
from qareporting.integration.utils import save_if_modified
from qareporting.integration.utils import to_list
from qareporting.resources.acceptance_testing_progress import get_acceptance_runs
from qareporting.utils import timestamp_to_dt
from qareporting.utils import utc_min_date
from . import testrail_client
from .utils import case_from_raw
from .utils import plan_from_raw
from .utils import run_from_raw
from .utils import test_from_raw

log = logging.getLogger(__name__)


class TestRailSync(SyncBase):

    event_code = EventLog.EVENT_TESTRAIL_SYNC
    client = None
    REQUESTS_CONCURRENCY = 6

    def __init__(self, *, fast_sync=False, project_ids=None):
        self.created_after = None
        self.fast_sync = fast_sync
        self.project_ids = project_ids

    def _setup(self):
        super()._setup()
        self.last_sync_finished = utc_min_date

        last_sync = self._get_last_successful_sync()
        if last_sync:
            self.last_sync_finished = last_sync.finished

    async def fetch(self, resource, *, use_paging=False):
        uri = 'get_' + resource
        try:
            if use_paging:
                return await to_list(self.client.collection(uri))
            else:
                return await self.client.send_get(uri)
        except Exception:
            log.exception('Fetching %s failed', resource)
            raise

    def _clean_cases(self):
        suite_ids = models.Suite.objects().distinct('id')
        project_ids = models.Project.objects().distinct('id')
        cases_to_delete = models.Case.objects(
            project_id__nin=project_ids,
            suite_ids__nin=suite_ids
        )
        count = cases_to_delete.count()
        if count:
            log.info('Deleting {}: {} objects'.format(
                cases_to_delete._document.__name__,
                count
            ))
        cases_to_delete.delete()

    @classmethod
    def _remove_others(cls, query, data):
        to_delete = query.filter(id__nin=[x['id'] for x in data])
        count = to_delete.count()
        if count:
            log.info('Deleting %s: %s objects',
                     query._document.__name__, count)
        to_delete.delete()

    async def get_users(self):
        raw_users = await self.fetch('users')

        for raw_user in raw_users:
            user = models.User(**raw_user)
            user.save()

        self._remove_others(models.User.objects, raw_users)

    async def get_case_types(self):
        raw_case_types = await self.fetch('case_types')

        for raw in raw_case_types:
            case_type = models.CaseType(**raw)
            case_type.save()

        self._remove_others(models.CaseType.objects, raw_case_types)

    async def get_statuses(self):
        raw_statuses = await self.fetch('statuses')

        for raw_status in raw_statuses:
            status = models.Status(**raw_status)
            status.save()

        self._remove_others(models.Status.objects, raw_statuses)

    async def get_priorities(self):
        raw_priorities = await self.fetch('priorities')

        for raw_priority in raw_priorities:
            priority = models.Priority(**raw_priority)
            priority.save()

        self._remove_others(models.Priority.objects, raw_priorities)

    async def get_milestones(self, project_id):
        raw_milestones = await self.fetch('milestones/{0}'.format(project_id))

        for raw_milestone in raw_milestones:

            completed_on = timestamp_to_dt(raw_milestone.pop('completed_on'))
            due_on = timestamp_to_dt(raw_milestone.pop('due_on'))

            milestone = models.Milestone(
                completed_on=completed_on,
                due_on=due_on,
                **raw_milestone
            )
            milestone.save()

        query = models.Milestone.objects(project=project_id)
        self._remove_others(query, raw_milestones)

    async def get_configs(self, project_id):
        raw_configs = await self.fetch('configs/{0}'.format(project_id))

        for raw_config in raw_configs:
            config = models.Config(**raw_config)
            config.save()

        query = models.Config.objects(project=project_id)
        self._remove_others(query, raw_configs)

    async def get_suite_sections(self, project_id, suite_id):
        raw_sections = await self.fetch('sections/{0}&suite_id={1}'.format(
            project_id, suite_id))

        for raw_section in raw_sections:
            section = models.Section(**raw_section)
            section.save()

        query = models.Section.objects(project=project_id, suite=suite_id)
        self._remove_others(query, raw_sections)

    async def get_suite_cases(self, project_id, suite_id):
        cases_url = 'cases/{0}&suite_id={1}'.format(project_id, suite_id)

        raw_cases = await self.fetch(cases_url)

        for raw_case in raw_cases:
            case = case_from_raw(raw_case, project_id)
            save_if_modified(case)

        query = models.Case.objects(project_id=project_id, suite=suite_id)
        self._remove_others(query, raw_cases)

    async def get_suites(self, project_id):
        raw_suites = await self.fetch('suites/{0}'.format(project_id))

        for raw_suite in raw_suites:

            completed_on = timestamp_to_dt(raw_suite.pop('completed_on'))
            suite = models.Suite(
                completed_on=completed_on,
                **raw_suite
            )
            suite.save()

        query = models.Suite.objects(project=project_id)
        self._remove_others(query, raw_suites)

        def tasks_gen():
            for i, raw in enumerate(raw_suites, 1):
                yield self.get_suite_sections(project_id, raw['id'])
                yield self.get_suite_cases(project_id, raw['id'])
                log.info('project %s: %s/%s suites synced',
                         project_id, i, len(raw_suites))

        await Pool(self.REQUESTS_CONCURRENCY, tasks_gen()).run()

    async def get_plans(self, project_id):
        plans_url = 'plans/{0}&is_completed=0'.format(project_id)
        plans = []
        raw_plans = await self.fetch(plans_url, use_paging=True)

        for raw_plan in raw_plans:
            plan = plan_from_raw(raw_plan)
            plan, is_saved = save_if_modified(
                plan, skip_fields=models.Plan.NON_API_FIELDS
            )
            plans.append(plan)

        query = models.Plan.objects(project=project_id, completed_on=None)
        self._remove_others(query, raw_plans)
        return plans

    async def get_runs_without_plan(self, project_id):
        raw_runs = await self.fetch(
            'runs/{0}&is_completed=0'.format(project_id), use_paging=True)

        for raw_run in raw_runs:
            run = run_from_raw(raw_run)
            save_if_modified(run, models.Run.NON_API_FIELDS)

        query = models.Run.objects(project=project_id, plan=None)
        self._remove_others(query, raw_runs)

    async def get_runs_in_plan(self, plan_id):
        raw_runs = []
        plan_with_run_entries = await self.fetch('plan/{0}'.format(plan_id))
        for entry in plan_with_run_entries.get('entries', []):
            for raw_run in entry['runs']:
                raw_runs.append(raw_run)

        for raw_run in raw_runs:
            run = run_from_raw(raw_run, raw_plan=plan_with_run_entries)
            save_if_modified(run, models.Run.NON_API_FIELDS)

        query = models.Run.objects(plan=plan_id)
        self._remove_others(query, raw_runs)

    async def get_plans_and_runs(self, project_id):

        plans = await self.get_plans(project_id)

        if self.fast_sync:
            plans_to_sync = [
                p for p in plans if p.modified_on > self.last_sync_finished
            ]
        else:
            plans_to_sync = plans

        skipped = len(plans) - len(plans_to_sync)
        log.info('project %s: %s plan_runs will be skipped',
                 project_id, skipped)

        def tasks_gen():
            for i, plan in enumerate(plans_to_sync, 1):
                yield self.get_runs_in_plan(plan.id)
                log.info('project %s: %s/%s plan_runs synced. (%s skipped)',
                         project_id, i, len(plans), skipped)

        await Pool(self.REQUESTS_CONCURRENCY, tasks_gen()).run()

        await self.get_runs_without_plan(project_id)

    async def get_tests_and_results_for_run(self, run_id):
        raw_tests = await self.fetch('tests/{0}'.format(run_id))
        raw_results = await self.fetch(
            'results_for_run/{0}'.format(run_id),
            use_paging=True
        )

        test_to_results = {}
        for raw_res in raw_results:
            test_to_results.setdefault(raw_res['test_id'], []).append(raw_res)

        for raw in raw_tests:
            test = test_from_raw(raw, test_to_results.get(raw['id'], []))
            test.save()

        query = models.Test.objects(run=run_id)
        self._remove_others(query, raw_tests)

    async def get_tests_and_results(self, project_id):

        runs_query = models.Run.objects.filter(project=project_id)
        total_count = runs_query.count()

        if self.fast_sync:
            runs_query = runs_query.filter(
                modified_on__gte=self.last_sync_finished,
            )

        to_sync_count = runs_query.count()
        log.info("project %s: %s runs tests/results will be processed. "
                 "(%s will be skipped)",
                 project_id, to_sync_count, total_count - to_sync_count)

        run_ids = runs_query.scalar('id')

        def tasks_gen():
            for i, run_id in enumerate(run_ids, 1):
                yield self.get_tests_and_results_for_run(run_id)
                log.info('project %s: %s/%s run tests/results synced.',
                         project_id, i, to_sync_count)

        await Pool(self.REQUESTS_CONCURRENCY, tasks_gen()).run()

    async def get_projects(self):
        raw_projects = await self.fetch('projects')

        for raw_project in raw_projects:
            project_id = raw_project['id']
            project_name = raw_project['name']

            if (self.project_ids is not None and
                    project_id not in self.project_ids):
                log.info('Skip project %s - "%s"', project_id, project_name)
                continue

            log.info('Sync Project "%s"', project_name)

            completed_on = timestamp_to_dt(raw_project.pop('completed_on'))
            proj = models.Project(
                completed_on=completed_on,
                **raw_project
            )
            proj.save()
            assert proj.pk == project_id

            await asyncio.wait([
                self.get_milestones(project_id),
                self.get_configs(project_id),
            ])
            await self.get_suites(project_id)
            await self.get_plans_and_runs(project_id)
            await self.get_tests_and_results(project_id)

        self._remove_others(models.Project.objects, raw_projects)

    async def get_collections(self):
        await asyncio.wait([
            self.get_users(),
            self.get_case_types(),
            self.get_statuses(),
            self.get_priorities(),
        ])
        await self.get_projects()

    def _sync(self):
        log.info('Run TestRail sync...')

        loop = asyncio.get_event_loop()
        with aiohttp.ClientSession(loop=loop) as session:
            self.client = testrail_client.TestRailClient(
                session,
                app.config['TESTRAIL_BASEURL'],
                app.config['TESTRAIL_USER'],
                app.config['TESTRAIL_PASSWORD'],
            )
            loop.run_until_complete(self.get_collections())

    def _sync_finally_action(self):
        acceptance_run_names = get_acceptance_runs()
        milestones = [x['milestone_id'] for x in acceptance_run_names]
        qa_teams = []
        run_names = []

        is_fake = any(isinstance(milestone, str) for milestone in milestones)

        if is_fake:
            fake_qa_teams = []
            fake_names = []
            fake_runs = [run for run in acceptance_run_names
                         if isinstance(run['milestone_id'], str)]
            for item in fake_runs:
                teams, names = zip(
                    *[(i['qa_team_id'], i['run_names']) for i in item['runs']]
                )
                fake_qa_teams.extend(list(teams))
                fake_names.extend(list(chain.from_iterable(names)))

            plan_ids = models.Plan.objects(
                fake_milestone__in=[run['milestone_id']
                                    for run in fake_runs]
            ).distinct('id')

            models.Run.objects(
                milestone__in=[run['milestone_id'].split('_')[0]
                               for run in fake_runs],
                qa_team__in=fake_qa_teams,
                name__in=fake_names,
                plan_id__in=plan_ids,
            ).update(
                set__run_type=models.Run.RUN_TYPE_ACCEPTANCE_ID
            )

        runs = [run for run in acceptance_run_names
                if isinstance(run['milestone_id'], int)]

        for item in runs:
            teams, names = zip(
                *[(i['qa_team_id'], i['run_names']) for i in item['runs']]
            )
            qa_teams.extend(list(teams))
            run_names.extend(list(chain.from_iterable(names)))

        models.Run.objects(
            milestone__in=milestones,
            qa_team__in=qa_teams,
            name__in=run_names,
        ).update(
            set__run_type=models.Run.RUN_TYPE_ACCEPTANCE_ID
        )
        self._clean_cases()
        with app.app_context():
            cache.clear()
        log.info('Cache has been cleared!')
