# TestRail API binding (API v2, available since TestRail 3.0)
#
# Learn more:
# http://docs.gurock.com/testrail-api2/start
# http://docs.gurock.com/testrail-api2/accessing
import asyncio
import logging
import urllib.parse
from collections import OrderedDict

import aiohttp

log = logging.getLogger(__name__)


TESTRAIL_MAX_ALLOWED_RESULTS_LIMIT = 250


class CollectionIterator:

    """Iteration of paginated TestRail api results."""

    RESULTS_LIMIT = TESTRAIL_MAX_ALLOWED_RESULTS_LIMIT

    def __init__(self, uri, client):
        self._client = client
        self._curr_entries = []
        self._curr_entry_index = 0
        self._next_chunk_uri = uri
        self._stop_flag = False

    @classmethod
    def _get_next_chunk_uri(cls, uri, shift):
        qs = OrderedDict(urllib.parse.parse_qsl(uri, True))
        offset = int(qs.get('offset', 0)) + shift
        qs['offset'] = str(offset)
        qs['limit'] = cls.RESULTS_LIMIT
        return urllib.parse.urlencode(qs)

    async def _load_next_entries_chunk(self):
        if self._stop_flag:
            return False

        shift = len(self._curr_entries)
        self._next_chunk_uri = self._get_next_chunk_uri(
            self._next_chunk_uri, shift)

        self._curr_entries = await self._client.send_get(
            self._next_chunk_uri)

        if len(self._curr_entries) < self.RESULTS_LIMIT:
            self._stop_flag = True

        if self._curr_entries:
            return True

        return False

    async def __aiter__(self):
        return self

    async def __anext__(self):
        if self._curr_entry_index >= len(self._curr_entries):
            is_loaded = await self._load_next_entries_chunk()
            if is_loaded:
                self._curr_entry_index = 0
            else:
                raise StopAsyncIteration

        result = self._curr_entries[self._curr_entry_index]
        self._curr_entry_index += 1
        return result


class TestRailClient(object):
    REQUEST_TIMEOUT = 180
    DELAY_TIMEOUT = 0.8

    def __init__(self, session, base_url, user, password):
        self.session = session
        if not base_url.endswith('/'):
            base_url += '/'
        self._url = base_url + 'index.php?/api/v2/'
        self.user = user
        self.password = password

    def collection(self, uri):
        return CollectionIterator(uri, self)

    async def send_get(self, uri):
        """Send Get.

        Issues a GET request (read) against the API and returns the result
        (as Python dict).

        :param uri: The API method to call including parameters (e.g.
            get_case/1)
        """
        return await self._send_request('get', uri)

    async def send_post(self, uri, **params):
        """Send POST.

        Issues a POST request (write) against the API and returns the result
        (as Python dict).

        :param uri: The API method to call including parameters (e.g.
            add_case/1)
        :param params: The data to submit as part of the request (as Python
            dict, strings must be UTF-8 encoded)
        """
        return await self._send_request('post', uri, **params)

    async def _send_request(self, method, uri, **params):
        url = self._url + uri
        auth = aiohttp.helpers.BasicAuth(self.user, self.password)
        headers = {'Content-type': 'application/json'}

        with aiohttp.Timeout(self.REQUEST_TIMEOUT):
            log.debug('Request: {0} {1}'.format(method, url))
            async with self.session.request(method,
                                            url,
                                            allow_redirects=False,
                                            auth=auth,
                                            headers=headers,
                                            **params) as response:
                if response.status == 429:  # Too Many Requests
                    retry_timeout = float(
                        response.headers.get('Retry-After')) + 1
                    log.warning('Too Many Requests. Request will be retried '
                                'after {0} seconds'.format(str(retry_timeout)))
                    await response.release()
                    await asyncio.sleep(retry_timeout)

                    return await self._send_request(method, uri, **params)
                elif response.status >= 300:
                    raise TestRailAPIError(
                        "Url: {} \n".format(url) +
                        "Wrong response from TestRail API:\n"
                        "status_code: {0.status}\n"
                        "headers: {0.headers}\n"
                        "content: '{0.content}'".format(response))

                result = await response.json()
                if result and 'error' in result:
                    log.warning(result)

                await asyncio.sleep(self.DELAY_TIMEOUT)
                return result


class TestRailAPIError(Exception):
    pass
