from qareporting.integration.testrail.models import Case
from qareporting.integration.testrail.models import CaseHistory
from qareporting.integration.testrail.models import Plan
from qareporting.integration.testrail.models import Run
from qareporting.integration.testrail.models import Status
from qareporting.integration.testrail.models import StatusCounts
from qareporting.integration.testrail.models import Test
from qareporting.integration.testrail.models import Project
from qareporting.utils import dt_now
from qareporting.utils import timestamp_to_dt
from qareporting.utils import utc_max_date

from . import consts


RUN_STATUSES_MAPPING = {
    Status.STATUS_PASSED: 'passed_count',
    Status.STATUS_FAILED: 'failed_count',
    Status.STATUS_BLOCKED: 'blocked_count',
    Status.STATUS_UNTESTED: 'untested_count',
    Status.STATUS_OTHER: 'retest_count',
    Status.STATUS_SKIPPED: 'custom_status1_count',
    Status.STATUS_IN_PROGRESS: 'custom_status2_count',
    Status.STATUS_PROD_FAILED: 'custom_status3_count',
    Status.STATUS_TEST_FAILED: 'custom_status4_count',
    Status.STATUS_INFRA_FAILED: 'custom_status5_count',
    Status.STATUS_FIXED: 'custom_status6_count',
    Status.STATUS_REGRESSION: 'custom_status7_count',
}


def run_from_raw(raw_run, raw_plan=None):

    counts_dict = {
        Status.INTERNAL_CODES[status_id]: raw_run[run_fld]
        for status_id, run_fld in RUN_STATUSES_MAPPING.items()
    }

    if raw_run['plan_id'] and not raw_plan:
        raise ValueError('Plan data required')

    created_on = timestamp_to_dt(raw_run.get('created_on'))
    if created_on is None:
        created_on = timestamp_to_dt(raw_plan['created_on'])

    createdby_id = raw_run.get('created_by')
    if createdby_id is None:
        createdby_id = raw_plan['created_by']

    run = Run(
        id=raw_run['id'],
        name=raw_run['name'],
        assignedto_id=raw_run['assignedto_id'],
        createdby_id=createdby_id,
        milestone_id=raw_run['milestone_id'],
        plan_id=raw_run['plan_id'],
        project_id=raw_run['project_id'],
        suite_id=raw_run['suite_id'],
        created_on=created_on,
        completed_on=timestamp_to_dt(raw_run['completed_on']),
        is_completed=raw_run['is_completed'],
        config=raw_run['config'],
        config_ids=raw_run['config_ids'],
        status_counts=StatusCounts(**counts_dict),
    )
    run.modified_on = dt_now()
    return run


def plan_from_raw(raw_plan):

    counts_dict = {
        Status.INTERNAL_CODES[status_id]: raw_plan[run_fld]
        for status_id, run_fld in RUN_STATUSES_MAPPING.items()
    }

    plan = Plan(
        id=raw_plan['id'],
        name=raw_plan['name'],
        assignedto_id=raw_plan.get('assignedto_id'),
        createdby_id=raw_plan.get('created_by'),
        milestone_id=raw_plan.get('milestone_id'),
        project_id=raw_plan.get('project_id'),
        created_on=timestamp_to_dt(raw_plan.get('created_on')),
        completed_on=timestamp_to_dt(raw_plan.get('completed_on')),
        is_completed=raw_plan.get('is_completed'),
        status_counts=StatusCounts(**counts_dict),
    )
    plan.modified_on = dt_now()
    return plan


def case_from_raw(raw_case, project_id):

    raw = raw_case.copy()

    refs = None
    refs_str = raw.pop('refs')
    if refs_str:
        refs = [r.strip() for r in refs_str.split(',')]

    custom_qa_team = raw.pop('custom_qa_team', None)

    # Hook for PCE team
    if custom_qa_team is None and project_id in Project.PCE_TEAM_PROJECTS:
        custom_qa_team = consts.QATeams.TEAM_PCE_ID

    case = Case(
        id=raw.pop('id'),
        casetype_id=raw.pop('type_id'),
        createdby_id=raw.pop('created_by'),
        updatedby_id=raw.pop('updated_by'),
        created_on=timestamp_to_dt(raw.pop('created_on')),
        updated_on=timestamp_to_dt(raw.pop('updated_on')),
        custom_qa_team=custom_qa_team,
        refs=refs,
        project_id=project_id,
        **raw,
    )
    return case


def case_history_from_case(case):

    def get_id(attr):
        return getattr(case._data.get(attr), 'id', None)

    case_hist = CaseHistory(
        case_id=case.id,
        suite_id=get_id('suite'),
        casetype_id=get_id('casetype'),
        milestone_id=get_id('milestone'),
        priority_id=get_id('priority'),
        title=case.title,
        created_on=case.created_on,
        updated_on=case.updated_on,
        refs=case.refs,
        custom_case_complexity=case.custom_case_complexity,
        custom_qa_team=case.custom_qa_team,
        custom_test_adoption=case.custom_test_adoption,
        section_id=get_id('section')
    )
    case_hist.version_begin = case.updated_on
    case_hist.version_end = utc_max_date
    return case_hist


def test_from_raw(raw_test, raw_results):

    test_id = raw_test['id']
    test_results = []
    for raw_res in raw_results:
        assert raw_res['test_id'] == test_id
        test_result = Test.Result(
            id=raw_res['id'],
            assignedto_id=raw_res['assignedto_id'],
            createdby_id=raw_res['created_by'],
            status_id=raw_res['status_id'],
            created_on=timestamp_to_dt(raw_res['created_on']),
            custom_launchpad_bug=raw_res['custom_launchpad_bug'],
            defects=raw_res['defects'],
        )
        test_results.append(test_result)

    test = Test(
        id=test_id,
        case_id=raw_test['case_id'],
        run_id=raw_test['run_id'],
        status_id=raw_test['status_id'],
        assignedto_id=raw_test['assignedto_id'],
        custom_test_group=raw_test['custom_test_group'],
        custom_qa_team=raw_test.get('custom_qa_team', None),
        results=test_results or None,
        casetype_id=raw_test['type_id']
    )
    return test
