import logging
import asyncio

log = logging.getLogger(__name__)


class Pool:

    """Execute `coros` sequence with maximum `size` parallel tasks."""

    skip_errors = False

    def __init__(self, size, coros, *, loop=None, skip_errors=False):
        """
        :param size: Concurrency limit. (As "workers" count in multiprocessing)
        :param coros: Coroutines sequence(any iterable) to execute.
        :param loop: Asyncio event loop.
        """
        assert size >= 1
        self.size = size
        self.coros = coros
        self.loop = loop or asyncio.get_event_loop()
        self.skip_errors = skip_errors
        self._current_tasks = {}
        self._no_more_tasks = False
        self._waiter = asyncio.Future(loop=self.loop)

    async def run(self):
        for i in range(self.size):
            self._add_task()

        await asyncio.wait_for(
            self._waiter, timeout=None, loop=self.loop)

    def _add_task(self):
        if self._no_more_tasks:
            return
        try:
            coro = next(self.coros)
        except StopIteration:
            self._no_more_tasks = True
            if not self._current_tasks:
                self._waiter.set_result(None)
        else:
            future = asyncio.ensure_future(coro)
            future.add_done_callback(self._on_task_done)
            key = id(future)
            assert key not in self._current_tasks
            self._current_tasks[key] = future

    def _on_task_done(self, done_future):
        key = id(done_future)
        task = self._current_tasks[key]
        del self._current_tasks[key]
        assert task is done_future

        exception = done_future.exception()
        if exception is not None and not self.skip_errors:
            self._waiter.set_exception(exception)
            return

        if self._no_more_tasks:
            if not self._current_tasks:
                self._waiter.set_result(None)
        elif len(self._current_tasks) < self.size:
            self._add_task()


async def to_list(async_iterator):
    result = []
    async for item in async_iterator:
        result.append(item)
    return result


def save_if_modified(new_obj, skip_fields=()):
    assert new_obj.id is not None, 'object must have id value'

    old_obj = new_obj._qs.with_id(new_obj.id)
    if old_obj is None:
        new_obj.save()
        return new_obj, True

    new_obj_fields = sorted(f for f in new_obj if f not in skip_fields)
    new_obj_as_son = new_obj.to_mongo(fields=new_obj_fields)
    new_sorted = sorted(new_obj_as_son.items())

    old_obj_fields = sorted(f for f in old_obj if f not in skip_fields)
    old_obj_as_son = old_obj.to_mongo(fields=old_obj_fields)
    old_sorted = sorted(old_obj_as_son.items())

    if old_sorted != new_sorted:
        old_obj_skipped_fields = [f for f in old_obj if f in skip_fields]
        if old_obj_skipped_fields:
            skipped_as_son = old_obj.to_mongo(fields=old_obj_skipped_fields)

            for key, val in skipped_as_son.items():
                setattr(new_obj, key, val)

        logging.debug(
            'Object %s is modified. saving....', repr(new_obj))
        new_obj.save()
        return new_obj, True

    return old_obj, False
