import urllib.parse

from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.testrail.models import (
    Run, Milestone, Plan, StatusCounts, TrAcceptanceDate)
from qareporting.resources.base import BaseResource
from qareporting.resources.bvt_swarm_report import (
    get_swarm_bvt_data, fake_milestone)
from qareporting.utils import dt_from_iso, utc_max_date, dt_now

WEEK = 7


def _get_filters(data):
    filters_result = []
    for index, start_date in enumerate(data):
        names = []
        start = start_date['date']
        finish = None
        if index < len(data) - 1:
            data_next = data[index + 1]
            finish = data_next['date']
        if index > 0:
            names = [data[x]['names'] for x in range(index + 1)]
        else:
            names.append(start_date['names'])
        filters_result.append(
            {
                'date_to': utc_max_date if finish is None else finish,
                'date_from': start,
                'names': names
            }
        )
    return filters_result


def _check_name(_list, run_names):
    return any(run_name in run_names for run_name in _list)


def _is_acceptance_plan(list_of_runs, run_names, is_any):
    if is_any:
        return any(_check_name(_list, run_names) for _list in list_of_runs)
    return all(_check_name(_list, run_names) for _list in list_of_runs)


def get_acceptance_runs(milestone_id=None, is_fake=False):

    pipeline = [
        {'$sort': {'milestone_id': -1}},
        {'$group': {
            '_id': {
                'milestone': '$milestone_id',
                'qa_team_id': '$qa_team'
            },
            'names': {
                '$addToSet': '$name',
            }
        }},
        {'$group': {
            '_id': '$_id.milestone',
            'runs': {
                '$push': {
                    'qa_team_id': '$_id.qa_team_id',
                    'run_names': '$names'
                }
            }
        }}
    ]

    fake_milestone_pipeline = [
        {'$sort': {'milestone_id': -1}},
        {'$group': {
            '_id': {
                'fake_milestone': '$fake_milestone',
                'qa_team_id': '$qa_team'
            },
            'names': {
                '$addToSet': '$name',
            }
        }},
        {'$group': {
            '_id': '$_id.fake_milestone',
            'runs': {
                '$push': {
                    'qa_team_id': '$_id.qa_team_id',
                    'run_names': '$names'
                }
            }
        }}
    ]

    if is_fake:
        fake_query_result = Run.objects(
            run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
            fake_milestone=milestone_id,
        ).aggregate(*fake_milestone_pipeline)
    elif milestone_id and not is_fake:
        query_result = Run.objects(
            run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
            milestone_id=milestone_id,
            fake_milestone__exists=False,
        ).aggregate(*pipeline)
        fake_query_result = Run.objects(
            run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
            milestone_id=milestone_id,
            fake_milestone__exists=True,
        ).aggregate(*fake_milestone_pipeline)
    else:
        query_result = Run.objects(
            run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
            fake_milestone__exists=False,
        ).aggregate(*pipeline)
        fake_query_result = Run.objects(
            run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
            fake_milestone__exists=True,
        ).aggregate(*fake_milestone_pipeline)

    result = []

    if milestone_id is None or isinstance(milestone_id, int):
        for group in query_result:
            result.append({
                'milestone_id': group['_id'],
                'runs': group['runs'],
            })

    for group in fake_query_result:
        result.append({
            'milestone_id': group['_id'],
            'runs': group['runs'],
        })

    for milestone in result:
        is_fake = isinstance(milestone['milestone_id'], str)
        if is_fake:
            date = TrAcceptanceDate.objects(
                milestone_id=milestone['milestone_id'].split('_')[0],
                is_fake_milestone__exists=is_fake
            ).only('qa_team_id', 'date').as_pymongo()
        else:
            date = TrAcceptanceDate.objects(
                milestone_id=milestone['milestone_id'],
                is_fake_milestone__exists=is_fake
            ).only('qa_team_id', 'date').as_pymongo()

            is_any = Milestone.objects(
                id=milestone['milestone_id']
            ).first().acceptance_any

            if is_any:
                milestone['is_any'] = True
            else:
                milestone['is_any'] = False

        start_dates = None

        if date:
            start_dates = {d['qa_team_id']: d['date'] for d in date}

        for team in milestone['runs']:
            if start_dates:
                team.update({'date': start_dates.get(
                    team['qa_team_id'], dt_now())})
            else:
                team.update({'date': dt_now()})

    return result


def get_acceptance(milestone_id, date_from=None,
                   date_to=None, last=None, is_fake=False):
    is_initial = (not date_from and not date_to)

    is_any = Milestone.objects(id=milestone_id).first().acceptance_any

    all_by_team = get_acceptance_runs(milestone_id, is_fake=is_fake)
    acceptance_all = []
    acceptance_start_date = []
    if all_by_team:
        for team in all_by_team[0]['runs']:
            acceptance_all.extend(team['run_names'])
            acceptance_start_date.append(dict(
                names=team['run_names'],
                team_id=team['qa_team_id'],
                date=team['date']))

    if not acceptance_all:
        return []

    sorted_data = sorted(acceptance_start_date, key=lambda x: x['date'])

    pipeline = [
        {'$sort': {'created_on': 1}},
        {'$group': {
            '_id': '$plan_id',
            'names': {
                '$addToSet': '$name'
            }
        }},
        {'$match': {
            'names': {
                '$in': acceptance_all
            }
        }}
    ]

    if is_initial or last:
        plans = Run.objects(
            milestone_id=milestone_id,
            created_on__gt=sorted_data[0]['date'],
        ).aggregate(*pipeline)
    else:
        plans = Run.objects(
            created_on__gte=date_from,
            created_on__lte=date_to,
            milestone_id=milestone_id,
        ).aggregate(*pipeline)

    if is_fake:
        plans = Run.objects(
            created_on__gt=sorted_data[0]['date'],
            fake_milestone=milestone_id,
        ).aggregate(*pipeline)

    plan_data = {}
    target_plans = []

    if is_initial:
        plans_query = Plan.objects.filter(
            created_on__gt=sorted_data[0]['date'],
        ).only('id', 'name', 'created_on').as_pymongo()
    else:
        plans_query = Plan.objects.filter(
            created_on__gte=date_from,
        ).only('id', 'name', 'created_on').as_pymongo()

    for plan in plans_query:
        plan_date = plan['created_on']
        plan_data[plan['_id']] = {
            'date': plan_date,
            'name': plan['name'],
            '_id': plan['_id'],
            Run.RUN_TYPE_ACCEPTANCE_ID: [],
        }

    for plan in plans:
        if plan['_id'] in plan_data:
            plan_data[plan['_id']].update({'run_names': plan['names']})

    acceptance_plans = [plan for plan in plan_data.values()
                        if plan.get('run_names')]

    filters = _get_filters(sorted_data)

    for plan in acceptance_plans:
        for index, _filter in enumerate(filters):
            if _filter['date_to'] > plan['date'] > _filter['date_from']:
                if _is_acceptance_plan(filters[index]['names'],
                                       plan['run_names'],
                                       is_any):
                    target_plans.append(plan)

    acceptance_plan_ids = [p['_id'] for p in target_plans]
    acceptance_plan_ids.sort(reverse=True)

    if is_initial:
        acceptance_plan_ids = acceptance_plan_ids[:WEEK]
    if last:
        acceptance_plan_ids = acceptance_plan_ids[:1]

    acceptance_runs = Run.objects(
        plan_id__in=acceptance_plan_ids,
        run_type=Run.RUN_TYPE_ACCEPTANCE_ID,
    )

    result = []

    for run in acceptance_runs:
        config_name = run.name
        status = run.status_counts.to_mongo().to_dict()
        date = plan_data[run.plan.pk]['date']
        plan_name = plan_data[run.plan.pk]['name']

        res = {
            'date': date.date().isoformat(),
            'config_name': config_name,
            'status_counts': status,
            'plan_name': plan_name,
            'qa_team_id': run.qa_team,
        }
        result.append(res)

    return result


class AcceptanceTestingProgress(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument('date_from', type=dt_from_iso)
    parser.add_argument('date_to', type=dt_from_iso)
    parser.add_argument('milestone_id', type=fake_milestone)

    def __init__(self):
        args = self.parser.parse_args()
        self.date_from = args.date_from
        self.date_to = args.date_to
        self.is_fake_milestone, self.milestone_id = args.milestone_id

    def get(self):
        if self.is_fake_milestone:
            return get_acceptance(
                self.milestone_id,
                is_fake=True,
                date_from=self.date_from,
                date_to=self.date_to
            )
        return get_acceptance(
            self.milestone_id,
            date_from=self.date_from,
            date_to=self.date_to
        )


class AcceptanceRuns(BaseResource):
    def get(self):
        return get_acceptance_runs()


class AcceptanceSwarmTempestSummary(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument('milestone', type=urllib.parse.unquote_plus)

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

    def get(self):

        result_fields = StatusCounts._fields_ordered

        empty_result = dict((field, 0) for field in result_fields)

        milestone_id = Milestone.objects(name=self.milestone).scalar('id')[0]

        last_acceptance = get_acceptance(milestone_id, last=True)

        acceptance_plan = None
        acceptance_date = None

        if last_acceptance:
            accept_sum = {key: sum(
                d['status_counts'][key] for d in last_acceptance
            )
                for key in last_acceptance[0]['status_counts']}

            acceptance_plan = last_acceptance[0]['plan_name']
            acceptance_date = last_acceptance[0]['date']
        else:
            accept_sum = empty_result

        plans = Plan.objects.filter(
            name=acceptance_plan
        )

        swarm_bvt_results = None

        if plans:
            swarm_bvt_results = get_swarm_bvt_data(plans)

            recent_tempest = Run.objects(
                run_type=Run.RUN_TYPE_TEMPEST_ID,
                plan_id=plans[0].pk,
            ).as_pymongo()

            if not recent_tempest:
                tempest_sum = empty_result
            else:
                tempest_sum = {key: sum(
                    d['status_counts'][key] for d in recent_tempest
                )
                   for key in recent_tempest[0]['status_counts']}
        else:
            tempest_sum = empty_result

        if not swarm_bvt_results:
            swarm_sum = empty_result
        else:
            swarm_sum = swarm_bvt_results[0].get('status_counts', {})

        result = {
            'acceptance': accept_sum,
            'swarm': swarm_sum,
            'tempest': tempest_sum,
        }

        total = {key: sum(d[key] for d in result.values())
                 for key in result['acceptance']}

        result.update({
            'total': total,
            'plan_name': acceptance_plan,
            'date': acceptance_date,
        })

        return result

api.add_resource(AcceptanceTestingProgress, '/testrail/acceptance_progress')
api.add_resource(AcceptanceRuns, '/testrail/acceptance_runs')
api.add_resource(
    AcceptanceSwarmTempestSummary,
    '/testrail/acceptance_swarm_tempest'
)
