from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.testrail.models import Case
from qareporting.integration.testrail.models import CaseHistory
from qareporting.integration.testrail.models import Milestone
from qareporting.integration.testrail.models import Project
from qareporting.integration.testrail.models import Suite
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso


def get_data_with_suite(data, suites_map, milestones_map):
    for item in data:
        if item['suite_id'] and suites_map.get(item['suite_id']):
            item['suite_name'] = suites_map.get(item['suite_id'])
            item['milestone_name'] = milestones_map.get(
                item['milestone_id'])
            yield item


def get_day_adoption_counts(moment, case_ids):

    aggregation_pipeline = [
        {'$group': {
            '_id': {
                'qa_team': '$custom_qa_team',
                'milestone_id': '$milestone_id',
                'adoption': '$custom_test_adoption',
                'suite_id': '$suite_id',
            },
            'count': {'$sum': 1},
        }},
        {'$group': {
            '_id': {
                'qa_team': '$_id.qa_team',
                'milestone_id': '$_id.milestone_id',
                'suite_id': '$_id.suite_id',
            },
            'case_adoption': {
                '$push': {
                    'adoption': '$_id.adoption',
                    'count': '$count',
                },
            },
        }}
    ]

    query = CaseHistory.objects.filter(
        version_begin__lte=moment,
        version_end__gt=moment,
        case_id__in=case_ids,
    ).aggregate(*aggregation_pipeline)

    for case in query:
        _id = case['_id']

        adoption_dict = {x.get('adoption'): x['count']
                         for x in case['case_adoption']}
        item = {
            'date': moment,
            'qa_team': _id.get('qa_team'),
            'milestone_id': _id.get('milestone_id'),
            'suite_id': _id.get('suite_id'),
            'adoption_counts': {
                'not_adopted': adoption_dict.get(Case.ADOPTION_NOT_ADOPTED, 0),
                'in_progress': adoption_dict.get(
                    Case.ADOPTION_IN_PROGRESS, 0),
                'adopted': adoption_dict.get(Case.ADOPTION_ADOPTED, 0),
                'n/a': adoption_dict.get(
                    None, 0)
            }
        }
        yield item


class AdoptionProgress(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument(
        'dates', type=lambda v: sorted(set(map(dt_from_iso, v.split(','))))
    )

    def __init__(self):
        self.args = self.parser.parse_args()

    def get_project_data(self, case_ids):
        for moment in self.args.dates:
            for item in get_day_adoption_counts(moment, case_ids):
                yield item

    def get(self):
        project_id = Project.objects(name='TCP project').first().id

        project_case_ids = Case.objects(
            project_id=project_id
        ).only('id').distinct('id')

        suites_map = dict(Suite.objects.scalar('id', 'name'))
        milestones_map = dict(Milestone.objects.scalar('id', 'name'))

        project_data = list(get_data_with_suite(
            self.get_project_data(project_case_ids),
            suites_map,
            milestones_map)
        )

        return project_data


api.add_resource(AdoptionProgress, '/testrail/adoption_progress')
