from flask_restful import reqparse

from qareporting.ext import api
from qareporting.ext import cache
from qareporting.integration.testrail.models import CaseHistory
from qareporting.integration.testrail.models import CaseType
from qareporting.integration.testrail.models import Milestone
from qareporting.integration.testrail.models import Suite
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_now, dt_from_iso
from qareporting.integration.testrail import consts
from qareporting.resources.adoption_progress import get_data_with_suite


def get_day_type_counts(moment, milestone_ids=None,
                        distinct_by_title=False, expand_suites=False,
                        dashboard_exclude=None):

    query = CaseHistory.objects.filter(
        version_begin__lte=moment,
        version_end__gt=moment,
    )

    if milestone_ids:
        query = query.filter(milestone_id__in=milestone_ids)

    if dashboard_exclude:
        query = query.filter(custom_qa_team__nin=[
            consts.QATeams.TEAM_CIBOT_ID,
            consts.QATeams.TEAM_FRAMEWORK_ID,
            consts.QATeams.TEAM_MAINTENANCE_ID,
        ])

    aggregate_pipeline = [
        {'$project': {'doc': '$$ROOT'}},
    ]

    if distinct_by_title:
        aggregate_pipeline += [
            {'$sort': {'doc.version_begin': -1}},
            {'$group': {
                '_id': {
                    'qa_team': '$doc.custom_qa_team',
                    'casetype_id': '$doc.casetype_id',
                    'title': '$doc.title',
                },
                'doc': {'$first': '$doc'},
            }},
        ]

    _id_group = {
        'qa_team': '$doc.custom_qa_team',
        'milestone_id': '$doc.milestone_id',
        'casetype_id': '$doc.casetype_id',
    }
    if expand_suites:
        _id_group['suite_id'] = '$doc.suite_id'

    aggregate_pipeline += [
        {'$group': {
            '_id': _id_group,
            'count': {'$sum': 1},
        }},
        {'$group': {
            '_id': {
                'qa_team': '$_id.qa_team',
                'milestone_id': '$_id.milestone_id',
                'suite_id': '$_id.suite_id',
            },
            'case_types': {
                '$push': {
                    'casetype_id': '$_id.casetype_id',
                    'count': '$count',
                },
            },
        }},
    ]

    raw_type_counts = query.aggregate(*aggregate_pipeline)

    result = []
    for raw_type_count in raw_type_counts:
        _id = raw_type_count['_id']

        types_dict = {x.get('casetype_id'): x['count']
                      for x in raw_type_count['case_types']}
        item = {
            'date': moment,
            'qa_team': _id.get('qa_team'),
            'milestone_id': _id.get('milestone_id'),
            'suite_id': _id.get('suite_id'),
            'type_counts': {
                'manual': types_dict.get(CaseType.MANUAL_ID, 0),
                'in_progress': types_dict.get(
                    CaseType.AUTOMATION_IN_PROGRESS_ID, 0),
                'automated': types_dict.get(CaseType.AUTOMATED_ID, 0),
                'not_automate': types_dict.get(
                    CaseType.NOT_AUTOMATE_ID, 0)
            }
        }
        result.append(item)

    return result


class AutomationProgressList(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument(
        'dates', type=lambda v: sorted(set(map(dt_from_iso, v.split(','))))
    )
    parser.add_argument('distinct', type=lambda x: x == 'true')

    def __init__(self):
        self.args = self.parser.parse_args()

    @classmethod
    @cache.memoize()
    def _get_day_type_counts(cls, moment, distinct_by_title=False):
        return get_day_type_counts(
            moment,
            distinct_by_title=distinct_by_title,
            expand_suites=True,
        )

    def get(self):
        res = []
        for current_date in self.args.dates:
            res.extend(
                self._get_day_type_counts(current_date, self.args.distinct)
            )

        suites_map = dict(Suite.objects.scalar('id', 'name'))
        milestones_map = dict(Milestone.objects.scalar('id', 'name'))

        return list(get_data_with_suite(res, suites_map, milestones_map))


class AutomationProgressSummary(BaseResource):

    def get(self, milestone):
        milestone = Milestone.objects.filter(name=milestone).first()
        milestone_plus = [21, 25]  # ids 3.0.0 Milestones
        if not milestone:
            return []

    # Exclude (CI-bot, Framework QA, Maintenance QA, Unknown)
    # team's results on dashboard, and add 3.0.0 Milestone - requirements
        return get_day_type_counts(
            dt_now(),
            milestone_ids=[milestone.id] + milestone_plus,
            dashboard_exclude=True
        )


api.add_resource(AutomationProgressList, '/testrail/automation_progress')

api.add_resource(
    AutomationProgressSummary,
    '/testrail/automation_progress/summary/<string:milestone>'
)
