import datetime
from collections import defaultdict

import pytz
from dateutil import rrule
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad.models import BugTaskHistory
from qareporting.integration.launchpad import consts
from qareporting.resources.base import BaseResource
from qareporting.integration.models import Iteration
from qareporting.utils import dt_from_iso, has_intersection

FUEL_NAME = consts.MajorQATeams.NAMES[consts.MajorQATeams.TEAM_FUEL_ID]


class LpBugTrends(BaseResource):
    AREA_TAGS = {
        'ISO build infra': ['area-build'],
        'CI': ['area-ci'],
        'DevOps': ['area-devops'],
        'Infra Apps': ['area-infra-apps'],
        'QA internal': ['area-qa', 'fuel-devops', 'system-tests'],
    }

    parser = reqparse.RequestParser()
    parser.add_argument('date_from', type=dt_from_iso)
    parser.add_argument('date_to', type=dt_from_iso)
    parser.add_argument('milestone', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

        self.date_from = pytz.utc.localize(datetime.datetime.combine(
            args.date_from.date(), datetime.time.min))

        self.date_to = pytz.utc.localize(datetime.datetime.combine(
            args.date_to.date(), datetime.time.max))

    @classmethod
    def _get_product_for_moment(cls, moment, milestone):
        query = cls._get_base_query(moment, milestone).filter(
            tags__nin=consts.PRODUCT_BUGS_EXCLUDE_TAGS,
        )

        pipeline = [
            {'$group': {
                '_id': {
                    'bug': '$bug_id',
                    'importance': '$importance',
                    'major_team_ids': '$major_team_ids',
                },
            }},
            {'$group': {
                '_id': {
                    'importance': '$_id.importance',
                    'major_team_ids': '$_id.major_team_ids',
                },
                'count': {'$sum': 1},
            }},
        ]

        result = {
            name: defaultdict(int)
            for name in consts.MajorQATeams.NAMES.values()
            if name not in cls.AREA_TAGS
            }

        for record in query.aggregate(*pipeline):
            key = record['_id']

            team_names = set(
                consts.MajorQATeams.NAMES.get(team_id)
                for team_id in key.get('major_team_ids', [])
            )

            for team_name in team_names:
                importance_key = key['importance'].lower()
                if team_name in result:
                    result[team_name][importance_key] += record['count']

        result['QA product'] = result.pop('QA')

        return result

    @classmethod
    def _get_related_for_moment(cls, moment, milestone):

        flatten_area_tags = [
            tag for tags in cls.AREA_TAGS.values() for tag in tags
            ]

        query = cls._get_base_query(moment, milestone).filter(
            tags__in=consts.RELATED_BUGS_INCLUDE_TAGS + flatten_area_tags,
        )

        pipeline = [
            {'$group': {
                '_id': {
                    'bug': '$bug_id',
                    'importance': '$importance',
                    'tags': '$tags',
                },
            }},
            {'$group': {
                '_id': {
                    'importance': '$_id.importance',
                    'tags': '$_id.tags',
                },
                'count': {'$sum': 1},
            }},
        ]

        result = {name: defaultdict(int) for name in cls.AREA_TAGS}

        for record in query.aggregate(*pipeline):
            key = record['_id']
            tags = set(key['tags'])

            team_names = set(
                name for name, area_tags in cls.AREA_TAGS.items()
                if set(area_tags) & tags
            )

            for team_name in team_names:
                result[team_name][key['importance'].lower()] += record['count']

        return result

    @classmethod
    def _get_base_query(cls, moment, milestone):
        query = BugTaskHistory.objects.filter(
            version_begin__lte=moment,
            version_end__gt=moment,
            milestone=milestone,
            status__in=[
                consts.STATUS_INCOMPLETE,
                consts.STATUS_CONFIRMED,
                consts.STATUS_TRIAGED,
                consts.STATUS_IN_PROGRESS,
            ],
            importance__in=[
                consts.IMPORTANCE_CRITICAL,
                consts.IMPORTANCE_HIGH,
                consts.IMPORTANCE_MEDIUM,
                consts.IMPORTANCE_LOW,
            ],
        )
        return query

    def get(self):
        result = []

        iterations_query = Iteration.objects.filter(
            milestone=self.milestone,
            date_begin__lt=self.date_to,
            date_end__gt=self.date_from,
        )

        for iteration in iterations_query:
            iteration_data = {
                'iteration': iteration.to_mongo(),
                'data': []
            }

            week_stops = rrule.rrule(
                freq=rrule.WEEKLY,
                dtstart=iteration.date_begin + datetime.timedelta(
                    days=7, hours=-2, seconds=-1),
                until=iteration.date_end,
            )

            for week_index, week_end_date in enumerate(week_stops, 1):
                # note: week_end_date can be arbitrary weekday

                if not self.date_from <= week_end_date < self.date_to:
                    continue

                iteration_data['data'].append({
                    'week': {
                        'week_index': week_index,
                        'week_end_date': week_end_date,
                    },
                    'data_product': self._get_product_for_moment(
                        week_end_date, self.milestone),
                    'data_related': self._get_related_for_moment(
                        week_end_date, self.milestone),
                })

            if iteration_data['data']:
                result.append(iteration_data)

        return result


api.add_resource(LpBugTrends, '/lp/bug_trends')
