import datetime

import pytz
from dateutil import rrule
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad.models import BugTaskHistory
from qareporting.integration.launchpad import consts
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_now, dt_from_iso, has_intersection


HIGH_AND_CRIT = [consts.IMPORTANCE_HIGH, consts.IMPORTANCE_CRITICAL]
FAKE_IMPORTANCE_OTHER = 'Other'
FAKE_STATUS_DEV_TO_VERIFY = 'Dev-to-verify'
FAKE_STATUS_OPEN = 'Open'


def get_for_moment(moment, milestones):

    query = BugTaskHistory.objects.filter(
        version_begin__lte=moment,
        version_end__gt=moment,
        status__in=consts.OPEN_STATUSES + [
            consts.STATUS_FIX_COMMITTED,
            consts.STATUS_FIX_RELEASED,
        ]
    )

    if milestones:
        query = query.filter(milestone__in=milestones)

    def has_tags(tags):
        return has_intersection('$tags', tags)

    _is_product = {'$not': has_tags(consts.PRODUCT_BUGS_EXCLUDE_TAGS)}
    _is_related = has_tags(consts.RELATED_BUGS_INCLUDE_TAGS)
    _is_internalqa = has_tags(consts.INTERNALQA_BUGS_INCLUDE_TAGS)

    _status = {'$cond': {
        'if': {'$setIsSubset': (['$status'], consts.OPEN_STATUSES)},
        'then': FAKE_STATUS_OPEN,
        'else': '$status',
    }}

    _importance = {'$cond': {
        'if': {'$setIsSubset': (['$importance'], HIGH_AND_CRIT)},
        'then': '$importance',
        'else': FAKE_IMPORTANCE_OTHER,
    }}

    _has_dtv_tag = has_tags(["dev-to-verify"])

    pipeline = [
        {'$group': {
            '_id': {
                'status': _status,
                'importance': _importance,
                'has_dtv_tag': _has_dtv_tag,
                'is_product': _is_product,
                'is_related': _is_related,
                'is_internalqa': _is_internalqa,
            },
            'count': {'$sum': 1},
        }},
        {'$match': {'$or': [
            {'_id.is_product': True},
            {'_id.is_related': True},
            {'_id.is_internalqa': True},
        ]}}
    ]

    agg_query = query.aggregate(*pipeline)

    result = []

    for record in agg_query:
        _id = record['_id']
        status = _id['status']

        if status == consts.STATUS_FIX_COMMITTED and _id['has_dtv_tag']:
            status = FAKE_STATUS_DEV_TO_VERIFY

        item = {
            'date': moment.date().isoformat(),
            'status': status,
            'importance': _id['importance'],
            'is_product': _id['is_product'],
            'is_related': _id['is_related'],
            'is_internalqa': _id['is_internalqa'],
            'count': record['count'],
        }
        result.append(item)

    return result


class LpVerificationProgressList(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument('date_from', type=dt_from_iso)
    parser.add_argument('date_to', type=dt_from_iso)
    parser.add_argument('milestones', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.date_from = args.date_from.date()
        self.date_to = args.date_to.date()
        assert self.date_from <= self.date_to
        assert (self.date_to - self.date_from)
        self.milestones = tuple(sorted(args.milestones.split()))

    def get(self):
        result = []

        dates = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=rrule.WE,
            dtstart=self.date_from,
            until=self.date_to + datetime.timedelta(days=1),
        )

        for date in dates:
            moment = datetime.datetime.combine(
                date, datetime.time.max).replace(tzinfo=pytz.utc)
            data = get_for_moment(moment, self.milestones)
            result.extend(data)

        return result


class VerificationProgressSummary(BaseResource):

    def get(self, milestone):
        return get_for_moment(dt_now(), [milestone])


api.add_resource(LpVerificationProgressList, '/lp/verification_progress')

api.add_resource(
    VerificationProgressSummary,
    '/lp/verification_progress/summary/<string:milestone>'
)
