import datetime

import pytz
from dateutil import rrule
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad.models import BugTaskHistory
from qareporting.integration.launchpad import consts
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso, has_intersection

HIGH_AND_CRIT = [consts.IMPORTANCE_HIGH, consts.IMPORTANCE_CRITICAL]
FAKE_IMPORTANCE_OTHER = 'Other'
FAKE_STATUS_OPEN = 'Open'
VERF_STATUS_ALL = 'all'
VERF_STATUS_PROD = 'prod'
VERF_STATUS_REL = 'rel'
VERF_STATUS_INTER = 'inter'

PERFORMANCE_TAG = 'scale'
PERFORMANCE_VERIFIERS = ['msemenov', 'mchernik',
                         'apodrepniy', 'sarkhipov',
                         'pshvetsov', 'ppetrov']

PCE = [
    consts.MajorQATeams.TEAM_PCE_ID,
]
FUEL = [
    consts.MajorQATeams.TEAM_FUEL_ID,
    consts.MajorQATeams.TEAM_WORKLOADS_ID,
    consts.MajorQATeams.TEAM_L23NETWORKING_ID,
    consts.MajorQATeams.TEAM_CUSTOMER_ENABLEMENT_ID,
    consts.MajorQATeams.TEAM_DOCS_ID,
    consts.MajorQATeams.TEAM_PLUGINS_ID,
]

MAINTENANCE = [consts.MajorQATeams.TEAM_MAINTENANCE_ID]
DOCS = [consts.MajorQATeams.TEAM_DOCS_ID]
INFRA = [consts.MajorQATeams.TEAM_INFRA_ID]

FAKE_MAP = {
    'fuel': 0,
    'mos': 1,
    'telco': 2,
    'performance': 4,
    'pce': 3,
    'maintenance': 5,
    'docs': 6,
    'infra': 7,
    'dev': 8,
    'unknown': 9
}

TEAMS = [
    consts.MajorQATeams.TEAM_FUEL_ID,
    consts.MajorQATeams.TEAM_MOS_ID,
    consts.MajorQATeams.TEAM_PCE_ID,
    consts.MajorQATeams.TEAM_QA_ID,
    consts.MajorQATeams.TEAM_MAINTENANCE_ID,
    consts.MajorQATeams.TEAM_WORKLOADS_ID,
    consts.MajorQATeams.TEAM_CUSTOMER_ENABLEMENT_ID
]

FUEL_EXCLUDE_TAGS_PROD = [
    'area-build',
    'area-ceph',
    'area-ci',
    'area-devops',
    'area-infra-apps',
    'area-linux',
    'area-mos',
    'area-qa',
    'ceph',
    'covered-by-bp',
    'dev-to-verify',
    'devops',
    'enhancement',
    'experimental',
    'feature-cgroups',
    'feature-cpu-pinning',
    'feature-granular-task-lcm-readiness',
    'feature-huge-pages',
    'feature-multipath-disks-suppor',
    'feature-openstack-config-change',
    'feature-reduced-footprint',
    'feature-ubuntu-bootstrap',
    'feature',
    'fuel-devops',
    'maintenance',
    'need-bp',
    'non-release',
    'scale',
    'system-tests',
    'team-mixed',
    'team-telco',
    'team-upgrades',
    'tech-debt',
    'vbox-scripts',
]

FUEL_EXCLUDE_TAGS_RELEASE = FUEL_EXCLUDE_TAGS_QA = FUEL_EXCLUDE_TAGS_ALL = [
    'area-ceph',
    'area-linux',
    'area-mos',
    'ceph',
    'dev-to-verify',
    'feature-cgroups',
    'feature-cpu-pinning',
    'feature-granular-task-lcm-readiness',
    'feature-huge-pages',
    'feature-multipath-disks-suppor',
    'feature-openstack-config-change',
    'feature-reduced-footprint',
    'feature-ubuntu-bootstrap',
    'maintenance',
    'scale',
    'team-mixed',
    'team-telco',
    'team-upgrades',
    'vbox-scripts',
]

FUEL_INCLUDE_TAGS_PROD = [
    'area-library',
    'area-python',
    'area-ui',
]

FUEL_INCLUDE_TAGS_RELEASE = [
    'area-build',
    'area-ci',
    'area-devops',
    'area-infra-apps',
    'area-qa',
    'fuel-devops',
    'system-tests',
]

FUEL_INCLUDE_TAGS_QA = [
    'area-qa',
    'fuel-devops',
    'system-tests',
]

MOS_EXCLUDE_TAGS_PROD = [
    'area-build',
    'area-ci',
    'area-devops',
    'area-infra-apps',
    'area-qa',
    'covered-by-bp',
    'dev-to-verify',
    'devops',
    'enhancement',
    'experimental',
    'feature-cgroups',
    'feature-cpu-pinning',
    'feature-granular-task-lcm-readiness',
    'feature-huge-pages',
    'feature-multipath-disks-suppor',
    'feature-openstack-config-change',
    'feature-reduced-footprint',
    'feature-ubuntu-bootstrap',
    'feature',
    'fuel-devops',
    'maintenance',
    'need-bp',
    'non-release',
    'scale',
    'system-tests',
    'team-mixed',
    'team-telco',
    'tech-debt',
]

MOS_EXCLUDE_TAGS_ALL = [
    'dev-to-verify',
    'feature-cgroups',
    'feature-cpu-pinning',
    'feature-granular-task-lcm-readiness',
    'feature-huge-pages',
    'feature-multipath-disks-suppor',
    'feature-openstack-config-change',
    'feature-reduced-footprint',
    'feature-ubuntu-bootstrap',
    'scale',
    'team-mixed',
    'team-telco',
]

MOS_INCLUDE_TAGS = [
    'area-ceilometer',
    'area-mos',
    'area-ironic',
]

TELCO_INCLUDE_TAGS = [
    'feature-cgroups',
    'feature-cpu-pinning',
    'feature-granular-task-lcm-readiness',
    'feature-huge-pages',
    'feature-multipath-disks-support',
    'feature-openstack-config-change',
    'feature-reduced-footprint',
    'feature-ubuntu-bootstrap',
]

PCE_EXCLUDE_TAGS = [
    'feature-cgroups',
    'feature-cpu-pinning',
    'feature-granular-task-lcm-readiness',
    'feature-huge-pages',
    'feature-multipath-disks-suppor',
    'feature-openstack-config-change',
    'feature-reduced-footprint',
    'feature-ubuntu-bootstrap',
    'team-mixed',
    'team-telco',
]


def get_team_id(data):
    if any(k for k in data.values()):
        if data['fuel']:
            return FAKE_MAP['fuel']
        if data['mos']:
            return FAKE_MAP['mos']
        if data['telco']:
            return FAKE_MAP['telco']
        if data['pce']:
            return FAKE_MAP['pce']
        if data['performance']:
            return FAKE_MAP['performance']
    else:
        return FAKE_MAP['unknown']


class LpVerificationProgressTeams(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('date_from', type=dt_from_iso)
    parser.add_argument('date_to', type=dt_from_iso)
    parser.add_argument('milestones', default='')
    parser.add_argument('verf_status', default=VERF_STATUS_ALL)

    def __init__(self):
        args = self.parser.parse_args()
        self.date_from = args.date_from.date()
        self.date_to = args.date_to.date()
        assert self.date_from <= self.date_to
        assert (self.date_to - self.date_from)
        self.milestones = tuple(sorted(args.milestones.split()))
        self.verf_status = args.verf_status

    def _get_for_moment(self, moment, milestones, verf_status):

        query = BugTaskHistory.objects.filter(
            version_begin__lte=moment,
            version_end__gt=moment,
            status__in=consts.OPEN_STATUSES + [
                consts.STATUS_FIX_COMMITTED,
                consts.STATUS_FIX_RELEASED
            ]
        )
        if milestones:
            query = query.filter(
                milestone__in=milestones)

        def has_tags(tags):
            return has_intersection('$tags', tags)

        _is_fuel = None
        _is_mos = None
        _is_telco = has_tags(TELCO_INCLUDE_TAGS)
        _is_pce = None
        _is_perform = _is_perform = has_tags(['scale'])

        if verf_status == VERF_STATUS_PROD:
            _is_fuel = {'$and': [{'$or': [
                {'$eq': [
                    '$dev_to_qa_team_id',
                    consts.MajorQATeams.TEAM_FUEL_ID
                ]},
                has_tags(FUEL_INCLUDE_TAGS_PROD),
                ]},
                {'$not': has_tags(FUEL_EXCLUDE_TAGS_PROD)}
            ]}
            _is_mos = {'$or': [
                {'$and': [
                    has_tags(MOS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_PROD)},
                    {'$eq': ['$dev_to_qa_team_id',
                             consts.MajorQATeams.TEAM_MOS_ID]}
                ]},
                {'$and': [
                    has_tags(MOS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_PROD)},
                    {'$setIsSubset': (
                        [consts.MajorQATeams.TEAM_MOS_ID], {
                            '$ifNull': ('$major_team_ids', [])})}
                ]}
            ]}
            _is_pce = {'$and': [
                {'$not': has_tags(PCE_EXCLUDE_TAGS)},
                {'$eq': [
                    '$dev_to_qa_team_id', consts.MajorQATeams.TEAM_PCE_ID
                ]}
            ]}

        elif verf_status == VERF_STATUS_REL:
            _is_fuel = {'$and': [
                has_tags(FUEL_INCLUDE_TAGS_RELEASE),
                {'$not': has_tags(FUEL_EXCLUDE_TAGS_RELEASE)},
                {'$eq': ['$dev_to_qa_team_id',
                         consts.MajorQATeams.TEAM_FUEL_ID]}
            ]}
            _is_mos = {'$or': [
                {'$and': [
                    has_tags(consts.RELATED_BUGS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_ALL)},
                    {'$eq': ['$dev_to_qa_team_id',
                             consts.MajorQATeams.TEAM_MOS_ID]}
                ]},
                {'$and': [
                    has_tags(consts.RELATED_BUGS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_ALL)},
                    {'$setIsSubset': (
                        [consts.MajorQATeams.TEAM_MOS_ID], {
                            '$ifNull': ('$major_team_ids', [])})}
                ]}
            ]}
        elif verf_status == VERF_STATUS_INTER:
            _is_fuel = {'$and': [
                has_tags(FUEL_INCLUDE_TAGS_QA),
                {'$not': has_tags(FUEL_EXCLUDE_TAGS_QA)},
                {'$eq': ['$dev_to_qa_team_id',
                         consts.MajorQATeams.TEAM_FUEL_ID]}
            ]}
            _is_mos = {'$or': [
                {'$and': [
                    has_tags(consts.INTERNALQA_BUGS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_ALL)},
                    {'$eq': ['$dev_to_qa_team_id',
                             consts.MajorQATeams.TEAM_MOS_ID]}
                ]},
                {'$and': [
                    has_tags(consts.INTERNALQA_BUGS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_ALL)},
                    {'$setIsSubset': (
                        [consts.MajorQATeams.TEAM_MOS_ID], {
                            '$ifNull': ('$major_team_ids', [])})}
                ]}
            ]}
        else:
            _is_fuel = {'$and': [
                {'$not': has_tags(FUEL_EXCLUDE_TAGS_ALL)},
                {'$eq': ['$dev_to_qa_team_id',
                         consts.MajorQATeams.TEAM_FUEL_ID]}
            ]}
            _is_mos = {'$or': [
                {'$and': [
                    has_tags(MOS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_ALL)},
                    {'$eq': ['$dev_to_qa_team_id',
                             consts.MajorQATeams.TEAM_MOS_ID]}
                ]},
                {'$and': [
                    has_tags(MOS_INCLUDE_TAGS),
                    {'$not': has_tags(MOS_EXCLUDE_TAGS_ALL)},
                    {'$setIsSubset': (
                        [consts.MajorQATeams.TEAM_MOS_ID], {
                            '$ifNull': ('$major_team_ids', [])})}
                ]}
            ]}
            _is_pce = {'$and': [
                {'$eq': [
                    '$dev_to_qa_team_id', consts.MajorQATeams.TEAM_PCE_ID
                ]}
            ]}

        _status = {'$cond': {
            'if': {'$setIsSubset': (['$status'], consts.OPEN_STATUSES)},
            'then': FAKE_STATUS_OPEN,
            'else': '$status',
        }}

        _importance = {'$cond': {
            'if': {'$setIsSubset': (['$importance'], HIGH_AND_CRIT)},
            'then': '$importance',
            'else': FAKE_IMPORTANCE_OTHER,
        }}

        pipeline = [
            {'$group': {
                '_id': {
                    'bug': '$bug_id',
                    'status': _status,
                    'importance': _importance,
                    'fuel': _is_fuel,
                    'mos': _is_mos,
                    'telco': _is_telco,
                    'pce': _is_pce,
                    'perform': _is_perform,
                }
            }},
            {'$group': {
                '_id': {
                    'status': '$_id.status',
                    'importance': '$_id.importance',
                    'fuel': '$_id.fuel',
                    'mos': '$_id.mos',
                    'telco': '$_id.telco',
                    'pce': '$_id.pce',
                    'performance': '$_id.perform',
                 },
                'count': {'$sum': 1}
            }}
        ]

        agg_query = query.aggregate(*pipeline)

        result = []
        for record in agg_query:
            _id = record['_id']
            status = _id['status']
            importance = _id['importance']
            item = {
                'team_id': get_team_id(_id),
                'date': moment.date().isoformat(),
                'status': status,
                'importance': importance,
                'count': record['count'],
            }
            result.append(item)
        return result

    def get(self):
        result = []

        dates = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=rrule.WE,
            dtstart=self.date_from,
            until=self.date_to + datetime.timedelta(days=1),
        )

        for date in dates:
            moment = datetime.datetime.combine(
                date, datetime.time.max
            ).replace(tzinfo=pytz.utc)
            data = self._get_for_moment(
                moment, self.milestones, self.verf_status
            )
            result.extend(data)

        return result


api.add_resource(
    LpVerificationProgressTeams, '/lp/verification_progress_teams'
)
