from collections import defaultdict
import datetime
from itertools import chain

from flask_restful import reqparse

from qareporting.ext import api
from qareporting.ext import cache
from qareporting.integration.testrail.models import (
    Plan, Run, Milestone, StatusCounts)
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso


def fake_milestone(string):
    if '_' in string:
        return True, string
    return False, int(string)


parser = reqparse.RequestParser()
parser.add_argument('date_from', type=dt_from_iso)
parser.add_argument('date_to', type=dt_from_iso)
parser.add_argument('milestone_id', type=fake_milestone)
parser.add_argument('milestone')

empty_status_counts = dict(
    (field, 0) for field in StatusCounts._fields_ordered
)


def get_swarm_bvt_data(plans):
    plan_data = {}

    for plan in plans:
        plan_date = plan.created_on.date().isoformat()
        plan_data[plan.pk] = {
            'date': plan_date,
            'name': plan.name,
            Run.RUN_TYPE_SWARM_ID: [],
            Run.RUN_NAME_SMOKE_BVT: []
        }

    swarm_runs = Run.objects(
        plan__in=tuple(plan_data),
        name__not__contains='Smoke/BVT',
        run_type=Run.RUN_TYPE_SWARM_ID,
    )

    for run in swarm_runs:
        plan_data[run.plan.pk][Run.RUN_TYPE_SWARM_ID] = (
            run.status_counts.to_mongo().to_dict())

    bvt_runs = Run.objects(
        plan__in=tuple(plan_data),
        name=Run.RUN_NAME_SMOKE_BVT,
    )

    for run in bvt_runs:
        plan_id = run.plan.pk
        bvt_statuses = run.status_counts.to_mongo().to_dict()
        plan_data[plan_id][Run.RUN_NAME_SMOKE_BVT] = bvt_statuses

    bvt_data = defaultdict(list)
    bvt_plan_names = defaultdict(list)

    for i in plan_data.values():
        if i[Run.RUN_NAME_SMOKE_BVT]:
            bvt_data[i['date']].append(i[Run.RUN_NAME_SMOKE_BVT])
            bvt_plan_names[i['date']].append(i['name'])

    swarm_data = {}

    for i in plan_data.values():
        if i[Run.RUN_TYPE_SWARM_ID]:
            swarm_data[i['date']] = {
                'name': i['name'], 'data': i[Run.RUN_TYPE_SWARM_ID]
            }
    result = []

    for bvt_date in bvt_data:
        if bvt_date in swarm_data:
            res = {
                'date': bvt_date,
                'milestone_name': swarm_data[bvt_date]['name'],
                'status_counts': swarm_data[bvt_date]['data'],
                'bvt_counts': bvt_data[bvt_date],
            }
        else:
            res = {
                'date': bvt_date,
                'milestone_name': bvt_plan_names[bvt_date][-1],
                'status_counts': empty_status_counts,
                'bvt_counts': bvt_data[bvt_date],
            }
        result.append(res)

    return result


def get_recent_seven(plans):
    plan_data = {}

    for plan in plans:
        plan_date = plan.created_on.date().isoformat()
        plan_data[plan.pk] = {
            'date': plan_date,
            'name': plan.name,
            Run.RUN_TYPE_SWARM_ID: [],
            Run.RUN_NAME_SMOKE_BVT: []
        }

    swarm_runs = Run.objects[:7](
        plan__in=tuple(plan_data),
        name__not__contains='Smoke/BVT',
        run_type=Run.RUN_TYPE_SWARM_ID,
    ).order_by('-created_on')

    if swarm_runs:
        count = swarm_runs.count()
        if count > 6:
            date_from = datetime.datetime.combine(
                swarm_runs[6].created_on, datetime.time.min)
        else:
            date_from = datetime.datetime.combine(
                swarm_runs[count - 1].created_on, datetime.time.min)
    else:
        return []

    bvt_runs = Run.objects(
        plan__in=tuple(plan_data),
        name=Run.RUN_NAME_SMOKE_BVT,
        run_type=Run.RUN_TYPE_SWARM_ID,
        created_on__gt=date_from,
    ).order_by('-created_on')

    for run in bvt_runs:
        plan_data[run.plan.pk][Run.RUN_NAME_SMOKE_BVT] = (
            run.status_counts.to_mongo().to_dict()
        )

    for run in swarm_runs:
        plan_data[run.plan.pk][Run.RUN_TYPE_SWARM_ID] = (
            run.status_counts.to_mongo().to_dict()
        )

    bvt_data = defaultdict(list)
    bvt_plan_names = defaultdict(list)

    for i in plan_data.values():
        if i[Run.RUN_NAME_SMOKE_BVT]:
            bvt_data[i['date']].append(i[Run.RUN_NAME_SMOKE_BVT])
            bvt_plan_names[i['date']].append(i['name'])

    swarm_data = {}

    for i in plan_data.values():
        if i[Run.RUN_TYPE_SWARM_ID]:
            swarm_data[i['date']] = {
                'name': i['name'], 'data': i[Run.RUN_TYPE_SWARM_ID]
            }

    result = {}

    for date in chain(bvt_data, swarm_data):
        if date in swarm_data and date in bvt_data:
            if date not in result:
                result[date] = {
                    'date': date,
                    'milestone_name': swarm_data[date]['name'],
                    'status_counts': swarm_data[date]['data'],
                    'bvt_counts': bvt_data[date],
                }
        elif date in swarm_data:
            if date not in result:
                result[date] = {
                    'date': date,
                    'milestone_name': swarm_data[date]['name'],
                    'status_counts': swarm_data[date]['data'],
                    'bvt_counts': [],
                }
        elif date in bvt_data:
            if date not in result:
                result[date] = {
                    'date': date,
                    'milestone_name': bvt_plan_names[date][-1],
                    'status_counts': empty_status_counts,
                    'bvt_counts': bvt_data[date],
                }

    return sorted(result.values(), key=lambda x: x['date'])[-7:]


class SwarmMetrics(BaseResource):
    def __init__(self):
        args = parser.parse_args()
        self.date_from = args.date_from
        self.date_to = args.date_to
        self.is_fake_milestone, self.milestone_id = args.milestone_id
        self.milestone = args.milestone

    @cache.memoize()
    def _get_data(self, date_from, date_to, milestone_id, is_fake):
        if not is_fake:
            plans = Plan.objects.filter(
                created_on__gte=date_from,
                created_on__lte=date_to,
                milestone=milestone_id,
                fake_milestone__exists=False,
            )
        else:
            plans = Plan.objects.filter(
                created_on__gte=date_from,
                created_on__lte=date_to,
                fake_milestone=milestone_id,
            )

        return get_swarm_bvt_data(plans)

    def get(self):
        return self._get_data(
            self.date_from,
            self.date_to,
            self.milestone_id,
            self.is_fake_milestone,
        )


class SwarmSummary(BaseResource):
    def __init__(self):
        args = parser.parse_args()
        self.date_from = args.date_from
        self.date_to = args.date_to
        self.milestone = args.milestone

    def get(self):
        milestone_ids = Milestone.objects(name=self.milestone).scalar('id')
        plans = Plan.objects.filter(
            created_on__gte=self.date_from,
            created_on__lte=self.date_to,
            milestone__in=milestone_ids
        )

        return get_swarm_bvt_data(plans)


class SwarmTempestSummary(BaseResource):
    def __init__(self):
        args = parser.parse_args()
        self.milestone = args.milestone

    def get(self):

        milestone_ids = Milestone.objects(name=self.milestone).scalar('id')

        last_swarm_and_tempest_pipeline = [
            {'$group': {'_id': '$plan_id',
                        'run_types': {'$addToSet': '$run_type'},
                        'date': {'$min': '$created_on'},
                        'milestone_ids': {'$addToSet': '$milestone_id'}
                        }},
            {'$match': {
                'run_types': {'$all': [
                    Run.RUN_TYPE_SWARM_ID, Run.RUN_TYPE_TEMPEST_ID
                ]},
                'milestone_ids': {'$all': [x for x in milestone_ids]}
            }},
            {'$sort': {'date': -1}},
            {'$limit': 1}
        ]

        recent_swarm_tempest = Run.objects(
        ).aggregate(*last_swarm_and_tempest_pipeline)

        result = {}

        if recent_swarm_tempest.alive:

            for item in recent_swarm_tempest:
                swarm_tempest_plan_id = item['_id']
                run_date = item['date']

            recent_swarm_day_max = datetime.datetime.combine(
                run_date, datetime.time.max)
            recent_swarm_day_min = datetime.datetime.combine(
                run_date, datetime.time.min)

            recent_tempest = Run.objects(
                run_type=Run.RUN_TYPE_TEMPEST_ID,
                plan_id=swarm_tempest_plan_id,
            )

            recent_swarm = Run.objects(
                run_type=Run.RUN_TYPE_SWARM_ID,
                name__contains='Swarm',
                milestone_id__in=milestone_ids,
                plan_id=swarm_tempest_plan_id,
            ).first()

            recent_bvt = Run.objects(
                name__contains=Run.RUN_NAME_SMOKE_BVT,
                created_on__lte=recent_swarm_day_max,
                created_on__gte=recent_swarm_day_min,
                milestone_id__in=milestone_ids
            )

            result['tempest_data'] = [
                dict(date=run_date.date().isoformat(),
                     config_name=run.name,
                     status_counts=run.status_counts.to_mongo().to_dict(),
                     plan_name=run.plan.name)
                for run in recent_tempest]

            if recent_swarm:
                result['swarm_data'] = [
                    dict(date=run_date.date().isoformat(),
                         config_name=recent_swarm.name,
                         status_counts=recent_swarm.status_counts.to_mongo().to_dict(),
                         plan_name=recent_swarm.plan.name)
                ]
            else:
                result['swarm_data'] = [
                    dict(date=run_date.date().isoformat(),
                         config_name='None',
                         status_counts=empty_status_counts,
                         plan_name='None')
                ]

            result['bvt_data'] = [
                dict(date=run_date.date().isoformat(),
                     config_name=run.name,
                     status_counts=run.status_counts.to_mongo().to_dict(),
                     plan_name=run.plan.name)
                for run in recent_bvt]

        else:
            return result

        return result


class SwarmRecentSeven(BaseResource):
    def __init__(self):
        args = parser.parse_args()
        self.is_fake_milestone, self.milestone_id = args.milestone_id

    def get(self):

        if not self.is_fake_milestone:
            plans = Plan.objects.filter(
                milestone=self.milestone_id,
                fake_milestone__exists=False,
            )
        else:
            plans = Plan.objects.filter(
                fake_milestone=self.milestone_id,
            )

        return get_recent_seven(plans)


# This resource used not only in BVT-SWARM report, but on Dashboard widget too.
api.add_resource(SwarmMetrics, '/testrail/swarm_metrics')
api.add_resource(SwarmSummary, '/testrail/swarm_summary')
api.add_resource(SwarmRecentSeven, '/testrail/swarm_recent_seven')
api.add_resource(SwarmTempestSummary, '/testrail/swam_tempest_summary')
