import re

from flask_restful import reqparse
from mongoengine.queryset.visitor import Q

from qareporting.ext import api
from qareporting.integration.launchpad import consts
from qareporting.integration.launchpad.models import Bug, BugTaskHistory
from qareporting.integration.testrail.models import (
    Run, Test, Status, Plan, Milestone)
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_now


OTHER_TYPE_ID = 1

RUN_TYPES = {
    OTHER_TYPE_ID: 'other',
    Run.RUN_TYPE_SWARM_ID: 'swarm_bvt',
    Run.RUN_TYPE_TEMPEST_ID: 'tempest',
    Run.RUN_TYPE_ACCEPTANCE_ID: 'acceptance',
}

FAILED_STATUSES = [
    Status.STATUS_FAILED,
    Status.STATUS_INFRA_FAILED,
    Status.STATUS_PROD_FAILED,
    Status.STATUS_TEST_FAILED,
    Status.STATUS_BLOCKED
]


def get_failure_drivers_data(milestone_id=None, plan_id=None, status='active'):

    if not plan_id:
        runs_with_bugs = Test.objects.filter(
            status_id__in=FAILED_STATUSES,
            results__custom_launchpad_bug__exists=True,
        ).distinct('run_id')

        plans_with_bugs = Run.objects.filter(id__in=runs_with_bugs)

        if milestone_id:
            plans_with_bugs = plans_with_bugs.filter(milestone_id=milestone_id)

        plans_with_bugs = tuple(plans_with_bugs.distinct('plan_id'))

        runs = Run.objects(plan__in=plans_with_bugs).as_pymongo()

        plans = Plan.objects(
            id__in=plans_with_bugs
        ).only('id', 'name').as_pymongo()

    else:
        runs = Run.objects(plan=plan_id).as_pymongo()

        plans = Plan.objects(
            id=plan_id
        ).only('id', 'name').as_pymongo()

    plan_map = {plan['_id']: plan['name'] for plan in plans}

    def get_run_type_id(run):
        value = run.get('run_type')
        if value in (None, Run.RUN_TYPE_MANUAL_ID):
            return OTHER_TYPE_ID
        return value

    run_map = {
        run['_id']: {
            'type_id': get_run_type_id(run),
            'milestone_id': run.get('milestone_id'),
            'plan_id': run.get('plan_id'),
            'run_name': run.get('name')}
        for run in runs}

    # The bug escalation computes by the first mention in TestRail
    bug_escalation_aggregate = [
        {'$sort': {'run_id': 1}},
        {'$unwind': '$results'},
        {'$group': {
            '_id': '$results.custom_launchpad_bug',
            'date': {'$min': "$results.created_on"}}},
    ]

    test_escalation_query = Test.objects.filter(
        status_id__in=FAILED_STATUSES,
        results__custom_launchpad_bug__exists=True,
    )

    bug_escalation = test_escalation_query.aggregate(*bug_escalation_aggregate)

    bug_escal_date_map = {
        bug['_id']: bug['date'] for bug in bug_escalation}

    test_aggregate_pipeline = [
        {'$unwind': '$results'},
        {'$group': {
            '_id': {'test': '$_id', 'run': '$run_id'},
            'bugs': {'$addToSet': '$results.custom_launchpad_bug'}}},
        {'$unwind': '$bugs'},
        {'$group': {
            '_id': {'run_id': '$_id.run', 'bug_url': '$bugs'},
            'count': {'$sum': 1}}},
        {'$sort': {'_id.run_id': 1}}
    ]

    test_query = Test.objects.filter(
        status_id__in=FAILED_STATUSES,
        results__custom_launchpad_bug__exists=True,
        run_id__in=tuple(run_map)
    )

    raw_tests = test_query.aggregate(*test_aggregate_pipeline)

    bug_pattern = re.compile(r'(fuel|mos)\/(.*)\/(?P<bug_id>\d+)\/?')

    bug_map = {}
    for test in raw_tests:
        link = test['_id']['bug_url']
        match_link = bug_pattern.search(link)
        if match_link:
            bug_id = int(match_link.group('bug_id'))
            counter = {
                RUN_TYPES[OTHER_TYPE_ID]: 0,
                RUN_TYPES[Run.RUN_TYPE_SWARM_ID]: 0,
                RUN_TYPES[Run.RUN_TYPE_TEMPEST_ID]: 0,
                RUN_TYPES[Run.RUN_TYPE_ACCEPTANCE_ID]: 0,
            }

            run_info = {
                RUN_TYPES[OTHER_TYPE_ID]: [],
                RUN_TYPES[Run.RUN_TYPE_SWARM_ID]: None,
                RUN_TYPES[Run.RUN_TYPE_TEMPEST_ID]: None,
                RUN_TYPES[Run.RUN_TYPE_ACCEPTANCE_ID]: None,
            }

            run_type = RUN_TYPES[run_map[test['_id']['run_id']]['type_id']]
            milestone_id = run_map[test['_id']['run_id']]['milestone_id']
            run_name = run_map[test['_id']['run_id']]['run_name']
            plan_name = plan_map[run_map[test['_id']['run_id']]['plan_id']]

            counter[run_type] = test['count']

            if run_type == RUN_TYPES[OTHER_TYPE_ID]:
                run_info[run_type].append(plan_name + ' - ' + run_name)
            else:
                run_info[run_type] = plan_name

            if bug_id in bug_map:
                bug_map[bug_id]['rate'].update(counter)
                runs = bug_map[bug_id]['runs']
                if run_type == RUN_TYPES[OTHER_TYPE_ID]:
                    if run_info[run_type] not in runs[run_type]:
                        runs[run_type] += run_info[run_type]
                else:
                    runs[run_type] = run_info[run_type]
                bug_map[bug_id]['runs'].update(runs)
                bug_map[bug_id]['milestone_id'] = milestone_id
                bug_map[bug_id]['iso_name'] = plan_name
            else:
                bug_map[bug_id] = dict(
                    iso_name=plan_name,
                    bug_id=bug_id,
                    rate=counter,
                    runs=run_info,
                    milestone_id=milestone_id,
                    bug_escalation=bug_escal_date_map[link])

    bug_reshape_pipeline = [
        {'$project': {
            '_id': 1,
            'title': 1,
            'user_comment': 1,
            'eta': 1,
            'date_fix_committed': {
                '$arrayElemAt': ['$tasks.date_fix_committed', 0]},
        }}
    ]

    result = {}

    tr_milestones = dict(Milestone.objects.scalar('name', 'id'))

    if status == 'fixed':
        bug_statuses = BugTaskHistory.objects(
            bug_id__in=tuple(bug_map),
            version_end__gt=dt_now(),
            status__in=consts.FIXED_STATUSES
        ).only(
            'status', 'importance',
            'assignee', 'bug_created',
            'major_team_ids', 'tags',
            'bug_id', 'milestone',
        ).as_pymongo()
    else:
        bug_statuses = BugTaskHistory.objects(
            Q(bug_id__in=tuple(bug_map),
              version_end__gt=dt_now(),
              status__in=consts.OPEN_STATUSES) |
            Q(status__in=consts.OPEN_STATUSES,
              version_end__gt=dt_now())
        ).only(
            'status', 'importance',
            'assignee', 'bug_created',
            'major_team_ids', 'tags',
            'bug_id', 'milestone',
        ).as_pymongo()

    lp_milestone_re = re.compile(r'^(\d+)(\.\d+)')

    for bug in bug_statuses:
        bug_id = bug['bug_id']

        team_id = next(iter(bug.get('major_team_ids', [])), None)

        if bug_id not in bug_map:
            lp_milestone = bug.get('milestone', '')

            match_milestone = lp_milestone_re.search(lp_milestone)

            if match_milestone:
                milestone_name = match_milestone.group()

                if milestone_name == '10.1':
                    milestone_name = '10.0'

                tr_milestone_id = tr_milestones.get(
                    milestone_name, milestone_name)

            else:

                tr_milestone_id = lp_milestone

            bug_map[bug_id] = {
                'bug_id': bug_id,
                'milestone_id': tr_milestone_id,
                'iso_name': 'none'
            }

        bug_map[bug_id].update(
            {
                'bug_status': bug['status'],
                'bug_importance': bug['importance'],
                'bug_reported': bug['bug_created'],
                'bug_assignee': bug.get('assignee'),
                'lp_milestone': bug.get('milestone'),
                'team': consts.MajorQATeams.NAMES.get(
                    team_id, 'Unknown'
                ),
                'tags': bug.get('tags', []),
            }
        )

    bugs_with_statuses = [bug_id for bug_id, val in bug_map.items()
                          if val.get('bug_status') and
                          val.get('bug_importance')]
    bugs = Bug.objects(
        id__in=bugs_with_statuses,
    ).aggregate(*bug_reshape_pipeline)

    for bug in bugs:
        result_bug = bug_map[bug['_id']]

        result_bug.update(
            {'bug_title': '#' + str(bug['_id']) + ' ' + bug['title'],
             'bug_fix_committed': bug.get('date_fix_committed'),
             'bug_ETA': bug.get('eta', 'n/a'),
             'bug_comment': bug.get('user_comment'),
             })
        result[bug['_id']] = result_bug

    return [bug for bug in result.values()]


class FailureMetrics(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument('milestone_id', type=int)
    parser.add_argument('status', type=str)

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone_id = args.milestone_id
        self.status = args.status

    def get(self):
        if self.milestone_id:
            return get_failure_drivers_data(self.milestone_id)
        return get_failure_drivers_data(status=self.status)


api.add_resource(FailureMetrics, '/testrail/failure_metrics')
