import re
from collections import defaultdict

from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.jira.models import Issue, JiraFilter
from qareporting.integration.launchpad.models import Bug
from qareporting.integration.testrail.models import (
    Case, Test, Status, Milestone, Run)
from qareporting.resources.base import BaseResource


def select_result(results):
    """Returns the latest result with status."""

    sorted_results = sorted(
        results,
        key=lambda x: x['created_on'],
        reverse=True
    )

    for res in sorted_results:
        if res.get('status_id'):
            return res

    return None


def get_bug_id(result):
    try:
        bug_id = int(result['custom_launchpad_bug'].rpartition('/')[-1])
    except:
        return None
    return bug_id


def get_cases(issues, bugs, version):
    milestone_ids = Milestone.objects(name=version).scalar('id')

    run_ids = list(Run.objects.filter(
        milestone_id__in=milestone_ids
    ).distinct('id'))

    cases = Case.objects.filter(
        refs__in=tuple(i.key for i in issues)
    ).as_pymongo()

    cases_ids = [x['_id'] for x in cases]

    case_last_test_query = Test.objects.aggregate(
        {'$match': {
            'run_id': {'$in': run_ids},
            'case_id': {'$in': cases_ids}
        }},
        {'$sort': {'results.created_on': -1}},
        {'$group': {'_id': '$case_id', 'results': {'$first': '$results'}}}
    )

    last_test_map = {x['_id']: x['results'] for x in case_last_test_query}

    result_cases_data = []
    for case in cases:
        case_id = case['_id']

        results = last_test_map.get(case_id, [])
        if results:
            result = results[0]  # get first(last by created_on)
            last_result_status_id = result.get('status_id')
            if result.get('custom_launchpad_bug'):
                bug_id = get_bug_id(result)
                if bug_id:
                    if bug_id in bugs:
                        bugs[bug_id].update([ref for ref in case['refs']])
                    else:
                        for ref in case['refs']:
                            if ref in (issue['key'] for issue in issues):
                                bugs[bug_id].add(ref)
            if last_result_status_id:
                result_cases_data.append({
                    'id': case_id,
                    'refs': case['refs'],
                    'last_result_status':
                        Status.INTERNAL_CODES[
                            last_result_status_id],
                })
        else:  # if case doesn't have any results
            result_cases_data.append({
                'id': case_id,
                'refs': case['refs'],
                'last_result_status': Status.INTERNAL_CODES[
                    Status.STATUS_UNTESTED],
            })

    return result_cases_data


def get_feature_testing_data(version=None):

    filter_id = None

    if version:
        filter_id = JiraFilter.objects(name=version).first()

    if filter_id:
        issues = tuple(Issue.objects(filter_ids=filter_id.filter_id))
    else:
        issues = tuple(Issue.objects)

    issues_with_bug_ids = [issue for issue in issues if
                           issue.launchpad_bug_ids]

    bugs = defaultdict(set)

    for issue in issues_with_bug_ids:
        bugs_list = issue.launchpad_bug_ids
        for bug in bugs_list:
            bugs[bug].add(issue.key)

    refs = defaultdict(list)

    cases = get_cases(issues, bugs, version)

    if version:
        exceptional_case = None

        if version == '10.0':
            exceptional_case = re.compile(r'[10.0|10.1]')

        if exceptional_case:
            bugs_data = Bug.objects(
                id__in=tuple(bugs),
                tasks__milestone=exceptional_case
            )
        else:
            bugs_data = Bug.objects(
                id__in=tuple(bugs),
                tasks__milestone__contains=version
            )

        for bug in bugs_data:
            tasks = bug.tasks
            for task in tasks:
                milestone = task.milestone
                add_bugs = None
                if version == '10.0':
                    if any(x in milestone for x in ('10.0', '10.1')):
                        add_bugs = True

                elif version in milestone:
                    add_bugs = True

                if add_bugs:
                    bug_refs = bugs[bug.pk]
                    for ref in bug_refs:
                        refs[ref].append({
                            'status': task.status,
                            'bug_id': bug.pk,
                        })
                    break

    return {
        'issues': issues,
        'cases': cases,
        'bugs': refs,
    }


class FeatureTestingProgress(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('prod_version', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.prod_version = args.prod_version

    def get(cls):
        return get_feature_testing_data(version=cls.prod_version)


class FeatureTestingSummary(BaseResource):
    def get(self, milestone):
        # currently this reports does not depend on milestone param.
        # It is hardcoded to "9.0".
        result = defaultdict(int)
        data = get_feature_testing_data(version=milestone)
        for case in data['cases']:
            status = case['last_result_status']
            if status:
                result[status] += 1
        return dict(result)


api.add_resource(FeatureTestingProgress, '/feature_testing_progress')

api.add_resource(
    FeatureTestingSummary,
    '/feature_testing_progress/summary/<string:milestone>'
)
