import re
from collections import defaultdict

from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad.consts import OPEN_STATUSES
from qareporting.integration.launchpad.models import BugTaskHistory
from qareporting.resources.base import BaseResource
from qareporting.resources.testrail import models
from qareporting.utils import dt_now

MAX_RUN_COUNT = 5

MIN_RUN_COUNT = 2


class FloatingIssues(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument('milestone', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

    def get(self):

        milestone_id = models.Milestone.objects.filter(
            name=self.milestone
        ).first().pk

        is_swarm = models.Run.objects.filter(
            milestone_id=milestone_id,
            name__contains='Swarm'
        )

        if is_swarm:
            runs = models.Run.objects[:5].filter(
                milestone_id=milestone_id,
                name__contains='Swarm',
            ).order_by('-created_on').only('id')
        else:
            runs = models.Run.objects[:5].filter(
                milestone_id=milestone_id
            ).order_by('-created_on').only('id')

        run_ids = [x.id for x in runs]

        tests_with_bugs = models.Test.objects.filter(
            run_id__in=run_ids,
            results__custom_launchpad_bug__exists=True,
        )

        bug_pattern = re.compile(r'(fuel|mos)\/(.*)\/'
                                 r'(?P<bug_id>\d\d\d\d\d\d\d)\/?$')
        runs = defaultdict(set)

        for test in tests_with_bugs:
            for res in test.results:
                if res.custom_launchpad_bug:
                    match_link = bug_pattern.search(res.custom_launchpad_bug)
                    if match_link:
                        runs[int(match_link.group('bug_id'))].add(test.run.id)
                        break

        bug_counter = {}

        for key, value in runs.items():
            count = len(value)
            if MIN_RUN_COUNT <= count < MAX_RUN_COUNT:
                bug_counter[key] = count

        bug_reporter_pipeline = [
            {'$sort': {'version_begin': 1}},
            {'$group': {'_id': '$bug_id',
                        'reporter': {
                            '$first': '$assignee_change_person'}
                        }
             }
        ]

        reporter_query = BugTaskHistory.objects.filter(
            bug_id__in=bug_counter.keys(),
            milestone=self.milestone,
        ).aggregate(*bug_reporter_pipeline)

        reporters = {bug['_id']: bug['reporter'] for bug
                     in reporter_query}

        bugs_info = BugTaskHistory.objects(
            version_end__gt=dt_now(),
            milestone=self.milestone,
            bug_id__in=bug_counter.keys(),
        ).only(
            'bug_id',
            'assignee',
            'title',
            'status',
            'importance',
            'version_begin',
        ).as_pymongo()

        result = []

        for bug in bugs_info:
            if bug['status'] in OPEN_STATUSES:
                res = {
                    'bug_id': bug['bug_id'],
                    'bug_assignee': bug.get('assignee', 'Unassigned'),
                    'bug_subject': bug['title'],
                    'bug_status': bug['status'],
                    'bug_importance': bug['importance'],
                    'bug_reporter': reporters[bug['bug_id']],
                    'bug_date': bug['version_begin'].date().isoformat(),
                    'bug_count': bug_counter[bug['bug_id']],
                }
                result.append(res)

        return result


api.add_resource(FloatingIssues, '/lp/floating_issues')
