import datetime
from collections import defaultdict

from bdateutil import relativedelta
from dateutil import rrule
from dateutil.relativedelta import relativedelta as rd
from dateutil.relativedelta import WE
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad.consts import (
    STATUS_INCOMPLETE, MajorQATeams)
from qareporting.integration.launchpad.models import (
    BugTaskHistory, LaunchpadUser)
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_now, get_top_week_day
from qareporting.utils import RUS, UA

FAKE_STATUS_OTHER = 'Other'
SLA = 3
NUMBERS_OF_WEEK = 24
HALF_YEAR_BACK = datetime.datetime.combine(
                    dt_now() - rd(weeks=+NUMBERS_OF_WEEK, weekday=WE(+1)),
                    datetime.time.min
                )


class LpIncompleteBugs(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

    def get(self):

        pipeline = [
            {'$sort': {'version_begin': 1}},
            {'$group': {
                '_id': '$bug_id',
                'statuses': {
                    '$push': {
                        'status': '$status',
                        'start': '$version_begin',
                        'end': '$version_end'
                    },
                },
            }
            },
            {'$match': {
                'statuses': {
                    '$elemMatch': {
                        'status': STATUS_INCOMPLETE
                    }
                }
            }
            }
        ]

        bugs = BugTaskHistory.objects.filter(
            version_begin__gt=HALF_YEAR_BACK,
            milestone=self.milestone,
        ).aggregate(*pipeline)

        bugs_map = {}
        current_bugs = defaultdict(list)
        rus_ua_holidays = RUS() + UA()

        def get_bdays(start, end):
            return relativedelta(
                end.isoformat(),
                start.isoformat(),
                holidays=rus_ua_holidays,
            ).bdays

        def set_bug_status(_id, start, end):
            if end.replace(tzinfo=None) > dt_now().replace(tzinfo=None):
                current_bugs[_id].append({
                    'date': start,
                    'bdays': get_bdays(start, dt_now()),
                })
            else:
                if _id in bugs_map:
                    bugs_map[_id].append({
                        'date': start,
                        'bdays': get_bdays(start, end),
                    })
                else:
                    bugs_map[_id] = [{
                        'date': start,
                        'bdays': get_bdays(start, end),
                    }]

        for bug in bugs:
            bug_id = bug['_id']
            statuses = bug['statuses']
            for index, status in enumerate(statuses):
                if status['status'] == STATUS_INCOMPLETE:
                    if index + 1 < len(statuses):
                        if statuses[index + 1]['status'] == STATUS_INCOMPLETE:
                            continue
                        set_bug_status(
                            bug_id,
                            status['start'],
                            statuses[index + 1]['start'],
                        )
                    else:
                        set_bug_status(
                            bug_id,
                            statuses[index]['start'],
                            statuses[index]['end'],
                        )

        dates = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=rrule.WE,
            dtstart=dt_now() - relativedelta(
                weeks=+NUMBERS_OF_WEEK, weekday=WE(+1)),
            until=dt_now() + relativedelta(
                weeks=+0, weekday=WE(+1)),
        )

        result = {
            date.date().isoformat(): {
                'within_sla': {'count': 0, 'bug_ids': []},
                'over_sla': {'count': 0, 'bug_ids': []},
            }
            for date in dates
            }

        for bug in bugs_map:
            bug_info = [(st['date'], st['bdays']) for st in bugs_map[bug]]
            for date, bdays in bug_info:
                top_week_day = get_top_week_day(date, dates)
                week = top_week_day.date().isoformat()
                sla_key = None
                if bdays <= SLA:
                    sla_key = 'within_sla'
                else:
                    sla_key = 'over_sla'
                cur_count = result[week][sla_key]['count']
                cur_bug_list = result[week][sla_key]['bug_ids']
                result[week][sla_key].update({
                    'count': cur_count + 1,
                    'bug_ids': cur_bug_list + [bug]
                })

        bug_reporter_pipeline = [
            {'$sort': {'version_begin': 1}},
            {'$group': {'_id': '$bug_id',
                        'reporter': {
                            '$first': '$assignee_change_person'}
                        }
             }
        ]

        reporter_query = BugTaskHistory.objects.filter(
            bug_id__in=current_bugs.keys(),
            milestone=self.milestone,
        ).aggregate(*bug_reporter_pipeline)

        reporters = {bug['_id']: bug['reporter'] for bug in reporter_query}

        cur_incomplete_bugs = BugTaskHistory.objects(
            version_end__gt=dt_now(),
            bug_id__in=current_bugs.keys(),
            milestone=self.milestone,
        ).only(
            'bug_id',
            'assignee',
            'title',
            'status',
            'importance',
            'status_change_person',
            'status_change_date',
            'version_begin',
        ).as_pymongo()

        cur_bugs = []

        user_query = LaunchpadUser.objects.all()

        user_map = {user.pk: user.major_team_ids for user in user_query}

        for bug in cur_incomplete_bugs:
            assignee = bug.get('assignee', 'Unassigned')
            reporter = reporters[bug['bug_id']]
            res = {
                'bug_id': bug['bug_id'],
                'bug_assignee': assignee,
                'bug_subject': bug['title'],
                'bug_status': bug['status'],
                'bug_importance': bug['importance'],
                'bug_changer': bug['status_change_person'],
                'bug_change_date': bug['status_change_date'],
                'bug_date': bug['version_begin'],
                'bdays': current_bugs[bug['bug_id']][0]['bdays'],
                'reporter': reporter,
                'assignee_team': MajorQATeams.CODES_IN_LP_REPORTS.get(
                    user_map.get(assignee, [MajorQATeams.TEAM_UNKNOWN])[:1][0]),
                'reporter_team': MajorQATeams.CODES_IN_LP_REPORTS.get(
                    user_map.get(reporter, [MajorQATeams.TEAM_UNKNOWN])[:1][0])
            }
            cur_bugs.append(res)

        result.update({'current_bugs': cur_bugs})

        return result


api.add_resource(LpIncompleteBugs, '/lp/incomplete_bugs')
