import datetime
from collections import defaultdict

import pytz
from dateutil import rrule
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.jira.models import BugHist
from qareporting.integration.models import Iteration
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso


def get_jira_teams():
    return list(BugHist.objects.all().distinct('delivery_team'))


class JiraBugTrends(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('date_from', type=dt_from_iso)
    parser.add_argument('date_to', type=dt_from_iso)
    parser.add_argument('milestone', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

        self.date_from = pytz.utc.localize(datetime.datetime.combine(
            args.date_from.date(), datetime.time.min))

        self.date_to = pytz.utc.localize(datetime.datetime.combine(
            args.date_to.date(), datetime.time.max))

    def _get_delivery_team_for_moment(self, moment, milestone):
        query = BugHist.objects.filter(
            version_start__lte=moment,
            version_end__gt=moment,
            fix_versions=milestone,
            status__in=[
                'Open',
            ],
        )

        pipeline = [
            {'$group': {
                '_id': {
                    'bug': '$bug_key',
                    'priority': '$priority',
                    'delivery_team': '$delivery_team',
                },
            }},
            {'$group': {
                '_id': {
                    'priority': '$_id.priority',
                    'delivery_team': '$_id.delivery_team',
                },
                'count': {'$sum': 1},
            }},
        ]

        team_list = get_jira_teams()
        results = {
            name: defaultdict(int) for name in team_list
            }

        for record in query.aggregate(*pipeline):
            key = record['_id']

            team_name = key.get('delivery_team')
            priority = key.get('priority')

            results[team_name][priority] += record['count']

        return results

    def get(self):
        result = []

        iterations_query = Iteration.objects.filter(
            milestone=self.milestone,
            date_begin__lt=self.date_to,
            date_end__gt=self.date_from,
        )

        for iteration in iterations_query:
            iteration_data = {
                'iteration': iteration.to_mongo(),
                'data': []
            }

            week_stops = rrule.rrule(
                freq=rrule.WEEKLY,
                dtstart=iteration.date_begin + datetime.timedelta(
                    days=7, hours=-2, seconds=-1),
                until=iteration.date_end,
            )

            for week_index, week_end_date in enumerate(week_stops, 1):
                # note: week_end_date can be arbitrary weekday

                if not self.date_from <= week_end_date < self.date_to:
                    continue

                iteration_data['data'].append({
                    'week': {
                        'week_index': week_index,
                        'week_end_date': week_end_date,
                    },
                    'delivery_team_data': self._get_delivery_team_for_moment(
                        week_end_date, self.milestone),
                })

            if iteration_data['data']:
                result.append(iteration_data)

        return result


api.add_resource(JiraBugTrends, '/jira/bug_trends')
