import re

from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.jira.models import Bug
from qareporting.integration.launchpad import consts
from qareporting.integration.testrail.models import (
    Run, Test, Status, Plan)
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso

OTHER_TYPE_ID = 1

FAILED_STATUSES = [
    Status.STATUS_FAILED,
    Status.STATUS_INFRA_FAILED,
    Status.STATUS_PROD_FAILED,
    Status.STATUS_TEST_FAILED,
    Status.STATUS_BLOCKED
]
RUN_TYPES = {
    OTHER_TYPE_ID: 'other',
    Run.RUN_TYPE_SWARM_ID: 'swarm_bvt',
    Run.RUN_TYPE_TEMPEST_ID: 'tempest',
    Run.RUN_TYPE_ACCEPTANCE_ID: 'acceptance',
}


def get_jira_failure_drivers(plan_id):
    runs = Run.objects.filter(
        plan_id=plan_id,
    ).scalar('id', 'name')

    run_ids = dict((run_id, name) for run_id, name in runs)

    bug_escalation_aggregate = [
        {'$sort': {'run_id': 1}},
        {'$unwind': '$results'},
        {'$group': {
            '_id': '$results.defects',
            'date': {'$min': "$results.created_on"}}},
    ]

    test_escalation_query = Test.objects.filter(
        status_id__in=FAILED_STATUSES,
        results__defects__exists=True,
    )

    bug_escalation = test_escalation_query.aggregate(*bug_escalation_aggregate)

    bug_escal_date_map = {
        bug['_id']: bug['date'] for bug in bug_escalation}

    test_aggregate_pipeline = [
        {'$unwind': '$results'},
        {'$sort': {'results.created_on': -1}},
        {'$group': {
            '_id': '$_id',
            'last_result': {'$first': "$results"},
            'run_id': {'$first': '$run_id'},
        }}
    ]

    test_query = Test.objects.filter(
        status_id__in=FAILED_STATUSES,
        results__defects__exists=True,
        run_id__in=tuple(run_ids)
    )

    raw_tests = test_query.aggregate(*test_aggregate_pipeline)

    bug_pattern = re.compile(r'PROD-(?P<bug_id>\d\d\d\d)$')

    bug_map = {}
    for test in raw_tests:
        link = test['last_result'].get('defects')
        if link:
            match_link = bug_pattern.search(link)
            if match_link:
                bug_id = 'PROD-' + match_link.group('bug_id')
                run_name = run_ids[test['run_id']]
                if bug_id in bug_map:
                    bug_map[bug_id]['rate'][run_name] += 1
                else:
                    run_counter = dict((name, 0) for name in run_ids.values())
                    run_counter[run_name] += 1
                    bug_map[bug_id] = dict(
                        bug_id=bug_id,
                        rate=run_counter,
                        bug_escalation=bug_escal_date_map[link]
                    )

    bug_statuses = Bug.objects(
        key__in=tuple(bug_map),
    ).only(
        'status', 'priority',
        'assignee', 'reporter',
        'major_team_ids', 'key',
        'summary', 'created_on',
    ).as_pymongo()

    for bug in bug_statuses:
        bug_id = bug['key']
        if bug_id in bug_map:
            team_id = next(iter(bug.get('major_team_ids', [])), None)
            bug_map[bug_id].update({
                'bug_status': bug['status'],
                'bug_priority': bug['priority'],
                'bug_assignee': bug.get('assignee', 'Unassigned'),
                'team': consts.MajorQATeams.NAMES.get(team_id, 'Unknown'),
                'summary': bug['summary'],
                'bug_created': bug['created_on']
            })

    return [bug for bug in bug_map.values()]


class FailureMetricsJira(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone_id', type=int)
    parser.add_argument('plan_id', type=int)

    def __init__(self):
        args = self.parser.parse_args()
        self.plan_id = args.plan_id

    def get(self):
        return get_jira_failure_drivers(self.plan_id)


class FailureNew(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('date_from', type=dt_from_iso)
    parser.add_argument('date_to', type=dt_from_iso)
    parser.add_argument(
        'milestone_ids',
        type=lambda v: map(int, v.split(',')))

    def __init__(self):
        args = self.parser.parse_args()
        self.date_from = args.date_from
        self.date_to = args.date_to
        self.milestone_ids = args.milestone_ids

    def get(self):

        plans = Plan.objects.filter(
            created_on__lte=self.date_to,
            created_on__gte=self.date_from,
            milestone_id__in=tuple(self.milestone_ids),
        ).scalar('id', 'name')

        plan_map = dict(plans)

        runs = Run.objects.filter(
            plan_id__in=tuple(plan_map)
        ).as_pymongo().scalar('id', 'plan_id')

        run_map = {}
        for run in runs:
            run_map[run['_id']] = run['plan_id']

        bug_escalation_aggregate = [
            {'$sort': {'run_id': 1}},
            {'$unwind': '$results'},
            {'$group': {
                '_id': '$results.defects',
                'date': {'$min': "$results.created_on"}}},
        ]

        test_escalation_query = Test.objects.filter(
            run_id__in=tuple(run_map),
            status_id__in=FAILED_STATUSES,
            results__defects__exists=True,
        )

        bug_escalation = test_escalation_query.aggregate(
            *bug_escalation_aggregate)

        bug_escal_date_map = {
            bug['_id']: bug['date'] for bug in bug_escalation}

        test_aggregate_pipeline = [
            {'$unwind': '$results'},
            {'$sort': {'results.created_on': -1}},
            {'$group': {
                '_id': '$_id',
                'last_result': {'$first': "$results"},
                'run_id': {'$first': '$run_id'},
            }}
        ]

        test_query = Test.objects.filter(
            status_id__in=FAILED_STATUSES,
            results__defects__exists=True,
            run_id__in=tuple(run_map)
        )

        raw_tests = test_query.aggregate(*test_aggregate_pipeline)
        bug_pattern = re.compile(r'PROD-(?P<bug_id>\d\d\d\d)$')
        bugs = {}

        for test in raw_tests:
            link = test['last_result'].get('defects')
            if link:
                match_link = bug_pattern.search(link)
                if match_link:
                    bug_type = {
                        'System': False,
                        'Component': False,
                        'Smoke': False,
                        'Smoke-Newton': False,
                        'Smoke-Mitaka': False,
                        'counter': 1
                    }

                    bug_id = 'PROD-' + match_link.group('bug_id')
                    plan_id = run_map[test['run_id']]
                    plan_name = plan_map[plan_id]
                    if 'System' in plan_name:
                        bug_type['System'] = True
                    elif 'Component' in plan_name:
                        bug_type['Component'] = True
                    elif 'Smoke-Newton' in plan_name:
                        bug_type['Smoke-Newton'] = True
                    elif 'Smoke-Mitaka' in plan_name:
                        bug_type['Smoke-Mitaka'] = True
                    elif 'Smoke' in plan_name:
                        bug_type['Smoke'] = True

                    if bug_id in bugs:
                        bugs[bug_id]['counter'] += 1
                    else:
                        bugs[bug_id] = bug_type

        bug_statuses = Bug.objects(
            key__in=tuple(bugs),
        ).only(
            'status', 'priority',
            'assignee', 'reporter',
            'major_team_ids', 'key',
            'summary', 'created_on',
            'delivery_team', 'components'
        ).as_pymongo()
        result = []
        for bug in bug_statuses:
            team_id = next(iter(bug.get('major_team_ids', [])), None)
            item = {
                'bug_id': bug['key'],
                'bug_status': bug['status'],
                'bug_priority': bug['priority'],
                'bug_assignee': bug.get('assignee', 'Unassigned'),
                'qa_team': consts.MajorQATeams.NAMES.get(team_id, 'Unknown'),
                'bug_created': bug['created_on'],
                'bug_title': bug['summary'],
                'delivery_team': bug.get('delivery_team'),
                'bug_component': bug.get('components'),
                'bug_escalation': bug_escal_date_map.get(bug['key']),
                'bug_types': bugs[bug['key']],
            }
            result.append(item)
        return result


api.add_resource(FailureMetricsJira, '/testrail/jira_failure_metrics')
api.add_resource(FailureNew, '/jira/jira_metrics')
