import datetime
from collections import defaultdict

from flask_restful import reqparse
from dateutil import rrule
from dateutil.relativedelta import relativedelta, WE

from qareporting.ext import api
from qareporting.resources.base import BaseResource
from qareporting.integration.jira.models import BugHist
from qareporting.utils import dt_now, get_top_week_day
from qareporting.resources.reopened_bugs_report import (
    REOPENED_PATTERN, NUMBERS_OF_WEEK
)


class JiraReopenedBugs(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

    def get_all_reopened(self):

        _status = {'$cond': {
            'if': {'$setIsSubset': (['$status'], ['Open', 'In Progress'])},
            'then': 'Open',
            'else': 'Close',
        }}

        historical_pipeline = [
            {'$sort': {'version_start': 1}},
            {'$group': {
                '_id': '$key',
                'status': {
                    '$push': _status
                },
                'version_begin': {
                    '$push': '$version_start'
                }}}
        ]

        bugs = BugHist.objects.filter(
            version_start__gt=datetime.datetime.combine(
                dt_now() - relativedelta(
                    weeks=+NUMBERS_OF_WEEK, weekday=WE(+1)),
                datetime.time.min),
            fix_versions=self.milestone
        ).aggregate(*historical_pipeline)

        reopened_bug_ids = set()
        reopened_bugs_by_date = defaultdict(list)

        for bug in bugs:
            statuses = bug['status']
            if REOPENED_PATTERN in zip(statuses, statuses[1:]):
                status_indexes = [
                    index + 1 for index, status in enumerate(statuses)
                    if list(REOPENED_PATTERN) == statuses[index:index + 2]
                    ]
                bug_dates = [bug['version_begin'][i] for i in status_indexes]
                for date in bug_dates:
                    reopened_bugs_by_date[date].append(bug['_id'])
                    reopened_bug_ids.add(bug['_id'])
            else:
                continue

        return reopened_bugs_by_date, reopened_bug_ids

    def get(self):

        dates = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=rrule.WE,
            dtstart=dt_now() - relativedelta(
                weeks=+NUMBERS_OF_WEEK, weekday=WE(+1)),
            until=dt_now() + relativedelta(
                weeks=+0, weekday=WE(+1)),
        )

        result = {
            date.date().isoformat(): {'count': 0, 'bug_ids': []}
            for date in dates
            }

        reopened_bugs, bug_keys = self.get_all_reopened()

        for date in reopened_bugs:
            top_week_day = get_top_week_day(date, dates)
            week = top_week_day.date().isoformat()
            cur_count = result[week]['count']
            cur_bug_list = result[week]['bug_ids']
            result[week].update({
                'count': cur_count + 1,
                'bug_ids': cur_bug_list + reopened_bugs[date],
            })

        cur_bugs = BugHist.objects.filter(
            version_end__gt=dt_now(),
            key__in=tuple(bug_keys),
            fix_versions=self.milestone,
            status__in=['Open', 'In Progress']
        ).as_pymongo()

        current_bugs = []

        for bug in cur_bugs:
            current_bugs.append({
                'bug_id': bug['key'],
                'bug_assignee': bug.get('assignee', 'Unassigned'),
                'bug_subject': bug['summary'],
                'bug_status': bug['status'],
                'bug_priority': bug['priority'],
                'bug_date': bug['version_start'].date().isoformat()
            })

        result.update({'current_bugs': current_bugs})

        return result

api.add_resource(JiraReopenedBugs, '/jira/reopened_bugs')
