import re
from collections import defaultdict
from itertools import chain

from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad.models import BugTaskHistory
from qareporting.resources.base import BaseResource
from qareporting.resources.testrail import models
from qareporting.utils import dt_now
from qareporting.integration.launchpad.consts import OPEN_STATUSES


FAILED_STATUSES = [
    models.Status.STATUS_BLOCKED,
    models.Status.STATUS_FAILED,
    models.Status.STATUS_PROD_FAILED,
    models.Status.STATUS_TEST_FAILED,
    models.Status.STATUS_INFRA_FAILED
]


class LpRegressionBugs(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone', type=str)

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

    def get(self):

        milestone_id = models.Milestone.objects.filter(
            name=self.milestone
        ).first().pk

        release_plan = models.Plan.objects.filter(
            name__contains='RC'
        ).order_by('-created_on').first()

        released_runs = models.Run.objects.filter(
            plan_id=release_plan.id
        )

        released_groups = models.Test.objects.filter(
            run_id__in=[run.pk for run in released_runs],
            status_id=models.Status.STATUS_PASSED,
        ).only('custom_test_group').as_pymongo()

        rel_group_list = set()

        for group in released_groups:
            if group.get('custom_test_group'):
                rel_group_list.add(group['custom_test_group'])

        bugs = defaultdict(int)
        bug_cases = {}

        bug_pattern = re.compile(r'(fuel|mos)\/(.*)\/(?P<bug_id>\d+)\/?')

        for i in range(1, 4):
            cur_swarm_runs = models.Run.objects[i:].filter(
                name__contains='Swarm',
                milestone_id=milestone_id
            ).order_by('-created_on').first()

            if not cur_swarm_runs:
                return []

            current_runs = models.Run.objects.filter(
                plan_id=cur_swarm_runs.plan.pk
            )

            pipeline = [
                {'$unwind': '$results'},
                {'$group': {
                    '_id': {
                        'url': '$results.custom_launchpad_bug',
                        'case': '$case_id',
                        'run': '$run_id',
                    },
                }
                }
            ]

            cur_tests = models.Test.objects.filter(
                run_id__in=[run.pk for run in current_runs],
                custom_test_group__in=rel_group_list,
                results__custom_launchpad_bug__exists=True,
                status_id__in=FAILED_STATUSES
            ).aggregate(*pipeline)

            for test in cur_tests:
                if isinstance(test.get('_id', {}).get('url'), str):
                    match_link = bug_pattern.search(test['_id']['url'])
                    if match_link:
                        bugs[int(match_link.group('bug_id'))] += 1
                        bug_cases[int(match_link.group('bug_id'))] = test.get(
                            '_id', {}).get('case')

        bug_reporter_pipeline = [
            {'$sort': {'version_begin': 1}},
            {'$group': {'_id': '$bug_id',
                        'reporter': {
                            '$first': '$assignee_change_person'}
                        }
             }
        ]

        reporter_query = BugTaskHistory.objects.filter(
            bug_id__in=bugs.keys(),
            milestone=self.milestone,
        ).aggregate(*bug_reporter_pipeline)

        reporter_query_tag_regression = BugTaskHistory.objects.filter(
            tags='regression',
            milestone=self.milestone,
        ).aggregate(*bug_reporter_pipeline)

        reporters = {bug['_id']: bug['reporter'] for bug
                     in reporter_query}

        reporters_tag = {bug['_id']: bug['reporter']
                         for bug in reporter_query_tag_regression}

        bugs.update({bug: 3 for bug in reporters_tag.keys()})

        reporters.update(reporters_tag)

        cases = models.Case.objects.filter(
            id__in=bug_cases.values()
        ).as_pymongo()

        cases_names = {case['_id']: case['title'] for case in cases}

        cur_regression_bugs = BugTaskHistory.objects(
            version_end__gt=dt_now(),
            milestone=self.milestone,
            bug_id__in=bugs.keys(),
        ).only(
            'bug_id',
            'assignee',
            'title',
            'status',
            'importance',
            'status_change_person',
            'version_begin',
        ).as_pymongo()

        cur_regression_bugs_with_tag = BugTaskHistory.objects(
            version_end__gt=dt_now(),
            milestone=self.milestone,
            tags='regression',
        ).only(
            'bug_id',
            'assignee',
            'title',
            'status',
            'importance',
            'status_change_person',
            'version_begin',
        ).as_pymongo()

        result = {}

        for bug in chain(cur_regression_bugs, cur_regression_bugs_with_tag):
            regression = None
            if bugs[bug['bug_id']] == 3:
                regression = True
            else:
                regression = False

            if bug['status'] in OPEN_STATUSES:
                res = {
                    'regression': regression,
                    'bug_id': bug['bug_id'],
                    'bug_assignee': bug.get('assignee', 'Unassigned'),
                    'bug_subject': bug['title'],
                    'bug_status': bug['status'],
                    'bug_importance': bug['importance'],
                    'bug_reporter': reporters[bug['bug_id']],
                    'bug_date': bug['version_begin'].date().isoformat(),
                    'last_case_id': bug_cases.get(bug['bug_id']),
                    'lase_case_title': cases_names.get(
                        bug_cases.get(bug['bug_id'])),
                }
                result[bug['bug_id']] = res

        return [bug for bug in result.values()]


api.add_resource(LpRegressionBugs, '/lp/regression_bugs')
