import datetime
from collections import defaultdict

from dateutil import rrule
from dateutil.relativedelta import relativedelta, WE
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.launchpad import consts
from qareporting.integration.launchpad.models import BugTaskHistory
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_now, get_top_week_day

FAKE_STATUS_OPEN = 'Open'
FAKE_STATUS_CLOSE = 'Close'
REOPENED_PATTERN = ('Close', 'Open')
NUMBERS_OF_WEEK = 24


class LpReopenedBugs(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone', default='')

    def __init__(self):
        args = self.parser.parse_args()
        self.milestone = args.milestone

    def get(self):

        _status = {'$cond': {
            'if': {'$setIsSubset': (['$status'], consts.OPEN_STATUSES + [
                consts.STATUS_NEW
            ])},
            'then': FAKE_STATUS_OPEN,
            'else': FAKE_STATUS_CLOSE,
        }}

        historical_pipeline = [
            {'$sort': {'version_begin': 1}},
            {'$group': {
                '_id': '$bug_id',
                'status': {
                    '$push': _status
                },
                'version_begin': {
                    '$push': '$version_begin'
                }}}
        ]

        bugs = BugTaskHistory.objects.filter(
            version_begin__gt=datetime.datetime.combine(
                dt_now() - relativedelta(
                    weeks=+NUMBERS_OF_WEEK, weekday=WE(+1)),
                datetime.time.min),
            status__in=consts.OPEN_STATUSES + [
                consts.STATUS_NEW,
                consts.STATUS_FIX_COMMITTED,
                consts.STATUS_FIX_RELEASED,
            ],
            milestone=self.milestone
        ).aggregate(*historical_pipeline)

        reopened_bug_ids = set()
        reopened_bugs_by_date = defaultdict(list)

        for bug in bugs:
            statuses = bug['status']
            if REOPENED_PATTERN in zip(statuses, statuses[1:]):
                status_indexes = [
                    index + 1 for index, status in enumerate(statuses)
                    if list(REOPENED_PATTERN) == statuses[index:index + 2]
                    ]
                bug_dates = [bug['version_begin'][i] for i in status_indexes]
                for date in bug_dates:
                    reopened_bugs_by_date[date].append(bug['_id'])
                    reopened_bug_ids.add(bug['_id'])
            else:
                continue

        cur_reopened_bugs = BugTaskHistory.objects(
            version_end__gt=dt_now(),
            bug_id__in=tuple(reopened_bug_ids),
            milestone=self.milestone,
        ).filter(
            status__in=consts.OPEN_STATUSES + [
                consts.STATUS_NEW,
                consts.STATUS_OPINION,
                consts.STATUS_EXPIRED
                ]
        ).only(
            'bug_id',
            'assignee',
            'title',
            'status',
            'importance',
            'status_change_person',
            'version_begin',
        ).as_pymongo()

        dates = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=rrule.WE,
            dtstart=dt_now() - relativedelta(
                weeks=+NUMBERS_OF_WEEK, weekday=WE(+1)),
            until=dt_now() + relativedelta(
                weeks=+0, weekday=WE(+1)),
        )

        result = {
            date.date().isoformat(): {'count': 0, 'bug_ids': []}
            for date in dates
        }

        for date in reopened_bugs_by_date:
            top_week_day = get_top_week_day(date, dates)
            week = top_week_day.date().isoformat()
            cur_count = result[week]['count']
            cur_bug_list = result[week]['bug_ids']
            result[week].update({
                'count': cur_count + 1,
                'bug_ids': cur_bug_list + reopened_bugs_by_date[date],
            })

        cur_bugs = []

        for bug in cur_reopened_bugs:
            res = {
                'bug_id': bug['bug_id'],
                'bug_assignee': bug.get('assignee', 'Unassigned'),
                'bug_subject': bug['title'],
                'bug_status': bug['status'],
                'bug_importance': bug['importance'],
                'bug_changer': bug['status_change_person'],
                'bug_date': bug['version_begin'].date().isoformat()
            }
            cur_bugs.append(res)

        result.update({'current_bugs': cur_bugs})

        return result


api.add_resource(LpReopenedBugs, '/lp/reopened_bugs')
