from flask_restful import fields
from flask_restful import marshal_with
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.ext import cache
from qareporting.integration.testrail import models
from qareporting.resources.base import BaseResource
from qareporting.resources.base import MappingField
from qareporting.utils import dt_from_iso


class RunStatuses(BaseResource):
    arg_parser = reqparse.RequestParser()
    arg_parser.add_argument('date_from', type=dt_from_iso)
    arg_parser.add_argument('date_to', type=dt_from_iso)

    def __init__(self):
        args = self.arg_parser.parse_args()
        self.date_from = args.date_from
        self.date_to = args.date_to

    run_fields = {
        'id': fields.Integer(attribute='_id'),
        'name': fields.String,
        'config': fields.String,
        'created_on': fields.DateTime(dt_format='iso8601'),
        'project': fields.Nested({
            'id': fields.Integer(attribute='_id'),
            'name': fields.String,
        }),
        'milestone': fields.Nested({
            'id': fields.Integer(attribute='_id'),
            'name': fields.String,
        }),
        'plan': fields.Nested({
            'id': fields.Integer(attribute='_id'),
            'name': fields.String,
            'created_on': fields.DateTime(dt_format='iso8601'),
        },),
        'qa_team': fields.Integer,
        'run_type_name': MappingField(
            dict(models.Run.RUN_TYPES), attribute='run_type'),
        'status_counts': fields.Raw,
        'casetype_counts': fields.Raw,
    }

    @marshal_with(run_fields)
    def get(self):
        agg_query = models.Test.objects(casetype__ne=None).aggregate(*[
            {'$group': {
                '_id': {
                    'run_id': '$run_id',
                    'casetype_id': '$casetype_id',
                },
                'count': {'$sum': 1},
            }},
            {'$group': {
                '_id': '$_id.run_id',
                'casetype_counts': {'$push': {
                    'casetype_id': '$_id.casetype_id',
                    'count': '$count',
                }},
            }}
        ])

        run_casetype_counts_map = {
            run['_id']: {
                item['casetype_id']: item['count']
                for item in run['casetype_counts']
            }
            for run in agg_query
        }

        def make_map(q):
            return {x['_id']: x for x in q.as_pymongo()}

        milestone_map = make_map(models.Milestone.objects)
        project_map = make_map(models.Project.objects)
        plan_map = make_map(models.Plan.objects)

        if self.date_to and self.date_from:
            data = tuple(models.Run.objects.filter(
                created_on__gte=self.date_from,
                created_on__lte=self.date_to
            ).as_pymongo())
        else:
            data = tuple(models.Run.objects.as_pymongo())

        for run in data:
            run['casetype_counts'] = run_casetype_counts_map.get(run['_id'])
            run['project'] = project_map.get(run['project_id'])
            run['milestone'] = milestone_map.get(run.get('milestone_id'))
            run['plan'] = plan_map.get(run.get('plan_id'))

        return data


api.add_resource(RunStatuses, '/testrail/run_statuses')
