from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.testrail.models import Case
from qareporting.integration.testrail.models import CaseHistory
from qareporting.integration.testrail.models import CaseType
from qareporting.integration.testrail.models import Milestone
from qareporting.integration.testrail.models import Section
from qareporting.integration.testrail.models import Suite
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso


def get_data_with_suite(data, suites_map, milestones_map, sections_map):
    for item in data:
        if item['suite_id'] and suites_map.get(item['suite_id']):
            item['suite_name'] = suites_map.get(item['suite_id'])
            item['milestone_name'] = milestones_map.get(
                item['milestone_id'], 'None')
            item['section_name'] = sections_map.get(item['section_id'])
            yield item


def get_day_adoption_counts(moment, case_ids):

    aggregation_pipeline = [
        {'$group': {
            '_id': {
                'milestone_id': '$milestone_id',
                'suite_id': '$suite_id',
                'section_id': '$section_id',
                'case_type': '$casetype_id',
            },
            'count': {'$sum': 1},
        }},
        {'$group': {
            '_id': {
                'milestone_id': '$_id.milestone_id',
                'suite_id': '$_id.suite_id',
                'section_id': '$_id.section_id',
            },
            'case_types': {
                '$push': {
                    'casetype_id': '$_id.case_type',
                    'count': '$count',
                },
            },
        }},
    ]

    query = CaseHistory.objects.filter(
        version_begin__lte=moment,
        version_end__gt=moment,
        case_id__in=case_ids,
    ).aggregate(*aggregation_pipeline)

    for case in query:
        _id = case['_id']

        types_dict = {x.get('casetype_id'): x['count']
                      for x in case['case_types']}
        item = {
            'date': moment,
            'section_id': _id.get('section_id'),
            'milestone_id': _id.get('milestone_id'),
            'suite_id': _id.get('suite_id'),
            'type_counts': {
                'manual': types_dict.get(CaseType.MANUAL_ID, 0),
                'in_progress': types_dict.get(
                    CaseType.AUTOMATION_IN_PROGRESS_ID, 0),
                'automated': types_dict.get(CaseType.AUTOMATED_ID, 0),
                'not_automate': types_dict.get(
                    CaseType.NOT_AUTOMATE_ID, 0),
                'ci_integration': types_dict.get(
                    CaseType.CI_INTEGRATED, 0)
            }
        }
        yield item


class SuiteProgress(BaseResource):

    parser = reqparse.RequestParser()
    parser.add_argument('started_from', type=dt_from_iso)
    parser.add_argument('project_id', type=int)
    parser.add_argument(
        'dates', type=lambda v: sorted(set(map(dt_from_iso, v.split(','))))
    )

    def __init__(self):
        self.args = self.parser.parse_args()

    def get_project_data(self, case_ids):
        for date in self.args.dates:
            for item in get_day_adoption_counts(date, case_ids):
                yield item

    def get(self):
        project_id = self.args.project_id

        project_case_ids = Case.objects(
            project_id=project_id
        ).only('id').distinct('id')

        suites_map = dict(Suite.objects(
            project_id=project_id
        ).scalar('id', 'name'))

        milestones_map = dict(Milestone.objects(
            project_id=project_id
        ).scalar('id', 'name'))

        sections_map = dict(Section.objects(
            suite_id__in=tuple(suites_map)
        ).scalar('id', 'name'))

        project_data = list(get_data_with_suite(
            self.get_project_data(project_case_ids),
            suites_map,
            milestones_map,
            sections_map)
        )

        return project_data


api.add_resource(SuiteProgress, '/testrail/suite_progress')
