import datetime

from dateutil import rrule
from flask_restful import reqparse

from qareporting.ext import api
from qareporting.integration.jira.models import BugHist
from qareporting.resources.base import BaseResource
from qareporting.utils import dt_from_iso

OPEN_STATUSES = [
    'Open',
    'In Progress',
    'In Development'
]

PRIORITY_BLOCKER = 1
PRIORITY_CRITICAL = 2
PRIORITY_MAJOR = 3
PRIORITY_NICE_TO_HAVE = 4
PRIORITY_SOME_DAY = 5
FAKE_PRIORITY = 99

PRIORITY_NAMES = {
    PRIORITY_BLOCKER: 'Blocker',
    PRIORITY_CRITICAL: 'Critical',
    PRIORITY_MAJOR: 'Major',
    PRIORITY_NICE_TO_HAVE: 'Nice to have',
    PRIORITY_SOME_DAY: 'Some day',
    FAKE_PRIORITY: 'Other',
}

TARGET_PRIORITY = [
    PRIORITY_BLOCKER,
    PRIORITY_CRITICAL,
    PRIORITY_MAJOR
]

PRIORITY = [
    'Blocker',
    'Critical',
    'Major',
]

parser = reqparse.RequestParser()
parser.add_argument('date_from', type=dt_from_iso)
parser.add_argument('date_to', type=dt_from_iso)
parser.add_argument('milestones', type=lambda x: x)

_status = {'$cond': {
    'if': {'$setIsSubset': (['$status'], OPEN_STATUSES)},
    'then': 'Open',
    'else': '$status',
}}
_priority = {'$cond': {
    'if': {'$setIsSubset': (['$priority'], PRIORITY)},
    'then': '$priority',
    'else': 'Other',
}}


def get_bug_records(pipeline, moment, milestones):
    query = BugHist.objects.filter(
        version_start__lte=moment,
        version_end__gt=moment,
        fix_versions__in=milestones,
    ).aggregate(*pipeline)

    for record in query:
        yield record


def get_bug_records_by_program(pipeline, moment, program):
    query = BugHist.objects.filter(
        version_start__lte=moment,
        version_end__gt=moment,
        program__in=program,
    ).aggregate(*pipeline)

    for record in query:
        yield record


class TcpMcpVerificationProgress(BaseResource):

    def __init__(self):
        args = parser.parse_args()
        self.date_from = args.date_from.date()
        self.date_to = args.date_to.date()
        self.milestones = args.milestones

        assert self.date_from <= self.date_to
        assert (self.date_to - self.date_from)

        self.date_set = TcpMcpVerificationProgress.get_date_set(
            self.date_from, self.date_to)

    @classmethod
    def get_date_set(cls, date_from, date_to):

        date_set = rrule.rruleset()
        dates = rrule.rrule(
            freq=rrule.WEEKLY,
            byweekday=rrule.WE,
            byhour=23,
            byminute=59,
            bysecond=59,
            dtstart=date_from,
            until=date_to,
        )

        date_set.rrule(dates)
        date_set.rdate(datetime.datetime.combine(
            date_to, datetime.time.max))

        return date_set

    def filter_bug_records(self, records, moment):
        for record in records:
            _id = record['_id']

            item = {
                'date': moment.date().isoformat(),
                'status': _id['status'],
                'priority': _id['priority'],
                'count': record['count'],
            }

            yield item

    def get_result(self, pipeline):

        result = []
        milestones = self.milestones.split(',')

        for date in self.date_set:

            data = list(self.filter_bug_records(
                get_bug_records(pipeline, date, milestones),
                date
            ))

            result.extend(data)

        return result

    def get(self):

        pipeline = [
            {'$group': {
                '_id': {
                    'status': _status,
                    'priority': _priority
                },
                'count': {'$sum': 1},
            }}
        ]

        return self.get_result(pipeline)


class TcpMcpVerificationProgressTeam(TcpMcpVerificationProgress):
    def filter_bug_records(self, records, moment):
        for record in records:
            _id = record['_id']

            item = {
                'date': moment.date().isoformat(),
                'status': _id['status'],
                'priority': _id['priority'],
                'team': _id.get('delivery_team', 'unknown'),
                'count': record['count'],
            }

            yield item

    def get(self):

        pipeline = [
            {'$group': {
                '_id': {
                    'status': _status,
                    'priority': _priority,
                    'delivery_team': '$delivery_team'
                },
                'count': {'$sum': 1},
            }}
        ]

        return self.get_result(pipeline)


class JiraVerificationProgress(TcpMcpVerificationProgress):

    def get_result(self, pipeline):

        result = []
        milestones = self.milestones.split(',')
        if 'none' in milestones:
            milestones = [None if x == 'none' else x for x in milestones]

        for date in self.date_set:

            data = list(self.filter_bug_records(
                get_bug_records_by_program(pipeline, date, milestones),
                date
            ))

            result.extend(data)
        return result

    def filter_bug_records(self, records, moment):
        for record in records:
            _id = record['_id']

            item = {
                'date': moment.date().isoformat(),
                'status': _id['status'],
                'priority': _id['priority'],
                'team': _id.get('delivery_team', 'unknown'),
                'program': _id.get('program', 'none'),
                'compatible_version': _id.get('compatible_version', 'none'),
                'component': _id.get('component', 'none'),
                'count': record['count'],
            }

            yield item

    def get(self):
        pipeline = [
            {'$unwind': {'path': '$compatible_versions',
                         'preserveNullAndEmptyArrays': True}},
            {'$unwind': {'path': '$components',
                         'preserveNullAndEmptyArrays': True}},
            {'$group': {
                '_id': {
                    'status': _status,
                    'priority': _priority,
                    'delivery_team': '$delivery_team',
                    'program': '$program',
                    'compatible_version': '$compatible_versions',
                    'component': '$components'
                },
                'count': {'$sum': 1},
            }}
        ]
        return self.get_result(pipeline)


api.add_resource(
    TcpMcpVerificationProgress,
    '/jira/tcp_mcp_verification_progress'
)

api.add_resource(
    TcpMcpVerificationProgressTeam,
    '/jira/tcp_mcp_verification_progress_team'
)

api.add_resource(
    JiraVerificationProgress,
    '/jira/jira_verification_progress'
)
