from flask_restful import reqparse

from qareporting.ext import api
from qareporting.ext import cache
from qareporting.integration.testrail.models import Run, Milestone, Plan
from qareporting.resources.base import BaseResource
from qareporting.resources.bvt_swarm_report import fake_milestone
from qareporting.utils import dt_from_iso

tempest_metrics_parser = reqparse.RequestParser()
tempest_metrics_parser.add_argument('date_from', type=dt_from_iso)
tempest_metrics_parser.add_argument('date_to', type=dt_from_iso)
tempest_metrics_parser.add_argument(
    'milestone_id', type=fake_milestone
)
tempest_metrics_parser.add_argument('milestone')


class TempestMetrics(BaseResource):
    def __init__(self):
        args = tempest_metrics_parser.parse_args()
        self.date_from = args.date_from
        self.date_to = args.date_to
        self.is_fake_milestone, self.milestone_id = args.milestone_id

    @cache.memoize()
    def _get_tempest_metrics(self, date_from, date_to, milestone_id, is_fake):
        if not is_fake:
            plans = Plan.objects.filter(
                created_on__gte=date_from,
                created_on__lte=date_to,
                milestone=milestone_id,
                fake_milestone__exists=False,
            )
        else:
            plans = Plan.objects.filter(
                created_on__gte=date_from,
                created_on__lte=date_to,
                fake_milestone=milestone_id,
            )

        plan_data = {}

        for plan in plans:
            plan_date = plan.created_on.date().isoformat()
            plan_data[plan.pk] = {
                'date': plan_date,
                'name': plan.name,
                Run.RUN_TYPE_TEMPEST_ID: [],
            }

        tempest_runs = Run.objects(
            plan_id__in=tuple(plan_data),
            run_type=Run.RUN_TYPE_TEMPEST_ID,
        )

        result = []

        for run in tempest_runs:
            config_name = run.name
            status = run.status_counts.to_mongo().to_dict()
            date = plan_data[run.plan.pk]['date']
            plan_name = plan_data[run.plan.pk]['name']

            res = {
                'date': date,
                'config_name': config_name,
                'status_counts': status,
                'plan_name': plan_name
            }
            result.append(res)

        return result

    def get(self):
        return self._get_tempest_metrics(
            self.date_from,
            self.date_to,
            self.milestone_id,
            self.is_fake_milestone,
        )


class TempestSummary(BaseResource):
    def __init__(self):
        args = tempest_metrics_parser.parse_args()
        self.milestone = args.milestone

    def get(self):
        milestone_ids = Milestone.objects(name=self.milestone).scalar('id')
        recent_swarm = Run.objects(
            name__contains='Swarm',
            milestone_id__in=milestone_ids,
        ).order_by('-created_on').first()

        recent_tempest = Run.objects(
            run_type=Run.RUN_TYPE_TEMPEST_ID,
            plan_id=recent_swarm.plan.pk,
        )

        if recent_tempest:
            return [dict(date=recent_swarm.plan.created_on.date().isoformat(),
                         config_name=run.name,
                         status_counts=run.status_counts.to_mongo().to_dict(),
                         plan_name=recent_swarm.plan.name)
                    for run in recent_tempest]
        else:
            return []


class TempestRecentSeven(BaseResource):
    def __init__(self):
        args = tempest_metrics_parser.parse_args()
        self.is_fake_milestone, self.milestone_id = args.milestone_id

    def get(self):

        if self.is_fake_milestone:
            plans = Plan.objects(
                fake_milestone=self.milestone_id,
            ).distinct('id')
        else:
            plans = Plan.objects(
                milestone_id=self.milestone_id,
                fake_milestone__exists=True,
            ).distinct('id')

        pipeline = [
            {'$group': {
                '_id': '$plan_id',
                'runs': {
                    '$push': {
                        'name': '$name',
                        'status_counts': '$status_counts',
                        'plan_id': '$plan_id',
                    }
                }
            }},
            {'$sort': {'_id': -1}},
            {'$limit': 7}
        ]

        if self.is_fake_milestone:
            last_seven_days = Run.objects(
                run_type=Run.RUN_TYPE_TEMPEST_ID,
                plan_id__in=tuple(plans)
            ).aggregate(*pipeline)
        else:
            last_seven_days = Run.objects(
                milestone_id=self.milestone_id,
                run_type=Run.RUN_TYPE_TEMPEST_ID,
                plan_id__not__in=tuple(plans)
            ).aggregate(*pipeline)

        result = []

        for run in last_seven_days:
            day_runs = run['runs']
            result.extend([{
                               'config_name': r['name'],
                               'plan_name': r['plan_id'],
                               'status_counts': r['status_counts'],
                           } for r in day_runs])

        plan_ids = [x['plan_name'] for x in result]

        plans = Plan.objects(
            id__in=plan_ids,
        ).only('name', 'created_on')

        plan_names = {plan.pk: {'name': plan.name, 'date': plan.created_on}
                      for plan in plans}

        for run in result:
            run.update({
                'plan_name': plan_names[run['plan_name']]['name'],
                'date': plan_names[run['plan_name']]['date'],
            })

        return result


# This resource used not only in Tempest report, but on Dashboard widget too.
api.add_resource(TempestMetrics, '/testrail/tempest_metrics')
api.add_resource(TempestSummary, '/testrail/tempest_summary')
api.add_resource(TempestRecentSeven, '/testrail/tempest_recent_seven')
