from flask_restful import reqparse
from natsort import natsorted

from qareporting.ext import api
from qareporting.ext import cache
from qareporting.integration.models import EventLog
from qareporting.integration.testrail import consts
from qareporting.integration.testrail import models
from qareporting.integration.testrail.models import TrMilestoneInFocus
from qareporting.resources.base import BaseResource
from qareporting.resources.bvt_swarm_report import fake_milestone


class UserList(BaseResource):
    @cache.memoize()
    def get(self):
        return list(models.User.objects)


class UserResource(BaseResource):
    @cache.memoize()
    def get(self, user_id):
        qs = models.User.objects
        return self.get_or_abort(user_id, qs)


class CaseTypeList(BaseResource):
    @cache.memoize()
    def get(self):
        return list(models.CaseType.objects)


class PriorityList(BaseResource):
    @cache.memoize()
    def get(self):
        return list(models.Priority.objects)


class MilestoneList(BaseResource):
    @cache.memoize()
    def get(self):
        real_list = models.Milestone.objects
        fake_milestones = models.Plan.objects(
            fake_milestone__exists=True
        ).only('fake_milestone').distinct('fake_milestone')
        fake_list = []
        for fmile in fake_milestones:
            real_milestone, mu_postfix = fmile.split('_')
            real_name = next(
                m.name for m in real_list if m.pk == int(real_milestone)
            )
            if real_name:
                fake_name = '{} MU{}'.format(real_name, mu_postfix)
                fake_list.append({
                    'id': fmile,
                    'name': fake_name,
                    'project': None,
                    'url': None,
                    "is_completed": False,
                })

        return natsorted(
            list(real_list) + fake_list,
            reverse=True,
            key=lambda m: (m['name'][0].isdigit(), m['name'])
        )


class QATeamsList(BaseResource):

    def get(self):
        return [
            {'id': i, 'name': name}
            for i, name in consts.QATeams.NAMES.items()
        ]


class CaseList(BaseResource):

    def get(self):
        target_fields = [
            'id', 'title', 'custom_qa_team',
            'milestone_id', 'priority_id', 'custom_case_complexity',
            'suite_id', 'section_id', 'created_on',
            'updated_on', 'estimate', 'estimate_forecast',
            'custom_test_group',
        ]

        cases_query = models.Case.objects.as_pymongo().scalar(*target_fields)

        milestone_query = models.Milestone.objects(
        ).scalar('id', 'name')

        priority_query = models.Priority.objects(
        ).scalar('id', 'name')

        suite_query = models.Suite.objects(
        ).scalar('id', 'name')

        section_query = models.Section.objects(
        ).scalar('id', 'name')

        milestone_map = dict(milestone_query)
        priority_map = dict(priority_query)
        suite_map = dict(suite_query)
        section_map = dict(section_query)
        case_complexity = dict(models.Case.CASE_COMPLEXITIES)

        result = []
        append = result.append

        for case in cases_query:
            item = {
                'id': case['_id'],
                'title': case['title'],
                'qa_team': case.get('custom_qa_team'),
                'milestone_name': milestone_map.get(
                    case.get('milestone_id'), 'None'
                ),
                'priority_name': priority_map.get(case.get('priority_id')),
                'complexity': case_complexity.get(case.get(
                    'custom_case_complexity'
                )),
                'suite': {
                    'id': case['suite_id'],
                    'name': suite_map.get(case['suite_id'])
                },
                'section': {
                    'name': section_map.get(case['section_id'])
                },
                'created_on': case['created_on'],
                'updated_on': case['updated_on'],
                'estimate': case.get('estimate'),
                'estimate_forecast': case.get('estimate_forecast'),
                'test_group': case.get('custom_test_group')
            }
            append(item)

        return result


class LastSync(BaseResource):

    def get(self):
        return EventLog.get_last(EventLog.EVENT_TESTRAIL_SYNC)


class MilestoneNameList(BaseResource):

    def get(self):
        l = list(models.Milestone.objects.all().scalar('name'))
        return natsorted(
            list(set(l)), reverse=True, key=lambda m: ('Plugin' not in m, m))


class TrMilestoneInFocusResource(BaseResource):
    def get(self):
        mileston = TrMilestoneInFocus.objects.all().first()
        return {'name': mileston.name}


class RunTypes(BaseResource):
    def get(self):
        return [{'id': i, 'name': name} for i, name in models.Run.RUN_TYPES]


class RunNamesList(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone_id', type=fake_milestone)
    parser.add_argument('qa_team', type=int)
    parser.add_argument('run_types', type=int, action='append')

    def __init__(self):
        args = self.parser.parse_args()
        self.is_fake_milestone, self.milestone_id = args.milestone_id
        self.qa_team = args.qa_team
        self.run_types = args.run_types

    def get(self):
        run_types = self.run_types

        if models.Run.RUN_TYPE_UNKNOWN_ID in run_types:
            run_types.append(None)

        if self.is_fake_milestone:
            plan_ids = models.Plan.objects(
                fake_milestone=self.milestone_id
            ).distinct('id')
            return list(models.Run.objects(
                plan_id__in=plan_ids,
                qa_team=self.qa_team,
                run_type__in=run_types,
            ).only('name').distinct('name'))

        return list(models.Run.objects(
            milestone_id=self.milestone_id,
            fake_milestone__exists=False,
            qa_team=self.qa_team,
            run_type__in=run_types,
        ).only('name').distinct('name'))


class PlanNameList(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('milestone_id', type=fake_milestone, required=True)

    def __init__(self):
        args = self.parser.parse_args()
        self.is_fake_milestone, self.milestone_id = args.milestone_id

    def get(self):

        if self.is_fake_milestone:
            return list(models.Plan.objects(
                fake_milestone=self.milestone_id
            ).order_by('-created_on').scalar('name'))

        return list(models.Plan.objects(
            milestone_id=self.milestone_id
        ).order_by('-created_on').scalar('name', 'id'))


class RunNameListByPlan(BaseResource):
    parser = reqparse.RequestParser()
    parser.add_argument('plan_id', type=int, required=True)

    def __init__(self):
        args = self.parser.parse_args()
        self.plan_id = args.plan_id

    def get(self):

        return list(models.Run.objects(
            plan_id=self.plan_id
        ).order_by('-created_on').scalar('name'))


class ProjectList(BaseResource):
    def get(self):
        query = models.Project.objects.scalar('id', 'name')
        return [{'id': i, 'name': name} for i, name in query]


api.add_resource(UserList, '/testrail/users')
api.add_resource(UserResource, '/testrail/user/<int:user_id>')
api.add_resource(CaseTypeList, '/testrail/case_types')
api.add_resource(PriorityList, '/testrail/priorities')
api.add_resource(MilestoneList, '/testrail/milestones')
api.add_resource(MilestoneNameList, '/testrail/milestone_names')
api.add_resource(QATeamsList, '/testrail/qa_teams')
api.add_resource(CaseList, '/testrail/cases')
api.add_resource(LastSync, '/testrail/sync/last')
api.add_resource(TrMilestoneInFocusResource, '/testrail/tr_milestone_in_focus')
api.add_resource(RunTypes, '/testrail/run_types')
api.add_resource(RunNamesList, '/testrail/run_list')
api.add_resource(PlanNameList, '/testrail/plans')
api.add_resource(RunNameListByPlan, '/testrail/runs')
api.add_resource(ProjectList, '/testrail/projects')
