import re
import datetime

import pytz
import dateutil
from bson import DBRef
from bson import ObjectId
from flask.json import JSONEncoder as FlaskJSONEncoder
from mongoengine.base import BaseDocument
from mongoengine.queryset.base import BaseQuerySet
from dateutil.relativedelta import relativedelta as rd
import holidays


utc_max_date = datetime.datetime.max.replace(
    tzinfo=pytz.utc,
    # replace microsecond by 0, because BSON date type doesn't support it.
    microsecond=0,
)

utc_min_date = datetime.datetime.max.replace(
    tzinfo=pytz.utc,
    # replace microsecond by 0, because BSON date type doesn't support it.
    microsecond=0,
)


def dt_now():
    return datetime.datetime.now(tz=pytz.utc)


def timestamp_to_dt(timestamp):
    try:
        return datetime.datetime.fromtimestamp(
            float(timestamp), tz=pytz.utc)
    except (ValueError, TypeError):
        return None


def dt_from_iso(dt):
    return dateutil.parser.parse(str(dt))


def daterange(date_from, date_to):
    day = datetime.timedelta(days=1)
    while date_from <= date_to:
        yield date_from
        date_from += day


def chunks(lst, n):
    """Split a list into evenly sized chunks."""
    return [lst[i:i + n] for i in range(0, len(lst), n)]


def slugify(text, delim='-'):
    """Generates an ASCII-only slug."""
    punct_re = re.compile(r'[\t !"#$%&\'()*\-/<=>?@\[\\\]^_`{|},.]+')
    result = []
    for word in punct_re.split(text.lower()):
        if word:
            result.append(word)
    return str(delim.join(result))


class LazyWrapper(object):
    def __init__(self, func):
        self.func = func

    def __call__(self):
        try:
            return self.value
        except AttributeError:
            self.value = self.func()
            return self.value


class MongoEngineJSONEncoder(FlaskJSONEncoder):
    """Override default Flask JSON encoder."""
    @staticmethod
    def _transform_document(obj):
        obj_mongo = obj.to_mongo(use_db_field=False)
        if '_id' in obj_mongo:
            del obj_mongo['_id']
        return obj_mongo

    def default(self, obj):
        if isinstance(obj, BaseDocument):
            return self._transform_document(obj)
        elif isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        elif isinstance(obj, (DBRef, ObjectId)):
            return str(obj)
        elif isinstance(obj, BaseQuerySet):
            return list(obj)
        return FlaskJSONEncoder.default(self, obj)


def has_intersection(array_field, values_list):
    """
    Boolean condition: Is document `array_field` values are intersected
    with `values_list`?. This is useful in mongo aggregate pipeline queries.

    :param array_field: array field in document
    :param values_list: list/tuple of values
    :return: dict with boolean condition.
    """
    assert isinstance(array_field, str) and array_field.startswith('$')
    assert isinstance(values_list, (tuple, list)) and values_list
    _array_field = {'$ifNull': (array_field, [])}
    return {
        '$gt': ({'$size': {'$setIntersection': (values_list, _array_field)}}, 0)
    }


def prefixed_collection_name(prefix=''):
    def get_name(model):
        if model._meta['abstract']:
            return None
        return prefix + ''.join(
            '_%s' % c if c.isupper() else c for c in model.__name__
        ).strip('_').lower()
    return get_name


def get_top_week_day(moment, week_range):

    def moment_in_range(_moment, top_week_margin):
        bottom_week_margin = top_week_margin - datetime.timedelta(days=7)
        return bottom_week_margin <= _moment <= top_week_margin

    for top_day in week_range:
        top_week_day = datetime.datetime.combine(
            top_day, datetime.time.max
        ).replace(tzinfo=pytz.utc)
        if moment_in_range(moment.replace(tzinfo=pytz.utc), top_week_day):
            return top_week_day


class RussianHolidays(holidays.HolidayBase):

    def __init__(self, **kwargs):
        self.country = 'RUS'
        holidays.HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):
        # New Year's Day
        if year > 1922:
            self[datetime.date(year, 1, 1)] = "New Year's Day"
            if year == 2015:
                for x in range(1, 12):
                    self[datetime.date(year, 1, x)] = "New Year's Days"
            if year == 2016:
                for x in range(1, 11):
                    self[datetime.date(year, 1, x)] = "New Year's Days"

        # 23 February
        if year > 1922:
            self[datetime.date(year, 2, 23)] = '23 February'
            if year == 2016:
                self[datetime.date(year, 2, 23) + rd(days=-1)] = '23 February'

        # 8 March
        if year > 1922:
            self[datetime.date(year, 3, 8)] = '8 March'
            if year == 2015:
                self[datetime.date(year, 3, 8) + rd(days=+1)] = '8 March'
            if year == 2016:
                self[datetime.date(year, 3, 8) - rd(days=+1)] = '8 March'

        # 1,2,3 May
        if year > 1922:
            self[datetime.date(year, 5, 1)] = "1 May"
            self[datetime.date(year, 5, 2)] = "2 May"
            self[datetime.date(year, 5, 3)] = "3 May"
            if year == 2015:
                self[datetime.date(year, 5, 3) + rd(days=+1)] = "1,2,3 May"

        # Victory Day
        if year > 1945:
            self[datetime.date(year, 5, 9)] = 'Victory Day'
            if datetime.date(year, 5, 9).weekday() == 5:
                self[datetime.date(year, 5, 9) + rd(days=+2)] = 'Victory Day'
            if datetime.date(year, 5, 9).weekday() == 6:
                self[datetime.date(year, 5, 9) + rd(days=+1)] = 'Victory Day'

        # Russia Day
        if year > 1992:
            name = 'Russia Day'
            self[datetime.date(year, 6, 12)] = name
            if datetime.date(year, 6, 12).weekday() == 5:
                self[datetime.date(year, 6, 12) + rd(days=+2)] = name
            if datetime.date(year, 6, 12).weekday() == 6:
                self[datetime.date(year, 6, 12) + rd(days=+1)] = name

        # Unity Day
        if year > 2005:
            name = 'Unity Day'
            self[datetime.date(year, 11, 4)] = name


class RUS(RussianHolidays):
    pass


class UkrainianHolidays(holidays.HolidayBase):

    def __init__(self, **kwargs):
        self.country = 'UA'
        holidays.HolidayBase.__init__(self, **kwargs)

    def _populate(self, year):

        # New Year's Day
        if year > 1922:
            self[datetime.date(year, 1, 1)] = "New Year's Day"

        # Christmas
        if year > 1922:
            self[datetime.date(year, 1, 7)] = 'Christmas'

        # 8 March
        if year > 1922:
            self[datetime.date(year, 3, 8)] = '8 March'

        # Easter
        if year == 2015:
            self[datetime.date(year, 4, 13)] = 'Easter'

        # Labor Day
        if year > 1922:
            name = 'Labor Day'
            self[datetime.date(year, 5, 1)] = name
            self[datetime.date(year, 5, 2)] = name
            if year == 2015:
                self[datetime.date(year, 5, 4)] = name
            if year == 2016:
                self[datetime.date(year, 5, 3)] = name

        # Victory Day
        if year > 1945:
            name = 'Victory Day'
            self[datetime.date(year, 5, 9)] = name
            if datetime.date(year, 5, 9).weekday() == 5:
                self[datetime.date(year, 5, 9) + rd(days=+2)] = name
            if datetime.date(year, 5, 9).weekday() == 6:
                self[datetime.date(year, 5, 9) + rd(days=+1)] = name

        # Pentecost
        if year == 2015:
            self[datetime.date(year, 6, 1)] = 'Pentecost'
        if year == 2016:
            self[datetime.date(year, 6, 20)] = 'Pentecost'

        # Constitution Day
        if year > 1996:
            name = 'Constitution Day'
            self[datetime.date(year, 6, 28)] = name
            if datetime.date(year, 6, 28).weekday() == 6:
                self[datetime.date(year, 6, 28) + rd(days=1)] = name

        # Independence Day
        if year > 1991:
            name = 'Independence Day'
            self[datetime.date(year, 8, 24)] = name
            if datetime.date(year, 8, 24).weekday() == 6:
                self[datetime.date(year, 8, 24) + rd(days=1)] = name

        # Defender of Ukraine Day
        if year >= 2015:
            name = 'Defender of Ukraine Day'
            self[datetime.date(year, 10, 14)] = name
            if datetime.date(year, 10, 14).weekday() == 6:
                self[datetime.date(year, 10, 14) + rd(days=1)] = name


class UA(UkrainianHolidays):
    pass
