import datetime
import json
import os
import unittest

from qareporting import create_app
from qareporting.integration.launchpad import HistoryBuilder
from qareporting.integration.launchpad import bug_from_raw
from qareporting.utils import dt_from_iso


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()

    raise TypeError("Type not serializable")


class TestSimple(unittest.TestCase):

    @classmethod
    def _load_bug_data(cls, name):
        filename = os.path.join(os.path.dirname(__file__), 'test_data', name)
        with open(filename, 'r') as f:
            return json.loads(f.read())

    @classmethod
    def _fix_expected_history_dates(cls, data):
        def date_hook(json_dict):
            for key, value in json_dict.items():
                if ('date' in key) or ('task_created' in key):
                    json_dict[key] = dt_from_iso(value)
                if key == 'interval':
                    json_dict[key] = [dt_from_iso(v) for v in value]
            return json_dict

        data['expected_history'] = json.loads(
            json.dumps(data['expected_history']),
            object_hook=date_hook,
        )

    def setUp(self):

        self.app = create_app('testing')
        self.app_context = self.app.app_context()
        self.app_context.push()

        self.bug_data = self._load_bug_data('bug_1335869.json')
        self._fix_expected_history_dates(self.bug_data)

        self.bug = bug_from_raw(
            self.bug_data['bug'],
            self.bug_data['tasks'],
            self.bug_data['activity']
        )
        self.hist_builder = HistoryBuilder(self.bug)
        self.history = self.hist_builder.get_history()
        # with open('/tmp/buildres.json', 'w') as f:
        #     f.write(json.dumps(self.history,
        #                        indent=4, sort_keys=True, default=json_serial))

    def test_history(self):
        expected_history = self.bug_data['expected_history']

        self.assertEqual(len(self.history), len(expected_history))
        for built, expected in zip(self.history, expected_history):

            self.assertDictEqual(built['bug'], expected['bug'])
            self.assertSequenceEqual(
                built['interval'], expected['interval'])

            self.assertEqual(len(built['tasks']), len(expected['tasks']))

            self.assertEqual(built['tasks'], expected['tasks'])

    def tearDown(self):
        self.app_context.pop()


